package de.uka.algo.generator.standalone.generators;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;

import de.uka.algo.generator.accessory.StandaloneDCRArguments;
import de.uka.algo.generator.standalone.graph.Cluster;
import de.uka.algo.generator.standalone.graph.DCRGraph;
import de.uka.algo.generator.standalone.graph.Node;
import de.uka.algo.generator.standalone.graph.clusterevents.ClusterEvent;
import de.uka.algo.generator.standalone.graph.clusterevents.ClusterEventManager;
import de.uka.algo.generator.standalone.graph.clusterevents.ImpossibleEventException;
import de.uka.algo.generator.util.Pair;
import de.uka.algo.generator.util.WeightedSelector;


/**
 * A generator of dynamic clustered random graphs.
 * 
 * @author <christian.staudt@ira.uka.de>
 */

public class DCRGenerator {

	Random rand;
	

	public DCRGenerator() {
		this.rand = new Random();
	}

	public DCRGraph generate(StandaloneDCRArguments args) {

		DCRGraph graph = new DCRGraph();
		

		boolean biasedClusterSelection;
		if (args.D_s == null)
			biasedClusterSelection = true;
		else
			biasedClusterSelection = false;

		boolean individualPIn;
		if (args.p_inList != null)
			individualPIn = true;
		else
			individualPIn = false;

		graph.setPOut(args.p_out);

		// create clusters
		if (individualPIn) {
			for (int i = 0; i < args.k; ++i) {
				graph.createCluster(args.p_inList.get(i));
			}
		} else {
			for (int i = 0; i < args.k; ++i) {
				graph.createCluster(args.p_in);
			}
		}

		// create cluster event manager
		ClusterEventManager clusterEventManager = graph
				.createClusterEventManager(args.enp, args.theta);
		if (individualPIn && args.enp) {
			clusterEventManager.makeGaussianEstimator(args.p_inList);
		}

		// create nodes
		if (biasedClusterSelection) {
			for (int i = 0; i < args.n; ++i) {
				Cluster c = graph.getRandomCluster(args.beta);
				graph.createNode(c);
			}
		} else {
			// weighted selection
			ArrayList<Cluster> clusterList = graph.getClusters();
			HashMap<Cluster,Double> sizeMap = new HashMap<Cluster,Double>();
			for (int i = 0; i < clusterList.size(); ++i) {
				sizeMap.put(clusterList.get(i), args.D_s.get(i));
			}
			WeightedSelector<Cluster> cSelector = new WeightedSelector<Cluster>(clusterList, sizeMap);
			
			for (int i = 0; i < args.n; ++i) {
				graph.createNode(cSelector.select());
			}
			
		}

		for (Pair<Node> uv : graph.getNodePairs()) {
			if (this.rand.nextDouble() <= graph.getEdgeProbability(uv)) {
				graph.createEdge(uv);
			}
		}
		
		// main loop
		//Rob control output
		int tenthCounter = 0;
		int tenthSize = (args.t_max / 10);
		boolean percInc = true; //has percentage been increased yet during this graph.t?
		//start = yes, to zero
		System.out.println("Progress: " + tenthCounter + "%, ");
		//\Rob
		
		for (int i = 1; graph.t < args.t_max; ++i) {
//			System.out.println(graph.t);
			//Rob control output
			if (((graph.t % tenthSize) == 0) && (!percInc)){
				System.out.println((++tenthCounter) * 10 + "%, ");
				percInc = true;
			}
			if (((graph.t % tenthSize) != 0) && (graph.t > 0)){
				percInc = false;
			}
			//\Rob
			
//			// DEBUG
//			System.out.print(i + " ");
//			if (i % 50 == 0) {
//				System.out.println();
//			}
//			// DEBUG
			
			
			if ((graph.edgeAdditionCounter + graph.edgeRemovalCounter >= args.eta) || (i == 1)) {
//				System.out.println("generator steps: " + i);


				
				graph.nextStep();
				
				graph.edgeAdditionCounter = 0;
				graph.edgeRemovalCounter = 0;
				
				
				// cluster events only once per timestep
				
				clusterEventManager.checkOngoingEvents();

				// cluster events
				if (rand.nextDouble() <= args.p_omega) {
					if (rand.nextDouble() <= args.p_mu) {
						try {
							ClusterEvent splitEvent = clusterEventManager.randomSplitEvent();
						} catch (ImpossibleEventException e) {
							// DEBUG
							System.out.println("no cluster to split");
							// DEBUG
						}
					} else {
						try {
							ClusterEvent mergeEvent = clusterEventManager.randomMergeEvent();
						} catch (ImpossibleEventException e) {
							// DEBUG
							System.out.println("no clusters to merge");
							// DEBUG
						}
					}
				}
				
			}



			
			// node and edge operations
			if (rand.nextDouble() >= args.p_chi) {
				// change nodes

				if (rand.nextDouble() <= args.p_nu) {
					Cluster c = graph.getRandomCluster(args.beta);
					Node u = graph.createNode(c);
					for (Node v : graph.getNodes()) {
						if (!v.equals(u)) {
							double p = graph.getEdgeProbability(u, v);
							if (rand.nextDouble() <= p) {
								graph.createEdge(u, v);
							}
						}
					}


				} else {
					if (graph.n > 0) {
						Node u = graph.getRandomNode();
						graph.removeNode(u);
					} else {
						System.out.println("could not delete node: graph has zero nodes");
					}

				}

			} else {
				// change edges
				if (graph.n >= 2) {
					

					
					if (graph.consistsOfDisjointCliques() && (! clusterEventManager.hasOngoingEvent())) {
						// do nothing in this case in order to prevent oscillation
						
						// artificially fire a time step event if generator cannot escape from this situation
						if (args.p_chi > 0.999) {
							graph.nextStep();
							graph.edgeAdditionCounter = 0;
							graph.edgeRemovalCounter = 0;
						} else {
							
						}
						
					} else {
						
						
						// decide whether to create or delete edge
						double edgePMass = graph.getPMassOfEdges();
						double complementPMass = graph.getPMassOfComplementEdges();
						
						//Robs suggestion for a global edge-densification parameter: 
						// don't use original masses, but modify them uniformly (i.e., their sum) by the densification_factor
						// densification_factor = by which factor shall all p_edge be modified per round
						// needs very small value! E.g. 2^(1/10K) = 1.0000693171203765 ( = all p_edge double after 10K steps) (not correct ! ;-))
						// this will work - but its dirty and does not comply with any nice assumptions or so
//						double edgePMass = densification_factor * graph.getPMassOfEdges();
//						double complementPMass = (1.0 / densification_factor) * graph.getPMassOfComplementEdges();
						
						ArrayList<Boolean> optionList = new ArrayList<Boolean>();
						optionList.add(true);
						optionList.add(false);
						HashMap<Boolean, Double> weightMap = new HashMap<Boolean, Double>();
						weightMap.put(true, complementPMass);
						weightMap.put(false, edgePMass);
						WeightedSelector<Boolean> optionSelector = new WeightedSelector<Boolean>(
								optionList, weightMap);
						boolean createEdge = optionSelector.select();
						if (createEdge) {
							if (!graph.isCompleteGraph()) {
								Pair<Node> uv = graph.selectUnconnectedPair();
								graph.createEdge(uv);
							}

						} else {
							if (!graph.isEdgelessGraph()) {
								Pair<Node> uv = graph.selectConnectedPair();
								graph.removeEdge(uv);
							}
						}
					}
					
				} else {
					// edge operation impossible: one or zero nodes
					System.out.println("edge operation impossible: one or zero nodes");
				}
			} // end change edges

		} // end main loop
		//Rob end control output
//		System.out.println();
		//\Rob

		// DEBUG
		System.out.println("Generator finished after " + graph.t + " timesteps.");
		// DEBUG
		
		return graph;

	}

}
