package de.uka.algo.generator.standalone.graph;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Random;

import de.uka.algo.generator.standalone.graph.clusterevents.ClusterEventManager;
import de.uka.algo.generator.standalone.graph.journaling.ClusteringJournal;
import de.uka.algo.generator.standalone.graph.journaling.GraphJournal;
import de.uka.algo.generator.standalone.graph.selecttree.EdgeSelectTree;
import de.uka.algo.generator.standalone.graph.selecttree.EdgeSelectTreeElement;
import de.uka.algo.generator.standalone.graph.selecttree.NodeSelectTree;
import de.uka.algo.generator.standalone.graph.selecttree.NodeSelectTreeElement;
import de.uka.algo.generator.util.BidiHashMap;

import de.uka.algo.generator.util.Pair;

/**
 * A custom data structure for a dynamic clustered random graph.
 * 
 * @author <christian.staudt @ ira.uka.de>
 * 
 */
public class DCRGraph {

	private NodeArray nodes;
	private EdgeTreeArray edges;
	private EdgeTreeArray complementEdges;
	private NodeSelectTree selectTree;
	private NodeSelectTree complementSelectTree;
	
	private HashMap<Integer, Integer> clusterSizeMap;
	private HashMap<Integer, Double> pMap;
	
	private ArrayList<Cluster> clusterList; // list of current clusters for selection
	
	private int clusterIdCounter;

	private EdgeCountMatrix edgeCount;
	private EdgeCountMatrix refEdgeCount;
	private Random rand;

	public int n; // number of nodes
	public int m; // number of edges
	public int t; // current time step

	private GraphJournal gJournal;
	private ClusteringJournal cJournal;
	private GraphLog log;

	private ClusterEventManager clusterEventManager;
	
	public int edgeAdditionCounter; // counts the number of edge events since the last time step
	public int edgeRemovalCounter;

	public DCRGraph() {
		this.nodes = new NodeArray(new Pair<Integer>(0, 0));

		this.edges = new EdgeTreeArray(this);
		this.complementEdges = new EdgeTreeArray(this);
		this.selectTree = new NodeSelectTree();
		this.complementSelectTree = new NodeSelectTree();

		this.clusterIdCounter = 1;
		
		this.clusterSizeMap = new HashMap<Integer, Integer>();
		this.pMap = new HashMap<Integer, Double>();
		this.clusterList = new ArrayList<Cluster>();
		

		this.edgeCount = new EdgeCountMatrix(3);
		this.refEdgeCount = new EdgeCountMatrix(3);

		this.rand = new Random();



		this.n = 0;
		this.m = 0;
		this.t = -1;

		this.gJournal = new GraphJournal();
		this.cJournal = new ClusteringJournal();
		this.log = new GraphLog(true);
		
		this.edgeAdditionCounter = 0;
		this.edgeRemovalCounter = 0;
		
	}

	/********** manipulating nodes and edges ************/

	public Node createNode(Cluster c) {

		++this.n;

		Cluster a = this.clusterEventManager.getAncestor(c);
		
		if (! a.equals(c)) {
			int as = this.clusterSizeMap.get(a.id);
			this.clusterSizeMap.put(a.id, as + 1);
		}



		Node u = this.nodes.createNode(new Pair<Integer>(c.id, a.id));

		// edges
		this.selectTree.add(0);
		this.complementSelectTree.add(0);
		this.edges.add(new EdgeSelectTree());
		this.complementEdges.add(new EdgeSelectTree());
		
		// add complement edges to all other nodes
		for (Node v : this.nodes) {
			if (!v.equals(u)) {
				EdgeSelectTree uComplementTree = this.complementEdges.getTree(u);
				if (uComplementTree != null) {
					uComplementTree.addEdge(v, this.getEdgeProbability(u, v));
				}
			} else {
			}
		}
		
		this.complementEdges.introduceNode(u);

		// update weight
		for (Node v : this.nodes) {
			EdgeSelectTree vComplementTree = this.complementEdges.getTree(v);
			this.complementSelectTree.updateWeight(v.id, vComplementTree.getTotalWeight());
		}
		
		// update cluster size
		int s = this.clusterSizeMap.get(c.id);
		this.clusterSizeMap.put(c.id, s+1);
		
		
		// journal
		this.gJournal.createNodeOp(c.id, a.id);
		// log
		this.log.logCreateNode(u.id, c.id, a.id);

		
		return u;
	}

	public void removeNode(Node u) {
		assert this.nodes.getNodeById(u.id) != null;

		// remove incident edges
		ArrayList<Edge> incidentEdges = new ArrayList<Edge>();
		for (Edge e : this.edges.getEdges(u)) {
			incidentEdges.add(e);
		}
		for (Edge e : incidentEdges) {
			this.removeEdge(e);
		}

		// remove complement edges
		for (Edge e : this.complementEdges.getEdges(u)) {
			this.complementEdges.getTree(e.v).removeEdge(u);
		}
		this.edges.delete(u);
		this.complementEdges.delete(u);
	

		// select trees
		this.selectTree.remove(u.id);
		this.complementSelectTree.remove(u.id);
		
		this.n--;
		
		int s = this.clusterSizeMap.get(u.data.x);
		this.clusterSizeMap.put(u.data.x, --s);
		
		if (u.data.x != u.data.y) {
			int rs = this.clusterSizeMap.get(u.data.y);
			this.clusterSizeMap.put(u.data.y, --rs);
		}

		
		this.nodes.deleteNode(u.id);
		
		
		// update weight
		for (Node v : this.nodes) {
			EdgeSelectTree vComplementTree = this.complementEdges.getTree(v);
			this.complementSelectTree.updateWeight(v.id, vComplementTree.getTotalWeight());
		}

		// journal
		this.gJournal.removeNodeOp(u.id);
		// log
		this.log.logDeleteNode(u.id);
		

	}

	/**
	 * Create an edge {u,v} with weight 1-p(u,v)
	 * 
	 * @param u
	 * @param v
	 * @return
	 */
	public Edge createEdge(Node u, Node v) {
		if ((!u.equals(v)) && !(this.hasEdge(u, v))) {

			this.m++;
			this.edgeAdditionCounter++;
			
			double p = this.getEdgeProbability(u, v);
			
			EdgeSelectTree uTree = this.edges.getTree(u);
			uTree.addEdge(v, 1 - p);
			EdgeSelectTree uComplementTree = this.complementEdges.getTree(u);
			uComplementTree.removeEdge(v);
			
			EdgeSelectTree vTree = this.edges.getTree(v);
			vTree.addEdge(u, 1 - p);
			EdgeSelectTree vComplementTree = this.complementEdges.getTree(v);
			vComplementTree.removeEdge(u);
			
			this.selectTree.updateWeight(u.id, uTree.getTotalWeight());
			this.selectTree.updateWeight(v.id, vTree.getTotalWeight());
			
			this.complementSelectTree.updateWeight(u.id, uComplementTree.getTotalWeight());
			this.complementSelectTree.updateWeight(v.id, vComplementTree.getTotalWeight());
			// update edge count
			Cluster c = this.getCluster(u);
			Cluster d = this.getCluster(v);

			this.edgeCount.update(c, d, 1);

			Cluster cr = this.getReferenceCluster(u);
			Cluster dr = this.getReferenceCluster(v);

			this.refEdgeCount.update(cr, dr, 1);

			// journal
			this.gJournal.createEdgeOp(u.id, v.id);
			// log
			this.log.logCreateEdge(u.id, v.id);
			

			Edge e = new Edge(u, v, 1 - p);
			
			assert this.m <= (this.n * (this.n-1)) / 2;

			
			return e;
		} else {
			return null;
		}
	}

	public Edge createEdge(Pair<Node> uv) {
		return this.createEdge(uv.x, uv.y);
	}

	/**
	 * Delete an edge {u,v}
	 * 
	 * @param u
	 * @param v
	 * @return
	 */
	public void removeEdge(Node u, Node v) {
		assert this.hasEdge(u, v);

		this.edgeRemovalCounter++;
		this.m--;
		double p = this.getEdgeProbability(u, v);
		
		EdgeSelectTree uTree = this.edges.getTree(u);
		uTree.removeEdge(v);
		
		EdgeSelectTree uComplementTree = this.complementEdges.getTree(u);
		uComplementTree.addEdge(v, p);
		
		EdgeSelectTree vTree = this.edges.getTree(v);
		vTree.removeEdge(u);
		
		EdgeSelectTree vComplementTree = this.complementEdges.getTree(v);
		vComplementTree.addEdge(u, p);
		
		// update select tree weight
		this.selectTree.updateWeight(u.id, uTree.getTotalWeight());
		this.selectTree.updateWeight(v.id, vTree.getTotalWeight());
		
		this.complementSelectTree.updateWeight(u.id, uComplementTree.getTotalWeight());
		this.complementSelectTree.updateWeight(v.id, vComplementTree.getTotalWeight());
		
		// update edge count
		Cluster c = this.getCluster(u);
		Cluster d = this.getCluster(v);
		Cluster cr = this.getReferenceCluster(u);
		Cluster dr = this.getReferenceCluster(v);
		this.edgeCount.update(c, d, -1);
		this.refEdgeCount.update(cr, dr, -1);

		// journal
		this.gJournal.removeEdgeOp(u.id, v.id);
		// log
		this.log.logDeleteEdge(u.id, v.id);
		

	}

	public void removeEdge(Edge e) {
		this.removeEdge(e.u, e.v);
	}

	public void removeEdge(Pair<Node> uv) {
		this.removeEdge(uv.x, uv.y);
	}

	public boolean hasEdge(Node u, Node v) {
		if (this.edges.getTree(u) != null && this.edges.getTree(v) != null) {
			return this.edges.getTree(u).hasEdgeTo(v)
					&& this.edges.getTree(v).hasEdgeTo(u);
		} else
			return false;
	}

	/************* Manipulating Clusters ******************/

	public void setPOut(double p_out) {
		this.pMap.put(0, p_out);
	}

	public double getPOut() {
		return this.pMap.get(0);
	}



	public Cluster createCluster(double p_in) {
		
		int id = this.clusterIdCounter;
		this.clusterIdCounter++;
		
		this.pMap.put(id, p_in);
		this.clusterSizeMap.put(id, 0);
		
		Cluster c = new Cluster(id, p_in);
		
		this.edgeCount.add(c);
		this.refEdgeCount.add(c);
		this.clusterList.add(c);
		
		return c;
	}


	public void removeCluster(Cluster c) {
		this.clusterList.remove(c);
		
	}


	public void removeReferenceCluster(Cluster rc) {
		this.pMap.remove(rc.id);
		this.clusterSizeMap.remove(rc.id);
	}

	public ClusterEventManager createClusterEventManager(boolean enp,
			double theta) {
		this.clusterEventManager = new ClusterEventManager(this, theta);
		return this.clusterEventManager;
	}

	/************* Acessing Objects ******************/

	public Iterable<Node> getNodes() {
		return this.nodes;
	}

	/**
	 * Get all node pairs {u,v} where u != v.
	 * 
	 * @return
	 */
	public Collection<Pair<Node>> getNodePairs() {
		return this.nodes.getNodePairs();

	}

	
	/**
	 * Select a random cluster (biased selection).
	 */
	public Cluster getRandomCluster(double bias) {
		double x = this.rand.nextDouble();
		int i = (int) Math.floor(Math.pow(x, bias) * (this.clusterList.size()) + 1);
		return this.clusterList.get(i-1);
	}
	

	public Node getRandomNode() {
		return this.nodes.getRandomNode();
	}

	/**
	 * Gets the cluster containing the given node.
	 * 
	 * @param u
	 * @return
	 */
	public Cluster getCluster(Node u) {
		int id = u.data.x;
		return new Cluster(id, this.pMap.get(id));
	}


	/**
	 * Gets the reference cluster for a node from nodes
	 * 
	 * @param u
	 * @return
	 */
	public Cluster getReferenceCluster(Node u) {
		int id = u.data.y;
		return new Cluster(id, this.pMap.get(id));
	}
	


	/**
	 * Returns the edge probability for two nodes.
	 * 
	 * @param u
	 * @param v
	 * @return
	 */
	public double getEdgeProbability(Node u, Node v) {
		if (u.data.x == v.data.x) {
			return this.pMap.get(u.data.x);
		} else {
			return this.pMap.get(0);
		}
	}
	

	/**
	 * @see DCRGraph#getEdgeProbability(Node, Node)
	 * @param uv
	 * @return
	 */
	public double getEdgeProbability(Pair<Node> uv) {
		return this.getEdgeProbability(uv.x, uv.y);
	}

	/**
	 * Weighted selection of a connected node pair.
	 * 
	 * @return
	 */
	public Pair<Node> selectConnectedPair() {
		assert this.n >= 2;
		
		int uid = this.selectTree.select();
		Node u = this.getNodeById(uid);
		
		assert u != null;
		assert ! (this.getDegree(u) == 0);
		
		EdgeSelectTree uTree = this.edges.getTree(u);
		int vid = uTree.select();

		Node v = this.getNodeById(vid);

		assert this.hasEdge(u, v);
		
		return new Pair<Node>(u,v);
		
	}

	/**
	 * Weighted selction of an unconnected node pair.
	 * 
	 * @return
	 */
	public Pair<Node> selectUnconnectedPair() {
		
		int uid = this.complementSelectTree.select();
		assert (uid != 0);
		
		Node u = this.getNodeById(uid);
		assert u != null;
		
		EdgeSelectTree uTree = this.complementEdges.getTree(u);

		assert ! (this.getDegree(u) == (this.n -1)); // <=> ! uTree.array == [null]

		
		int vid = uTree.select();
		Node v = this.getNodeById(vid);
		assert v != null;
		
		return new Pair<Node>(u,v);
	}

	public boolean isCompleteGraph() {
		return (this.m == ((this.n * (this.n-1)) / 2));
	}
	
	public boolean isEdgelessGraph() {
		return this.m == 0;
	}
	

	/**
	 * Return list of current clusters.
	 * 
	 * @return
	 */
	public ArrayList<Cluster> getClusters() {
		return this.clusterList;
	}



	/********************************/

	private void moveNode(Node u, Cluster d) {
		
		this.nodes.updateNodeData(u, new Pair<Integer>(d.id, u.data.y));
		

		
		this.gJournal.setClusterOp(u.id, d.id);
		this.log.logMoveNode(u.id, d.id);
	}

	private void moveReferenceNode(Node u, Cluster d) {
		this.nodes.updateNodeData(u, new Pair<Integer>(u.data.x, d.id));

		this.gJournal.setRefClusterOp(u.id, d.id);
		this.log.logMoveRefNode(u.id, d.id);
	}

	/********** expected values **********/


	/**
	 * Expected number of edges
	 * 
	 * @return
	 */
	public double expectedTotalEdges() {
		
		double expectedM = 0.0;
		int intraPairs = 0;
		int totalSize = 0;
		for (Cluster c : this.clusterList) {
			int size = this.clusterSizeMap.get(c.id);
			double p_in = this.pMap.get(c.id);
			
			int pairs = ((size * (size - 1)) / 2);
			intraPairs = intraPairs + pairs;
			expectedM = expectedM + pairs * p_in;
			totalSize = totalSize + size;
		}

		int allPairs = ((totalSize) * (totalSize - 1)) / 2;
		int interPairs = allPairs - intraPairs;
		double p_out = this.pMap.get(0);
		double interEdges = interPairs * p_out;

		expectedM = expectedM + interEdges;
		return expectedM;
	}

	/**
	 * Returns the degree of a node.
	 * 
	 * @param v
	 * @return
	 */
	public int getDegree(Node v) {
		return this.edges.getTree(v).size();
	}

	/**
	 * Return the number of nodes in a cluster
	 * 
	 * @param c
	 * @return
	 */
	public int getClusterSize(Cluster c) {
		
		return this.clusterSizeMap.get(c.id);
	}



	/**
	 * Return the current number of edges between two clusters.
	 * 
	 * @param c
	 * @param d
	 * @return
	 */
	public int getEdgeCountBetween(Cluster c, Cluster d) {
		return this.edgeCount.get(c, d);
	}

	public int getEdgeCountBetweenReference(Cluster cr, Cluster dr) {
		return this.refEdgeCount.get(cr, dr);
	}

	/******** Cluster Manipulation *****/

	/**
	 * Split a cluster.
	 */
	public Pair<Cluster> splitCluster(Cluster c, Pair<Double> p_in, double beta) {
		
		// DEBUG
		System.out.println(this.t + ": splitting cluster " + c);
		// DEBUG
		

		Cluster d = this.createCluster(p_in.x);
		Cluster e = this.createCluster(p_in.y);
		
		int dSize = 0;
		int eSize = 0;
		
		for (Node u : this.nodes) {
			if (this.getCluster(u).equals(c)) {
				int i = (int) Math.floor((this.rand.nextDouble() * beta) * 2);
				if (i == 0) {
					this.moveNode(u, d);
					++dSize;
				} else {
					this.moveNode(u, e);
					++eSize;
				}
			}
		}
		
		this.clusterSizeMap.put(d.id, dSize);
		this.clusterSizeMap.put(e.id, eSize);
		
		// re-weight edges formerly belonging to the split cluster
		double p_out = this.getPOut();
		
		for (Node u : this.nodes) {
			Cluster f = this.getCluster(u);
			Cluster fr = this.getReferenceCluster(u);
			if (fr.equals(c)) {
				// if node u belonged to cluster c
				
				EdgeSelectTree uTree = this.edges.getTree(u);
				for (EdgeSelectTreeElement element : uTree) {
					Node v = this.getNodeById(element.target);
					Cluster g = this.getCluster(v);
					if (g.equals(d)) {
						if (f.equals(d)) {
							// u and v in d
							uTree.removeEdge(v);
							uTree.addEdge(v, 1 - p_in.x);
						} else if (f.equals(e)) {
							// u and v in different clusters
							uTree.removeEdge(v);
							uTree.addEdge(v, 1 - p_out);
						} else {
							// edge weights are not affected
						}
					} else if (g.equals(e)) {
						if (f.equals(e)) {
							// u and v in e
							uTree.removeEdge(v);
							uTree.addEdge(v, 1 - p_in.y);
						} else if (f.equals(d)) {
							// u and v in different clusters
							uTree.removeEdge(v);
							uTree.addEdge(v, 1 - p_out);
						} else {
							// edge weights are not affected
						}
					} else {
						// v not in d or e => edge weights not affected
					}
				}
				
				EdgeSelectTree uComplementTree = this.complementEdges.getTree(u);
				for (EdgeSelectTreeElement element : uComplementTree) {
					Node v = this.getNodeById(element.target);
					Cluster g = this.getCluster(v);
					if (g.equals(d)) {
						if (f.equals(d)) {
							// u and v in d
							uComplementTree.removeEdge(v);
							uComplementTree.addEdge(v, p_in.x);
						} else if (f.equals(e)) {
							// u and v in different clusters
							uComplementTree.removeEdge(v);
							uComplementTree.addEdge(v, p_out);
						} else {
							// edge weights are not affected
						}
					} else if (g.equals(e)) {
						if (f.equals(e)) {
							// u and v in e
							uComplementTree.removeEdge(v);
							uComplementTree.addEdge(v, p_in.y);
						} else if (f.equals(d)) {
							// u and v in different clusters
							uComplementTree.removeEdge(v);
							uComplementTree.addEdge(v, p_out);
						} else {
							// edge weights are not affected
						}
					} else {
						// v not in d or e => edge weights not affected
					}
				}
				
				// update NodeSelectTree weights
				this.selectTree.updateWeight(u.id, uTree.getTotalWeight());
				this.complementSelectTree.updateWeight(u.id, uComplementTree.getTotalWeight());
				
			}
		}
		

		this.removeCluster(c);
		
		this.recountEdges();

		
		this.log.logSplitCluster(c.id, d.id, e.id);
		
		this.cJournal.splitOp(c.id, d.id, e.id);
		
		return new Pair<Cluster>(d, e);
	}

	/**
	 * Split a cluster in the reference clustering.
	 * 
	 * @param c
	 */
	public void splitReferenceCluster(Cluster c) {
		for (Node u : this.nodes) {
			if (this.getReferenceCluster(u).equals(c)) {
				Cluster d = this.getCluster(u);
				this.moveReferenceNode(u, d);
			}
		}
		
		this.refRecountEdges();
		
		this.log.logSplitRefCluster(c.id);
		
	}

	/**
	 * Merge two clusters.
	 * 
	 * @param cd
	 * @param p_in
	 * @return
	 */
	public Cluster mergeClusters(Pair<Cluster> cd, double p_in) {
		assert (!cd.x.equals(cd.y));

		// DEBUG
		System.out.println(this.t + ": merging clusters " + cd.x + " " + cd.y);
		// DEBUG
		

		
		// re-weight edges belonging to the resulting cluster
		
		for (Node u : this.nodes) {
			Cluster e = this.getCluster(u);
			if (e.equals(cd.x) || e.equals(cd.y)) {
				// if node u belongs to one of the merging clusters
				
				EdgeSelectTree uTree = this.edges.getTree(u);
				for (EdgeSelectTreeElement element : uTree) {
					Node v = this.getNodeById(element.target);
					Cluster f = this.getCluster(v);
					if (f.equals(cd.x) || f.equals(cd.y)) {
						uTree.removeEdge(v);
						uTree.addEdge(v, 1 - p_in);
					}
				}
				
				EdgeSelectTree uComplementTree = this.complementEdges.getTree(u);
				for (EdgeSelectTreeElement element : uComplementTree) {
					Node v = this.getNodeById(element.target);
					Cluster f = this.getCluster(v);
					if (f.equals(cd.x) || f.equals(cd.y)) {
						uComplementTree.removeEdge(v);
						uComplementTree.removeEdge(v);
						uComplementTree.addEdge(v, p_in);
					}
				}
				
				// update NodeSelectTree weights
				this.selectTree.updateWeight(u.id, uTree.getTotalWeight());
				this.complementSelectTree.updateWeight(u.id, uComplementTree.getTotalWeight());
			}
		}
		

		Cluster e = this.createCluster(p_in);
		
		int eSize = 0;
		
		for (Node u : this.nodes) {
			if ((this.getCluster(u).equals(cd.x))) {
				this.moveNode(u, e);
				++eSize;
			} else if ((this.getCluster(u).equals(cd.y))) {
				this.moveNode(u, e);
				++eSize;
			}
		}
		
		// update cluster size
		this.clusterSizeMap.put(e.id, eSize);

		
		this.removeCluster(cd.x);
		this.removeCluster(cd.y);
		
		
		// recount edges
		this.recountEdges();
				
		this.log.logMergeClusters(cd.x.id, cd.y.id, e.id);
		
		// journal
		this.cJournal.mergeOp(cd.x.id, cd.y.id, e.id);

		return new Cluster(e.id, this.pMap.get(e.id));
	}

	/**
	 * Merges two clusters in the reference clustering.
	 * 
	 * @param cd
	 */
	public Cluster mergeReferenceClusters(Pair<Cluster> cd, Cluster e) {
		
		for (Node u : this.nodes) {
			if (this.getReferenceCluster(u).equals(cd.x)) {
				this.moveReferenceNode(u, e);
			} else if (this.getReferenceCluster(u).equals(cd.y)) {
				this.moveReferenceNode(u, e);
			}
		}
		

		this.removeReferenceCluster(cd.x);
		this.removeReferenceCluster(cd.y);

		this.refRecountEdges();
				
		this.log.logMergeRefClusters(cd.x.id, cd.y.id);
		
		return new Cluster(e.id, this.pMap.get(e.id));

	}

	/*************/

	public String toString() {
		String nodes = this.nodes.toString();
		String edges = "";
		for (EdgeSelectTree est : this.edges.getEdgeSelectTrees()) {
				edges = edges + est.toString();
		}
		String cedges = "";
		for (EdgeSelectTree est : this.complementEdges.getEdgeSelectTrees()) {
				cedges = cedges + est.toString();
		}
		return nodes + "\n" + edges + "\n" + cedges + "\n";
	}



	/**
	 * Set journal to the next timestep.
	 */
	public void nextStep() {
		this.gJournal.nextStepOp();
		this.cJournal.nextStepOp();
		this.log.logNextStep();
		this.t++;
	}

	/**
	 * Get the journal from which the graph can be reconstructed.
	 * 
	 * @return
	 */
	public GraphJournal getGraphJournal() {
		return this.gJournal;
	}
	
	public ClusteringJournal getClusteringJournal() {
		return this.cJournal;
	}
	
	
	
	public StringBuffer getLogBuffer() {
		return this.log.getStringBuffer();
	}
	
	public GraphLog getGraphLog() {
		return this.log;
	}

	
	
	/**
	 * Get all edges.
	 * @return
	 */
	public ArrayList<Edge> getEdges() {
		// TODO: implement more elegant solution using iterators
		
		HashSet<Node> scanned = new HashSet<Node>();
		ArrayList<Edge> edges = new ArrayList<Edge>();
		for (Node u : this.getNodes()) {
			for (Edge edge : this.edges.getEdges(u)) {
				if (! scanned.contains(edge.v)) {
					edges.add(edge);
				}
			}
			scanned.add(u);
		}
		
		int edgesSize = edges.size();
		assert edgesSize == this.m;
		
		return edges;
	}
	
	
	
	private void recountEdges() {
		this.edgeCount.resetAll();
		
		for (Edge edge : this.getEdges()) {
			
			Cluster s = this.getCluster(edge.u);
			Cluster t = this.getCluster(edge.v);
			int count = this.edgeCount.get(s, t);
			this.edgeCount.set(s, t, count + 1);
		}
		
		int sum = this.edgeCount.getSum();
		assert  sum == this.m;
	}
	
	private void refRecountEdges() {
		this.refEdgeCount.resetAll();
		
		for (Edge edge : this.getEdges()) {
			Cluster s = this.getReferenceCluster(edge.u);
			Cluster t = this.getReferenceCluster(edge.v);
			int count = this.refEdgeCount.get(s, t);
			this.refEdgeCount.set(s, t, count + 1);
		}
		
		int sum = this.refEdgeCount.getSum();
		assert  sum == this.m;
	}
	
	
	/**
	 * check data structure for consistency
	 */
	public void checkGraph() {
		assert this.m == this.edgeCount.getSum();
		
		// check NodeSelectTrees for consistency
		for (int i = 1; i < this.selectTree.array.size(); ++i) {
			NodeSelectTreeElement e = this.selectTree.array.get(i);
			if (this.getNodeById(i) != null) {
				Node u = this.getNodeById(i);
				EdgeSelectTree uTree = this.edges.getTree(u);
				assert e.thisWeight == uTree.getTotalWeight();
			}
		}
		
		for (int i = 1; i < this.complementSelectTree.array.size(); ++i) {
			NodeSelectTreeElement e = this.complementSelectTree.array.get(i);
			if (this.getNodeById(i) != null) {
				Node u = this.getNodeById(i);
				EdgeSelectTree uComplementTree = this.complementEdges.getTree(u);
				assert e.thisWeight == uComplementTree.getTotalWeight();
			}
		}
		
		// check cluster size
		int size = 0;
		for (Cluster c : this.clusterList) {
			size += this.clusterSizeMap.get(c.id);
		}
		assert this.n == size;
		
	}
	
	
	public Node getNodeById(int id) {
		return this.nodes.getNodeById(id);
	}

	
	public double getPMassOfEdges() {
		return this.selectTree.getTotalWeight();
	}
	
	
	public double getPMassOfComplementEdges() {
		return this.complementSelectTree.getTotalWeight();
	}
	
	
	/**
	 * Test whether the graph consists of disjoint cliques.
	 * @return
	 */
	public boolean consistsOfDisjointCliques() {
		boolean result;
		int sum = 0;

		for (Cluster c : this.clusterList) {
			int size = this.getClusterSize(c);
			int maxEdgeCount = (size * (size-1)) / 2;
			int intraClusterEdgeCount = this.edgeCount.get(c, c);
			sum += intraClusterEdgeCount;
			if  (intraClusterEdgeCount != maxEdgeCount) {
				return false;
			}
		}
		
		if ((sum - this.m) != 0) {
			// there are inter-cluster edges
			return false;
		}
		return true;
	}
	

}
