package de.uka.algo.generator.standalone.graph.clusterevents;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Random;

import de.uka.algo.generator.standalone.graph.Cluster;
import de.uka.algo.generator.standalone.graph.DCRGraph;
import de.uka.algo.generator.standalone.graph.GraphLog;
import de.uka.algo.generator.standalone.graph.journaling.ClusteringJournal;
import de.uka.algo.generator.standalone.graph.journaling.GraphJournal;
import de.uka.algo.generator.util.GaussianEstimator;
import de.uka.algo.generator.util.Pair;



public class ClusterEventManager {
	
	private final DCRGraph graph;
	
    private final ArrayList<ClusterEvent> ongoingEvents;
    private final ArrayList<ClusterEvent> completedEvents;
    private final ArrayList<Cluster> engaged;
    private final ArrayList<Cluster> disengaged;
    private final HashMap<Cluster, Cluster> splitAncestors;
    private final HashMap<Cluster, Pair<Cluster>> mergeAncestors;
    private final HashMap<Cluster, ClusterEvent.EventType> productOf;
    private Random rand;
    private GaussianEstimator gaussianEstimator;
    
    private GraphLog log;
    private ClusteringJournal cJournal;
    
	private double theta;
    
    public ClusterEventManager(DCRGraph graph, double theta) {
    	
    	this.graph = graph;
    	this.ongoingEvents = new ArrayList<ClusterEvent>();
    	this.completedEvents = new ArrayList<ClusterEvent>();
    	this.engaged = new ArrayList<Cluster>();
    	this.disengaged = new ArrayList<Cluster>();
    	for (Cluster c : this.graph.getClusters()) {
    		this.disengaged.add(c);
    	}
    	this.splitAncestors = new HashMap<Cluster, Cluster>();
    	this.mergeAncestors = new HashMap<Cluster, Pair<Cluster>>();
    	this.productOf = new HashMap<Cluster, ClusterEvent.EventType>();
    	this.rand = new Random();
    	
    	this.theta = theta;
    	
    	this.log = graph.getGraphLog();
    	this.cJournal = graph.getClusteringJournal();
    }
    
    
    public void makeGaussianEstimator(Collection<Double> p_inList) {
    	this.gaussianEstimator = new GaussianEstimator(p_inList);
    }

	
    /**
     * Try to split a randomly selected cluster.
     * @throws ImpossibleEventException
     */
	public ClusterEvent randomSplitEvent() throws ImpossibleEventException {
		
		if (this.disengaged.size() >= 1) {
			// get clusters
			int i = this.rand.nextInt(this.disengaged.size());
			Cluster c = this.disengaged.get(i);
			// get p_in
			Pair<Double> p_in;
			if (this.gaussianEstimator != null) {
				p_in = new Pair<Double>(this.gaussianEstimator.next(), this.gaussianEstimator.next());
			} else {
				p_in = new Pair<Double>(c.p_in, c.p_in);
			}
			// split cluster
			Pair<Cluster> de = this.graph.splitCluster(c, p_in, 1.0);
			// engage
			this.engageForSplit(c, de);
			ClusterSplitEvent clusterSplitEvent = new ClusterSplitEvent(c, de, p_in);
			this.ongoingEvents.add(clusterSplitEvent);
			this.productOf.put(de.x, ClusterEvent.EventType.SPLIT);
			this.productOf.put(de.y, ClusterEvent.EventType.SPLIT);
			this.splitAncestors.put(de.x, c);
			this.splitAncestors.put(de.y, c);
			
			return clusterSplitEvent;
			
		} else throw new ImpossibleEventException();
		
	}
	
	/**
	 * Try to merge two randomly selected clusters.
	 * @throws ImpossibleEventException
	 */
	public ClusterEvent randomMergeEvent() throws ImpossibleEventException {
		if (this.disengaged.size() >= 2) {
			// get two available clusters
			int i = this.rand.nextInt(this.disengaged.size());
			int j;
			do {
				j = this.rand.nextInt(this.disengaged.size());
			} while (i == j);

			Pair<Cluster> cd = new Pair<Cluster>(this.disengaged.get(i), this.disengaged.get(j));
			// get p_in
			double p_in;
			if (this.gaussianEstimator != null) {
				p_in = this.gaussianEstimator.next();
			} else {
				p_in = (cd.x.p_in + cd.y.p_in) / 2.0;
			}
			// merge clusters
			Cluster e = this.graph.mergeClusters(cd, p_in);
			// engage 
			this.engageForMerge(cd, e);
			ClusterMergeEvent clusterMergeEvent = new ClusterMergeEvent(cd, e, p_in);
			this.ongoingEvents.add(clusterMergeEvent);
			this.productOf.put(e, ClusterEvent.EventType.MERGE);
			this.mergeAncestors.put(e, cd);
			return clusterMergeEvent;
			
		} else throw new ImpossibleEventException();
		
	}
	
	
	public void checkOngoingEvents() {
		ArrayList<ClusterEvent> complete = new ArrayList<ClusterEvent>();
    	for (ClusterEvent ev : this.ongoingEvents) {
    		if (this.isComplete(ev)) {
    			// DEBUG
    			System.out.println(graph.t + ": event " + ev + " is complete");
    			// DEBUG
    			complete.add(ev);
    			this.establishEvent(ev);

    		}
    	}
		this.ongoingEvents.removeAll(complete);
		this.completedEvents.addAll(complete);
    	
	}
	
	
	
    private boolean isComplete(ClusterEvent ev) {
        if (ev.getType() == ClusterEvent.EventType.MERGE) {
            ClusterMergeEvent mergeEv = (ClusterMergeEvent) ev;
            Cluster c = mergeEv.cd.x;
            Cluster d = mergeEv.cd.y;
            Cluster e = mergeEv.e;

            int cSize = this.graph.getClusterSize(c);
            int dSize = this.graph.getClusterSize(d);
            double p_out = this.graph.getPOut();
            double p_in = e.p_in;
            
            double a =  cSize * dSize * p_out;
            double b = cSize * dSize * p_in;
            double count = graph.getEdgeCountBetweenReference(c, d);

            
            double value = ((this.theta * a ) + ((1 - this.theta) * b));
            boolean complete;
            
            // DEBUG
//            double percentage = (count / value) * 100;
//            System.out.println(graph.t + ": event " + ev + ": " +  percentage + " %");
            // DEBUG
            
            if (count >= value ) {
                complete = true;
            } else complete = false;
            
            this.log.logMergeCompleteness(mergeEv, complete, this.graph.n, cSize, dSize, p_out, p_in, a, b, count, theta, value);

            return complete;
            
        } else if (ev.getType() == ClusterEvent.EventType.SPLIT) {
        	ClusterSplitEvent splitEv = (ClusterSplitEvent) ev;
        	Cluster c = splitEv.c;
        	Cluster d = splitEv.de.x;
        	Cluster e = splitEv.de.y;
        	

        	int dSize = this.graph.getClusterSize(d);
			int eSize = this.graph.getClusterSize(e);
			double p_in = c.p_in;
			double p_out = this.graph.getPOut();
			
			double b = dSize * eSize * p_in;
        	double a = dSize * eSize * p_out;
        	double count = this.graph.getEdgeCountBetween(d, e);
        	
        	double value = ((this.theta * b) + ((1 - this.theta) * a));
        	boolean complete;
        	if ( count <= value ) {
        		complete = true;
        	} else complete = false;
        	
        	this.log.logSplitCompleteness(splitEv, complete, this.graph.n, dSize, eSize, p_out, p_in, a, b, count, theta, value);
        	
        	return complete;
        }
        return false;
    }
	
    
    private void engageForSplit(Cluster c, Pair<Cluster> de) {
    	disengaged.remove(c);
    	engaged.add(de.x);
    	engaged.add(de.y);
    }
    
    private void disengageForSplit(Pair<Cluster> de) {
    	engaged.remove(de.x);
    	engaged.remove(de.y);
    	disengaged.add(de.x);
    	disengaged.add(de.y);
    }
    
    private void engageForMerge(Pair<Cluster> cd, Cluster result) {
    	disengaged.remove(cd.x);
    	disengaged.remove(cd.y);
    	engaged.add(result);
    }
    
    private void disengageForMerge(Cluster result) {
    	assert (! disengaged.contains(result));
    	engaged.remove(result);
    	disengaged.add(result);
    }
	
	
	private void establishEvent(ClusterEvent ev) {
		
		if (ev.getType() == ClusterEvent.EventType.SPLIT) {
			ClusterSplitEvent splitEv = (ClusterSplitEvent) ev;

			this.graph.splitReferenceCluster(splitEv.c);
			// disengage
			this.disengageForSplit(splitEv.de);
			// journal
			this.cJournal.splitDone(splitEv.c.id, splitEv.de.x.id, splitEv.de.y.id);
			
		} else if (ev.getType() == ClusterEvent.EventType.MERGE) {
			ClusterMergeEvent mergeEv = (ClusterMergeEvent) ev;
			this.graph.mergeReferenceClusters(mergeEv.cd, mergeEv.e);
			//disengage
			this.disengageForMerge(mergeEv.e);
			// journal
			this.cJournal.mergeDone(mergeEv.cd.x.id, mergeEv.cd.y.id, mergeEv.e.id);

		}
	}
	
	
	public Cluster getAncestor(Cluster c) {
		if (this.isEngaged(c)) {
			if (this.productOf.containsKey(c)) {
				if (this.productOf.get(c) == ClusterEvent.EventType.MERGE) {
					Pair<Cluster> pair = this.mergeAncestors.get(c);
					if (this.rand.nextDouble() < 0.5) {
						return pair.x;
					} else {
						return pair.y;
					}
				} else if (this.productOf.get(c) == ClusterEvent.EventType.SPLIT) {
					return this.splitAncestors.get(c);
				}
			} else {
				return c;
			}
		} else {
			return c;
		}
		
		return null;
	}
	
	
	public boolean isEngaged(Cluster c) {
		for (Cluster d : this.engaged) {
			if (d.equals(c)) {
				return true;
			}
		}
		return false;
	}
	
	
	public boolean hasOngoingEvent() {
		return (! this.ongoingEvents.isEmpty());
	}
	
	
}
