package de.uka.algo.generator.util;

import java.util.ArrayList;
import java.util.HashMap;

/**
 * A generic, dynamic upper triangular matrix for
 * associating a pair {x,y} of keys with a value.
 * 
 * @author <christian.staudt @ ira.uka.de>
 *
 * @param <Key>
 * @param <Entry>
 */
public class DynamicTriangularMatrix<Key, Entry> {
	
	ArrayList<ArrayList<Entry>> matrix;
	int dimension;
	HashMap<Key,Integer> forwardMap;
	HashMap<Integer,Key> reverseMap;
	Entry nullEntry;
	
	public DynamicTriangularMatrix (int dim, Entry nullEntry) {
		this.nullEntry = nullEntry;
		this.matrix = new ArrayList<ArrayList<Entry>>();
		for (int i = 0; i <= dim; ++i) {
			ArrayList<Entry> column = new ArrayList<Entry>();
			for (int j = 0; j <= i; ++j) {
				column.add(this.nullEntry);
			}
			this.matrix.add(column);
		}
		this.dimension = dim;
		this.forwardMap = new HashMap<Key,Integer>();
		this.reverseMap = new HashMap<Integer,Key>();
		for (int i = 1; i <= dim; ++i) {
			this.reverseMap.put(i, null);
		}
	}
	
	private int getDimension() {
		return this.dimension;
	}

	/**
	 * Get the entry for the key pair {x,y}.
	 * @param x
	 * @param y
	 * @return
	 */
	public Entry get(Key x, Key y) {
		int i = this.forwardMap.get(x);
		int j = this.forwardMap.get(y);
		try {
			if (i >= j) {
				return this.matrix.get(i).get(j);
			} else {
				return this.matrix.get(j).get(i);
			}
		} catch (IndexOutOfBoundsException ex) {
			// TODO Auto-generated catch block
			ex.printStackTrace();
			System.exit(1);
			return null;
		}
	}
	
	/**
	 * Set the entry for the key pair {x,y}.
	 * @param x
	 * @param y
	 * @param e
	 */
	public void set(Key x, Key y, Entry e) {
		assert this.forwardMap.containsKey(x) && this.forwardMap.containsKey(y);
		int i = this.forwardMap.get(x);
		int j = this.forwardMap.get(y);
		try {
			if (i >= j) {
				this.matrix.get(i).set(j, e);
			} else {
				this.matrix.get(j).set(i, e);
			}
		} catch (IndexOutOfBoundsException ex) {
			ex.printStackTrace();
			System.exit(1);
		}

	}
	

	
	
	/**
	 * Add a new key, resize if necessary.
	 * @param x
	 */
	public void add(Key x) {
		if (this.forwardMap.containsKey(x)) {
			System.out.println("Key already exists");
			return;
		}
		
		// find a vacant index
		for (int i : this.reverseMap.keySet()) {
			if (this.reverseMap.get(i) == null) {
				this.reverseMap.put(i, x);
				this.forwardMap.put(x, i);
				return;
			}
		}
		// no vacant key found
		this.dimension++;
		ArrayList<Entry> newColumn = new ArrayList<Entry>();
		for (int j = 0; j <= this.dimension; ++j) {
			newColumn.add(this.nullEntry);
		}
		this.matrix.add(newColumn);
		
		this.forwardMap.put(x, this.dimension);
		this.reverseMap.put(this.dimension, x);
		
	}
	
	/**
	 * Remove a key and reset its entries.
	 * @param x
	 */
	public void remove(Key x) {
		for (Key y : this.forwardMap.keySet()) {
			this.set(x, y, this.nullEntry);
		}
		int i = this.forwardMap.get(x);
		this.reverseMap.put(i, null);
		this.forwardMap.remove(x);
	}
	
	
	
	/**
	 * Reset all entries to nullEntry.
	 */
	public void resetAll() {
		for (int i = 0; i < this.matrix.size(); ++i) {
			for (int j = 0; j < this.matrix.get(i).size(); ++j) {
				this.matrix.get(i).set(j, this.nullEntry);
			}
		}
	}
	
	
	
	public ArrayList<Entry> getEntries() {
		ArrayList<Entry> entries = new ArrayList<Entry>();
		for (int i = 0; i < this.matrix.size(); ++i) {
			entries.addAll(this.matrix.get(i));
		}
		return entries;
	}
	
	
	
	


}
