package de.uka.algo.generator.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;


/**
 * A generic Union-Find data structure with additional operations, 
 * implemented as a disjoint-set forest.
 * 
 * @author Christian Staudt
 *
 * @param <Element>
 */
public class UnionFind<Element> {
    
    private ArrayList<Element> roots;
    private HashMap<Element, Boolean> isRoot;
    private HashMap<Element, Element> parent;
    private HashMap<Element, Integer> rank;
    private final boolean pathCompression;
    private final boolean unionByRank;
    
    
    /**
     * Normal constructor.
     * @param collection
     * @param pathCompression
     * @param unionByRank
     */
    public UnionFind(Collection<Element> collection, boolean pathCompression, boolean unionByRank) {
    	this.roots = new ArrayList<Element>();
    	this.parent = new HashMap<Element, Element>();
    	this.rank = new HashMap<Element, Integer>();
    	this.isRoot = new HashMap<Element, Boolean>();
        this.pathCompression = pathCompression;
        this.unionByRank = unionByRank;
        if (collection != null) {
	        for (Element e : collection) {
	            add(e);
	        }
        }

    }
    
    /**
     * Deep copy constructor.
     * @param other
     */
    public UnionFind(UnionFind<Element> other) {
    	this.roots = new ArrayList<Element>();
    	for (Element root : other.roots) {
    		this.roots.add(root);
    	}
    	this.isRoot = new HashMap<Element,Boolean>();
    	for (Element e : other.isRoot.keySet()) {
    		this.isRoot.put(e, other.isRoot.get(e));
    	}
    	this.parent = new HashMap<Element,Element>();
    	for (Element e : other.parent.keySet()) {
    		this.parent.put(e, other.parent.get(e));
    	}
    	this.unionByRank = other.unionByRank;
    	if (this.unionByRank) {
    		this.rank = new HashMap<Element,Integer>();
    		for (Element e : other.rank.keySet()) {
    			this.rank.put(e, other.rank.get(e));
    		}
    	}
    	this.pathCompression = other.pathCompression;
    }
    
    
    /**
     * Makes element a root.
     * @param e
     */
    private void makeRoot(Element e) {
        parent.put(e, e);
        roots.add(e);
        isRoot.put(e, true);
        if (unionByRank) {
            rank.put(e, 0);
        }
    }
    
    /**
     * Finds the parent for a given element.
     * @param e
     * @return
     */
    public Element find(Element e) throws ElementNotFoundException {
        if (! parent.containsKey(e)) throw new ElementNotFoundException();
        
        if (parent.get(e) == e) {
            return e;
        } else {
            Element e_ = find(parent.get(e));
            if (pathCompression) {
                parent.put(e, e_);
            }
            return e_;
        }
    }
    
    
    /**
     * Links two elements.
     * @param e
     * @param 
     */
    public Element link(Element e, Element f) {
        parent.put(e, f);
        if (isRoot.get(e)) {
            isRoot.put(e, false);
            roots.remove(e);
        }
        return f;
    }
    
    
    /**
     * Unite the set containing e with the 
     * set containing f.
     * @param e
     * @param f
     */
    public Element union(Element e, Element f) {
        try {
            if (find(e) != find(f)) {
                if (unionByRank) {
                    int i = rank.get(e);
                    int j = rank.get(f);
                    if (i < j) {
                        return link(find(e),find(f));
                    } else {
                        if (i == j) {
                            rank.put(e, i + 1);
                        }
                        return link(find(f), find(e));
                    }
                } else {
                    return link(find(e), find(f));
                }
            } else {
            	return null;
            } 
        } catch (ElementNotFoundException e1) {
            e1.printStackTrace();
            return null;
        }
    }
    
    /**
     * Cut the subtree rooted in the given element
     * out and make it a new tree.
     * @param e
     */
    private void cut(Element e) {
        assert contains(e);
        // TODO: how does this affect the rank?
        makeRoot(e);
    }

    
    /**
     * Return the set of roots.
     * @return
     */
    public ArrayList<Element> getRepresentatives() {
        return roots;
    }
    
   
    
    /**
     * Check whether element has been added.
     * @param e
     * @return
     */
    public boolean contains(Element e) {
        return parent.containsKey(e);
    }
    
    
    public boolean isRepresentative(Element e) {
        assert contains(e);
        return isRoot.get(e);
    }
    
    /**
     * Add an element as a new root.
     * @param e
     */
    public void add(Element e) {
        assert (! contains(e));
        makeRoot(e);
    }
    
    /**
     * Add an element and link it to another one.
     * @param e
     * @param f
     */
    public void addAndLink(Element e, Element f) {
        assert (contains(f));
        add(e);
        link(e,f);
    }
    
    
    /**
     * Remove an element and handle its children
     * according to "adoption".
     * @param e
     * @param adoption
     */
    public void remove(Element e, boolean adoption) {
        assert contains(e);
        if (adoption) {
            Element a = getChildOf(e);
            if (!(a == null)) {
                // let a adopt the children of e
                adopt(e, a);
                // if e was a root, a becomes a root
                if (isRepresentative(e)) {
                	makeRoot(a);
                } else {
                	try {
                    	union(a, find(e));
                	} catch (ElementNotFoundException ex) {
                		ex.printStackTrace();
                	}
                }
                delete(e);
            } else {
                // in case e has no children, simply delete it
                delete(e);
            }
        } else {
            // cut out children of e
            explode(e);
            delete(e);
        }
    }
    
    /**
     * Delete element from the data containers.
     * @param e
     */
    private void delete(Element e) {
        if (isRoot.get(e)) {
            roots.remove(e);
        }
        isRoot.remove(e);
        parent.remove(e);
        rank.remove(e);
    }
    
    
    /**
     * Return an arbitrary child of the
     * given element.
     * @return
     */
    public Element getChildOf(Element p) {
        for (Element e : parent.keySet()) {
            if (parent.get(e) == p) {
                return e;
            }
        }
        return null;
    }
    
    
    /**
     * Cut out all the subtrees rooted in
     * the element's children and make them new trees.
     * @param e
     */
    private void explode(Element e) {
        for (Element x : parent.keySet()) {
            if (parent.get(x) == e) {
                cut(x);
            } 
        }
    }
    
    
    /**
     * Let element a "adopt" all children of element p,
     * i.e. a becomes their new parent.
     * @param p
     * @param a
     */
    private void adopt(Element p, Element a) {
        for (Element x : parent.keySet()) {
            if (parent.get(x) == p) {
                link(x, a);
            }
        }
    }
    
    
    /**
     * Union of every two roots.
     */
    public void pairing() {
        ArrayList<Element> oldRoots = new ArrayList<Element>();
        oldRoots.addAll(roots);
        for (int i = 0; i < oldRoots.size(); i += 2) {
            union(oldRoots.get(i), oldRoots.get(i+1));
        }
    }
    
    
    /**
     * Print the data structure.
     */
    public void print() {
    	System.out.print("roots: ");
    	for (Element e : this.getRepresentatives()) {
    		System.out.print(e.toString() + ", ");
    	}
    	System.out.println();
    	System.out.println("groups: ");
    	for (Element e : parent.keySet()) {
        	System.out.print(e.toString() + " ");
        	System.out.print("=> ");
        	try {
        		Element p = this.find(e);
        		System.out.print(p.toString() + ", ");
        	} catch (ElementNotFoundException ex) {
        		System.out.print("? ");
        	}
        	
    	}
    	System.out.println();
    	System.out.println("links: ");
    	for (Element e : parent.keySet()) {
        	System.out.print(e.toString() + " ");
        	System.out.print("-> ");
    		Element p = parent.get(e);
    		System.out.print(p.toString() + ", ");
        	
    	}
    }
    
    
    public String toOrderedPairs() {
    	String s = "{";
    	for (Element e : this.parent.keySet()) {
    		s = s + "{" + e.toString() + "," + this.parent.get(e).toString() +  "},";
    	}
    	s = s + "}";
    	s = s.replace("},}", "}}");
    	return s;
    	
    }
    
    
    
}
