/* 
 * Integration: May 2, 2006 (katz)
 * Reviewed: May 2, 2006 (katz)
 *
 * Copyright 2004-2006.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering;

import y.base.ListCell;
import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeList;
import de.uka.algo.util.datastructures.parasitic.SlotObject;

/**
 * A cluster is a non-empty set of nodes belonging to the same graph. Clusters
 * aggregate to clusterings, which ensure properties like disjointness,
 * synchronize changes and manage access to data (clustermaps) stored 'besides'
 * the clusters. Therefore, neither constructors nor any of the modifying
 * methods are public. They may only be called by their respective clustering.
 * 
 * @see de.uka.algo.clustering.Clustering
 * @see de.uka.algo.clustering.ClusterMap
 * 
 * @author Bastian Katz, mailto:katz@ira.uka.de
 */
public final class Cluster extends SlotObject {

	/* ===== fields ===== */
	/* error messages */
	private static final String DIFFERENT_GRAPH_ERROR = "node belongs to different graph.";

	/* the list of contained nodes */
	private NodeList memberList = null;

	/* the clusters id, volatile */
	int id = 0;

	/* ===== constructor ===== */
	/**
	 * Instantiates a new (empty) cluster to a given clustering in O(number of
	 * slots for ClusterMaps). May only be called by the clustering to contain
	 * the new cluster.
	 * 
	 * @param parent
	 *            the calling clustering
	 */
	protected Cluster(Clustering parent) {
		super(parent);
		this.memberList = new NodeList();
	}

	/* ===== getters and queries ===== */

	public boolean hasLabel() {
		if (getLabel()==null) return false;
		return true;
	}
	
	public String getLabel() {
		if (!getClustering().hasLabels()) return null;
		return getClustering().getClusterLabels().get(this);
	}
	
	public void setLabel(String name) {
		getClustering().getClusterLabels().set(this, name);
	}
	/**
	 * Returns the current cluster's index. Note: The index may change if the
	 * set of clusters changes in any way. Takes O(number of clusters) if
	 * clusters need to be renumbered, O(1) otherwise.
	 * 
	 * @return the current cluster's index
	 */
	public int getIndex() {
		getClustering().assureNumbering();
		return id;
	}

	/**
	 * Returns the cluster's cardinality in O(1).
	 * 
	 * @return the number of nodes
	 */
	public int size() {
		return memberList.size();
	}

	/**
	 * Returns a cursor over all nodes of a cluster in O(1) (?)
	 * 
	 * @return a cursor over all contained nodes
	 */
	public NodeCursor members() {
		return memberList.nodes();
	}

	/**
	 * Returns a member of the cluster in O(1)
	 * 
	 * @return a node contained in the cluster
	 */
	public Node aMember() {
		return memberList.firstNode();
	}

	/**
	 * Returns the clustering a cluster belongs to in O(1).
	 * 
	 * @return the cluster's clustering
	 */
	public Clustering getClustering() {
		return (Clustering) parent;
	}
	
	public String toString() {
		if (!hasLabel()) {
			String delim = "";
			String result = "Cluster #:"+getIndex()+", "+size()+" node(s):[";
			for (NodeCursor nc = members();nc.ok(); nc.next()) {
				result += delim+nc.node().index();
				delim = ", ";
			}
			return result+"]";
		} else {
			return getLabel();
		}			
	}
	
	/**
	 * Provides access to the cluster's list of members. May only be called for
	 * cluster deletion by its clustering.
	 * 
	 * @see Clustering
	 * 
	 * @return the cluster's member list
	 */
	protected NodeList getMemberList() {
		return memberList;
	}
	
	/* ===== internal parts (add/remove nodes) ===== */

	/**
	 * Adds a node to a cluster in O(1). May only be called by the clustering
	 * containing the cluster.
	 * 
	 * @param aNode
	 *            the node to add
	 * @return the cell containing the new node in the cluster's member list
	 */
	protected ListCell add(Node aNode) {
		if (aNode.getGraph() != getClustering().getGraph()) {
			throw new RuntimeException(DIFFERENT_GRAPH_ERROR);
		}
		return memberList.addLast(aNode);
	}

	/**
	 * Removes a node from a cluster in O(1). May only be called by the
	 * clustering containing the cluster.
	 * 
	 * @param aMemberCell
	 */
	protected void remove(ListCell aMemberCell) {
		memberList.removeCell(aMemberCell);
	}
}