package de.uka.algo.clustering;

import y.base.Graph;
import de.uka.algo.util.GYCursor;

/**
 * The ClusterSizeListener supports a clustering with basic functions such as
 * entropy. Unlike other listeners, it does not register itself a the
 * clustering, but is instantiated and registered by the clustering itself in a
 * list of privileged listeners.  
 * 
 * @author Bastian Katz, katz@kit.edu
 * 
 */
final class ClusterSizeListener extends DefaultListener {

	private boolean valid = false;
	private int failureEdgeCount = 0;
	private double entropy = 0.0;
	private double edgeEntropy = 0.0;

	ClusterSizeListener(Clustering clustering) {
		super(clustering);
		// the cluster size listener is added to list of privileged listeners
	}

	private void validate() {
		if (valid)
			return;
		Graph graph = clustering.getGraph();
		failureEdgeCount = 0;
		entropy = 0.0;
		edgeEntropy = 0.0;
		for (GYCursor<Cluster> cc = clustering.clusters(); cc.ok(); cc.next()) {
			Cluster c = cc.active();
			double p = (double) (c.size()) / (double) (graph.N());
			double pe = (double) (clustering.getBorderEdgeCount(c) + 2 * clustering
					.getInnerEdgeCount(c))
					/ (double) (2 * graph.E());
			entropy -= p * Math.log(p) / Math.log(2);
			edgeEntropy -= pe * Math.log(pe) / Math.log(2);
			failureEdgeCount += c.size() * (c.size() - 1);
		}
		failureEdgeCount /= 2;
		failureEdgeCount += graph.edgeCount() - 2
				* clustering.getInnerEdgeCount();
		valid = true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.clustering.DefaultListener#defaultEvent()
	 */
	@Override
	protected void defaultEvent() {
		valid = false;
	}

	@Override
	protected String id() {
		return "ClusterSizeListener";
	}

	@Override
	protected void policyChanged(boolean policy) {
		// nothing to do here
	}

	double getEdgeEntropy() {
		validate();
		return edgeEntropy;
	}

	double getEntropy() {
		validate();
		return entropy;
	}

	int getFailureEdgeCount() {
		validate();
		return failureEdgeCount;
	}
}
