/* 
 * Integration: May 2, 2006 (katz)
 * Reviewed: May 3, 2006 (katz)
 *
 * Copyright 2004-2006.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering;

import y.base.DataProvider;
import y.base.Edge;
import y.base.EdgeMap;
import y.base.EdgeCursor;
import y.base.Graph;
import y.base.GraphEvent;
import y.base.GraphListener;
import y.base.ListCell;
import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeList;
import y.base.NodeMap;
import y.view.Graph2D;
import de.uka.algo.clustering.io.ClusteringIOHandler;
import de.uka.algo.graphs.GraphInterpretation;
import de.uka.algo.graphs.GraphInterpretationListener;
import de.uka.algo.graphs.SimpleGraph;
import de.uka.algo.graphs.WeightedGraph;
import de.uka.algo.util.GYCursor;
import de.uka.algo.util.GYList;
import de.uka.algo.util.datastructures.parasitic.SlotDisposer;

/**
 * A clustering is a partition of a graph (with an interpretation), i.e. a set
 * of nonempty clusters plus a list of nodes that are currently not assigned to
 * any cluster. To avoid inconsistency when working with changing graphs, a
 * clustering listens to the graph and its interpretation to notice node/edge
 * deletion/insertion as well as changing weights. Moreover, it uses
 * NodeMaps/EdgeMaps also registered at the graph. If a clustering is no longer
 * in use while the graph still is, <code>dispose()</code> should be called to
 * make the clustering quit listening and free its maps. Note: As long as the
 * graph is in use, the garbage collection cannot wipe out the clustering unless
 * you called <code>dispose()</code>. A Clustering also implements a
 * DataProvider for the graph's nodes, mapping nodes from the same cluster to
 * its cluster/it's clusters ID.
 * 
 * @see y.base.Graph
 * @see de.uka.algo.graphs.GraphInterpretation
 * @see de.uka.algo.clustering.Cluster
 * 
 * @author Bastian Katz, @ bastian.katz@kit.edu
 */
public class Clustering extends SlotDisposer<Cluster> implements GraphListener,
		GraphInterpretationListener, Cloneable, DataProvider {

	/* ===== fields ===== */
	/* the underlying graph */
	private Graph graph = null;

	/* the underlying graph's interpretation */
	private GraphInterpretation interpretation = null;

	/* a list of nodes that are currently not assigned */
	private NodeList pool = null;

	/* the mapping from a node to its cluster */
	private NodeMap clusters = null;

	/* the mapping from a node to its cell in a member list */
	private NodeMap memberCells = null;

	/* the list of listeners */
	private ClusteringListenerList listenerList = null;

	/* the list of privileged listeners */
	private ClusteringListenerList internalListeners = null;

	/* the current listening policy (see below). */
	private boolean policy = ClusteringListener.CONSTANTHANDLING;

	/* the ClusterArray, constructed on demand, reset to null if void */
	private Cluster[] clusterArray = null;

	/* the listener for the most basic index functions */
	private ClusterSizeListener clusterSizeListener = null;

	/*
	 * remember, if the clustering was disposed, to handle comparisons
	 * respectively
	 */
	private boolean disposed = false;

	private ClusterMap<String> clusterNames = null;

	/* ===== constructors ===== */

	/**
	 * Standard constructor. Instantiates a new clustering (no nodes assigned to
	 * clusters) for a given graph with an interpretation.
	 * 
	 * @param aGraph
	 *            the graph to cluster
	 * @param anInterpretation
	 *            an interpretation of the given graph
	 */
	public Clustering(Graph aGraph, GraphInterpretation anInterpretation) {
		/* registering at graph and listener */
		this.graph = aGraph;
		aGraph.addGraphListener(this);
		this.interpretation = anInterpretation;
		anInterpretation.addListener(this);

		/* set up listener lists */
		listenerList = new ClusteringListenerList(this);
		internalListeners = new ClusteringListenerList(this);
		listenerList.addClusteringListener(internalListeners); // internals

		/* add the private cluster size listener to internal listeners */
		clusterSizeListener = new ClusterSizeListener(this);
		internalListeners.addClusteringListener(clusterSizeListener);

		/* set up node- and edgemaps */
		clusters = getGraph().createNodeMap(); // disposed by dispose()
		memberCells = getGraph().createNodeMap(); // disposed by dispose()

		/* setting up the pool of not assigned nodes */
		pool = new NodeList();
		for (NodeCursor nc = getGraph().nodes(); nc.ok(); nc.next()) {
			Node aNode = nc.node();
			ListCell aMemberCell = pool.addLast(aNode);
			memberCells.set(aNode, aMemberCell);
		}
	}

	public Clustering(Graph2D graph, boolean extractWeightsFromLabel) {
		this(graph, GraphInterpretation.getInterpretation(graph,
				extractWeightsFromLabel));
	}

	@Deprecated
	public Clustering(Graph2D graph) {
		this(graph, GraphInterpretation.getInterpretation(graph, true));
	}

	/**
	 * Convenience call for standard constructor. Creates an interpretation that
	 * is undirected and unweighted.
	 * 
	 * @param aGraph
	 *            the graph to cluster
	 */
	public Clustering(Graph aGraph) {
		this(aGraph, new SimpleGraph(aGraph));
	}

	/**
	 * Convenience call for standard constructor. Creates an interpretation that
	 * is undirected and weighted.
	 * 
	 * @param aGraph
	 *            the graph to cluster
	 * @param weights
	 *            edge weights as EdgeMap
	 */
	public Clustering(Graph aGraph, EdgeMap weights) {
		this(aGraph, new WeightedGraph(aGraph, weights));
	}
	
	// TODO IO-constructor.

	/* ===== getters ===== */

	/**
	 * Returns the current policy for dependent ClusteringListeners.
	 * 
	 * @see de.uka.algo.clustering.ClusteringListener
	 * 
	 * @return the current policy for dependent ClusteringListeners.
	 */
	public boolean getPolicy() {
		return policy;
	}

	/**
	 * Returns the underlying Graph.
	 * 
	 * @return the underlying Graph
	 */
	public Graph getGraph() {
		return graph;
	}

	/**
	 * Returns the interpretation of the underlying graph.
	 * 
	 * @see GraphInterpretation
	 * 
	 * @return the interpretation of the underlying graph
	 */
	public GraphInterpretation getInterpretation() {
		return interpretation;
	}

	/**
	 * Returns the number of clusters.
	 * 
	 * @return the number of clusters
	 */
	public int clusterCount() {
		return objectCount();
	}

	/**
	 * Returns the number of clusters.
	 * 
	 * @return the number of clusters
	 */
	public int size() {
		return objectCount();
	}

	/**
	 * Returns the number of nodes not assigned to any cluster.
	 * 
	 * @return the number of nodes not assigned to any cluster
	 */
	public int poolSize() {
		return pool.size();
	}

	/**
	 * Returns a ClusterCursor iterating over all clusters.
	 * 
	 * @return a ClusterCursor iterating over all clusters
	 */
	public GYCursor<Cluster> clusters() {
		return objectList.elements();
	}

	/**
	 * Returns a NodeCursor iterating over all nodes which are not assigned to
	 * any cluster.
	 * 
	 * @return a NodeCursor iterating over all nodes which are not assigned to
	 *         any cluster
	 */
	public NodeCursor poolNodes() {
		return pool.nodes();
	}

	/**
	 * Returns the cluster containing the given node (null if none) in O(1).
	 * 
	 * @param aNode
	 *            the node to look for
	 * @return the cluster containing the node
	 */
	public Cluster getCluster(Node aNode) {
		return (Cluster) clusters.get(aNode);
	}

	/**
	 * Returns the status of a given edge as intra-cluster edge. Nodes that are
	 * not assigned to any cluster count as singletons, i.e. their incident
	 * edges are inter-cluster edges.
	 * 
	 * @param anEdge
	 *            given edge
	 * @return <code>true</code>, if a given edge is an intra-cluster edge;
	 *         <code>false</code> otherwise
	 */
	public boolean isIntraClusterEdge(Edge anEdge) {
		return ((clusters.get(anEdge.source()) == clusters.get(anEdge.target())) && (clusters
				.get(anEdge.source()) != null));
	}

	/**
	 * Returns the status of a given edge as inter-cluster edge.
	 * 
	 * @param anEdge
	 *            given edge
	 * @return <code>true</code>, if a given edge is an inter-cluster edge;
	 *         <code>false</code> otherwise
	 */
	public boolean isInterClusterEdge(Edge anEdge) {
		return !isIntraClusterEdge(anEdge);
	}

	/**
	 * Returns the status of two given nodes with respect to the equivalence
	 * relation induced by the clustering.
	 * 
	 * @param aNode
	 *            given node (non-null)
	 * @param otherNode
	 *            other given node (non-null)
	 * @return <code>true</code>, if both nodes belong to the same cluster;
	 *         <code>false</code> otherwise
	 */
	public boolean belongToSameCluster(Node aNode, Node otherNode) {
		return ((clusters.get(aNode) == clusters.get(otherNode)) && (clusters
				.get(aNode) != null));
	}

	public double getEdgeEntropy() {
		return clusterSizeListener.getEdgeEntropy();
	}

	public double getEntropy() {
		return clusterSizeListener.getEntropy();
	}

	public int getFailureEdgeCount() {
		return clusterSizeListener.getFailureEdgeCount();
	}

	/* helper methods */

	private void put(Node aNode, Cluster target) {
		ListCell newCell = target.add(aNode);
		memberCells.set(aNode, newCell);
		clusters.set(aNode, target);
	}

	private void putPool(Node aNode) {
		ListCell newMemberCell = pool.addLast(aNode);
		memberCells.set(aNode, newMemberCell);
		clusters.set(aNode, null);
	}

	private void get(Node aNode, Cluster source) {
		source.remove(((ListCell) memberCells.get(aNode)));
	}

	private void getPool(Node aNode) {
		pool.removeCell(((ListCell) memberCells.get(aNode)));
	}

	/* ===== operations ===== */

	/**
	 * Sets the listeners' policy.
	 * 
	 * @see de.uka.algo.clustering.ClusteringListener
	 * 
	 * @param newPolicy
	 */
	public void setPolicy(boolean newPolicy) {
		if (newPolicy == getPolicy())
			return;
		this.policy = newPolicy;

		/* notify all listeners */
		listenerList.policyChanged(newPolicy);
	}

	/**
	 * Adds nodes residing in the pool to a cluster.
	 * 
	 * @param aCluster
	 *            the cluster to add nodes to
	 * @param nodesToAdd
	 *            a cursor iterating over the nodes to be added
	 */
	public void add(Cluster aCluster, NodeCursor nodesToAdd) {
		if (aCluster.getClustering() != this) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_ADDING_TO_CLUSTER,
					ClusteringOperationException.ER_EMPTY_NODELIST, aCluster);
		}

		/* notify all listeners */
		listenerList.addEvent(ClusteringListener.PRE, aCluster, nodesToAdd);

		/* adding nodes */
		nodesToAdd.toFirst();
		for (; nodesToAdd.ok(); nodesToAdd.next()) {
			Node nodeToAdd = nodesToAdd.node();
			if (getCluster(nodeToAdd) != null) {
				if (getCluster(nodeToAdd) != aCluster) {
					GYList<Cluster> involvedClusters = new GYList<Cluster>();
					involvedClusters.add(aCluster);
					involvedClusters.add(getCluster(nodeToAdd));
					throw new ClusteringOperationException(
							ClusteringOperationException.OP_ADDING_TO_CLUSTER,
							ClusteringOperationException.ER_NODE_ALREADY_ASSIGNED,
							nodeToAdd, involvedClusters.elements());
				} else {
					int countInList = 0;
					nodesToAdd.toFirst();
					for (; nodesToAdd.ok(); nodesToAdd.next()) {
						if (nodesToAdd.node() == nodeToAdd) {
							countInList++;
						}
					}
					if (countInList > 1) {
						throw new ClusteringOperationException(
								ClusteringOperationException.OP_ADDING_TO_CLUSTER,
								ClusteringOperationException.ER_NODE_TWICE_IN_LIST,
								nodeToAdd, aCluster);
					} else {
						throw new ClusteringOperationException(
								ClusteringOperationException.OP_ADDING_TO_CLUSTER,
								ClusteringOperationException.ER_NODE_ALREADY_IN_CLUSTER,
								nodeToAdd, aCluster);

					}

				}
			}
			getPool(nodeToAdd);
			put(nodeToAdd, aCluster);
		}

		/* notify all listeners */
		listenerList.addEvent(ClusteringListener.POST, aCluster, nodesToAdd);
	}

	/**
	 * Deletes a cluster, i.e. removes the cluster and releases all nodes to the
	 * pool. The cluster must belong to the clustering.
	 * 
	 * @param aCluster
	 *            the cluster to delete
	 */
	public void delete(Cluster aCluster) {
		/* checking preconditions: does cluster belong to clustering? */
		if (aCluster.getClustering() != this) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_CLUSTER_DELETION,
					ClusteringOperationException.ER_CLUSTER_NOT_PART_OF_CLUSTERING,
					aCluster);
		}

		/* notify all listeners before performing operation */
		listenerList.deleteEvent(ClusteringListener.PRE, aCluster);

		/* shifting nodes to pool */
		NodeList aMemberList = aCluster.getMemberList();
		while (!aMemberList.isEmpty()) {
			Node aNode = aMemberList.popNode();
			putPool(aNode);
		}

		/* remove (empty) cluster */
		remove(aCluster);

		/* invalidate cluster numbering */
		clusterArray = null;

		/* notify all listeners after successful operation */
		listenerList.deleteEvent(ClusteringListener.POST, null);
	}

	/**
	 * Extracts nodes from a Cluster.
	 * 
	 * @param extractFrom
	 *            cluster to extract nodes from
	 * @param nodesToExtract
	 *            a cursor iterating over all nodes to extract. Note: If the
	 *            cursor spans all the clusters' nodes, it will be deleted
	 *            instead.
	 */
	public void extract(Cluster extractFrom, NodeCursor nodesToExtract) {
		/* checking preconditions: does clustering belong to clustering? */
		if (extractFrom.getClustering() != this) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_EXTRACT_FROM_CLUSTER,
					ClusteringOperationException.ER_CLUSTER_NOT_PART_OF_CLUSTERING,
					extractFrom);
		}

		/* checking for real extraction or implicit deletion */
		boolean deletion = (extractFrom.size() == nodesToExtract.size());

		/*
		 * notify all listeners before performing operation, if it's a real
		 * extraction
		 */
		if (!deletion)
			listenerList.extractEvent(ClusteringListener.PRE, extractFrom,
					nodesToExtract);

		for (nodesToExtract.toFirst(); nodesToExtract.ok(); nodesToExtract
				.next()) {
			Node aNode = nodesToExtract.node();
			if (clusters.get(aNode) != extractFrom) {
				if (hasDoubleOccurrence(nodesToExtract, aNode)) {
					throw new ClusteringOperationException(
							ClusteringOperationException.OP_EXTRACT_FROM_CLUSTER,
							ClusteringOperationException.ER_NODE_TWICE_IN_LIST,
							aNode, extractFrom);
				} else {
					throw new ClusteringOperationException(
							ClusteringOperationException.OP_EXTRACT_FROM_CLUSTER,
							ClusteringOperationException.ER_NODE_NOT_IN_CLUSTER,
							aNode, extractFrom);
				}
			}
			/*
			 * if the last node within the cluster is about to be extracted,
			 * delete it
			 */
			if (extractFrom.size() == 1) {
				delete(extractFrom);
			} else {
				get(aNode, extractFrom);
				putPool(aNode);
			}
		}

		/*
		 * notify all listeners after successful operation, in case of real
		 * extraction
		 */
		if (!deletion)
			listenerList.extractEvent(ClusteringListener.POST, extractFrom,
					nodesToExtract);
	}

	/**
	 * Merges two clusters. Does nothing if the two clusters are identical.
	 * 
	 * @param aCluster
	 *            the first cluster to merge
	 * @param anotherCluster
	 *            the second cluster to merge
	 * @return the resulting cluster (re-using one that's larger beforehand, the
	 *         first on ties)
	 */
	public Cluster merge(Cluster aCluster, Cluster anotherCluster) {
		if (aCluster.getClustering() != this) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_MERGE_TWO_CLUSTERS,
					ClusteringOperationException.ER_CLUSTER_NOT_PART_OF_CLUSTERING,
					aCluster);
		}
		if (anotherCluster.getClustering() != this) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_MERGE_TWO_CLUSTERS,
					ClusteringOperationException.ER_CLUSTER_NOT_PART_OF_CLUSTERING,
					anotherCluster);
		}
		if (aCluster == anotherCluster) {
			return aCluster;
		}

		/* ensure that aCluster isn't smaller than anotherCluster */
		if (aCluster.size() < anotherCluster.size()) {
			Cluster cb = aCluster;
			aCluster = anotherCluster;
			anotherCluster = cb;
		}

		/* notifying all listeners in advance */
		listenerList.mergeEvent(ClusteringListener.PRE, aCluster,
				anotherCluster);

		/* merging clusters */
		NodeList aMemberList = anotherCluster.getMemberList();
		while (!aMemberList.isEmpty()) {
			Node aNode = aMemberList.popNode();
			put(aNode, aCluster);
		}
		/* removing cluster from */
		remove(anotherCluster);

		/* invalidate cluster numbering */
		clusterArray = null;

		/* notifying all listeners in advance */
		listenerList.mergeEvent(ClusteringListener.POST, aCluster, null);

		return aCluster;
	}

	/**
	 * Merges the clusters of two given nodes.
	 * 
	 * @param aNode
	 *            a node contained in the first cluster to merge
	 * @param anotherNode
	 *            a node contained in the second cluster to merge
	 * @return the merged cluster
	 */
	public Cluster merge(Node aNode, Node anotherNode) {
		return merge(getCluster(aNode), getCluster(anotherNode));
	}

	/**
	 * Merges all clusters which have at least one node in the given list. The
	 * list must be non-empty. All contained nodes must be assigned to a
	 * cluster, but not necessarily to different clusters. If all nodes belong
	 * to the same cluster, the clustering remains unchanged.
	 * 
	 * @param nodesToMerge
	 *            the nodes which clusters are to be merged (may be changed).
	 * 
	 * @return resulting cluster.
	 */
	public Cluster merge(NodeCursor nodesToMerge) {
		/* checking precondition: list must be non-empty */
		if (nodesToMerge.size() == 0) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_MERGE_MANY_CLUSTERS,
					ClusteringOperationException.ER_EMPTY_NODELIST);
		}

		/* reset node list */
		nodesToMerge.toFirst();

		/* preliminary result, correct ii only one cluster affected */
		Cluster result = getCluster(nodesToMerge.node());

		/* make sure that first node is assigned to a cluster */
		if (result == null) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_MERGE_MANY_CLUSTERS,
					ClusteringOperationException.ER_NODE_NOT_IN_ANY_CLUSTER,
					nodesToMerge.node());
		}

		/* traverse list of nodes */
		if (nodesToMerge.size() > 1) {
			for (nodesToMerge.next(); nodesToMerge.ok(); nodesToMerge.next()) {
				Cluster clusterOfCurrentNode = getCluster(nodesToMerge.node());
				/* make sure that current node is assigned to a cluster */
				if (result == null) {
					throw new ClusteringOperationException(
							ClusteringOperationException.OP_MERGE_MANY_CLUSTERS,
							ClusteringOperationException.ER_NODE_NOT_IN_ANY_CLUSTER,
							nodesToMerge.node());
				}

				/* merge if clusters are not equal, set result */
				if (result != clusterOfCurrentNode) {
					result = merge(result, clusterOfCurrentNode);
				}
			}
		}

		/* return resulting cluster */
		return result;
	}

	/**
	 * Moves a node from a cluster to another cluster.
	 * 
	 * @param from
	 *            the cluster to take nodes from
	 * @param to
	 *            the cluster to move nodes to
	 * @param node
	 *            the node to move
	 */
	public void move(Cluster from, Cluster to, Node node) {
		NodeList nl = new NodeList();
		nl.add(node);
		this.move(from, to, nl.nodes());
	}

	/**
	 * Moves nodes from a cluster to another cluster.
	 * 
	 * @param from
	 *            the cluster to take nodes from
	 * @param to
	 *            the cluster to move nodes to
	 * @param nodesToMove
	 *            a cursor iteration over the nodes to move. Note: the cursor
	 *            must not contain all nodes residing in <code>to</code>.
	 */
	public void move(Cluster from, Cluster to, NodeCursor nodesToMove) {
		if (from.getClustering() != this) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_MOVE_NODES,
					ClusteringOperationException.ER_CLUSTER_NOT_PART_OF_CLUSTERING,
					from);
		}
		if (to.getClustering() != this) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_MOVE_NODES,
					ClusteringOperationException.ER_CLUSTER_NOT_PART_OF_CLUSTERING,
					to);
		}
		if (nodesToMove.size() == from.size()) {
			GYList<Cluster> operands = new GYList<Cluster>();
			operands.add(from);
			operands.add(to);
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_MOVE_NODES,
					ClusteringOperationException.ER_FULL_NODELIST, operands
							.elements());
		}
		if (from == to) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_MOVE_NODES,
					ClusteringOperationException.ER_IDENTICAL_CLUSTERS, from);
		}

		/* notify all listeners */
		listenerList.moveEvent(ClusteringListener.PRE, from, to, nodesToMove);

		/* move nodes */
		nodesToMove.toFirst();
		for (; nodesToMove.ok(); nodesToMove.next()) {
			Node nodeToMove = nodesToMove.node();
			if (clusters.get(nodeToMove) != from) {
				throw new ClusteringOperationException(
						ClusteringOperationException.OP_MOVE_NODES,
						ClusteringOperationException.ER_NODE_NOT_IN_CLUSTER,
						nodeToMove, from);
			}
			get(nodeToMove, from);
			put(nodeToMove, to);
		}

		/* notify all listeners */
		listenerList.moveEvent(ClusteringListener.POST, from, to, nodesToMove);
	}

	public void add(Cluster c, Node node) {
		NodeList nl = new NodeList();
		nl.add(node);
		add(c, nl.nodes());
	}

	public void extract(Cluster c, Node n) {
		NodeList nl = new NodeList();
		nl.add(n);
		extract(c, nl.nodes());
	}

	/**
	 * Resets the clustering to the state after instantiation, i.e. without any
	 * clusters.
	 */
	public void reset() {

		/* notify all listeners */
		listenerList.resetEvent(ClusteringListener.PRE);

		/* reset all fields */
		newObjectList();
		pool = new NodeList();
		for (NodeCursor nc = getGraph().nodes(); nc.ok(); nc.next()) {
			Node aNode = nc.node();
			putPool(aNode);
		}
		clusterArray = null;

		/* notify all listeners */
		listenerList.resetEvent(ClusteringListener.POST);
	}

	/**
	 * Resets the clustering to a state where all nodes are contained in one
	 * large cluster. Note that this is only possible, if the graph contains at
	 * least one node.
	 */
	public Cluster resetLargeCluster() {
		/* check precondition: graph must be non-empty */
		if (getGraph().nodeCount() == 0) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_RESET_LARGE_CLUSTER,
					ClusteringOperationException.ER_GRAPH_EMPTY);
		}

		/* notify all listeners */
		listenerList.resetLargeClusterEvent(ClusteringListener.PRE, null);

		newObjectList();
		Cluster bigCluster = new Cluster(this);
		pool = new NodeList();

		/* put nodes into big cluster */
		for (NodeCursor nc = getGraph().nodes(); nc.ok(); nc.next()) {
			Node aNode = nc.node();
			put(aNode, bigCluster);
		}
		clusterArray = null;

		/* notify all listeners */
		listenerList
				.resetLargeClusterEvent(ClusteringListener.POST, bigCluster);
		return bigCluster;
	}

	/**
	 * Resets the clustering to state where all nodes are contained in an own
	 * cluster.
	 */
	public void resetSingletons() {

		/* notify all listeners */
		listenerList.resetSingletonsEvent(ClusteringListener.PRE);

		/* reset lists */
		newObjectList();
		pool = new NodeList();

		/* create clusters */
		for (NodeCursor nc = getGraph().nodes(); nc.ok(); nc.next()) {
			Node aNode = nc.node();
			Cluster aCluster = new Cluster(this);
			put(aNode, aCluster);
		}
		clusterArray = null;

		/* notify all listeners */
		listenerList.resetSingletonsEvent(ClusteringListener.POST);
	}

	/**
	 * Resets the clustering by a given array of cluster indizes. Nodes having
	 * the same, non-negative cluster index are clustered to the same cluster,
	 * nodes with negative cluster index are not assigned to a cluster. Empty
	 * clusters (i.e. unused cluster indizes) are allowed, but the runtime is
	 * linear in the number of nodes and the maximum cluster index.
	 * 
	 * @param clusterIndizes
	 *            array containing the cluster index for every node, indexed by
	 *            the nodes' indizes.
	 */
	public void resetToClustering(int[] clusterIndizes) {
		if (clusterIndizes.length != graph.N()) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_RESET_INTEGER_ARRAY,
					ClusteringOperationException.ER_ARRAY_SIZE_DOESNT_MATCH_NODECOUNT);
		}
		reset();
		int numberOfClusters = 0;
		for (int clusterIndex : clusterIndizes) {
			if (numberOfClusters < clusterIndex + 1) {
				numberOfClusters = clusterIndex + 1;
			}
		}
		Cluster[] clusters = new Cluster[numberOfClusters];
		for (Node n : graph.getNodeArray()) {
			int clusterIndex = clusterIndizes[n.index()];
			if (clusterIndex < 0)
				continue;
			if (clusters[clusterIndex] == null) {				
				clusters[clusterIndex] = newCluster(n);
			} else {
				add(clusters[clusterIndex], n);
			}
		}
	}

	/**
	 * Creates a new Cluster.
	 * 
	 * @param nodesToAdd
	 *            nonempty list of nodes (currently residing in the pool) to
	 *            build the new cluster.
	 * @return the new cluster
	 */
	public Cluster newCluster(NodeCursor nodesToAdd) {
		if (nodesToAdd.size() == 0) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_CLUSTER_CREATION,
					ClusteringOperationException.ER_EMPTY_NODELIST);
		}

		/* notify all listeners (post) */
		listenerList.newClusterEvent(ClusteringListener.PRE, null);

		/* create new cluster */
		nodesToAdd.toFirst();

		Cluster newCluster = new Cluster(this);
		for (; nodesToAdd.ok(); nodesToAdd.next()) {
			Node nodeToAdd = nodesToAdd.node();
			if (clusters.get(nodeToAdd) != null) {
				if (getCluster(nodeToAdd) == newCluster) {
					throw new ClusteringOperationException(
							ClusteringOperationException.OP_CLUSTER_CREATION,
							ClusteringOperationException.ER_NODE_TWICE_IN_LIST,
							nodeToAdd);
				} else {
					throw new ClusteringOperationException(
							ClusteringOperationException.OP_CLUSTER_CREATION,
							ClusteringOperationException.ER_NODE_ALREADY_ASSIGNED,
							nodeToAdd, getCluster(nodeToAdd));
				}
			}
			getPool(nodeToAdd);
			put(nodeToAdd, newCluster);
		}
		clusterArray = null;

		/* notify all listeners (post) */
		listenerList.newClusterEvent(ClusteringListener.POST, newCluster);

		return newCluster;
	}

	public void pool2Singletons() {
		while (!pool.isEmpty()) {
			newCluster(pool.firstNode());
		}
	}

	public Cluster newCluster(Node node) {
		NodeList nl = new NodeList();
		nl.add(node);
		return newCluster(nl.nodes());
	}

	/**
	 * Splits a cluster into two parts.
	 * 
	 * @param toSplit
	 *            the cluster to split
	 * @param newClustersNodes
	 *            the clusters to be contained in the new cluster
	 * @return the created cluster
	 */
	public Cluster split(Cluster toSplit, NodeCursor newClustersNodes) {
		if (toSplit.getClustering() != this) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_SPLIT_CLUSTER,
					ClusteringOperationException.ER_CLUSTER_NOT_PART_OF_CLUSTERING,
					toSplit);
		}
		if (newClustersNodes.size() == 0) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_SPLIT_CLUSTER,
					ClusteringOperationException.ER_EMPTY_NODELIST, toSplit);
		}
		if (newClustersNodes.size() == toSplit.size()) {
			throw new ClusteringOperationException(
					ClusteringOperationException.OP_SPLIT_CLUSTER,
					ClusteringOperationException.ER_FULL_NODELIST, toSplit);
		}

		/* notify all listeners */
		listenerList.splitEvent(ClusteringListener.PRE, toSplit, null);

		/* create new cluster */
		newClustersNodes.toFirst();
		Cluster newCluster = new Cluster(this);
		for (; newClustersNodes.ok(); newClustersNodes.next()) {
			Node aNode = newClustersNodes.node();
			if (clusters.get(aNode) != toSplit) {
				throw new ClusteringOperationException(
						ClusteringOperationException.OP_SPLIT_CLUSTER,
						ClusteringOperationException.ER_NODE_NOT_IN_CLUSTER,
						aNode, toSplit);
			}
			get(aNode, toSplit);
			put(aNode, newCluster);
		}
		clusterArray = null;

		/* notify all listeners */
		listenerList.splitEvent(ClusteringListener.POST, toSplit, newCluster);
		return newCluster;
	}

	/* cluster map management */

	/**
	 * Creates a ClusterMap. The returned map performs all operations in O(1).
	 * Note: Unused maps should be released by a call of
	 * <code>disposeClusterMap</code>
	 * 
	 * @see Clustering#disposeClusterMap(ClusterMap)
	 * 
	 * @return a new ClusterMap
	 * @deprecated use a direct call of <code>new ClusterMap()</code> instead
	 */
	public ClusterMap<Object> createClusterMap() {
		return new ClusterMap<Object>(this);
	}

	/**
	 * Disposes a ClusterMap in <i>O(1) </i>. Note: Disposing a map must be the
	 * last operation involving the map.
	 * 
	 * @param mapToDispose
	 *            unused map
	 * @deprecated use a direct call of <code>ClusterMap.dispose()</code>
	 *             instead
	 * 
	 */
	@SuppressWarnings("unchecked")
	public void disposeClusterMap(ClusterMap mapToDispose) {
		((ClusterMap) mapToDispose).dispose();
	}

	/* other */

	/**
	 * By disposing the clustering, it stops listening to the respective graph
	 * and the interpretation. This is necessary for the garbage collector to
	 * sweep the memory used by the clustering. In addition, NodeMaps used by
	 * the clustering are disposed as well as the listeners registered with the
	 * clustering itself.
	 */
	public void dispose() {
		/* mark this clustering disposed */
		this.disposed = true;

		/* make listeners quit listening */
		listenerList.disposeEvent();

		/* dispose maps used by clustering */
		graph.disposeNodeMap(clusters);
		graph.disposeNodeMap(memberCells);

		/* stop listening to graph and interpretation */
		graph.removeGraphListener(this);
		interpretation.removeListener(this);
	}

	/**
	 * Registers a new ClusteringListener.
	 * 
	 * @param aListener
	 */
	public void addListener(ClusteringListener aListener) {
		listenerList.addClusteringListener(aListener);
	}

	/**
	 * Removes a ClusteringListener.
	 * 
	 * @param aListener
	 */
	public void removeClusteringListener(ClusteringListener aListener) {
		listenerList.removeClusteringListener(aListener);
	}

	/**
	 * Returns a ClusteringListener of a given type, if such a Listener is
	 * registered. Note that the Listener can be removed (by whoever registered
	 * it) without notice.
	 * 
	 * @param aClass
	 *            a Class
	 * @return a ClusteringListener of the given type
	 */
	public ClusteringListener getClusteringListener(String id) {
		return listenerList.getClusteringListener(id);
	}

	/**
	 * Disposes all listeners. Make sure to call this method only if there is no
	 * way to use registered listeners any longer. Be careful.
	 */
	public void disposeAllListeners() {
		listenerList.disposeAllListeners();
	}

	/**
	 * Clones the clustering. All clusters are the same afterwards, and the new
	 * clustering belongs to the same graph and interpretation.
	 * 
	 * Internal data structures are created from scratch.
	 * 
	 * @return the cloned clustering.
	 */
	public Clustering clone() {
		Clustering cloned = new Clustering(this.graph, this.interpretation);
		GYCursor<Cluster> clusters = this.clusters();
		for (clusters.toFirst(); clusters.ok(); clusters.next()) {
			cloned.newCluster(clusters.active().members());
		}
		return cloned;
	}

	public void setToClustering(Clustering otherClustering) {
		if (this.graph != otherClustering.graph)
			throw new RuntimeException("Clusterings must belong to same graph.");
		this.reset();
		for (GYCursor<Cluster> cc = otherClustering.clusters(); cc.ok(); cc
				.next())
			this.newCluster(cc.active().members());
	}

	/**
	 * Clones the clustering. All clusters are the same afterwards. Be careful.
	 * if the Graph is cloned as well, the new graph is simple!
	 * 
	 * @param cloneGraph
	 *            the graph belonging to the clustering is copied if true
	 * 
	 * @return the cloned clustering
	 */

	public Clustering clone(boolean cloneGraph) {
		if (!cloneGraph)
			return clone();
		Graph g = new Graph();

		// copy the nodes
		Node[] gNodes = new Node[graph.nodeCount()];
		for (NodeCursor nc = graph.nodes(); nc.ok(); nc.next()) {
			gNodes[nc.node().index()] = g.createNode();
		}

		// copy the edges
		for (EdgeCursor ec = graph.edges(); ec.ok(); ec.next()) {
			g.createEdge(gNodes[ec.edge().source().index()], gNodes[ec.edge()
					.target().index()]);
		}

		Clustering cloned = new Clustering(g);

		NodeList nl = new NodeList();
		GYCursor<Cluster> clusters = this.clusters();
		for (clusters.toFirst(); clusters.ok(); clusters.next()) {
			for (NodeCursor members = clusters.active().members(); members.ok(); members
					.next()) {
				nl.add(gNodes[members.node().index()]);
			}
			cloned.newCluster(nl.nodes());
			nl.clear();
		}
		return cloned;
	}

	/**
	 * Writes the clustering to a file. The clustering must not contain 'free'
	 * nodes, i.e. all nodes must be assigned to a cluster. This is a
	 * convenience call to
	 * <code>ClusteringIOHandler().write(clustering, filename)</code>.
	 * 
	 * @see ClusteringIOHandler
	 * 
	 * @param filename
	 *            output filename
	 */
	public void write(String filename) {
		new ClusteringIOHandler().write(this, filename);
	}

	/* ===== basic index functions ===== */

	/* internal EdgePartitionListener for basic index functions (on demand) */
	private EdgePartitionListener edgePartitionListener = null;

	/**
	 * Assures that an EdgePartitionListener is (internally) added to provide
	 * basic index functions if one of these functions is called.
	 */
	private void assureEPL() {
		if (edgePartitionListener == null) {
			edgePartitionListener = new EdgePartitionListener(this);
			internalListeners.addClusteringListener(edgePartitionListener);
		}
	}

	/**
	 * Returns the sum over all egde weights.
	 * 
	 * @return the sum of all edge weights
	 */
	public double getOverAllWeight() {
		assureEPL();
		return edgePartitionListener.getOverAllWeight();
	}

	/**
	 * Returns the number of edges having both adjacent nodes in the specified
	 * cluster.
	 * 
	 * @param cluster
	 *            the cluster
	 * @return the number of edges having both adjacent nodes in cluster
	 */
	public int getInnerEdgeCount(Cluster cluster) {
		assureEPL();
		return edgePartitionListener.getInnerEdgeCount(cluster);
	}

	/**
	 * Returns a cursor iterating over all edges having both adjacent nodes in
	 * the specified cluster.
	 * 
	 * @param cluster
	 *            the cluster
	 * @return cursor iterating over all edges having both adjacent nodes in the
	 *         specified cluster.
	 */
	public EdgeCursor innerEdges(Cluster cluster) {
		assureEPL();
		return edgePartitionListener.innerEdges(cluster);
	}

	/**
	 * Returns a cursor iterating over all edges having one (and only one)
	 * adjacent node in the specified cluster (those edges cross the cluster's
	 * border).
	 * 
	 * @param cluster
	 *            the cluster
	 * @return a cursor iterating over all edges having one (and only one)
	 *         adjacent node in cluster
	 */
	public EdgeCursor borderEdges(Cluster cluster) {
		assureEPL();
		return edgePartitionListener.borderEdges(cluster);
	}

	/**
	 * Returns the number of edges crossing a cluster's border, i.e. the number
	 * of edges having one (and only one) adjacent nodes in a specified cluster.
	 * 
	 * @param cluster
	 *            the cluster
	 * @return the number of edges crossing the cluster's border
	 */
	public int getBorderEdgeCount(Cluster cluster) {
		assureEPL();
		return edgePartitionListener.getBorderEdgeCount(cluster);
	}

	/**
	 * Returns the weight sum of all edges not having both adjacent nodes in the
	 * same cluster.
	 * 
	 * @return the weight sum of all edges not having both adjacent nodes in the
	 *         same cluster
	 */
	public double getOuterWeight() {
		assureEPL();
		return edgePartitionListener.getOuterWeight();
	}

	/**
	 * Returns the weight sum of all edges having both incident nodes in the
	 * specified cluster.
	 * 
	 * @param cluster
	 *            the cluster
	 * @return the weight sum of all edges having both incident nodes in the
	 *         specified cluster.
	 */
	public double getInnerWeight(Cluster cluster) {
		assureEPL();
		return edgePartitionListener.getInnerWeight(cluster);
	}

	/**
	 * Returns the weight sum of edges having both incident nodes in the same
	 * cluster.
	 * 
	 * @return the weight sum of edges having both incident nodes in the same
	 *         cluster
	 */
	public double getInnerWeight() {
		assureEPL();
		return edgePartitionListener.getInnerWeight();
	}

	/**
	 * Returns the weight sum of edges having one (and only one) incident node
	 * in the given cluster.
	 * 
	 * @param cluster
	 *            the cluster
	 * @return the weight sum of edges having one (and only one) incident node
	 *         in the given cluster.
	 */
	public double getBorderWeight(Cluster cluster) {
		assureEPL();
		return edgePartitionListener.getBorderWeight(cluster);
	}

	/**
	 * Returns the number of edges having both incident nodes in the same
	 * cluster.
	 * 
	 * @return the number of edges having both incident nodes in the same
	 *         cluster
	 */
	public int getInnerEdgeCount() {
		assureEPL();
		return edgePartitionListener.getInnerEdgeCount();
	}

	/**
	 * Returns the number of edges not having both incident nodes in the same
	 * cluster.
	 * 
	 * @return the number of edges not having both incident nodes in the same
	 *         cluster
	 */
	public int getOuterEdgeCount() {
		assureEPL();
		return edgePartitionListener.getOuterEdgeCount();
	}

	/* ===== enumerating Clusters ===== */

	public boolean hasLabels() {
		return clusterNames != null;
	}

	protected ClusterMap<String> getClusterLabels() {
		if (clusterNames == null) {
			clusterNames = new ClusterMap<String>(this);
		}
		return clusterNames;
	}

	/**
	 * @return the Clusters as array.
	 */
	public Cluster[] getClusterArray() {
		assureNumbering();
		return clusterArray;
	}

	/**
	 * @param index
	 *            the index of the Cluster
	 * @return the indicated Cluster
	 */
	public Cluster getCluster(int index) {
		assureNumbering();
		return clusterArray[index];
	}

	/**
	 * Assures that Clusters know their correct number.
	 */
	void assureNumbering() {
		if (clusterArray != null)
			return;
		clusterArray = new Cluster[size()];
		int index = 0;
		for (GYCursor<Cluster> yc = clusters(); yc.ok(); yc.next()) {
			Cluster current = yc.active();
			clusterArray[index] = current;
			current.id = index;
			index++;
		}
	}

	/* implementations (graph&interpretation listening) */

	/*
	 * (non-Javadoc)
	 * 
	 * @see y.base.GraphListener#onGraphEvent(y.base.GraphEvent)
	 */
	public void onGraphEvent(GraphEvent e) {
		switch (e.getType()) {

		/* high-level events are not supported */
		case GraphEvent.POST_EVENT:
			break;
		case GraphEvent.PRE_EVENT:
			break;

		/* handled events */
		case GraphEvent.EDGE_CREATION:
			listenerList.edgeCreationEvent((Edge) e.getData());
			break;
		case GraphEvent.NODE_CREATION:
			Node newNode = (Node) e.getData();
			putPool(newNode);
			listenerList.nodeCreationEvent(newNode);
			break;
		case GraphEvent.PRE_NODE_REMOVAL:
			Node nodeToRemove = (Node) e.getData();
			listenerList.nodeRemovalEvent(ClusteringListener.PRE, nodeToRemove);
			Cluster removeFrom = getCluster(nodeToRemove);
			if (removeFrom != null && removeFrom.size() == 1) {
				delete(removeFrom);
				removeFrom = null;
			}
			if (removeFrom == null) {
				getPool(nodeToRemove);
			} else {
				get(nodeToRemove, removeFrom);
			}
			break;
		case GraphEvent.POST_NODE_REMOVAL:
			nodeToRemove = (Node) e.getData();
			listenerList
					.nodeRemovalEvent(ClusteringListener.POST, nodeToRemove);
			break;
		case GraphEvent.PRE_EDGE_REMOVAL:
			listenerList.edgeRemovalEvent(ClusteringListener.PRE, (Edge) e
					.getData());
			break;
		case GraphEvent.POST_EDGE_REMOVAL:
			listenerList.edgeRemovalEvent(ClusteringListener.POST, (Edge) e
					.getData());
			break;

		/* not implemented events (throw exception like default) */
		case GraphEvent.EDGE_REINSERTION:
		case GraphEvent.NODE_REINSERTION:
		case GraphEvent.POST_EDGE_CHANGE:
		case GraphEvent.PRE_EDGE_CHANGE:
		case GraphEvent.SUBGRAPH_INSERTION:
		case GraphEvent.SUBGRAPH_REMOVAL:
		default:
			throw new RuntimeException(
					"NOT IMPLEMENTED: Graph event handling in Clustering: UNKNOWN");
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * de.uka.algo.graphs.GraphInterpretationListener#weightChangeEvent(y.base
	 * .Edge, double, double)
	 */
	public void weightChangeEvent(Edge edge, double oldWeight, double newWeigth) {
		listenerList.weightChangedEvent(edge, oldWeight, newWeigth);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.graphs.GraphInterpretationListener#newWeightsEvent()
	 */
	public void newWeightsEvent() {
		listenerList.newWeightsEvent();
	}

	public boolean equals(Object obj) {
		if (disposed)
			return obj == this;
		return equalsHelper(obj) && ((Clustering) obj).equalsHelper(this);
	}

	private boolean equalsHelper(Object obj) {
		if (!(obj instanceof Clustering)) {
			return false;
		}
		Clustering compareTo = (Clustering) obj;
		if (getGraph() != compareTo.getGraph()) {
			return false;
		}
		for (GYCursor<Cluster> gyc = clusters(); gyc.ok(); gyc.next()) {
			Cluster myCluster = gyc.active();
			Cluster hisCluster = compareTo.getCluster(myCluster.aMember());
			for (NodeCursor nc = myCluster.members(); nc.ok(); nc.next()) {
				Node n = nc.node();
				if (compareTo.getCluster(n) != hisCluster) {
					return false;
				}
			}
		}
		for (NodeCursor nc = poolNodes(); nc.ok(); nc.next()) {
			if (compareTo.getCluster(nc.node()) != null) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Returns the cluster of a node, null if none.
	 */
	public Object get(Object arg0) {
		if (!(arg0 instanceof Node)) {
			throw new RuntimeException("unsupported type, must be node");
		}
		return getCluster((Node) arg0);
	}

	/**
	 * Returns whether a node is assigned to a cluster.
	 */
	public boolean getBool(Object arg0) {
		return get(arg0) != null;
	}

	/**
	 * Returns the id of the node's cluster, -1 if none. Note: Cluster IDs may
	 * change if any operation is performed which changes the number of
	 * clusters.
	 */
	public double getDouble(Object arg0) {
		return getInt(arg0);
	}

	/**
	 * Returns the id of the node's cluster, -1 if none. Note: Cluster IDs may
	 * change if any operation is performed which changes the number of
	 * clusters.
	 */
	public int getInt(Object arg0) {
		Cluster c = (Cluster) get(arg0);
		return (c == null) ? -1 : c.getIndex();
	}

	private boolean hasDoubleOccurrence(NodeCursor aNodeCursor, Node aNode) {
		int count = 0;
		for (aNodeCursor.toFirst(); aNodeCursor.ok(); aNodeCursor.ok()) {
			if (aNodeCursor.node() == aNode)
				count++;
		}
		return count > 1;
	}

	/**
	 * Copies the given clustering into this clustering. In order to do this,
	 * the two underlying graphs have to be identical.
	 * 
	 * @param otherClustering
	 *            source clustering that is copied
	 */
	public void copyClusterAssignmentFrom(Clustering otherClustering) {
		if (this.graph != otherClustering.graph) {
			throw new RuntimeException(
					"cannot copy cluster assignment when the underlying graph is different");
		}
		// reset this clustering to singletons
		this.resetSingletons();
		// merge the members of the clusters of the other clustering in this
		// clustering
		for (Cluster c : otherClustering.getClusterArray()) {
			this.merge(c.members());
		}
	}

	/**
	 * Copies this clustering into the given clustering. In order to do this,
	 * the two underlying graphs have to be identical.
	 * 
	 * @param otherClustering
	 *            target clustering where the copy is stored
	 */
	public void copyClusterAssignmentTo(Clustering otherClustering) {
		otherClustering.copyClusterAssignmentFrom(this);
	}

}