/* 
 * Integration: May 2, 2006 (katz)
 * Reviewed: May 4, 2006 (katz)
 *
 * Copyright 2004-2006.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering;

import java.lang.ref.Reference;
import java.lang.ref.WeakReference;

import y.base.Edge;
import y.base.Node;
import y.base.NodeCursor;

/**
 * A ClusteringListener is an object registered with a clustering being notified
 * if the graph, the graph's interpretation or the clustering changes. By
 * convention, a ClusteringListener should handle events with respect to the
 * policy flag (which can be read by <code>getPolicy()</code>). If it is set
 * to <code>CONSTANTHANDLING</code> (default), handling events should not take
 * more than constant time, i.e., O(1). More complex operations than have to be
 * performed on demand. If the policy is <code>UNRESTRICTED</code>, there is
 * no restriction. If the policy changes, a notification is sent as well.<br>
 * While it is still recommended to remove listeners <b>that have explicitly
 * been registered and are no longer in use</b>, listeners are registered
 * <i>weakly</i>, i.e., listeners which are only referenced by the respective
 * clustering are finalized.
 * 
 * Each method with the parameter notificationTime is called once with 
 * ClusteringListener.PRE (before the event has been completed)
 * and once with ClusteringListener.POST (after the
 * event has been completed). Remember to check the parameter.
 * 
 * @author Bastian Katz, mailto:katz@ira.uka.de
 */
public abstract class ClusteringListener extends ClusteringReference {

	/* ===== fields ===== */

	/* notification time constants */
	/* notification time: before action */
	public static final boolean PRE = true;

	/* notification time: after action */
	public static final boolean POST = false;

	/* handling policy: O(1) handling */
	public static final boolean CONSTANTHANDLING = true;

	/* handling policy: no restrictions */
	public static final boolean UNRESTRICTED = false;

	/* a weak reference to itself */
	private Reference<ClusteringListener> myReference;
	
	/* flag to indicate that the listener has been disposed */
	protected boolean disposed = false;


	/* constructor */

	/**
	 * Instantiates a new ClusteringListener. Does <b>not <b>register it as
	 * listener. Subclasses may want to register other listeners first.
	 * 
	 * @param clustering
	 *            the clustering
	 * 
	 */
	protected ClusteringListener(Clustering clustering) {
		super(clustering);
		myReference = new WeakReference<ClusteringListener>(this);
	}

	/* getters and queries */

	/**
	 * Returns the current handling policy.
	 * 
	 * @return the current handling policy
	 */
	public boolean getPolicy() {
		return clustering.getPolicy();
	}

	/**
	 * Returns a unique weak reference to itself
	 * 
	 * @return a unique weak reference to itself
	 */
	protected Reference<ClusteringListener> getReference() {
		return myReference;
	}


	/**
	 * Method called when the clustering is disposed. Override this method to
	 * dispose objects like NodeMaps here. <br>
	 * Note: This method is also called at finalization.
	 */
	protected void disposeEvent() {
		disposed = true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#finalize()
	 */
	protected void finalize() throws Throwable {
		if (!disposed)
			disposeEvent();
		super.finalize();
	}
	
	abstract protected String id();

	abstract protected void policyChanged(boolean policy);

	abstract protected void newClusterEvent(boolean notificationTime, Cluster newCluster);

	
	abstract protected void deleteEvent(boolean notificationTime, Cluster deletedCluster);

	abstract protected void addEvent(boolean notificationTime, Cluster to,
			NodeCursor nc);

	
	abstract protected void extractEvent(boolean notificationTime, Cluster from,
			NodeCursor nc);

	abstract protected void resetEvent(boolean notificationTime);

	
	abstract protected void resetSingletonsEvent(boolean notificationTime);

	abstract protected void resetLargeClusterEvent(boolean notificationTime,
			Cluster c);

	abstract protected void splitEvent(boolean notificationTime, Cluster a,
			Cluster b);

	abstract protected void mergeEvent(boolean notificationTime, Cluster a,
			Cluster b);

	abstract protected void moveEvent(boolean notificationTime, Cluster from,
			Cluster to, NodeCursor nc);

	abstract protected void edgeCreationEvent(Edge newEdge);

	abstract protected void nodeCreationEvent(Node newNode);

	abstract protected void nodeRemovalEvent(boolean notificationTime,
			Node nodeToRemove);

	abstract protected void edgeRemovalEvent(boolean notificationTime,
			Edge edgeToRemove);

	abstract protected void weightChangedEvent(Edge e, double oldWeight,
			double newWeight);

	abstract protected void newWeightsEvent();
}