/* 
 * Integration: May 4, 2006 (katz)
 *
 * Copyright 2004-2006.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering;

import java.lang.ref.WeakReference;

import y.base.Edge;
import y.base.Node;
import y.base.NodeCursor;
import y.base.YCursor;
import de.uka.algo.util.GYList;

/**
 * A ClusteringListenerList is a helper class to broadcast events to all
 * ClusteringListeners registered with a Clustering. It stores listeners in a
 * list of weak references which allows garbage collection to wipe out listeners
 * after all <i>other </i> references are lost, i.e. if you can't use a listener
 * anymore, it will be unregistered by the garbage collection. Listeners will be
 * notified in insertion order. ListenerLists are only to be used to
 * broadcast events thrown by the clustering and to find registered listeners.
 * ClusteringListenerLists should only be instantiated by the clustering.
 * 
 * @author Bastian Katz, mailto:katz@ira.uka.de
 */
final class ClusteringListenerList extends ClusteringListener {

	/* fields */

	/* the list of weak references to other listeners */
	private GYList<ClusteringListener> listenerList = new GYList<ClusteringListener>();

	/* constructor */

	/**
	 * Instantiates a new ClusteringListenerList. Should only be called by the
	 * clustering.
	 * 
	 * @param clustering
	 *            the clustering to listen to.
	 */
	ClusteringListenerList(Clustering clustering) {
		super(clustering);
	}

	/* adding and removing listeners */

	void addClusteringListener(ClusteringListener listener) {
		listenerList.addLast(listener.getReference());
	}

	void removeClusteringListener(ClusteringListener listener) {
		listenerList.remove(listener.getReference());
	}

	/* listener query (by class) */

	@SuppressWarnings("unchecked")
	ClusteringListener getClusteringListener(String id) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify = ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify != null) {
				String id2 = toNotify.id();
				if (id2.equals(id)) {
					return toNotify;
				}
			}
		}
		return null;
	}

	/* event propagation */

	@SuppressWarnings("unchecked")
	protected void newClusterEvent(boolean notificationTime, Cluster c) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify = ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				toNotify.newClusterEvent(notificationTime, c);
			}
		}
	}

	@SuppressWarnings("unchecked")
	protected void deleteEvent(boolean notificationTime, Cluster c) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify =  ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				toNotify.deleteEvent(notificationTime, c);
			}
		}
	}

	@SuppressWarnings("unchecked")
	protected void addEvent(boolean notificationTime, Cluster c, NodeCursor nc) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify = ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				nc.toFirst();
				toNotify.addEvent(notificationTime, c, nc);
			}
		}
	}

	@SuppressWarnings("unchecked")
	protected void extractEvent(boolean notificationTime, Cluster c,
			NodeCursor nc) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify = ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				nc.toFirst();
				toNotify.extractEvent(notificationTime, c, nc);
			}
		}
	}

	@SuppressWarnings("unchecked")
	protected void resetEvent(boolean notificationTime) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify = ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				toNotify.resetEvent(notificationTime);
			}
		}
	}

	@SuppressWarnings("unchecked")
	protected void resetSingletonsEvent(boolean notificationTime) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify =  ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				toNotify.resetSingletonsEvent(notificationTime);
			}
		}
	}

	@SuppressWarnings("unchecked")
	protected void resetLargeClusterEvent(boolean notificationTime, Cluster c) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify = ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				toNotify.resetLargeClusterEvent(notificationTime, c);
			}
		}
	}

	@SuppressWarnings("unchecked")
	protected void splitEvent(boolean notificationTime, Cluster a, Cluster b) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify = ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				toNotify.splitEvent(notificationTime, a, b);
			}
		}
	}

	@SuppressWarnings("unchecked")
	protected void mergeEvent(boolean notificationTime, Cluster a, Cluster b) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify = ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				toNotify.mergeEvent(notificationTime, a, b);
			}
		}
	}

	@SuppressWarnings("unchecked")
	protected void moveEvent(boolean notificationTime, Cluster a, Cluster b,
			NodeCursor nc) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {			
			ClusteringListener toNotify =  ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				nc.toFirst();
				toNotify.moveEvent(notificationTime, a, b, nc);
			}
		}
	}

	@SuppressWarnings("unchecked")
	protected void disposeEvent() {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify = ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				toNotify.disposeEvent();
			}
		}
	}

	@SuppressWarnings("unchecked")
	protected void policyChanged(boolean policy) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify = ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				toNotify.policyChanged(policy);
			}
		}
	}

	/**
	 * 
	 */
	@SuppressWarnings("unchecked")
	public void disposeAllListeners() {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify = ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify != null) {
				toNotify.disposeEvent();
			}
			listenerList.remove(yc.current());
		}
	}

	@SuppressWarnings("unchecked")
	protected void edgeCreationEvent(Edge newEdge) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify = ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				toNotify.edgeCreationEvent(newEdge);
			}
		}
	}

	@SuppressWarnings("unchecked")
	protected void nodeCreationEvent(Node newNode) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify = ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				toNotify.nodeCreationEvent(newNode);
			}
		}
	}

	@SuppressWarnings("unchecked")
	protected void nodeRemovalEvent(boolean notificationTime, Node nodeToRemove) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify = ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				toNotify.nodeRemovalEvent(notificationTime, nodeToRemove);
			}
		}
	}

	@SuppressWarnings("unchecked")
	protected void edgeRemovalEvent(boolean notificationTime, Edge edgeToRemove) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify =  ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				toNotify.edgeRemovalEvent(notificationTime, edgeToRemove);
			}
		}

	}

	@SuppressWarnings("unchecked")
	@Override
	protected void newWeightsEvent() {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify =  ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				toNotify.newWeightsEvent();
			}
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	protected void weightChangedEvent(Edge e, double oldWeight, double newWeight) {
		for (YCursor yc = listenerList.cursor(); yc.ok(); yc.next()) {
			ClusteringListener toNotify =  ((WeakReference<ClusteringListener>) yc
					.current()).get();
			if (toNotify == null) {
				listenerList.remove(yc.current());
			} else {
				toNotify.weightChangedEvent(e, oldWeight, newWeight);
			}
		}
	}

	@Override
	protected String id() {		
		return "ListenerList";
	}


}