package de.uka.algo.clustering;

import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeList;
import de.uka.algo.util.GYCursor;
import de.uka.algo.util.GYList;

/**
 * A ClusteringOperationException in thrown whenever an operation on a
 * clustering cannot be performed due to invalid parameters, such as unexpected
 * null parameters, violated preconditions and so on. Clustering operations
 * should check rigorously for violated preconditions to provide helpful
 * feedback.
 * 
 * @author Bastian Katz, mailto:katz@kit.edu
 * 
 */
public class ClusteringOperationException extends RuntimeException {

	private static final long serialVersionUID = 1L;
	static final String OP_CLUSTER_CREATION = "cluster creation";
	static final String OP_RESET_LARGE_CLUSTER = "reset to large cluster";
	static final String OP_CLUSTER_DELETION = "cluster deletion";
	static final String OP_ADDING_TO_CLUSTER = "adding nodes to cluster";
	static final String OP_MERGE_MANY_CLUSTERS = "merging many clusters";
	static final String OP_MERGE_TWO_CLUSTERS = "merging two clusters";
	static final String OP_MOVE_NODES = "moving nodes to different cluster";
	static final String OP_SPLIT_CLUSTER = "splitting cluster";
	static final String OP_EXTRACT_FROM_CLUSTER = "extracting nodes from cluster";
	static final String OP_RESET_INTEGER_ARRAY = "reset clustering by integer array";
	static final String ER_NODE_TWICE_IN_LIST = "node appeared twice in node list";
	static final String ER_GRAPH_EMPTY = "graph is empty";
	static final String ER_IDENTICAL_CLUSTERS = "clusters are equal";
	static final String ER_CLUSTER_NOT_PART_OF_CLUSTERING = "cluster does not belong to clustering";
	static final String ER_EMPTY_NODELIST = "node list empty";
	static final String ER_FULL_NODELIST = "node list contains all nodes from cluster";
	static final String ER_NODE_ALREADY_ASSIGNED = "node already assigned to different cluster";
	static final String ER_NODE_ALREADY_IN_CLUSTER = "node already assigned to this cluster";
	static final String ER_NODE_NOT_IN_CLUSTER = "node not assigned to given cluster";
	static final String ER_NODE_NOT_IN_ANY_CLUSTER = "node not assigned to any cluster";
	static final String ER_ARRAY_SIZE_DOESNT_MATCH_NODECOUNT = "array size doesnt match nodecount";

	private String operation;
	private String error;

	private NodeCursor involvedNodes;
	private GYCursor<Cluster> involvedClusters;

	protected ClusteringOperationException(String operation, String error,
			NodeCursor involvedNodes, GYCursor<Cluster> involvedClusters) {
		this.operation = operation;
		this.error = error;
		this.involvedNodes = involvedNodes;
		this.involvedClusters = involvedClusters;
	}

	ClusteringOperationException(String operation, String error) {
		this(operation, error, getNodeCursor(null), getClusterList(null));
	}

	ClusteringOperationException(String operation, String error,
			Node involvedNode, Cluster involvedCluster) {
		this(operation, error, getNodeCursor(involvedNode),
				getClusterList(involvedCluster));
	}

	ClusteringOperationException(String operation, String error,
			Node involvedNode, GYCursor<Cluster> involvedCluster) {
		this(operation, error, getNodeCursor(involvedNode), involvedCluster);
	}

	ClusteringOperationException(String operation, String error,
			GYCursor<Cluster> involvedClusters) {
		this(operation, error, getNodeCursor(null), involvedClusters);
	}

	ClusteringOperationException(String operation, String error,
			Cluster involvedCluster) {
		this(operation, error, getNodeCursor(null),
				getClusterList(involvedCluster));
	}

	ClusteringOperationException(String operation, String error,
			Node involvedNode) {
		this(operation, error, getNodeCursor(involvedNode),
				getClusterList(null));
	}

	private static GYCursor<Cluster> getClusterList(Cluster involvedCluster) {
		GYList<Cluster> gl = new GYList<Cluster>();
		if (involvedCluster != null) {
			gl.add(involvedCluster);
		}
		return gl.elements();
	}

	private static NodeCursor getNodeCursor(Node n) {
		NodeList nl = new NodeList();
		if (n != null) {
			nl.add(n);
		}
		return nl.nodes();
	}

	@Override
	public String toString() {
		String errorMsg = "\n  ================================================";
		errorMsg += "\n  Clustering operation failed!";
		errorMsg += "\n    failing operation..: " + operation;
		errorMsg += "\n    error..............: " + error;
		if (involvedNodes != null) {
			involvedNodes.toFirst();
			errorMsg += "\n    node(s)............: ";
			for (; involvedNodes.ok(); involvedNodes.next()) {
				errorMsg += involvedNodes.node() + " ";
			}
		}
		if (involvedClusters != null) {
			involvedClusters.toFirst();
			errorMsg += "\n    cluster(s).........: ";
			for (; involvedClusters.ok(); involvedClusters.next()) {
				errorMsg += involvedClusters.active() + " ";
			}
		}
		errorMsg += "\n  ------------------------------------------------";
		errorMsg += "\n  please refer to the online-API for more details."
				+ "\n    <http://i11www.iti.uni-karlsruhe.de/~katz/clusteringTools/>";
		errorMsg += "\n  ================================================";
		return errorMsg;
	}
}
