/* 
 * Integration: May 4, 2006 (katz)
 *
 * Copyright 2004-2006.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.EdgeList;
import y.base.EdgeMap;
import y.base.ListCell;
import y.base.Node;
import y.base.NodeCursor;
import de.uka.algo.util.GYCursor;

/**
 * The class {@link EdgePartitionListener} provides additional functionality for
 * clusterings, such as counting inner cluster edges etc. An object of this
 * class is instantiated <i>on demand</i> and attached to the clustering. It
 * fully supports update policies, i.e., will add only constant overhead to any
 * clustering operation if the policy is set to
 * {@code ClusteringListener.CONSTANTHANDLING}, but will perform smart updates
 * if the policy is set to {@code ClusteringListener.UNRESTRICTED}.
 * 
 * @author Bastian Katz, mailto:katz@ira.uka.de
 */
final class EdgePartitionListener extends DefaultListener {

	/* ===== fields ===== */

	/**
	 * Flag indicating whether the internal structure currently reflects the
	 * clustering's structure properly. May be false if the policy is set to
	 * <i>constant handling</i>.
	 */
	private boolean valid = false;

	/**
	 * edges are organized in lists - one border list and one inner list for
	 * every cluster. If an edge is in a cluster's border list, the edge's
	 * source being in that cluster, the edgeMap {@code sourceCell} contains the
	 * respective listCell. If the edge's target resides in the cluster, the
	 * respective cell is stored in {@code targetCell} for the respective edge.
	 * If an edge is an inner edge for some cluster, both maps contain the same
	 * cell, namely the cell containing the edge in the respective list. The
	 * maps contain {@code null}s if one or both endpoints are not assigned to
	 * a cluster.
	 */
	private EdgeMap sourceCell = null;
	private EdgeMap targetCell = null;

	/** number of edges currently covered by the clustering */
	private int innerEdgeCount;

	/** sum of all edge weights */
	private double overAllWeight;

	/** sum of border edge weights */
	private double outerWeight;

	/** cluster map containing {@code EdgeIndexInfos} */
	private ClusterMap<ClusterInfo> info = null;

	/* constructor */

	/**
	 * Instantiates a new instance. This constructor may only be called by
	 * respective clustering if there is none assigned to the clustering.
	 * 
	 * @param clustering
	 *            the clustering to instantiate the
	 *            {@code EdgePartitionListener} for
	 */
	EdgePartitionListener(Clustering clustering) {
		super(clustering);
		// the cluster size listener is added to list of privileged listeners by
		// the clustering itself

		// get or instantiate required maps
		info = new ClusterMap<ClusterInfo>(clustering);
		sourceCell = clustering.getGraph().createEdgeMap(); // disposed at
		// finalize
		targetCell = clustering.getGraph().createEdgeMap(); // disposed at
		// finalize
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.clustering.ClusteringListener#policyChanged(boolean)
	 */
	@Override
	protected void policyChanged(boolean policy) {
		/* if policy changes to unrestricted, the structure must be valid */
		if (policy == UNRESTRICTED) {
			assertValidity();
		}
	}

	/* queries */

	/**
	 * Returns the sum of all egde weights.
	 * 
	 * @return the sum of all edge weights
	 */
	double getOverAllWeight() {
		assertValidity(); //make sure the data is up to date
		return overAllWeight;
	}

	/**
	 * Returns the number of edges having both adjacent nodes in the specified
	 * cluster.
	 * 
	 * @param cluster
	 *            the cluster
	 * @return the number of edges having both adjacent nodes in cluster
	 */
	int getInnerEdgeCount(Cluster cluster) {
		assertValidity(); //make sure the data is up to date
		return get(cluster).inner.size();
	}

	/**
	 * Returns a cursor iterating over all edges having both adjacent nodes in
	 * the specified cluster.
	 * 
	 * @param cluster
	 *            the cluster
	 * @return cursor iterating over all edges having both adjacent nodes in the
	 *         specified cluster.
	 */
	EdgeCursor innerEdges(Cluster cluster) {
		assertValidity(); //make sure the data is up to date
		return get(cluster).inner.edges();
	}

	/**
	 * Returns a cursor iterating over all edges having one (and only one)
	 * adjacent node in the specified cluster (those edges cross the cluster's
	 * border).
	 * 
	 * @param cluster
	 *            the cluster
	 * @return a cursor iterating over all edges having one (and only one)
	 *         adjacent node in cluster
	 */
	EdgeCursor borderEdges(Cluster cluster) {
		assertValidity(); //make sure the data is up to date
		return get(cluster).border.edges();
	}

	/**
	 * Returns the number of edges crossing a cluster's border, i.e. the number
	 * of edges having one (and only one) adjacent nodes in a specified cluster.
	 * 
	 * @param cluster
	 *            the cluster
	 * @return the number of edges crossing the cluster's border
	 */
	int getBorderEdgeCount(Cluster cluster) {
		assertValidity(); //make sure the data is up to date
		return get(cluster).border.size();
	}

	/**
	 * Returns the weight sum of all edges not having both adjacent nodes in the
	 * same cluster.
	 * 
	 * @return the weight sum of all edges not having both adjacent nodes in the
	 *         same cluster
	 */
	double getOuterWeight() {
		assertValidity(); //make sure the data is up to date
		return outerWeight;
	}

	/**
	 * Returns the weight sum over all nodes having both adjacent nodes in the
	 * specified clusters.
	 * 
	 * @param cluster
	 *            the cluster
	 * @return the weight sum over all nodes having both adjacent nodes in
	 *         cluster
	 */
	double getInnerWeight(Cluster cluster) {
		assertValidity(); //make sure the data is up to date
		return get(cluster).innerWeight;
	}

	/**
	 * Returns the weight sum over all edges having both adjacent nodes in the
	 * same cluster.
	 * 
	 * @return the weight sum over all edges having both adjacent nodes in the
	 *         same cluster
	 */
	double getInnerWeight() {
		assertValidity(); //make sure the data is up to date
		return getOverAllWeight() - getOuterWeight();
	}

	/**
	 * Returns the weight sum over all edges having one (and only one) adjacent
	 * node in the given cluster.
	 * 
	 * @param cluster
	 *            the cluster
	 * @return the weight sum over all edges having one (and only one) adjacent
	 *         node in the given cluster.
	 */
	double getBorderWeight(Cluster cluster) {
		assertValidity(); //make sure the data is up to date
		return get(cluster).borderWeight;
	}

	/**
	 * Returns the number of edges having both adjacent nodes in the same
	 * cluster.
	 * 
	 * @return the number of edges having both adjacent nodes in the same
	 *         cluster
	 */
	int getInnerEdgeCount() {
		assertValidity(); //make sure the data is up to date
		return innerEdgeCount;
	}

	/**
	 * Returns the number of edges not having both adjacent nodes in the same
	 * cluster.
	 * 
	 * @return the number of edges not having both adjacent nodes in the same
	 *         cluster
	 */
	int getOuterEdgeCount() {
		assertValidity(); //make sure the data is up to date
		return clustering.getGraph().edgeCount() - innerEdgeCount;
	}

	/**
	 * Returns the number of add and remove Edge operations to transform the
	 * graph into a clustergraph according to the clustering. See Cluster
	 * Editing Problem for further details.
	 * 
	 * @return the number of add and remove Edge operations to transform the
	 *         graph into a clustergraph according to the clustering
	 */
	int getFailureEdgeCount() { //TODO: move to updates and reset! (katz, low priority)
		assertValidity();
		int result = 0;
		for (GYCursor<Cluster> cc = clustering.clusters(); cc.ok(); cc.next()) {
			Cluster c = cc.active();
			result += c.size() * (c.size() - 1);
		}
		result /= 2;
		result += clustering.getGraph().edgeCount() - 2 * innerEdgeCount;
		return result;
	}

	/* helper methods */

	/**
	 * Returns the ClusterInfo assigned to the given Cluster.
	 * 
	 * @param cluster
	 *            the cluster
	 * @return the ClusterInfo assigned to the cluster
	 */
	ClusterInfo get(Cluster cluster) {
		return info.get(cluster);
	}

	private void assertValidity() {
		if (!valid)
			reset();
	}

	protected void reset() {
		for (GYCursor<Cluster> cc = clustering.clusters(); cc.ok(); cc.next()) {
			Cluster aCluster = cc.active();
			info.set(aCluster, new ClusterInfo());
		}
		overAllWeight = 0;
		innerEdgeCount = 0;
		outerWeight = 0;

		double sumInnerWeight = 0;

		// inner and border
		for (EdgeCursor ec = clustering.getGraph().edges(); ec.ok(); ec.next()) {
			Edge anEdge = ec.edge();
			double weight = clustering.getInterpretation().getWeight(anEdge);

			overAllWeight += weight;

			Node source = anEdge.source();
			Node target = anEdge.target();
			Cluster sourceCluster = clustering.getCluster(source);
			Cluster targetCluster = clustering.getCluster(target);

			if (sourceCluster == targetCluster && sourceCluster != null) {
				ListCell aCell = get(sourceCluster).inner.addLast(anEdge);
				innerEdgeCount++;
				sourceCell.set(anEdge, aCell);
				targetCell.set(anEdge, aCell);
				get(sourceCluster).innerWeight += weight;
				sumInnerWeight += weight;
			} else {
				if (sourceCluster != null) {
					sourceCell.set(anEdge, get(sourceCluster).border
							.addLast(anEdge));
					get(sourceCluster).borderWeight += weight;
				}
				if (targetCluster != null) {
					targetCell.set(anEdge, get(targetCluster).border
							.addLast(anEdge));
					get(targetCluster).borderWeight += weight;
				}
			}
		}
		outerWeight = overAllWeight - sumInnerWeight;
		valid = true;
	}

	/* event handling */

	protected void defaultEvent() {
		if (getPolicy() == CONSTANTHANDLING) {
			valid = false;
		} else {
			reset();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.clustering.base.ClusteringListener#newClusterEvent(boolean,
	 *      de.uka.algo.clustering.base.Clustering,
	 *      de.uka.algo.clustering.base.Cluster)
	 */
	protected void newClusterEvent(boolean notificationTime, Cluster c) {
		if (getPolicy() == CONSTANTHANDLING) {
			valid = false;
			return;
		}
		if (notificationTime == PRE)
			return;
		Clustering source = getClustering();
		ClusterInfo clusterInfo = new ClusterInfo();
		info.set(c, clusterInfo);
		for (NodeCursor nc = c.members(); nc.ok(); nc.next()) {
			Node aNode = nc.node();
			for (EdgeCursor ec = aNode.outEdges(); ec.ok(); ec.next()) {
				Edge anEdge = ec.edge();
				double weight = source.getInterpretation().getWeight(anEdge);
				Node targetNode = anEdge.target();
				Cluster targetCluster = source.getCluster(targetNode);
				if (targetCluster == c) {
					ListCell aCell = clusterInfo.inner.addLast(anEdge);
					innerEdgeCount++;
					clusterInfo.innerWeight += weight;
					sourceCell.set(anEdge, aCell);
					targetCell.set(anEdge, aCell);
				} else {
					ListCell aCell = clusterInfo.border.addLast(anEdge);
					clusterInfo.borderWeight += weight;
					sourceCell.set(anEdge, aCell);
				}
			}
			for (EdgeCursor ec = aNode.inEdges(); ec.ok(); ec.next()) {
				Edge anEdge = ec.edge();
				double weight = source.getInterpretation().getWeight(anEdge);
				Node sourceNode = anEdge.source();
				Cluster sourceCluster = source.getCluster(sourceNode);
				if (sourceCluster != c) {
					ListCell aCell = clusterInfo.border.addLast(anEdge);
					clusterInfo.borderWeight += weight;
					targetCell.set(anEdge, aCell);
				}
			}
		}
		outerWeight -= clusterInfo.innerWeight;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.clustering.base.ClusteringListener#deleteEvent(boolean,
	 *      de.uka.algo.clustering.base.Clustering,
	 *      de.uka.algo.clustering.base.Cluster)
	 */
	protected void deleteEvent(boolean notificationTime, Cluster c) {
		if (getPolicy() == CONSTANTHANDLING) {
			valid = false;
			return;
		}

		if (notificationTime == POST)
			return;
		ClusterInfo clusterInfo = get(c);
		innerEdgeCount -= getInnerEdgeCount(c);
		outerWeight += clusterInfo.innerWeight;
		for (EdgeCursor ec = clusterInfo.inner.edges(); ec.ok(); ec.next()) {
			Edge anEdge = ec.edge();
			sourceCell.set(anEdge, null);
			targetCell.set(anEdge, null);
		}
		for (EdgeCursor ec = clusterInfo.border.edges(); ec.ok(); ec.next()) {
			Edge anEdge = ec.edge();
			Cluster sourceCluster = getClustering().getCluster(anEdge.source());
			if (sourceCluster == c) {
				sourceCell.set(anEdge, null);
			} else {
				targetCell.set(anEdge, null);
			}
		}
		info.set(c, null);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.clustering.base.ClusteringListener#addEvent(boolean,
	 *      de.uka.algo.clustering.base.Clustering,
	 *      de.uka.algo.clustering.base.Cluster, y.base.NodeCursor)
	 */
	protected void addEvent(boolean notificationTime, Cluster toCluster,
			NodeCursor nc) {
		if (getPolicy() == CONSTANTHANDLING) {
			valid = false;
			return;
		}

		if (notificationTime == PRE)
			return;
		ClusterInfo clusterInfo = get(toCluster);
		outerWeight += clusterInfo.innerWeight;

		for (; nc.ok(); nc.next()) {
			Node aNode = nc.node();
			for (EdgeCursor ec = aNode.outEdges(); ec.ok(); ec.next()) {
				Edge anEdge = ec.edge();
				double weight = getClustering().getInterpretation().getWeight(
						anEdge);
				Node targetNode = anEdge.target();
				Cluster targetCluster = getClustering().getCluster(targetNode);
				if (sourceCell.get(anEdge) == null) {
					if (targetCluster != toCluster) {
						ListCell aCell = clusterInfo.border.addLast(anEdge);
						clusterInfo.borderWeight += weight;
						sourceCell.set(anEdge, aCell);
					}
				} else {
					if (targetCell.get(anEdge) != null) {
						clusterInfo.border.removeCell((ListCell) targetCell
								.get(aNode));
						clusterInfo.borderWeight -= weight;
						ListCell aCell = clusterInfo.inner.addLast(anEdge);
						innerEdgeCount++;
						sourceCell.set(anEdge, aCell);
						targetCell.set(anEdge, aCell);
						clusterInfo.innerWeight += weight;
					}
				}
			}
			for (EdgeCursor ec = aNode.inEdges(); ec.ok(); ec.next()) {
				Edge anEdge = ec.edge();
				double weight = getClustering().getInterpretation().getWeight(
						anEdge);
				Node sourceNode = anEdge.source();
				Cluster sourceCluster = getClustering().getCluster(sourceNode);
				if (targetCell.get(anEdge) == null) {
					if (sourceCluster != toCluster) {
						ListCell aCell = clusterInfo.border.addLast(anEdge);
						clusterInfo.borderWeight += weight;

						targetCell.set(anEdge, aCell);
					}
				} else {
					if (sourceCell.get(anEdge) != null) {
						clusterInfo.border.removeCell((ListCell) sourceCell
								.get(aNode));
						clusterInfo.borderWeight -= weight;
						ListCell aCell = clusterInfo.inner.addLast(anEdge);
						innerEdgeCount++;
						sourceCell.set(anEdge, aCell);
						targetCell.set(anEdge, aCell);
						clusterInfo.innerWeight += weight;
					}
				}

			}
		}
		outerWeight -= clusterInfo.innerWeight;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.clustering.base.ClusteringListener#extractEvent(boolean,
	 *      de.uka.algo.clustering.base.Clustering,
	 *      de.uka.algo.clustering.base.Cluster, y.base.NodeCursor)
	 */
	protected void extractEvent(boolean notificationTime, Cluster fromCluster,
			NodeCursor nc) {
		if (getPolicy() == CONSTANTHANDLING) {
			valid = false;
			return;
		}

		if (notificationTime == PRE)
			return;
		Clustering source = getClustering();
		ClusterInfo clusterInfo = get(fromCluster);
		outerWeight += clusterInfo.innerWeight;
		for (; nc.ok(); nc.next()) {
			Node aNode = nc.node();
			for (EdgeCursor ec = aNode.outEdges(); ec.ok(); ec.next()) {
				Edge anEdge = ec.edge();
				double weight = source.getInterpretation().getWeight(anEdge);

				Node targetNode = anEdge.target();
				Cluster targetCluster = source.getCluster(targetNode);
				if (sourceCell.get(anEdge) != null) { // edge not done
					if (targetCluster == fromCluster) { // inner(c) to
						// border(c-)
						clusterInfo.inner.removeCell((ListCell) sourceCell
								.get(anEdge));
						innerEdgeCount--;
						clusterInfo.innerWeight -= weight;
						ListCell aCell = clusterInfo.border.addLast(anEdge);
						clusterInfo.borderWeight += weight;
						targetCell.set(anEdge, aCell);
						sourceCell.set(anEdge, null);
					} else if (targetCluster != null) { // bord(cx) to bord(x)
						clusterInfo.border.removeCell((ListCell) sourceCell
								.get(anEdge));
						clusterInfo.borderWeight -= weight;
						sourceCell.set(anEdge, null);
					} else if (targetCell.get(anEdge) == null) { // bord(c-)--
						clusterInfo.border.removeCell((ListCell) sourceCell
								.get(anEdge));
						clusterInfo.borderWeight -= weight;
						sourceCell.set(anEdge, null);
					} else { // inner(c) to --
						clusterInfo.inner.removeCell((ListCell) sourceCell
								.get(anEdge));
						innerEdgeCount--;
						targetCell.set(anEdge, null);
						sourceCell.set(anEdge, null);
					}
				}
			}

			for (EdgeCursor ec = aNode.inEdges(); ec.ok(); ec.next()) {
				Edge anEdge = ec.edge();
				double weight = source.getInterpretation().getWeight(anEdge);
				Node sourceNode = anEdge.source();
				Cluster sourceCluster = source.getCluster(sourceNode);
				if (targetCell.get(anEdge) != null) { // edge not done
					if (sourceCluster == fromCluster) { // inner(c) to
						// border(c-)
						clusterInfo.inner.removeCell((ListCell) targetCell
								.get(anEdge));
						innerEdgeCount--;
						clusterInfo.innerWeight -= weight;
						ListCell aCell = clusterInfo.border.addLast(anEdge);
						clusterInfo.borderWeight += weight;
						sourceCell.set(anEdge, aCell);
						targetCell.set(anEdge, null);
					} else if (sourceCluster != null) { // bord(cx) to bord(x)
						clusterInfo.border.removeCell((ListCell) targetCell
								.get(anEdge));
						clusterInfo.borderWeight -= weight;
						targetCell.set(anEdge, null);
					} else if (sourceCell.get(anEdge) == null) { // bord(c-)--
						clusterInfo.border.removeCell((ListCell) targetCell
								.get(anEdge));
						clusterInfo.borderWeight -= weight;
						targetCell.set(anEdge, null);
					} else { // inner(c) to --
						clusterInfo.inner.removeCell((ListCell) targetCell
								.get(anEdge));
						innerEdgeCount--;
						sourceCell.set(anEdge, null);
						targetCell.set(anEdge, null);
					}
				}
			}

		}
		outerWeight -= clusterInfo.innerWeight;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.clustering.base.ClusteringListener#resetEvent(boolean,
	 *      de.uka.algo.clustering.base.Clustering)
	 */
	protected void resetEvent(boolean notificationTime) {
		if (getPolicy() == CONSTANTHANDLING) {
			valid = false;
			return;
		}
		if (notificationTime == PRE)
			return;
		Clustering cl = getClustering();
		for (EdgeCursor ec = cl.getGraph().edges(); ec.ok(); ec.next()) {
			Edge anEdge = ec.edge();
			sourceCell.set(anEdge, null);
			targetCell.set(anEdge, null);
		}
		outerWeight = overAllWeight;
		innerEdgeCount = 0;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.clustering.base.ClusteringListener#resetSingletonsEvent(boolean,
	 *      de.uka.algo.clustering.base.Clustering)
	 */
	protected void resetSingletonsEvent(boolean notificationTime) {
		if (getPolicy() == CONSTANTHANDLING) {
			valid = false;
			return;
		}

		if (notificationTime == PRE)
			return;
		Clustering cl = getClustering();
		outerWeight = overAllWeight;
		innerEdgeCount = 0;
		for (GYCursor<Cluster> cc = cl.clusters(); cc.ok(); cc.next()) {
			Cluster aCluster = cc.active();
			info.set(aCluster, new ClusterInfo());

			Node aNode = aCluster.members().node();
			for (EdgeCursor ec = aNode.outEdges(); ec.ok(); ec.next()) {
				Edge anEdge = ec.edge();
				if (anEdge.target() == aNode) {
					ListCell aCell = get(aCluster).inner.addLast(anEdge);
					innerEdgeCount++;
					sourceCell.set(anEdge, aCell);
					targetCell.set(anEdge, aCell);
					get(aCluster).innerWeight += cl.getInterpretation()
							.getWeight(anEdge);
				} else {
					ListCell aCell = get(aCluster).border.addLast(anEdge);
					sourceCell.set(anEdge, aCell);
					get(aCluster).borderWeight += cl.getInterpretation()
							.getWeight(anEdge);
				}
			}
			for (EdgeCursor ec = aNode.inEdges(); ec.ok(); ec.next()) {
				Edge anEdge = ec.edge();
				if (anEdge.source() != aNode) {
					ListCell aCell = get(aCluster).border.addLast(anEdge);
					targetCell.set(anEdge, aCell);
					get(aCluster).borderWeight += cl.getInterpretation()
							.getWeight(anEdge);
				}
			}
			outerWeight -= get(aCluster).innerWeight;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.clustering.base.ClusteringListener#resetLargeClusterEvent(boolean,
	 *      de.uka.algo.clustering.base.Clustering,
	 *      de.uka.algo.clustering.base.Cluster)
	 */
	protected void resetLargeClusterEvent(boolean notificationTime, Cluster c) {
		if (getPolicy() == CONSTANTHANDLING) {
			valid = false;
			return;
		}

		if (notificationTime == PRE)
			return;
		Clustering cl = getClustering();
		info.set(c, new ClusterInfo());
		for (EdgeCursor ec = cl.getGraph().edges(); ec.ok(); ec.next()) {
			Edge anEdge = ec.edge();
			ListCell aCell = get(c).inner.addLast(anEdge);
			innerEdgeCount++;
			sourceCell.set(anEdge, aCell);
			targetCell.set(anEdge, aCell);
		}
		get(c).innerWeight = overAllWeight;
		outerWeight = 0;
		innerEdgeCount = clustering.getGraph().edgeCount();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.clustering.base.ClusteringListener#splitEvent(boolean,
	 *      de.uka.algo.clustering.base.Clustering,
	 *      de.uka.algo.clustering.base.Cluster,
	 *      de.uka.algo.clustering.base.Cluster)
	 */
	protected void splitEvent(boolean notificationTime, Cluster a, Cluster b) {
		if (getPolicy() == CONSTANTHANDLING) {
			valid = false;
			return;
		}

		if (notificationTime == PRE)
			return;
		Clustering clustering = getClustering();
		info.set(b, new ClusterInfo());
		for (EdgeCursor ec = get(a).border.edges(); ec.ok(); ec.next()) {
			Edge anEdge = ec.edge();
			double weight = clustering.getInterpretation().getWeight(anEdge);
			Node sourceNode = anEdge.source();
			Node targetNode = anEdge.target();
			Cluster sourceCluster = clustering.getCluster(sourceNode);
			Cluster targetCluster = clustering.getCluster(targetNode);
			if (sourceCluster == b || targetCluster == b) {
				get(a).border.removeAt(ec);
				get(a).borderWeight -= weight;
				ListCell aCell = get(b).border.addLast(anEdge);
				get(b).borderWeight += weight;
				if (sourceCluster == b) {
					sourceCell.set(anEdge, aCell);
				} else {
					targetCell.set(anEdge, aCell);
				}
			}
		}
		for (EdgeCursor ec = get(a).inner.edges(); ec.ok(); ec.next()) {
			Edge anEdge = ec.edge();
			double weight = clustering.getInterpretation().getWeight(anEdge);
			Node sourceNode = anEdge.source();
			Node targetNode = anEdge.target();
			Cluster sourceCluster = clustering.getCluster(sourceNode);
			Cluster targetCluster = clustering.getCluster(targetNode);
			if (sourceCluster != targetCluster) {
				get(a).inner.removeAt(ec);
				innerEdgeCount--;
				get(a).innerWeight -= weight;
				outerWeight += weight;
				ListCell aCell = get(a).border.addLast(anEdge);
				get(a).borderWeight += weight;
				ListCell bCell = get(b).border.addLast(anEdge);
				get(b).borderWeight += weight;
				if (sourceCluster == a) {
					sourceCell.set(anEdge, aCell);
					targetCell.set(anEdge, bCell);
				} else {
					sourceCell.set(anEdge, bCell);
					targetCell.set(anEdge, aCell);
				}
			} else if (sourceCluster == b) {
				get(a).inner.removeAt(ec);
				innerEdgeCount--;
				get(a).innerWeight -= weight;
				ListCell aCell = get(b).inner.addLast(anEdge);
				innerEdgeCount++;
				get(b).innerWeight += weight;
				sourceCell.set(anEdge, aCell);
				targetCell.set(anEdge, aCell);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.clustering.base.ClusteringListener#mergeEvent(boolean,
	 *      de.uka.algo.clustering.base.Clustering,
	 *      de.uka.algo.clustering.base.Cluster,
	 *      de.uka.algo.clustering.base.Cluster)
	 */
	// TODO: buggy?
	// protected void mergeEvent(boolean notficationTime, Cluster big,
	// Cluster small) {
	// if (getPolicy() == CONSTANTHANDLING) {
	// valid = false;
	// return;
	// }
	//
	// if (notficationTime == POST)
	// return;
	// ClusterInfo bigInfo = get(big);
	// ClusterInfo smallInfo = get(small);
	// outerWeight += bigInfo.innerWeight;
	// outerWeight += smallInfo.innerWeight;
	//
	// Clustering cl =getClustering();
	// //copy inner edges
	// bigInfo.innerWeight += smallInfo.innerWeight;
	// for (EdgeCursor ec = smallInfo.inner.edges(); ec.ok(); ec.next()) {
	// Edge anEdge = ec.edge();
	// ListCell aCell = bigInfo.inner.addLast(anEdge);
	// innerEdgeCount++;
	// sourceCell.set(anEdge, aCell);
	// targetCell.set(anEdge, aCell);
	// }
	// //check borders (big)
	// for (EdgeCursor ec = bigInfo.border.edges(); ec.ok(); ec.next()) {
	// Edge anEdge = ec.edge();
	// double weight = cl.getInterpretation().getWeight(anEdge);
	// Node sourceNode = anEdge.source();
	// Node targetNode = anEdge.target();
	// if (cl.getCluster(targetNode) == small
	// || cl.getCluster(sourceNode) == small) {
	// bigInfo.border.removeAt(ec);
	// bigInfo.borderWeight -= weight;
	// ListCell aCell = bigInfo.inner.addLast(anEdge);
	// innerEdgeCount++;
	// bigInfo.innerWeight += weight;
	// sourceCell.set(anEdge, aCell);
	// targetCell.set(anEdge, aCell);
	// }
	// }
	// //check borders small
	// for (EdgeCursor ec = smallInfo.border.edges(); ec.ok(); ec.next()) {
	// Edge anEdge = ec.edge();
	// double weight = cl.getInterpretation().getWeight(anEdge);
	// Node sourceNode = anEdge.source();
	// Node targetNode = anEdge.target();
	// if (!(cl.getCluster(targetNode) == big
	// || cl.getCluster(sourceNode) == big)) {
	// ListCell aCell = bigInfo.border.addLast(anEdge);
	// bigInfo.borderWeight += weight;
	// if (cl.getCluster(sourceNode) == small) {
	// sourceCell.set(anEdge, aCell);
	// } else {
	// targetCell.set(anEdge, aCell);
	// }
	// }
	// }
	// outerWeight -= bigInfo.innerWeight;
	// //delete smallInfo
	// info.set(small, null);
	// }
	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.clustering.base.ClusteringListener#moveEvent(boolean,
	 *      de.uka.algo.clustering.base.Clustering,
	 *      de.uka.algo.clustering.base.Cluster,
	 *      de.uka.algo.clustering.base.Cluster, y.base.NodeCursor)
	 */
	protected void moveEvent(boolean notificationTime, Cluster from,
			Cluster to, NodeCursor nc) {
		if (getPolicy() == CONSTANTHANDLING) {
			valid = false;
			return;
		}

		if (notificationTime == POST)
			return;
		Clustering source = getClustering();
		ClusterInfo fromData = get(from);
		ClusterInfo toData = get(to);
		outerWeight += toData.innerWeight;
		outerWeight += fromData.innerWeight;
		for (; nc.ok(); nc.next()) {
			Node aNode = nc.node();
			for (EdgeCursor ec = aNode.outEdges(); ec.ok(); ec.next()) {
				Edge anEdge = ec.edge();
				double weight = source.getInterpretation().getWeight(anEdge);
				Node targetNode = anEdge.target();
				Cluster targetCluster = source.getCluster(targetNode);
				/* inner(from) -> border(to/from) */
				if (targetCluster == from) {
					/* removing from ONE lists */
					fromData.inner
							.removeCell((ListCell) sourceCell.get(anEdge));

					/* adding to both lists */
					ListCell fromCell = fromData.border.addLast(anEdge);
					ListCell toCell = toData.border.addLast(anEdge);
					sourceCell.set(anEdge, toCell);
					targetCell.set(anEdge, fromCell);

					/* adjusting weights */
					fromData.innerWeight -= weight;
					innerEdgeCount--;
					fromData.borderWeight += weight;
					toData.borderWeight += weight;
				}

				/* border(from/to) -> inner(to) */
				if (targetCluster == to) {
					/* removing from both lists */
					fromData.border.removeCell((ListCell) sourceCell
							.get(anEdge));
					toData.border.removeCell((ListCell) targetCell.get(anEdge));

					/* adding to ONE lists */
					ListCell cell = toData.inner.addLast(anEdge);
					sourceCell.set(anEdge, cell);
					targetCell.set(anEdge, cell);

					fromData.borderWeight -= weight;
					toData.borderWeight -= weight;
					toData.innerWeight += weight;
					innerEdgeCount++;
				}

				/* border(from/x) -> border(to/x) */
				if (targetCluster != from && targetCluster != to) {
					/* remove from border list */
					fromData.border.removeCell((ListCell) sourceCell
							.get(anEdge));
					/* adding to border list */
					sourceCell.set(anEdge, toData.border.addLast(anEdge));

					fromData.borderWeight -= weight;
					toData.borderWeight += weight;
				}
			}
			for (EdgeCursor ec = aNode.inEdges(); ec.ok(); ec.next()) {
				Edge anEdge = ec.edge();
				double weight = source.getInterpretation().getWeight(anEdge);
				Node targetNode = anEdge.source();
				Cluster targetCluster = source.getCluster(targetNode);
				/* inner(from) -> border(to/from) */
				if (targetCluster == from) {
					/* removing from ONE lists */
					fromData.inner
							.removeCell((ListCell) targetCell.get(anEdge));

					/* adding to both lists */
					ListCell fromCell = fromData.border.addLast(anEdge);
					ListCell toCell = toData.border.addLast(anEdge);
					targetCell.set(anEdge, toCell);
					sourceCell.set(anEdge, fromCell);

					/* adjusting weights */
					fromData.innerWeight -= weight;
					innerEdgeCount--;
					fromData.borderWeight += weight;
					toData.borderWeight += weight;
				}

				/* border(from/to) -> inner(to) */
				if (targetCluster == to) {
					/* removing from both lists */
					fromData.border.removeCell((ListCell) targetCell
							.get(anEdge));
					toData.border.removeCell((ListCell) sourceCell.get(anEdge));

					/* adding to ONE lists */
					ListCell cell = toData.inner.addLast(anEdge);
					targetCell.set(anEdge, cell);
					sourceCell.set(anEdge, cell);

					fromData.borderWeight -= weight;
					toData.borderWeight -= weight;
					toData.innerWeight += weight;
					innerEdgeCount++;
				}

				/* border(from/x) -> border(to/x) */
				if (targetCluster != from && targetCluster != to) {
					/* remove from border list */
					fromData.border.removeCell((ListCell) targetCell
							.get(anEdge));
					/* adding to border list */
					targetCell.set(anEdge, toData.border.addLast(anEdge));

					fromData.borderWeight -= weight;
					toData.borderWeight += weight;
				}
			}
		}
		outerWeight -= toData.innerWeight;
		outerWeight -= fromData.innerWeight;
	}

	protected void disposeEvent() {
		if (disposed)
			return;
		clustering.getGraph().disposeEdgeMap(sourceCell);
		clustering.getGraph().disposeEdgeMap(targetCell);
		info.dispose();
		super.disposeEvent();
	}

	/* inner classes */

	/**
	 * Struct to store cluster-related data.
	 * 
	 * @author Bastian Katz
	 */
	protected static class ClusterInfo {
		protected EdgeList inner = new EdgeList();
		protected EdgeList border = new EdgeList();
		protected double innerWeight = 0;
		protected double borderWeight = 0;
	}

	@Override
	protected String id() {
		return "EdgePartitionListener";
	}
}