/* 
 * Integration: May 4, 2006 (katz)
 *
 * Copyright 2004-2006.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering.addon;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.EdgeMap;
import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeMap;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.ClusterMap;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.DefaultListener;
import de.uka.algo.graphs.GraphInterpretation;

/**
 * The class Subgraphs implements an addon to clusterings that provides
 * cluster-induced subgraphs with interpretations and reverse mappings.
 * 
 * @author Bastian Katz
 */
public final class Subgraphs extends DefaultListener {

	private static final String ID = "Subgraphs";
	
	/** maps the graph's nodes to their images in a subgraph */
	NodeMap forwardNodes;

	/** maps the graph's edges to their images in a subgraph (if any) */
	EdgeMap forwardEdges;

	/** stores the data structure used for the clusters' subgraphs */
	ClusterMap<ClusterInfo> info;

	/**
	 * Instantiates a new subgraphs-addon
	 * 
	 * @param clustering
	 *            the clustering to listen to
	 */
	protected Subgraphs(Clustering clustering) {
		super(clustering);
		clustering.addListener(this);
		forwardNodes = clustering.getGraph().createNodeMap(); // is disposed
		forwardEdges = clustering.getGraph().createEdgeMap(); // is disposed
		info = new ClusterMap<ClusterInfo>(clustering);
	}

	private ClusterInfo validate(Cluster cluster) {
		ClusterInfo result = info.get(cluster);
		if (result == null) {
			result = new ClusterInfo(cluster);
			info.set(cluster, result);
		}
		return result;
	}

	private void invalidate(Cluster cluster) {
		info.set(cluster, null);
	}

	/**
	 * Returns the subgraph induced by a given cluster
	 * 
	 * @param aCluster
	 *            the cluster
	 * @return the induced subgraph
	 */
	public Graph getGraph(Cluster aCluster) {
		ClusterInfo info = validate(aCluster);
		return info.subgraph;
	}

	/**
	 * Returns an interpretation for the subgraph returned by
	 * <code>getGraph()</code> (dynamic type and weights as before)
	 * 
	 * @param aCluster
	 *            the cluster
	 * @return the interpretation
	 */
	public GraphInterpretation getInterpretation(Cluster aCluster) {
		ClusterInfo info = validate(aCluster);
		return info.interpretation;
	}

	/**
	 * Returns a reverse node mapping for the subgraph returned by
	 * <code>getGraph()</code>.
	 * 
	 * @return a reverse node mapping for the subgraph returned by
	 *         <code>getGraph()</code>.
	 */
	public NodeMap getReverseNodes(Cluster aCluster) {
		ClusterInfo info = validate(aCluster);
		return info.reverseNodes;
	}

	public static Subgraphs getInstance(Clustering cl) {
		Subgraphs sub = null;
		sub = (Subgraphs) cl.getClusteringListener(ID);
		if (sub == null) {
			return new Subgraphs(cl);
		}
		return sub;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.clustering.base.ClusteringListener#disposeEvent()
	 */
	protected void disposeEvent() {
		clustering.getGraph().disposeNodeMap(forwardNodes);
		clustering.getGraph().disposeEdgeMap(forwardEdges);
		info.dispose();
		super.disposeEvent();
	}

	@Override
	protected void policyChanged(boolean policy) {
		// there's nothing to do here
	}

	@Override
	protected void addEvent(boolean notificationTime, Cluster c, NodeCursor nc) {
		if (notificationTime == POST)
			invalidate(c);
	}

	@Override
	protected void defaultEvent() {
		// nothing to do here
	}

	@Override
	protected void edgeCreationEvent(Edge newEdge) {
		Cluster cluster = getClustering().getCluster(newEdge.source());
		if (cluster != null
				&& cluster == getClustering().getCluster(newEdge.target())) {
			invalidate(cluster);
		}
	}

	@Override
	protected void edgeRemovalEvent(boolean notificationTime, Edge edgeToRemove) {
		if (notificationTime == PRE)
			return;
		Cluster cluster = getClustering().getCluster(edgeToRemove.source());
		if (cluster != null
				&& cluster == getClustering().getCluster(edgeToRemove.target())) {
			invalidate(cluster);
		}
	}

	@Override
	protected void extractEvent(boolean notificationTime, Cluster c,
			NodeCursor nc) {
		if (notificationTime == POST)
			invalidate(c);
	}

	@Override
	protected void mergeEvent(boolean notificationTime, Cluster a, Cluster b) {
		if (notificationTime == POST)
			invalidate(a);
	}

	@Override
	protected void moveEvent(boolean notificationTime, Cluster a, Cluster b,
			NodeCursor nc) {
		if (notificationTime == POST) {
			invalidate(a);
			invalidate(b);
		}
	}

	@Override
	protected void nodeRemovalEvent(boolean notificationTime, Node nodeToRemove) {
		if (notificationTime == POST)
			return;
		Cluster c = getClustering().getCluster(nodeToRemove);
		if (c != null)
			invalidate(c);
	}

	@Override
	protected void splitEvent(boolean notificationTime, Cluster a, Cluster b) {
		if (notificationTime == POST) {
			invalidate(a);
			invalidate(b);
		}
	}

	private class ClusterInfo {
		Graph subgraph = null;

		GraphInterpretation interpretation = null;

		NodeMap reverseNodes = null;

		private EdgeMap reverseEdges = null;

		private Cluster cluster = null;

		ClusterInfo(Cluster aCluster) {
			this.cluster = aCluster;
			subgraph = new Graph();

			// setting up graph
			reverseNodes = subgraph.createNodeMap();
			reverseEdges = subgraph.createEdgeMap();
			for (NodeCursor nc = cluster.members(); nc.ok(); nc.next()) {
				Node aNode = nc.node();
				Node newNode = subgraph.createNode();
				forwardNodes.set(aNode, newNode);
				reverseNodes.set(newNode, aNode);
			}

			for (NodeCursor nc = cluster.members(); nc.ok(); nc.next()) {
				Node node = nc.node();
				for (EdgeCursor ec = node.outEdges(); ec.ok(); ec.next()) {
					Edge anEdge = ec.edge();
					Node targetNode = anEdge.target();
					Cluster targetCluster = getClustering().getCluster(
							targetNode);
					if (targetCluster == cluster) {
						Edge newEdge = subgraph
								.createEdge((Node) forwardNodes.get(node),
										(Node) forwardNodes.get(targetNode));
						forwardEdges.set(anEdge, newEdge);
						reverseEdges.set(newEdge, anEdge);
					}
				}
			}

			interpretation = getClustering().getInterpretation()
					.getPartialInterpretation(subgraph, reverseEdges);

		}

		protected void finalize() {
			subgraph.disposeNodeMap(reverseNodes);
			subgraph.disposeEdgeMap(reverseEdges);
		}
	}

	@Override
	protected String id() {
		return ID;
	}
}