/*
 * Created on Jan 8, 2005
 * Reviewed 1.1 (Jan 10, 2005)
 *
 * Copyright 2004.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering.algorithms;

import y.base.NodeMap;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.addon.Subgraphs;
import de.uka.algo.util.GYCursor;
import de.uka.algo.util.GYList;
import de.uka.algo.util.MappedNodeCursor;

/**
 * A RefiningStage is a Clusterer that iteratively clusters all subgraphs
 * induced by the clusters od the given clustering and splits the clusters
 * according to that. That is, a Refining stage performs the following
 * algorithm: <br>
 * 1. Set up graphs for all subgraphs induced by unmarked clusters (if any) <br>
 * 2. Pass their clusterings (no nodes assigned to clusters) to the core
 * algorithm <br>
 * 3. If the core returns a large cluster containing all nodes, mark the
 * corresponding cluster in the input clustering. Else extract nodes
 * corresponding to free nodes and split the cluster according to the clusters
 * returned by the core algorithm. <br>
 * 4. Repeat steps 1-3 n times <br>
 * 
 * @author Bastian Katz
 */
public class RefiningWrapper extends Wrapper {

    /** number of iterations (depth of the splitting-tree) */
    private int iterations;

    /**
     * Instantiates a new RefiningStage with a core algorithm and a number of
     * iterations
     * 
     * @param core
     *            the core algorithm
     * @param iterations
     *            the (maximum) number of iterations
     */
    public RefiningWrapper(Algorithm core, int iterations) {
        super(core);
        this.iterations = iterations;
    }

    /**
     * Refines a clustering by a core algorithm, which is used to iteratively
     * cluster clusters. Postcondition: Nodes that were not assigned to a
     * cluster when the refining stage is called, are not assigned to a cluster
     * afterwards.
     */
    public void run(Clustering aClustering) {

        /* need to extract subgraphs to cut */
        Subgraphs subgraphs = Subgraphs.getInstance(aClustering);

        /* initialize clusterlist */
        GYList<Cluster> toSplit = new GYList<Cluster>(aClustering.clusters());
        GYList<Cluster> nextToSplit = new GYList<Cluster>();
        int iteration = 1;
        while (!toSplit.isEmpty()) {
            Cluster a = toSplit.popElement();
            if (toSplit.isEmpty() && iteration < iterations) {
                iteration++;
                toSplit = nextToSplit;
                nextToSplit = new GYList<Cluster>();
            }
            Clustering subClustering = new Clustering(subgraphs.getGraph(a), subgraphs
                    .getInterpretation(a));
            NodeMap rev = subgraphs.getReverseNodes(a);
            
            pass(subClustering);

            subClustering.disposeAllListeners();

            if (subClustering.poolSize() == a.size()) {
                aClustering.delete(a);
            } else {
                boolean performedChange = false;
                /* extract free nodes */
                if (subClustering.poolSize() != 0) {
                    aClustering.extract(a, new MappedNodeCursor(subClustering
                            .poolNodes(), subgraphs.getReverseNodes(a)));
                    performedChange = true;
                }
                /* split clusters from aCluster */
                if (subClustering.clusterCount() > 1) {
                    GYCursor<Cluster> cc = subClustering.clusters();
                    cc.next();
                    for (; cc.ok(); cc.next()) {
                        Cluster subCluster = cc.active();
                        Cluster newCluster = aClustering.split(a,
                                new MappedNodeCursor(subCluster.members(), rev));
                        nextToSplit.add(newCluster);
                    }
                    performedChange = true;
                }
                if (performedChange)
                    nextToSplit.add(a);
            }
            subClustering.dispose();
        }
    }
}