/*
 * Created on Jul 6, 2004
 * Reviewed 1.2 (Jan 10, 2005)
 * 
 * Copyright 2004.  --- All rights resevered. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering.algorithms;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.NodeList;
import y.base.YCursor;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.index.Index;
import de.uka.algo.clustering.index.IndexFactory;

/**
 * A SingleLinkageStage tries to optimize a clustering only by a set of merging
 * operations: <br>
 * 1. Store the clusterings quality and the clustering <br>
 * 2. Sort edges (done by implementations - abstract) <br>
 * 3. For all edges (in sorted order) do steps 4-5 <br>
 * 4. If the edge connects two clusters, merge them <br>
 * 5. If the clusterings quality is higher than the stored quality, store bot
 * new quality and the current clustering <br>
 * 6. Reset the clustering to the stored clustering <br>
 * 
 * @author Bastian Katz
 *  
 */
public abstract class SingleLinkage implements Algorithm {

    /** the edge ordering algorithm */
    abstract protected EdgeCursor sortedEdges(Clustering aClustering);

    /** the quality index to optimize */
    abstract protected IndexFactory quality();

    /**
     * Changes the clustering as described above. Does only merge clusters.
     */
    public void run(Clustering aClustering) {

        /* get index */
        Index qualityIndex = quality().getIndex(aClustering);

        /* store initial values */
        double bestQuality = qualityIndex.getValue();
        NodeList[] members = new NodeList[aClustering.clusterCount()];
        int clusterNumber = 0;
        for (YCursor c = aClustering.clusters(); c.ok(); c.next(), clusterNumber++) {
            Cluster cluster = (Cluster) c.current();
            members[clusterNumber] = new NodeList(cluster.members());
        }

        /* get edge ordering */
        EdgeCursor ec = sortedEdges(aClustering);

        /* merge clusters iteratively */
        for (; ec.ok() && aClustering.clusterCount() > 1; ec.next()) {
            Edge anEdge = ec.edge();
            Cluster source = aClustering.getCluster(anEdge.source());
            Cluster target = aClustering.getCluster(anEdge.target());

            if (source != target & source != null & target != null) {
                aClustering.merge(source, target);
                double thisQuality = qualityIndex.getValue();

                /* store better clustering */
                if (thisQuality > bestQuality) {
                    bestQuality = thisQuality;
                    members = new NodeList[aClustering.clusterCount()];
                    clusterNumber = 0;
                    for (YCursor c = aClustering.clusters(); c.ok(); c.next(), clusterNumber++) {
                        Cluster cluster = (Cluster) c.current();
                        members[clusterNumber] = new NodeList(cluster.members());
                    }
                }
            }

        }

        /* restore best clustering */
        aClustering.reset();
        for (clusterNumber = 0; clusterNumber < members.length; clusterNumber++) {
            aClustering.newCluster(members[clusterNumber].nodes());
        }
    }
}