package de.uka.algo.clustering.comparison;

import y.base.Node;
import y.base.NodeCursor;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.DefaultListener;
import de.uka.algo.util.GYCursor;

/**
 * A Confusion Matrix of two clusterings cl1 and cl2 is a |cl1|x|cl2| matrix
 * with each entry representing the number of nodes in the intersection of
 * the corresponding clusters.
 * 
 * @author Daniel Delling
 * 
 */
public class ConfusionMatrix {

	boolean valid = false;

	ConfusionListener cl1Listener;

	ConfusionListener cl2Listener;

	Clustering cl1;

	Clustering cl2;

	int[][] confusionMatrix;
	

	private ConfusionMatrix(Clustering a, Clustering b) {
		cl1Listener = new ConfusionListener(a, b, this);
		cl2Listener = new ConfusionListener(b, a, this);
		cl1 = a;
		cl2 = b;
	}

	/**
	 * returns the confusion matrix of clustering a and b.
	 * 
	 * @param a
	 *            Clustering 1
	 * @param b
	 *            Clustering 2
	 * @return the confusion matrix
	 */
	public static ConfusionMatrix getConfusionMatrix(Clustering a, Clustering b) {
		ConfusionListener l1 = (ConfusionListener) a
				.getClusteringListener(ConfusionListener.staticID(a, b));
		if (l1 != null) {
			return l1.parent;
		}
		return new ConfusionMatrix(a, b);
	}

	private void calculateConfusionMatrix() {
		confusionMatrix = new int[cl1.clusterCount()][cl2.clusterCount()];
		for (GYCursor<Cluster> cc = cl1.clusters(); cc.ok(); cc.next()) {
			Cluster c = cc.active();
			for (NodeCursor nc = c.members(); nc.ok(); nc.next()) {
				Node n = nc.node();
				/* a node in c@cl1 might be unclustered in cl2 */ 
				if (cl2.getCluster(n) != null)
					confusionMatrix[c.getIndex()][cl2.getCluster(n).getIndex()]++;
			}
		}

	}

	private void validate() {
		if (valid)
			return;
		calculateConfusionMatrix();
		valid = true;
	}

	/**
	 * checks, whether the confusion matrix is still valid
	 * 
	 * @return true if the matrix is valid
	 */
	public boolean isValid() {
		return valid;
	}

	/**
	 * return the number of elements in the intersection of cluster c1 and c2
	 * 
	 * @param c1
	 *            a Cluster of clustering 1
	 * @param c2
	 *            a Cluster of clustering 2
	 * @return number of elements in the intersection of c1 and c2
	 */
	public int getIntersectionCardinality(Cluster c1, Cluster c2) {
		validate();
		if (c1.getClustering() == cl1 && c2.getClustering() == cl2) {
			return confusionMatrix[c1.getIndex()][c2.getIndex()];
		} else if (c2.getClustering() == cl1 && c1.getClustering() == cl2) {
			return confusionMatrix[c2.getIndex()][c1.getIndex()];
		} else {
			throw new RuntimeException(
					"c1 and c2 must not be null and must belong to different clusterings");
		}

	}

	/**
	 * returns the number of elements in the maximum intersection with the other
	 * clustering
	 * 
	 * @param c
	 *            the cluster
	 * @return number of elements
	 */
	public int getMaximumIntersectionToCluster(Cluster c) {
		validate();
		if (c.getClustering() == cl1) {
			int max = 0;
			for (int i = 0; i < cl2.clusterCount(); i++) {
				max = Math.max(max, confusionMatrix[c.getIndex()][i]);
			}
			return max;
		} else if (c.getClustering() == cl2) {
			int max = 0;
			for (int i = 0; i < cl1.clusterCount(); i++) {
				max = Math.max(max, confusionMatrix[i][c.getIndex()]);
			}
			return max;
		} else {
			throw new RuntimeException(
					"c must belong to clustering1 or clustering2!");
		}
	}
	
	
	
	/**
	 * The sum of all entries in the matrix.
	 * If two clusterings with different V are compared,
	 * this is the smaller |V|.
	 * @return	sum of all entries
	 */
	public int N() {
		this.validate();
		int n = 0;
		for (int[] a : this.confusionMatrix) {
			for (int i : a) {
				n += i;
			}
		}
		return n;
	}
	
	
	

	private static class ConfusionListener extends DefaultListener {

		ConfusionMatrix parent;

		private Clustering opposite;

		protected ConfusionListener(Clustering clustering, Clustering opposite,
				ConfusionMatrix parent) {
			super(clustering);
			this.parent = parent;
			this.opposite = opposite;
			clustering.addListener(this);
		}

		@Override
		protected void defaultEvent() {
			parent.valid = false;
		}

		@Override
		protected void policyChanged(boolean policy) {
			// do nothing
		}

		@Override
		protected String id() {
			return staticID(getClustering(), opposite);
		}

		protected static String staticID(Clustering clustering,
				Clustering opposite) {
			return "ConfusionListener:" + clustering + "" + opposite;
		}
		
		
		


	}

}
