package de.uka.algo.clustering.comparison;

import y.base.Edge;
import y.base.EdgeCursor;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.DefaultListener;

/**
 * This class implements the connected counting pair sets according to two
 * clusterings. Each pair of nodes connected by an edge in the underlying graph
 * is classified whether they are in the same (or different) cluster with
 * respect to both clusterings
 * 
 * @author Daniel Delling
 * 
 */
public class ConnectedCountingPairSet {

	boolean valid = false;

	ConnectedCountingListener cl1Listener;

	ConnectedCountingListener cl2Listener;

	Clustering cl1;

	Clustering cl2;

	private int e11 = 0;

	private int e01 = 0;

	private int e10 = 0;

	private int e00 = 0;

	private ConnectedCountingPairSet(Clustering a, Clustering b) {
		cl1Listener = new ConnectedCountingListener(a, b, this);
		cl2Listener = new ConnectedCountingListener(b, a, this);
		cl1 = a;
		cl2 = b;

	}

	/**
	 * creates a new ConnectedCountingPairSet if needed. If there already exists
	 * a (valid) counting pair set with respect to clustering a and b, the
	 * existing is returned.
	 * 
	 * @param a
	 *            Clustering a
	 * @param b
	 *            Clustering b
	 * @return a ConnectedCountingPairSet
	 */
	public static ConnectedCountingPairSet getConnectedCountingPairSet(
			Clustering a, Clustering b) {
		ConnectedCountingListener l1 = (ConnectedCountingListener) a
				.getClusteringListener(ConnectedCountingListener.staticID(b));
		if (l1 != null) {
			return l1.parent;
		}
		return new ConnectedCountingPairSet(a, b);
	}

	private void calculateCountingPairSets() {
		e11 = 0;
		e10 = 0;
		e01 = 0;
		e00 = 0;
		for (EdgeCursor ec = cl1.getGraph().edges(); ec.ok(); ec.next()) {
			Edge e = ec.edge();
	
			/* edges which do not exist in both graphs will be ignored */
			if (cl1.isIntraClusterEdge(e)) {
				// E11
				if (cl2.isIntraClusterEdge(e)) {
					e11++;
				} else {
					e10++;
				}
			} else { // E0i
				// E01
				if (cl2.isIntraClusterEdge(e)) {
					e01++;
				} else {
					e00++;
				}
			}
			


		}
	}

	private void validate() {
		if (valid)
			return;
		calculateCountingPairSets();
		valid = true;
	}

	/**
	 * returns the number of connected pairs in different clusters with respect
	 * to both clusterings
	 * 
	 * @return number of pairs
	 */
	public int getE00() {
		validate();
		return e00;
	}

	/**
	 * returns the number of connected pairs in different clusters according to
	 * cl1 and in same clusterings according to cl2
	 * 
	 * @return number of pairs
	 */
	public int getE01() {
		validate();
		return e01;
	}

	/**
	 * returns the number of connected pairs in the same clusters according to
	 * cl1 and different clusterings according to cl2
	 * 
	 * @return number of pairs
	 */
	public int getE10() {
		validate();
		return e10;
	}

	/**
	 * returns the number of connected pairs in the same clusters with respect
	 * to both clusterings
	 * 
	 * @return number of pairs
	 */
	public int getE11() {
		validate();
		return e11;
	}

	public int getNumberOfAllConnectedPairs() {
		validate();
		return e00 + e01 + e10 + e11;
	}

	public static class ConnectedCountingListener extends DefaultListener {

		ConnectedCountingPairSet parent;

		private Clustering opposite;

		protected ConnectedCountingListener(Clustering clustering,
				Clustering opposite, ConnectedCountingPairSet parent) {
			super(clustering);
			this.parent = parent;
			this.opposite = opposite;
			clustering.addListener(this);
		}

		@Override
		protected void defaultEvent() {
			parent.valid = false;
		}

		@Override
		protected void policyChanged(boolean policy) {
			// do nothing
		}

		@Override
		protected String id() {
			return staticID(opposite);
		}

		protected static String staticID(Clustering opposite) {
			return "ConnectedCountingListener:" + opposite;
		}

	}

}
