/*
 * 
 */

package de.uka.algo.clustering.comparison;

import y.base.Node;
import y.base.NodeCursor;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.DefaultListener;

/**
 * This class implements the counting pair sets according to two clusterings.
 * Each pair fo nodes is classified whether they are in the same (or different)
 * cluster with respect to both clusterings
 * 
 * @author Daniel Delling
 * 
 */
public class CountingPairSet {

	boolean valid = false;

	CountingListener cl1Listener;

	CountingListener cl2Listener;

	Clustering cl1;

	Clustering cl2;

	private int n11 = 0;

	private int n01 = 0;

	private int n10 = 0;

	private int n00 = 0;

	private CountingPairSet(Clustering a, Clustering b) {
		cl1Listener = new CountingListener(a, b, this);
		cl2Listener = new CountingListener(b, a, this);
		cl1 = a;
		cl2 = b;
		
	}

	/**
	 * creates a new CountingPairSet if needed. If there already exists a (valid) counting pair set with respect to clustering a and b, the existing is returned.
	 * @param a Clustering a
	 * @param b Clustering b
	 * @return a CountingPairSet
	 */
	public static CountingPairSet getCountingPairSet(Clustering a, Clustering b) {
		CountingListener l1 = (CountingListener) a
				.getClusteringListener(CountingListener.staticID(b));
		if (l1 != null) {
			return l1.parent;
		}
		return new CountingPairSet(a, b);
	}

	private void calculateCountingPairSets() {
		n11 = 0;
		n10 = 0;
		n01 = 0;
		n00 = 0;

		for (NodeCursor nc1 = cl1.getGraph().nodes(); nc1.ok(); nc1.next()) {
			for (NodeCursor nc2 = cl2.getGraph().nodes(); nc2.node() != nc1
					.node(); nc2.next()) {
				Node n1 = nc1.node();
				Node n2 = nc2.node();

				Cluster c11 = cl1.getCluster(n1);
				Cluster c12 = cl1.getCluster(n2);
				Cluster c21 = cl2.getCluster(n1);
				Cluster c22 = cl2.getCluster(n2);

				/**
				 * Modified counting in order to take care of unclustered nodes.
				 * @author Christian Staudt
				 */
				
				if ((c11 == null) || (c21 == null) || (c12 == null) || (c22 == null)) {
					// ignore pairs with unclustered nodes
				} else {
					if (c11 == c12) { // n1x
						if (c21 == c22) { // n11
							n11++;
						} else { // n10
							n10++;
						}
					} else {// n0x
						if (c21 == c22) { // n01
							n01++;
						} else { // n00
							n00++;
						}
					}
				}
				
				//
				
				
//				if (c11 == c12) { // n1x
//					if (c21 == c22) { // n11
//						n11++;
//					} else { // n10
//						n10++;
//					}
//				} else {// n0x
//					if (c21 == c22) { // n01
//						n01++;
//					} else { // n00
//						n00++;
//					}
//				}
				
			}
		}
	}

	private void validate() {
		if (valid)
			return;
		calculateCountingPairSets();
		valid = true;
	}
	
	/**
	 * returns the number of pairs in different clusters with respect to both clusterings
	 * @return number of pairs
	 */
	public int getN00() {
		validate();
		return n00;
	}

	/**
	 * returns the number of pairs in different clusters according to cl1 and in same clusterings according to cl2
	 * @return number of pairs
	 */
	public int getN01() {
		validate();
		return n01;
	}

	/**
	 * returns the number of pairs in the same clusters according to cl1 and different clusterings according to cl2
	 * @return number of pairs
	 */
	public int getN10() {
		validate();
		return n10;
	}
	
	/**
	 * returns the number of pairs in the same clusters with respect to both clusterings
	 * @return number of pairs
	 */
	public int getN11() {
		validate();
		return n11;
	}

	public int getNumberOfAllPairs() {
		validate();
		return n00+n01+n10+n11;
	}
	
	public static class CountingListener extends DefaultListener {

		CountingPairSet parent;
		private Clustering opposite;

		protected CountingListener(Clustering clustering, Clustering opposite, CountingPairSet parent) {
			super(clustering);
			this.parent = parent; 
			this.opposite = opposite;
			clustering.addListener(this);
		}

		@Override
		protected void defaultEvent() {
			parent.valid = false;
		}

		@Override
		protected void policyChanged(boolean policy) {
			// do nothing
		}

		@Override
		protected String id() {
			return staticID(opposite);
		}

		protected static String staticID(Clustering opposite) {
			return "CountingListener:"+opposite;
		}

	}

}
