package de.uka.algo.clustering.comparison;

import y.base.Node;
import y.base.NodeCursor;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.DefaultListener;
import de.uka.algo.util.GYCursor;
/**
 * A Degree Confusion Matrix of two clusterings cl1 and cl2 is a |cl1|x|cl2| matrix
 * with each entry representing the sum of degrees of the node in the intersection of
 * the corresponding clusters.
 * 
 * @author Daniel Delling
 * 
 */
public class DegreeConfusionMatrix {

	boolean valid = false;

	DegreeConfusionListener cl1Listener;

	DegreeConfusionListener cl2Listener;

	Clustering cl1;

	Clustering cl2;

	int[][] degreeConfusionMatrix;

	private DegreeConfusionMatrix(Clustering a, Clustering b) {
		cl1Listener = new DegreeConfusionListener(a, b, this);
		cl2Listener = new DegreeConfusionListener(b, a, this);
		cl1 = a;
		cl2 = b;
	}

	public static DegreeConfusionMatrix getDegreeConfusionMatrix(Clustering a, Clustering b) {
		DegreeConfusionListener l1 = (DegreeConfusionListener) a
				.getClusteringListener(DegreeConfusionListener.staticID(a, b));
		if (l1 != null) {
			return l1.parent;
		}
		return new DegreeConfusionMatrix(a, b);
	}

	/**
	 * returns the degree confusion matrix of clustering a and b.
	 * 
	 * @param a
	 *            Clustering 1
	 * @param b
	 *            Clustering 2
	 * @return the degree confusion matrix
	 */
	private void calculateDegreeConfusionMatrix() {
		degreeConfusionMatrix = new int[cl1.clusterCount()][cl2.clusterCount()];
		for (GYCursor<Cluster> cc = cl1.clusters(); cc.ok(); cc.next()) {
			Cluster c = cc.active();
			for (NodeCursor nc = c.members(); nc.ok(); nc.next()) {
				Node n = nc.node();
				// TODO: check for correctness
				if (cl2.getCluster(n) != null) {
					// node exists in 2nd clustering
					degreeConfusionMatrix[c.getIndex()][cl2.getCluster(n).getIndex()] += n.degree();
				}
				// ?
			}
		}

	}

	private void validate() {
		if (valid)
			return;
		calculateDegreeConfusionMatrix();
		valid = true;
	}

	/**
	 * checks, whether the degree confusion matrix is still valid
	 * 
	 * @return true if the matrix is valid
	 */
	public boolean isValid() {
		return valid;
	}

	/**
	 * return the sum of degrees of the nodes in the intersection of cluster c1 and c2
	 * 
	 * @param c1
	 *            a Cluster of clustering 1
	 * @param c2
	 *            a Cluster of clustering 2
	 * @return sum of degrees of the nodes in the intersection of c1 and c2
	 */
	public int getIntersectionCardinality(Cluster c1, Cluster c2) {
		validate();
		if (c1.getClustering() == cl1 && c2.getClustering() == cl2) {
			return degreeConfusionMatrix[c1.getIndex()][c2.getIndex()];
		} else if (c2.getClustering() == cl1 && c1.getClustering() == cl2) {
			return degreeConfusionMatrix[c2.getIndex()][c1.getIndex()];
		} else {
			throw new RuntimeException(
					"c1 and c2 must not be null and must belong to different clusterings");
		}

	}

	/**
	 * returns the sum of degrees of the nodes in the maximum intersection with the other
	 * clustering
	 * 
	 * @param c
	 *            the cluster
	 * @return the sum of degrees of the nodes
	 */
	public int getMaximumIntersectionToCluster(Cluster c) {
		validate();
		if (c.getClustering() == cl1) {
			int max = 0;
			for (int i = 0; i < cl2.clusterCount(); i++) {
				max = Math.max(max, degreeConfusionMatrix[c.getIndex()][i]);
			}
			return max;
		} else if (c.getClustering() == cl2) {
			int max = 0;
			for (int i = 0; i < cl1.clusterCount(); i++) {
				max = Math.max(max, degreeConfusionMatrix[i][c.getIndex()]);
			}
			return max;
		} else {
			throw new RuntimeException(
					"c must belong to clustering1 or clustering2!");
		}
	}

	private static class DegreeConfusionListener extends DefaultListener {

		DegreeConfusionMatrix parent;

		private Clustering opposite;

		protected DegreeConfusionListener(Clustering clustering, Clustering opposite,
				DegreeConfusionMatrix parent) {
			super(clustering);
			this.parent = parent;
			this.opposite = opposite;
			clustering.addListener(this);
		}

		@Override
		protected void defaultEvent() {
			parent.valid = false;
		}

		@Override
		protected void policyChanged(boolean policy) {
			// do nothing
		}

		@Override
		protected String id() {
			return staticID(getClustering(), opposite);
		}

		protected static String staticID(Clustering clustering,
				Clustering opposite) {
			return "DegreeConfusionListener:" + clustering + "" + opposite;
		}

	}

}
