package de.uka.algo.clustering.comparison;

import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.DefaultListener;
import de.uka.algo.util.GYCursor;

/**
 * The MutualEdgeInformation of two clusterings is the loss of (edgebased)
 * uncertanity if one clustering is known.
 * 
 * @author Daniel Delling
 * 
 */
public class MutualEdgeInformation {

	boolean valid = false;

	MutuaEdgelListener cl1Listener;

	MutuaEdgelListener cl2Listener;

	Clustering cl1;

	Clustering cl2;

	private double mutualEdgeInfo = 0;

	DegreeConfusionMatrix dcm;

	private MutualEdgeInformation(Clustering a, Clustering b) {
		cl1Listener = new MutuaEdgelListener(a, b, this);
		cl2Listener = new MutuaEdgelListener(b, a, this);
		cl1 = a;
		cl2 = b;
		dcm = DegreeConfusionMatrix.getDegreeConfusionMatrix(cl1, cl2);
	}

	/**
	 * returns the MutualEdgeInformation of two clusterings
	 * 
	 * @param a
	 *            the first clustering
	 * @param b
	 *            the second clustering
	 * @return the mutual edgebased information
	 */
	public static MutualEdgeInformation getMutualEdgeInformation(Clustering a,
			Clustering b) {
		MutuaEdgelListener l1 = (MutuaEdgelListener) a
				.getClusteringListener(MutuaEdgelListener.staticID(b));
		if (l1 != null) {
			return l1.parent;
		}
		return new MutualEdgeInformation(a, b);
	}

	private void calculateMutualInformation() {
		double p_i = 0;
		double p_j = 0;
		double p_ij = 0;
		mutualEdgeInfo = 0;
		for (GYCursor<Cluster> cc1 = cl1.clusters(); cc1.ok(); cc1.next()) {
			Cluster c1 = cc1.active();
			int c1DegreeSum = 2 * cl1.getInnerEdgeCount(c1)
					+ cl1.getBorderEdgeCount(c1);
			p_i = (double) c1DegreeSum / (double) (2 * cl1.getGraph().E());
			for (GYCursor<Cluster> cc2 = cl2.clusters(); cc2.ok(); cc2.next()) {
				Cluster c2 = cc2.active();
				int c2DegreeSum = 2 * cl2.getInnerEdgeCount(c2)
						+ cl2.getBorderEdgeCount(c2);
				p_j = (double) c2DegreeSum / (double) (2 * cl2.getGraph().E());
				p_ij = (double) dcm.getIntersectionCardinality(c1, c2)
						/ (double) (2 * cl2.getGraph().E());
				if (p_ij != 0.0)
					mutualEdgeInfo += p_ij * Math.log10(p_ij / (p_i * p_j))
							/ Math.log10(2);
			}
		}

	}

	private void validate() {
		if (valid)
			return;
		calculateMutualInformation();
		valid = true;
	}

	/**
	 * return the valid value of the edgeBased mutual information
	 * @return the edgebased mutual information
	 */
	public double getMutualEdgeInfo() {
		validate();
		return mutualEdgeInfo;
	}

	public static class MutuaEdgelListener extends DefaultListener {

		MutualEdgeInformation parent;

		private Clustering opposite;

		protected MutuaEdgelListener(Clustering clustering,
				Clustering opposite, MutualEdgeInformation parent) {
			super(clustering);
			this.parent = parent;
			this.opposite = opposite;
			clustering.addListener(this);
		}

		@Override
		protected void defaultEvent() {
			parent.valid = false;
		}

		@Override
		protected void policyChanged(boolean policy) {
			// do nothing
		}

		@Override
		protected String id() {
			return staticID(opposite);
		}

		protected static String staticID(Clustering opposite) {
			return "MutualEdgeListener:" + opposite;
		}

	}

}
