package de.uka.algo.clustering.comparison;

import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.DefaultListener;
import de.uka.algo.util.GYCursor;

/**
 * The MutualInformation of two clusterings is the loss of uncertanity if one
 * clustering is known.
 * 
 * @author Daniel Delling
 * 
 */
public class MutualInformation {

	boolean valid = false;

	MutualListener cl1Listener;

	MutualListener cl2Listener;

	Clustering cl1;

	Clustering cl2;

	private double mutualInfo = 0;

	ConfusionMatrix cm;

	private MutualInformation(Clustering a, Clustering b) {
		cl1Listener = new MutualListener(a, b, this);
		cl2Listener = new MutualListener(b, a, this);
		cl1 = a;
		cl2 = b;
		cm = ConfusionMatrix.getConfusionMatrix(cl1, cl2);
	}

	/**
	 * returns the MutualInformation of two clusterings
	 * 
	 * @param a
	 *            the first clustering
	 * @param b
	 *            the second clustering
	 * @return the mutual information
	 */
	public static MutualInformation getMutualInformation(Clustering a,
			Clustering b) {
		MutualListener l1 = (MutualListener) a
				.getClusteringListener(MutualListener.staticID(b));
		if (l1 != null) {
			return l1.parent;
		}
		return new MutualInformation(a, b);
	}

	private void calculateMutualInformation() {
		double p_i = 0;
		double p_j = 0;
		double p_ij = 0;
		mutualInfo = 0;
		for (GYCursor<Cluster> cc1 = cl1.clusters(); cc1.ok(); cc1.next()) {
			Cluster c1 = cc1.active();
			p_i = (double) c1.size() / (double) cl1.getGraph().N();
			for (GYCursor<Cluster> cc2 = cl2.clusters(); cc2.ok(); cc2.next()) {
				Cluster c2 = cc2.active();
				p_j = (double) c2.size() / (double) cl2.getGraph().N();
				p_ij = (double) cm.getIntersectionCardinality(c1, c2)
						/ (double) cl2.getGraph().N();
				if (p_ij != 0.0)
					mutualInfo += p_ij * Math.log10(p_ij / (p_i * p_j))
							/ Math.log10(2);
			}
		}

	}

	private void validate() {
		if (valid)
			return;
		calculateMutualInformation();
		valid = true;
	}

	/**
	 * return the mutual information as double
	 * 
	 * @return the mutual information
	 */
	public double getMutualInfo() {
		validate();
		return mutualInfo;
	}

	public static class MutualListener extends DefaultListener {

		MutualInformation parent;

		private Clustering opposite;

		protected MutualListener(Clustering clustering, Clustering opposite,
				MutualInformation parent) {
			super(clustering);
			this.parent = parent;
			this.opposite = opposite;
			clustering.addListener(this);
		}

		@Override
		protected void defaultEvent() {
			parent.valid = false;
		}

		@Override
		protected void policyChanged(boolean policy) {
			// do nothing
		}

		@Override
		protected String id() {
			return staticID(opposite);
		}

		protected static String staticID(Clustering opposite) {
			return "MutualListener:" + opposite;
		}

	}

}
