/*
 * created Apr 28, 2006
 *
 * Copyright 2006.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 * 
 */
package de.uka.algo.clustering.generators;

import java.util.Random;

import y.base.Graph;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.graphs.GraphInterpretation;

/**
 * super class for preclustered graphs. A clustered graph generates an initial
 * clustering. The significance is tunable by the parameters. This superclass
 * provides the possibility of embedding nodes in order to create clusterings
 * based on geometrical properties.
 * 
 * @author Daniel Delling
 * 
 */
public abstract class ClusteredGraph extends Clustering {

	protected Random ran;

	/**
	 * standard constructer, the clustered graph will be interpreted as
	 * unweighted
	 * 
	 * @param g
	 *            the graph the clustering should be assigned to
	 */
	public ClusteredGraph(Graph g) {
		super(g);
		ran = new Random();
		g.clear();
	}

	/**
	 * creates a new clustered graph with given interpretation
	 * 
	 * @param g
	 *            the graph the clustering is assigned to
	 * @param gi
	 *            the interpretation of the graph (e.g. weighted or undirected)
	 */
	public ClusteredGraph(Graph g, GraphInterpretation gi) {
		super(g, gi);
		ran = new Random();
		g.clear();
	}

	/**
	 * returns Information about the implemented Generator
	 * 
	 * @return the identification String of the Generator
	 */
	public abstract String getID();

	@Override
	public String toString() {
		return getID();
	}

	/**
	 * Returns a randomly chosen numbers of clusters from the interval
	 * <code>[log(n), sqrt(n)]</code>, where <code>n</code> is the number
	 * of nodes.
	 * 
	 * @param numberOfNodes
	 *            given number of nodes
	 * @return randomly chosen number of clusters.
	 */
	protected int getRandomNumberOfClusters(int numberOfNodes) {
		int k = (int) (Math.floor(Math.sqrt(numberOfNodes))
				- Math.floor(Math.log10(numberOfNodes)) + 1);
		k = ran.nextInt(k) + (int) Math.floor(Math.log10(numberOfNodes));
		return Math.max(k, 2);
	}
}
