/*
 * Created on May 31, 2006
 *
 * Copyright 2006. --- All rights resevered. 
 * Algorithmic Group, Karlsruhe, Germany.
 * 
 * project: base
 * package: de.uka.algo.clustering.generators
 */
package de.uka.algo.clustering.generators;

import de.uka.algo.graphs.GraphInterpretation;
import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeMap;
import y.geom.YPoint;

/**
 * @author Marco Gaertler
 * 
 */
public abstract class EmbeddedClusteredGraph extends ClusteredGraph {

	private NodeMap xCoord;

	private NodeMap yCoord;

	public EmbeddedClusteredGraph(Graph graph) {
		super(graph);
		xCoord = graph.createNodeMap();
		yCoord = graph.createNodeMap();
	}

	public EmbeddedClusteredGraph(Graph graph, GraphInterpretation gi) {
		super(graph, gi);
		xCoord = graph.createNodeMap();
		yCoord = graph.createNodeMap();
	}

	@Override
	public void dispose() {
		getGraph().disposeNodeMap(xCoord);
		getGraph().disposeNodeMap(yCoord);
		super.dispose();
	}

	/**
	 * Assigns new coordinates to a given node.
	 * 
	 * @param v
	 *            given node
	 * @param x
	 *            new x-coordinate
	 * @param y
	 *            new y-coordinate
	 */
	public void setPosition(Node v, double x, double y) {
		if (v != null) {
			xCoord.setDouble(v, x);
			yCoord.setDouble(v, y);
		}
	}

	/**
	 * Assigns new coordinates to a given node.
	 * 
	 * @param v
	 *            given node
	 * @param p
	 *            new position
	 */
	public void setPosition(Node v, YPoint p) {
		if (v != null && p != null) {
			setPosition(v, p.getX(), p.getY());
		}
	}

	/**
	 * Returns the position of a give node.
	 * 
	 * @param v
	 *            given node
	 * @return position as <code>double</code> array of length 2
	 */
	public double[] getPosition(Node v) {
		return new double[] { xCoord.getDouble(v), yCoord.getDouble(v) };
	}

	/**
	 * Returns the position of a give node.
	 * 
	 * @param v
	 *            given node
	 * @return position as <code>YPoint</code>
	 */
	public YPoint getPositionAsPoint(Node v) {
		return new YPoint(xCoord.getDouble(v), yCoord.getDouble(v));
	}

	/**
	 * Returns the Euclidean distance between two given nodes.
	 * 
	 * @param v
	 *            first node
	 * @param w
	 *            second node
	 * @return Euclidean distance between node <code>v</code> and
	 *         <code>w</code>
	 */
	protected double distance(Node v, Node w) {
		if (v != null && w != null) {
			return YPoint.distance(xCoord.getDouble(v), yCoord.getDouble(v),
					xCoord.getDouble(w), yCoord.getDouble(w));
		}
		return Double.MAX_VALUE;
	}

	/**
	 * Returns the Euclidean distance between a position and a given node.
	 * 
	 * @param p
	 *            given position
	 * @param v
	 *            given node
	 * @return Euclidean distance between node <code>w</code> and position
	 *         <code>p</code>
	 */
	protected double distance(YPoint p, Node w) {
		if (p != null && w != null) {
			return YPoint.distance(p, getPositionAsPoint(w));
		}
		return Double.MAX_VALUE;
	}

	/**
	 * Rescales the embedding to a suitable size.
	 */
	protected void rescale() {
		double scale = Math.sqrt(getGraph().nodeCount());
		for (NodeCursor c = getGraph().nodes(); c.ok(); c.next()) {
			YPoint p = getPositionAsPoint(c.node());
			p = new YPoint( scale * p.getX() , scale * p.getY() );
			setPosition(c.node(), p);
		}
	}

}
