/*
 * Created on May 31, 2006
 *
 * Copyright 2006. --- All rights resevered. 
 * Algorithmic Group, Karlsruhe, Germany.
 * 
 * project: base
 * package: de.uka.algo.clustering.generators
 */
package de.uka.algo.clustering.generators;

import java.util.List;
import java.util.Vector;

import y.base.Edge;
import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeList;
import y.geom.YPoint;
import de.uka.algo.graphs.GraphInterpretation;

/**
 * @author Marco Gaertler
 * 
 */
public abstract class GeometricClusteredGraph extends EmbeddedClusteredGraph {

	protected double stretchFactor;

	public GeometricClusteredGraph(Graph graph) {
		super(graph);
		stretchFactor = 100.0;
	}

	public GeometricClusteredGraph(Graph graph, GraphInterpretation gi) {
		super(graph, gi);
		stretchFactor = 100.0;
	}

	/**
	 * Returns random position in the square <code>[0, 100]x[0,100]</code>
	 * 
	 * @return random position
	 */
	protected double[] getRandomPosition() {
		YPoint p = getRandomPositionAsPoint();
		return new double[] { p.getX(), p.getY() };
	}

	/**
	 * Returns random position in the square <code>[0, 100]x[0,100]</code>.
	 * 
	 * @return random position
	 */
	protected YPoint getRandomPositionAsPoint() {
		return new YPoint(stretchFactor * ran.nextDouble(), stretchFactor
				* ran.nextDouble());
	}

	/**
	 * Returns the node (from a given set) with the minimum distance to a given
	 * position.
	 * 
	 * @param pos
	 *            given position
	 * @param possibleNodes
	 *            candidates of nodes
	 * @return node with minimum distance
	 */
	protected Node getClosestNode(YPoint pos, NodeCursor possibleNodes) {
		Node closestNode = null;
		double distance = Double.MAX_VALUE;

		if (possibleNodes == null)
			possibleNodes = getGraph().nodes();

		for (possibleNodes.toFirst(); possibleNodes.ok(); possibleNodes.next()) {
			Node current = possibleNodes.node();
			double currentDistance = distance(pos, current);
			if (distance > currentDistance) {
				closestNode = current;
				distance = currentDistance;
			}
		}
		return closestNode;
	}

	/**
	 * Returns the node with the minimum distance to a given position.
	 * 
	 * @param pos
	 *            given position
	 * @return node with minimum distance
	 */
	protected Node getClosestNode(YPoint pos) {
		return getClosestNode(pos, getGraph().nodes());
	}

	/**
	 * Returns the node (from a given set) with the minimum distance to a given
	 * position.
	 * 
	 * @param pos
	 *            given position
	 * @param possibleNodes
	 *            candidates of nodes
	 * @return node with minimum distance
	 */
	protected YPoint getClosestNode(YPoint pos, List<YPoint> otherPositions) {
		YPoint closestPoint = null;
		double distance = Double.MAX_VALUE;

		if (otherPositions == null)
			return null;

		for (YPoint currentP : otherPositions) {
			if (currentP.distanceTo(pos) < distance) {
				closestPoint = currentP;
				distance = currentP.distanceTo(pos);
			}
		}

		return closestPoint;
	}

	/**
	 * Returns a list with random positions. The
	 * pairwise distance between these positions is at least
	 * <code>minimumDistance</code>.
	 * 
	 * @param k
	 *            number of positions 
	 * @param minimumDistance
	 *            minimum distance between elements
	 * @return list with positions 
	 */
	protected List<YPoint> getRandomPositionDistribution(int k,
			double minimumDistance) {
		List<YPoint> clusterCenterPositions = new Vector<YPoint>(k);

		for (int i = 0; i < k; ++i) {
			boolean valid = true;
			do {
				valid = true;
				// create new position
				YPoint currentPos = getRandomPositionAsPoint();
				// check distance to other points
				for (YPoint other : clusterCenterPositions) {
					valid &= (currentPos.distanceTo(other) >= minimumDistance);
				}
				if (valid)
					clusterCenterPositions.add(currentPos);
			} while (!valid);
		}

		return clusterCenterPositions;
	}

	/**
	 * Creates new nodes that are placed at the given positions and returns the
	 * set of created nodes.
	 * 
	 * @param positions
	 *            given positions
	 * @return list of nodes that were created
	 */
	protected NodeList insertNodesAtPositions(List<YPoint> positions) {
		NodeList nodes = new NodeList();

		for (YPoint p : positions) {
			Node v = getGraph().createNode();
			nodes.add(v);
			setPosition(v, p);
		}

		return nodes;
	};

	/**
	 * Creates edges between the given nodes that are closer than
	 * <code>maxDistance</code>
	 * 
	 * @param nodes
	 *            given node set
	 * @param maxDistance
	 *            maximum distance between two nodes that should be connected by
	 *            an edge
	 */
	protected void createEdgeset(NodeList nodes, double maxDistance) {
		for (NodeCursor c = nodes.nodes(); c.ok(); c.next()) {
			Node v = c.node();

			boolean reachedCurrentNode = false;
			for (NodeCursor c2 = nodes.nodes(); c2.ok(); c2.next()) {
				if (!reachedCurrentNode) {
					reachedCurrentNode = (c2.node() == v);
				} else {
					if (distance(v, c2.node()) < maxDistance) {
						Edge e = v.getEdge(c2.node());
						if (e == null)
							getGraph().createEdge(v, c2.node());
					}
				}
			}
		}
	}

	/**
	 * Creates a voronoi embedding with <code>k</code> center nodes and
	 * <code>n</code> that are placed in the cells. The pairwise distance
	 * between centers is at least <code>minDistance</code>.
	 * 
	 * @param k
	 *            number of centers
	 * @param n
	 *            number of nodes in the cell
	 * @param minDistance
	 *            minimum distance between centers
	 * @return list of all centers
	 */
	protected NodeList createVoronoiEmbedding(int k, int n, double minDistance) {
		List<YPoint> attractorPos = getRandomPositionDistribution(k, minDistance);
		NodeList attractorNodes = insertNodesAtPositions(attractorPos);
		pool2Singletons();

		for (int i = 0; i < n; ++i) {
			boolean validPosition = true;
			do {
				validPosition = true;
				YPoint pos = getRandomPositionAsPoint();
				Node clusterCenter = getClosestNode(pos, attractorNodes.nodes());
				double distance = distance(pos, clusterCenter);
				double p = ran.nextDouble()*minDistance;
				if (distance > p) {
					validPosition = false;
				} else {
					Node v = getGraph().createNode();
					setPosition(v, pos);
					add(getCluster(clusterCenter), v);
				}
			} while (!validPosition);
		}

		return attractorNodes;
	}

}