/*
 * Created on Apr 2, 2008
 *
 * Copyright 2008. --- All rights resevered. 
 * Algorithmic Group, Karlsruhe, Germany.
 * 
 * project: base
 * package: de.uka.algo.clustering.index
 */
package de.uka.algo.clustering.index;

import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.util.GYCursor;

/**
 * This class is a template for clustering indices that basically evaluate a
 * quality function for each cluster and then agglomerate this information into
 * a single value. Current supported types are <i>best case</i>, <i>average
 * case</i> and <i>worst case</i>.
 * 
 * @author Marco Gaertler
 * 
 */
public abstract class AgglomerationIndex extends ClusterBufferedIndex<double[]> {

	/**
	 * Supported types of agglomeration.
	 */
	public enum Operation {
		Minimum, Average, Maximum
	};

	protected Operation op;

	/**
	 * Standard constructor taking a (reference) clustering and a type of
	 * agglomeration.
	 * 
	 * @param clustering
	 *            reference clustering
	 * @param op
	 *            type of agglomeration.
	 */
	protected AgglomerationIndex(Clustering clustering, Operation op) {
		super(clustering);
		this.op = op;
	}

	/**
	 * Returns the agglomerated value of a given list of values.
	 * 
	 * @param values
	 *            list of values
	 * @return agglomerated results according to given type. If an empty list or
	 *         no type is given, then 0 is returned.
	 */
	protected double merge(GYCursor<double[]> values, boolean isWeighted) {

		int index = (isWeighted ? 1 : 0);
		double result = 0;

		switch (op) {
		case Minimum:
		case Maximum:
			boolean first = true;
			for (GYCursor<Cluster> cc = clustering.clusters(); cc.ok(); cc
					.next()) {
				Cluster c = cc.active();
				double currentValue = info.get(c)[index];
				if (first) {
					result = currentValue;
					first = false;
				} else {
					if (op == Operation.Minimum) {
						if (currentValue < result)
							result = currentValue;
					} else {
						if (currentValue > result)
							result = currentValue;
					}
				}
			}
			return result;
		case Average:
			double size = 0;
			for (GYCursor<Cluster> cc = clustering.clusters(); cc.ok(); cc
					.next()) {
				Cluster c = cc.active();
				double currentValue = info.get(c)[index];
				result += currentValue;
				++size;
			}
			return result / (size);
		}
		return 0;
	}

	@Override
	protected double[] getData(Cluster c) {
		double[] results = new double[2];
		results[0] = getClusterValue(c, false);
		results[1] = getClusterValue(c, true);
		return results;
	}

	@Override
	protected double unweightedValue(GYCursor<double[]> infos) {
		return postprocessResult(merge(infos, false), false);
	}

	@Override
	protected double value(GYCursor<double[]> infos) {
		return postprocessResult(merge(infos, true), true);
	}

	/**
	 * Abstract method that returns the quality of a given cluster.
	 * 
	 * @param c
	 *            given cluster
	 * @param isWeighted
	 *            flag indicating if (edge-)weights should be considered. If
	 *            <code>true</code> then weights should be used, else not.
	 * @return the quality of the given cluster
	 * @note The cluster will always be contained in the references clustering.
	 */
	protected abstract double getClusterValue(Cluster c, boolean isWeighted);

	/**
	 * Method to post process the agglomerated result.
	 * 
	 * @param value
	 *            agglomerated result of the quality of the clusters
	 * @param isWeighted
	 *            flag indicating if (edge-)weights should be considered. If
	 *            <code>true</code> then weights should be used, else not.
	 * @return the modified result
	 * @note By default this function will be the identity.
	 */
	protected double postprocessResult(double value, boolean isWeighted) {
		return value;
	}

}
