/*
 * Created on 14.10.2004
 *
 * Copyright 2004.  --- All rights resevered.
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering.index;

import de.uka.algo.clustering.Clustering;

/**
 * A MixedIndex is an Index returning the mean (arithmetic or geometric) of some
 * other indices.
 *
 * @author Bastian Katz
 */
public class MixedIndex implements Index {

	private Index[] indizes;

	double[] weights;

	boolean geometric;

	protected MixedIndex(Clustering clustering, IndexFactory[] factories,
			double[] weights, boolean geometric) {
		indizes = new Index[factories.length];
		this.weights = (double[]) weights.clone();
		for (int i = 0; i < factories.length; i++) {
			if (weights[i] != 0.0)
				indizes[i] = factories[i].getIndex(clustering);
		}
		this.geometric = geometric;
	}

	public double getValue() {
		if (indizes.length == 0) return 1.0;
		double exp = 0.0;
		double sum = 0.0;

		if (geometric) {
			for (int i = 0; i < indizes.length; i++) {
				if (weights[i] != 0.0) {
					double single = indizes[i].getValue();
					if (single == 0.0) {
						return 0.0;
					}
					exp += weights[i] * Math.log(single);
					sum += weights[i];
				}
			}
			return Math.abs(sum) > 1e-8 ? Math.exp(exp / sum) : 1.0;
		} else {
			for (int i = 0; i < indizes.length; i++) {
			    if (indizes[i] == null)
			        continue;
				exp += weights[i] * indizes[i].getValue();
				sum += weights[i];
			}
			return Math.abs(sum) > 1e-8 ? exp / sum : 1.0;
		}

	}

	/*
	 * (non-Javadoc)
	 *
	 * @see de.uka.algo.clustering.base.Index#getUnweightedValue()
	 */
	public double getUnweightedValue() {
		if (indizes.length == 0) return 1.0;
		double exp = 0.0;
		double sum = 0.0;

		if (geometric) {
			for (int i = 0; i < indizes.length; i++) {
				if (weights[i] != 0.0) {
					double single = indizes[i].getUnweightedValue();
					if (single == 0.0)
						return 0;
					exp += weights[i] * Math.log(single);
					sum += weights[i];
				}
			}
			return Math.exp(exp / sum);
		} else {
			for (int i = 0; i < indizes.length; i++) {
				exp += weights[i] * indizes[i].getUnweightedValue();
				sum += weights[i];
			}
			return exp / sum;
		}

	}

	public static class Factory implements IndexFactory {

		private IndexFactory[] factories;

		private double[] weights;

		private boolean geometric;

		public Factory(boolean geometric, IndexFactory[] factories,
				double[] weights) {
			this.factories = (IndexFactory[]) factories.clone();
			this.weights = (double[]) weights.clone();
			this.geometric = geometric;
		};

		/*
		 * (non-Javadoc)
		 *
		 * @see de.uka.algo.clustering.addon.IndexFactory#getIndex(de.uka.algo.clustering.base.Clustering,
		 *      java.lang.Object)
		 */
		public Index getIndex(Clustering cl) {
			return new MixedIndex(cl, factories, weights, geometric);
		}

		public MixedIndex getInstance(Clustering cl) {
			return (MixedIndex) getIndex(cl);
		}

		public String toString() {
			double esum = 0;
			for (int i = 0; i < weights.length; i++) {
				esum += weights[i];
			}

			String result = "MixedIndex[";
			if (geometric) {
				result += "(" + esum + ")-root(";
				boolean a = false;
				for (int i = 0; i < weights.length; i++) {
					if (weights[i] != 0.0) {
						if (a)
							result += "*";
						a = true;
						result += factories[i] + "^(" + weights[i] + ")";
					}
				}
				result += ")";
			} else {
				result = "(";
				boolean a = false;
				for (int i = 0; i < weights.length; i++) {
					if (weights[i] != 0.0) {
						if (a)
							result += "+";
						a = true;
						result += weights[i] + "*" + factories[i];
					}
				}
				result += ")/" + esum;
			}
			return result+"]";
		}
	}

	public String toString() {
		double esum = 0;
		for (int i = 0; i < weights.length; i++) {
			esum += weights[i];
		}

		String result = "MixedIndex[";
		if (geometric) {
			result += "(" + esum + ")-root(";
			boolean a = false;
			for (int i = 0; i < weights.length; i++) {
				if (weights[i] != 0.0) {
					if (a)
						result += "*";
					a = true;
					result += indizes[i] + "^(" + weights[i] + ")";
				}
			}
			result += ")";
		} else {
			result = "(";
			boolean a = false;
			for (int i = 0; i < weights.length; i++) {
				if (weights[i] != 0.0) {
					if (a)
						result += "+";
					a = true;
					result += weights[i] + "*" + indizes[i];
				}
			}
			result += ")/" + esum;
		}
		return result + "]";
	}
}
