/*
 * Created on 09.10.2004
 *
 */
package de.uka.algo.clustering.io;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import y.view.Graph2D;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.generators.EmbeddedClusteredGraph;
import de.uka.algo.graphs.GraphInterpretation;

/**
 * This class is for I/O of Clusterings. The clustering is written as graphml
 * file. The nodes belonging to the same cluster have the same color and number.
 * It is possible to save more than one clustering for a graph
 * 
 * 
 * @author Daniel Delling
 * 
 */
public class ClusteringIOHandler {

	/**
	 * reads clusterings from a file. If the file contains edgeweights, the
	 * graph interpretation is set to weighted.
	 * 
	 * @param file
	 *            the file
	 * @return the Clusterings as linked list containing the graph
	 */
	public LinkedList<Clustering> read(Graph g, String file) {
		g.clear();
		// Use an instance of ourselves as the SAX event handler
		XMLHandler handler = new XMLHandler(g);
		// Use the default (non-validating) parser
		SAXParserFactory factory = SAXParserFactory.newInstance();
		try {
			// Parse the input
			SAXParser saxParser = factory.newSAXParser();
			saxParser.parse(new File(file), handler);

		} catch (Throwable t) {
			t.printStackTrace();
		}

		LinkedList<Clustering> cl = handler.getClustering();
		return cl;

	}

	/**
	 * reads only the first clustering contained in the file
	 * 
	 * @param file
	 *            the file that should be readed
	 * @return a clustering containng the graph
	 */
	public Clustering readSingle(String file) {
		return read(new Graph(), file).getFirst();
	}

	/**
	 * writes a single clustering with an underlying graph to a file. If the
	 * graph is weighted the edgeweights are saved as well
	 * 
	 * @param aClustering
	 * @param file
	 */
	public void write(Clustering aClustering, String file) {
		LinkedList<Clustering> cl = new LinkedList<Clustering>();
		cl.add(aClustering);
		this.write(cl, file);
	}

	/**
	 * prints several clustering to a file. All clusterings must belong to the
	 * same graph! If the graph is weighted, the edgeweights are stored as well
	 * 
	 * @param cl
	 *            the clustering
	 * @param file
	 *            the file
	 */
	public void write(LinkedList<Clustering> clusterings, String file) {
		for (Iterator<Clustering> ic = clusterings.iterator(); ic.hasNext();) {
			Clustering c = ic.next();
			if (c.poolSize() != 0) {
				throw new ClusteringIOException(ClusteringIOException.OP_WRITE,
						ClusteringIOException.ER_POOL_NOT_EMPTY, null, null);
			}
		}

		boolean graph2d = false;
		Graph2D g2 = null;
		Graph g = clusterings.getFirst().getGraph();
		if (g instanceof Graph2D) {
			g2 = (Graph2D) g;
			graph2d = true;
		}

		boolean embedded = false;
		EmbeddedClusteredGraph ecc = null;
		if (clusterings.getFirst() instanceof EmbeddedClusteredGraph) {
			ecc = (EmbeddedClusteredGraph) (clusterings.getFirst());
			embedded = true;
		}

		GraphInterpretation gi = clusterings.getFirst().getInterpretation();
		try {

			FileWriter fw = new FileWriter(file);
			BufferedWriter bw = new BufferedWriter(fw);
			bw.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
			bw
					.write("<graphml xmlns=\"http://graphml.graphdrawing.org/xmlns/graphml\"  xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n");
			bw
					.write("xsi:schemaLocation=\"http://graphml.graphdrawing.org/xmlns/graphml http://www.yworks.com/xml/schema/graphml/1.0/ygraphml.xsd\" xmlns:y=\"http://www.yworks.com/xml/graphml\">\n");
			for (int i = 0; i < clusterings.size(); i++) {
				bw.write(" <key id=\"clustering" + i
						+ "\" for=\"node\" type=\"int\"/>\n");
				if (clusterings.get(i).hasLabels()) {					
					bw.write(" <key id=\"nameOfCluster" + i
							+ "\" for=\"node\" type=\"string\"/>\n");
				}
			}
			if (graph2d) {
				bw
						.write(" <key id=\"d0\" for=\"node\" yfiles.type=\"nodegraphics\"/>\n");
			}
			if (gi.isWeighted()) {
				bw
						.write("   <key id=\"d1\" for=\"edge\" yfiles.type=\"edgegraphics\"/>\n");
			}
			bw.write(" <graph id=\"G\" edgedefault=\"undirected\">\n");

			// knoten erzeugen

			for (NodeCursor nc = clusterings.getFirst().getGraph().nodes(); nc
					.ok(); nc.next()) {
				Node n = nc.node();
				bw.write("  <node id=\"n" + n.index() + "\">\n");
				for (int i = 0; i < clusterings.size(); i++) {
					if (clusterings.get(i).hasLabels()) {	
						String name = clusterings.get(i).getCluster(n).getLabel();
						if (name !=null) {
						bw.write("   <data key=\"nameOfCluster" + i + "\" cluster=\""
								+ name
								+ "\"></data>\n");
						}
					}
					bw.write("   <data key=\"clustering" + i + "\" cluster=\""
							+ clusterings.get(i).getCluster(n).getIndex()
							+ "\"></data>\n");
				}
				
				
				if (graph2d) {
					bw.write("   <data key=\"d0\">\n");
					bw.write("    <y:ShapeNode>\n");
					bw.write("     <y:Geometry  x=\"" + g2.getCenterX(n)
							+ "\" y=\"" + g2.getCenterY(n)
							+ "\" width=\"30.0\" height=\"30.0\"/>\n");
					bw.write("     <y:Fill color=\"#"
							+ Integer.toHexString(
									g2.getRealizer(n).getFillColor().getRGB())
									.substring(2)
							+ "\"  transparent=\"false\"/>\n");
					bw
							.write("     <y:BorderStyle type=\"line\" width=\"5.0\" color=\"#"
									+ Integer.toHexString(
											g2.getRealizer(n).getLineColor()
													.getRGB()).substring(2)
									+ "\"/>\n");

					// TODO: experimental node labels (für robert, den hagger)
					bw
							.write("<y:NodeLabel x=\"-2.5\" y=\"4.273193359375\" width=\"35.0\" height=\"21.45361328125\" visible=\"true\" alignment=\"center\" fontFamily=\"Dialog\" fontSize=\"15\" fontStyle=\"plain\" textColor=\"#000000\" hasBackgroundColor=\"false\" hasLineColor=\"false\" modelName=\"internal\" modelPosition=\"c\" autoSizePolicy=\"content\">"
									+ g2.getLabelText(n) + "</y:NodeLabel>");

					bw.write("     <y:Shape type=\"rectangle\"/>\n");
					bw.write("     </y:ShapeNode>\n");
					bw.write("   </data> \n");

				} else if (embedded) {
					bw.write("   <data key=\"d0\">\n");
					bw.write("    <y:ShapeNode>\n");
					bw.write("     <y:Geometry  x=\"" + ecc.getPosition(n)[0]
							+ "\" y=\"" + ecc.getPosition(n)[1]
							+ "\" width=\"30.0\" height=\"30.0\"/>\n");
					bw.write("     </y:ShapeNode>\n");
					bw.write("   </data> \n");
				}
				bw.write("  </node>\n");
			}

			// kanten erzeugen
			Edge e;
			for (EdgeCursor ec = g.edges(); ec.ok(); ec.next()) {
				e = ec.edge();
				bw.write("  <edge id=\"e" + e.index() + "\" source=\"n"
						+ e.source().index() + "\" target=\"n"
						+ e.target().index() + "\">\n");
				if (gi.isWeighted() == true) {
					bw.write("   <data key=\"d1\">\n"
							+ "    <y:PolyLineEdge>\n" + "     <y:EdgeLabel>"
							+ gi.getWeight(e) + "</y:EdgeLabel>\n"
							+ "    </y:PolyLineEdge>\n" + "   </data>\n");
				}
				bw.write("  </edge>\n");
			}

			bw.write("</graph>\n");
			bw.write("</graphml>\n");
			bw.close();
		} catch (IOException e) {
			System.out.println("Failed writing file " + file);
		}

	}
}
