package de.uka.algo.clustering.visualization;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Locale;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.Graph;
import y.view.Graph2D;
import de.uka.algo.clustering.Clustering;

/**
 * A Graph2DFacade is the super class for all classes to store or retrieve
 * clusterings from Graph2D objects by means of visual attributes.<br>
 * Subclasses should implement either the interface
 * <code>Graph2DExtractor</code> (to retrieve a clustering from a Graph2D) or
 * the interface <code>Graph2DVisualizer</code> or both. In case they
 * implement both, they ought to be implemented such <code>visualize</code>
 * and <code>extract</code> rely on the same semantics, i.e., calling
 * <code>extract</code> after <code>visualize</code> should return the
 * original clustering and vice versa (up to the interpretation, which may
 * vary). <br>
 * The class Graph2DFacade also serves as registration class for newly
 * implemented facade classes.
 * 
 * @author Bastian Katz, bastian.katz@ira.uka.de
 */
public abstract class Graph2DFacade {

	/** list of registered facades */
	private static LinkedList<Graph2DFacade> facades = new LinkedList<Graph2DFacade>();

	/**
	 * Adds a Graph2DFacade to the list of facades that are displayed in yEd
	 * modules.
	 * 
	 * @param facadeToAdd
	 *            the facade to add. Should implement Graph2DVisualizer,
	 *            Graph2DExtractor or both.
	 */
	public static void addVisualization(Graph2DFacade facadeToAdd) {
		facades.add(facadeToAdd);
	}

	/**
	 * @return all registered Graph2DFacades that implement the interface
	 *         Graph2DExtractor
	 */
	public static Graph2DExtractor[] getExtractors() {
		LinkedList<Graph2DExtractor> resultAsList = new LinkedList<Graph2DExtractor>();
		for (Iterator<Graph2DFacade> registeredFacades = facades.iterator(); registeredFacades
				.hasNext();) {
			Graph2DFacade currentFacade = registeredFacades.next();
			if (currentFacade instanceof Graph2DExtractor) {
				resultAsList.add((Graph2DExtractor) currentFacade);
			}
		}
		return resultAsList.toArray(new Graph2DExtractor[0]);
	}

	/**
	 * @return all registered Graph2DFacades that implement the interface
	 *         Graph2DVisualizer
	 */
	public static Graph2DVisualizer[] getVisualizer() {
		LinkedList<Graph2DVisualizer> resultAsList = new LinkedList<Graph2DVisualizer>();
		for (Iterator<Graph2DFacade> registeredFacades = facades.iterator(); registeredFacades
				.hasNext();) {
			Graph2DFacade currentFacade = registeredFacades.next();
			if (currentFacade instanceof Graph2DVisualizer) {
				resultAsList.add((Graph2DVisualizer) currentFacade);
			}
		}
		return resultAsList.toArray(new Graph2DVisualizer[0]);
	}

	protected static Graph2D getGraph2D(Clustering aClustering) {
		Graph graph = aClustering.getGraph();
		if (graph instanceof Graph2D) {
			return (Graph2D) graph;
		}
		throw new RuntimeException("Visualizations must operate on Graph2D.");
	}
	
	/**
	 * Visualizes edge weights of the respective clustering's interpretation as
	 * edge labels. By convention, an empty label means "1".
	 * 
	 * @param clustering
	 *            the clustering to visualize weights from
	 * @param performRounding
	 *            whether to round weights to 3 decimals for improved
	 *            readability
	 */
	protected void setEdgeLabels(Clustering clustering,
			boolean performRounding) {
		Graph2D graph = getGraph2D(clustering);
		for (EdgeCursor ec = graph.edges(); ec.ok(); ec.next()) {
			Edge currentEdge = ec.edge();
			double value = clustering.getInterpretation()
					.getWeight(currentEdge);
			String label = "";
			if (value != 1.0) {
				if (performRounding) {
					label = String.format(Locale.US, "%.3f", value);
				} else {
					label = String.format(Locale.US, "%f", value);
				}
			}
			graph.setLabelText(currentEdge, label);
		}
	}
}
