/*
 * Created on 08.08.2004
 * Reviewed 1.7 (Mar 26, 2005)
 *
 * Copyright 2004.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.graphs;

import java.util.LinkedList;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.EdgeMap;
import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import y.view.Graph2D;

/**
 * A GraphInterpretation extends a graph by type information
 * (directed/undirected) and edge weights. It also allows listeners to register
 * for events like changing weights.
 * 
 * @author Bastian Katz, katz@kit.edu
 */
public abstract class GraphInterpretation {

	/** the underlying graph */
	protected Graph graph = null;

	/** the registered listeners */
	protected LinkedList<GraphInterpretationListener> listeners = null;

	/**
	 * Instantiates a GraphInterpretation for a given graph.
	 * 
	 * @param aGraph
	 *            the underlying graph
	 */
	protected GraphInterpretation(Graph aGraph) {
		this.graph = aGraph;
		listeners = new LinkedList<GraphInterpretationListener>();
	}

	/**
	 * Registers a listener for weight changes.
	 * 
	 * @param aListener
	 *            the listener to register
	 */
	public void addListener(GraphInterpretationListener aListener) {
		listeners.add(aListener);
	}

	/**
	 * Removes a listener.
	 * 
	 * @param aListener
	 *            the listener to remove
	 */
	public void removeListener(GraphInterpretationListener aListener) {
		listeners.remove(aListener);
	}

	/**
	 * Gets the underlying graph.
	 * 
	 * @return the underlying graph
	 */
	public Graph getGraph() {
		return graph;
	}

	/**
	 * Returns the weight for an edge.
	 * 
	 * @param anEdge
	 *            an edge in the underlying graph
	 * @return the edge's weight, 0 for a null argument
	 */
	public abstract double getWeight(Edge anEdge);

	/**
	 * Returns a GraphInterpretation of the same dynamic type for a subgraph
	 * given by a graph and a mapping from the subgraph's edges to the graphs
	 * edges.
	 * 
	 * @param subgraph
	 *            a graph (subgraph of the interpretation's underlying graph)
	 * @param map
	 *            an edgeMap mapping edges from the subgraph to their respective
	 *            'preimages'
	 * @return A GraphInterpretation of the same dynamic type s.t. all edges in
	 *         the subgraph have the same weight as the edges the given map maps
	 *         them onto.
	 */
	public abstract GraphInterpretation getPartialInterpretation(
			Graph subgraph, EdgeMap map);

	/**
	 * @return Whether the graph is weighted, i.e. provides non-trivial weights
	 *         for the edges
	 */
	public abstract boolean isWeighted();

	/**
	 * @return whether the graph is directed
	 */
	public abstract boolean isDirected();

	/**
	 * Returns the adjacency between the two nodes. Note that the definition of
	 * the adjacency is different for directed and undirected graphs.
	 * 
	 * @param source
	 *            the source node
	 * @param target
	 *            the target node
	 * @return the adjacency
	 */
	public abstract double getAdjacency(Node source, Node target);

	/**
	 * Returns the outdegree of a given node. Note that the definition of the
	 * outdegree is different for directed and undirected graphs.
	 * 
	 * @param aNode
	 *            the node
	 * @return the node's outdegree
	 */
	public abstract double getOutDegree(Node aNode);

	/**
	 * Returns the outdegree of a given node. Note that the definition of the
	 * outdegree is different for directed and undirected graphs.
	 * 
	 * @param aNode
	 *            the node
	 * @return the node's outdegree
	 */
	public double getOutDegree(int index) {
		return getOutDegree(graph.getNodeArray()[index]);
	}

	/**
	 * Returns the maximum outdegree.
	 * 
	 * @return the maximum outdegree (over all nodes)
	 */
	public double getMaxOutDegree() {
		double result = 0;
		for (NodeCursor nc = graph.nodes(); nc.ok(); nc.next()) {
			double thisDegree = getOutDegree(nc.node());
			if (thisDegree > result)
				result = thisDegree;
		}
		return result;
	}

	/**
	 * Returns the indegree of a given node. Note that the definition of the
	 * indegree is different for directed and undirected graphs.
	 * 
	 * @param aNode
	 *            the node
	 * @return the node's indegree
	 */
	public abstract double getInDegree(Node aNode);

	/**
	 * Returns the indegree of a given node. Note that the definition of the
	 * indegree is different for directed and undirected graphs.
	 * 
	 * @param aNode
	 *            the node's index
	 * @return the node's indegree
	 */
	public double getInDegree(int index) {
		return getInDegree(graph.getNodeArray()[index]);
	}

	/**
	 * Returns the maximum indegree.
	 * 
	 * @return the maximum indegree (over all nodes)
	 */
	public double getMaxInDegree() {
		double result = 0;
		for (NodeCursor nc = graph.nodes(); nc.ok(); nc.next()) {
			double thisDegree = getInDegree(nc.node());
			if (thisDegree > result)
				result = thisDegree;
		}
		return result;
	}

	/**
	 * Returns the adjacency between the two nodes. Note that the definition of
	 * the adjacency is different for directed and undirected graphs.
	 * 
	 * @param sourceIndex
	 *            the index of the source node
	 * @param targetIndex
	 *            the index of the target node
	 * @return the adjacency
	 */
	public double getAdjacency(int sourceIndex, int targetIndex) {
		Node[] nodeArray = graph.getNodeArray();
		return getAdjacency(nodeArray[sourceIndex], nodeArray[targetIndex]);
	}

	/**
	 * Returns a node's degree. Note that the definition of the adjacency is
	 * different for directed and undirected graphs.
	 * 
	 * @param aNode
	 *            the node
	 * @return the node's degree
	 */
	public abstract double getDegree(Node aNode);

	/**
	 * Returns a node's degree. Note that the definition of the adjacency is
	 * different for directed and undirected graphs.
	 * 
	 * @param aNodeIndex
	 *            the node's index
	 * @return the node's degree
	 */
	public double getDegree(int aNodeIndex) {
		return getDegree(graph.getNodeArray()[aNodeIndex]);
	}

	/**
	 * Returns the maximum degree.
	 * 
	 * @return the maximum degree (over all nodes)
	 */
	public double getMaxDegree() {
		double result = 0;
		for (NodeCursor nc = graph.nodes(); nc.ok(); nc.next()) {
			double thisDegree = getDegree(nc.node());
			if (thisDegree > result)
				result = thisDegree;
		}
		return result;
	}

	/**
	 * Returns the sum of all edge weights.
	 * 
	 * @return the sum of all edge weights
	 */
	public double getEdgeWeightSum() {
		double result = 0;
		for (EdgeCursor ec = graph.edges(); ec.ok(); ec.next()) {
			result += getWeight(ec.edge());
		}
		return result;
	}

	/**
	 * Returns an undirected GraphInterpretation for a given Graph2D, optionally
	 * using edge labels as weights (empty labels are treated as 1.0).
	 * 
	 * @param graph
	 *            the graph
	 * @param useWeights
	 *            flag whether to extract edge weights
	 * @return a SimpleGraph if useweights=false or all labels are empty, a
	 *         WeightedGraph otherwise.
	 */
	public static GraphInterpretation getInterpretation(Graph2D graph,
			boolean useWeights) {
		boolean foundWeights = false;
		GraphInterpretation result = null;
		if (useWeights) {
			EdgeMap e = graph.createEdgeMap();
			for (EdgeCursor ec = graph.edges(); ec.ok(); ec.next()) {
				Edge edge = ec.edge();
				double v = 1.0;
				if (!graph.getLabelText(edge).equals("")) {
					foundWeights = true;
					try {
						v = Double.parseDouble(graph.getLabelText(edge));
					} catch (NumberFormatException nfe) {
						throw new RuntimeException(
								"cannot parse edge labels as weights.");

					}
				}
				e.setDouble(edge, v);
			}
			if (foundWeights) {
				result = new WeightedGraph(graph, e);
			}
			graph.disposeEdgeMap(e);
		}
		result = (result == null ? new SimpleGraph(graph) : result);
		return result;
	}

}