/*
 * Created on Aug 24, 2004
 *
 * Copyright 2004.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.graphs;

import java.util.Iterator;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.EdgeMap;
import y.base.Graph;
import y.base.Node;

/**
 * A WeightedGraph is an undirected graph providing additional, non-static edge
 * weights.
 * 
 * @author Bastian Katz
 * 
 */
public class WeightedGraph extends UndirectedGraph {

	/** edge weights as doubles */
	private EdgeMap weight = null;

	/**
	 * Instantiates a weighted, undirected interpretation of the given graph,
	 * taking weights provided by an EdgeMap. Note: Edge weights are stored in a
	 * local copy of the given EdgeMap. Thus, changing the EdgeMap afterwards
	 * does not have any effect on the interpretation. Changing the
	 * interpretation is only possible via calls of <code>changeWeight</code>
	 * which will notify registered listeners to avoid inconsistent data.
	 * 
	 * @param aGraph
	 *            the underlying graph
	 * @param weights
	 *            an EdgeMap providing edge weights for the given graph as
	 *            doubles
	 */
	public WeightedGraph(Graph aGraph, EdgeMap weights) {
		super(aGraph);
		
		// create map to store weigths internally; is disposed at finalize()
		this.weight = aGraph.createEdgeMap(); 
		
		//copy weights
		for (EdgeCursor nc = aGraph.edges(); nc.ok(); nc.next()) {
			Edge currentEdge = nc.edge();
			this.weight.setDouble(currentEdge, weights.getDouble(currentEdge));
		}
	}
	
	/**
	 * Creates a WeightedGraph out of an existing graph interpretation.
	 * The new interpretation is a copy of the old one.
	 * 
	 * @param anInterpretation
	 *            the graph interpretation to copy
	 */	
	public WeightedGraph(GraphInterpretation anInterpretation) {
		super(anInterpretation.getGraph().createCopy());
		weight = graph.createEdgeMap();
		for (int i = 0; i<graph.E(); i++) {
			double w = anInterpretation.getWeight(anInterpretation.graph.getEdgeArray()[i]);
			Edge e = graph.getEdgeArray()[i];
			weight.setDouble(e, w);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#finalize()
	 */
	protected void finalize() {
		graph.disposeEdgeMap(weight);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.graphs.GraphInterpretation#getWeight(y.base.Edge)
	 */
	public double getWeight(Edge anEdge) {
		return weight.getDouble(anEdge);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.graphs.GraphInterpretation#getPartialInterpretation(y.base.Graph,
	 *      y.base.EdgeMap)
	 */
	public GraphInterpretation getPartialInterpretation(Graph subgraph,
			EdgeMap map) {
		
		//set up edgemap to temporarily store weights
		EdgeMap partialWeight = subgraph.createEdgeMap();
		for (EdgeCursor ec = subgraph.edges(); ec.ok(); ec.next()) {
			Edge anEdge = ec.edge();
			Edge fromEdge = (Edge) map.get(anEdge);
			partialWeight.setDouble(anEdge, weight.getDouble(fromEdge));
		}
		
		//create and return partial interpretation, release EdgeMap
		WeightedGraph result = new WeightedGraph(subgraph, partialWeight);
		subgraph.disposeEdgeMap(partialWeight);
		return result;
	}

	/* (non-Javadoc)
	 * @see de.uka.algo.graphs.GraphInterpretation#getAdjacency(y.base.Node, y.base.Node)
	 */
	public double getAdjacency(Node source, Node target) {
		double result = 0;

		//summing weights for incident edges, counting loops twice
		for (EdgeCursor ec = source.edges(); ec.ok(); ec.next()) {
			Edge anEdge = ec.edge();
			if (anEdge.opposite(source) == target) {
				result += weight.getDouble(anEdge);
			}
		}

		//for reflexive adjacency, correct doubly counted loops
		if (source==target) result /= 2.0;
		return result;
	}

	/* (non-Javadoc)
	 * @see de.uka.algo.graphs.GraphInterpretation#isWeighted()
	 */
	public boolean isWeighted() {
		return true;
	}

	/* (non-Javadoc)
	 * @see de.uka.algo.graphs.UndirectedGraph#getDegree(y.base.Node)
	 */
	@Override
	public double getDegree(Node aNode) {
		double result = 0.0;
		
		//sum all outgoing edges (this includes loops)
		for (EdgeCursor ec = aNode.outEdges(); ec.ok(); ec.next()) {
			result += weight.getDouble(ec.edge());
		}
		
		//sum all incoming edges excluding loops
		for (EdgeCursor ec = aNode.inEdges(); ec.ok(); ec.next()) {
			Edge edge = ec.edge();
			if (edge.source() != aNode)
				result += weight.getDouble(edge);
		}
		
		return result;
	}

	/**
	 * Assigns a new weight to an edge. If the new weight differs from the old
	 * one, listeners (such as clusterings and their listeners) are notified.
	 * 
	 * @param anEdge
	 *            the edge
	 * @param newWeight
	 *            the new weight
	 * @return whether the new weight differs from the old weight
	 */
	public boolean setWeight(Edge anEdge, double newWeight) {
		//do nothing if the weight does not change.
		if (weight.getDouble(anEdge) == newWeight)
			return false;
		
		//notify all listeners about the upcoming change
		for (Iterator<GraphInterpretationListener> gyc = listeners.iterator(); gyc
				.hasNext();) {
			gyc.next().weightChangeEvent(anEdge, this.weight.getDouble(anEdge),
					newWeight);
		}
		//there is no post-event for weight changes
		
		//set the new value and return true
		weight.setDouble(anEdge, newWeight);		
		return true;
	}	
}