package de.uka.algo.arxivGen;

import java.sql.Timestamp;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map.Entry;

import y.base.Edge;
import y.base.Node;
import y.view.Graph2D;

/**
 * can be used to generate a list of edge or node events with edge timeouts
 * usage: 
 * 0. new ...
 * 1. sorted by time, enqueue all or many events by enqueue(new event(int a,int b,Timestamp T))
 * 1.a if almsost sorted use almsotSortedEnque
 * 1.b if unsorted, after all enqueues are done, use sortEvents();
 * 2. optionally configure and set stuff
 * 3. init();
 * 4. while (hasMoreEvents()) nextEvents gives a LinkedList of Strings of announcements
 * 5. optionally getGraph or getNodeWeights gives internal data structure back 
 */

/**
 * @author robert
 * 
 */
public class eventTimeOutScheduler {

	public static class ScheduleEvent {
		int s;
		int t;
		Timestamp T;
		String id;
		double weight;

		public ScheduleEvent(int a, int b, Timestamp stamp, String anId, double aWeight) {
			s = Math.min(a, b);
			t = Math.max(a, b);
			T = stamp;
			id = anId;
			weight = aWeight;
			// System.out.println("event created: " + s + ", " + t + ", " + T);
		}

		public ScheduleEvent(int a, int b, long millis, String anId, double aWeight) {
			this(a, b, new Timestamp(millis), anId, aWeight);
		}

		public ScheduleEvent(int a, int b, Timestamp stamp) {
			this(a, b, stamp, null, 1.0);
		}

		public ScheduleEvent(int a, int b, long millis) {
			this(a, b, new Timestamp(millis), null, 1.0);
		}

		public String verbose() {
			return ("event " + this.toString() + " is " + s + " -> " + t + " at " + T + ", id: " + id + ", weight: " + weight);
		}
	}

	public static class Ante implements Comparator<ScheduleEvent> {

		public int compare(ScheduleEvent sev1, ScheduleEvent sev2) {
			if (sev1.T.before(sev2.T)) {
				return -1;
			}
			if (sev1.T.after(sev2.T)) {
				return 1;
			}
			return 0;
		}
	}

	static LinkedList<ScheduleEvent> proactiveEvents = new LinkedList<ScheduleEvent>();
	private long timeOut = (1000L * 60 * 60 * 24 * 365 * 2); // millis until timeout
	private boolean noDisconnecteds = true; // delete deg0-nodes?
	static LinkedList<ScheduleEvent> timeoutEvents = new LinkedList<ScheduleEvent>();
	private Timestamp nextProactive, nextTimeout, T;
	private long maxmillis = Long.MAX_VALUE;
	private Timestamp verynever = new Timestamp(maxmillis);
	private boolean moreEvents; // are there any more events?
	private HashMap<Integer, HashMap<Integer, Double>> adjList = new HashMap<Integer, HashMap<Integer, Double>>();
	private HashMap<Integer, Double> nodeWeight = new HashMap<Integer, Double>();
	private int totalEdgeCount = 0, maxEverEdgeCount = 0, maxEverN = 0;
	private ScheduleEvent currentEvent, checkEvent;
	private double currentWeight, w, little = 0.00000001, maxEverEdgeWeight = 0.0, totalEdgeWeight = 0.0;
	private int s, t;
	// private String id; //not used atm
	private int ncCount = 0, ndCount = 0, ecCount = 0, edCount = 0, wiCount = 0, wdCount = 0;
	private boolean timeOutOn = true;
	private boolean stopAfterPro = true;
	private boolean intWeights = false;
	// private Timestamp lastT = new Timestamp(0L); //not used atm
	private int maxCompound = 0;
	private static int timeStepCounter = 0;
	private boolean paperPerTimeStep = false;
	private int paperCount = 0;

	// assumes sorted insertions! (earliest first)
	public void enqueue(ScheduleEvent ev) {
		proactiveEvents.add(ev);
		if (timeOutOn) {
			ScheduleEvent timedOutEvent = new ScheduleEvent(ev.s, ev.t, (ev.T.getTime() + (long) timeOut), ev.id, ev.weight);
			timeoutEvents.add(timedOutEvent);
		}
	}

	// assumes almost sorted insertions! (earliest first)
	// for each insertion performs linear search from back, before inserting, must be cheap!
	public void almostSortedEnqueue(ScheduleEvent ev) {
		int skipBack = 0;
		if (proactiveEvents.isEmpty()) {
			proactiveEvents.add(ev);
		} else {
			while (proactiveEvents.get(proactiveEvents.size() - skipBack - 1).T.after(ev.T)) { // must be
				// inserted
				// earlier!
				++skipBack;
				if (skipBack == proactiveEvents.size()) {
					proactiveEvents.add(proactiveEvents.size() - skipBack, ev);
					break;
				}
			}
			proactiveEvents.add(proactiveEvents.size() - skipBack, ev);
		}
		if (timeOutOn) {
			ScheduleEvent timedOutEvent = new ScheduleEvent(ev.s, ev.t, (ev.T.getTime() + (long) timeOut), ev.id, ev.weight);
			timeoutEvents.add(timeoutEvents.size() - skipBack, timedOutEvent);
		}
	}

	public void sortEvents() {
		Ante anteComparator = new Ante();
		Collections.sort(proactiveEvents, anteComparator);
		Collections.sort(timeoutEvents, anteComparator);
	}

	public boolean checkSorting() {
		boolean sorted = true;
		Timestamp lastTS = new Timestamp(0L);
		ScheduleEvent sev;
		for (Iterator<ScheduleEvent> it = proactiveEvents.iterator(); it.hasNext();) {
			sev = it.next();
			if (sev.T.before(lastTS)) {
				sorted = false;
				System.out.println(sev.T + " in pro is before " + lastTS);
				return sorted;
			}
			lastTS = sev.T;
		}
		lastTS = new Timestamp(0L);
		for (Iterator<ScheduleEvent> it = timeoutEvents.iterator(); it.hasNext();) {
			sev = it.next();
			if (sev.T.before(lastTS)) {
				sorted = false;
				System.out.println(sev.T + " in TO is before " + lastTS);
				return sorted;
			}
			lastTS = sev.T;
		}
		return sorted;
	}

	public void setIntWeights(boolean intWeightsOn) {
		intWeights = intWeightsOn;
	}

	public void setTimeOutOn(boolean TOOn) {
		timeOutOn = TOOn;
	}

	public boolean hasMoreEvents() {
		return moreEvents;
	}

	public HashMap<Integer, HashMap<Integer, Double>> getGraph() {
		return adjList;
	}

	public HashMap<Integer, Double> getNodeWeights() {
		return nodeWeight;
	}

	public void setTimeOut(long aTimeOut) {
		timeOut = aTimeOut;
		System.out.println("timeout is " + timeOut);
	}

	public void setTimeOutYears(int TO) {
		timeOut = (1000L * 60 * 60 * 24 * 365 * TO);
		System.out.println("timeout is " + timeOut);
	}

	public void setNoDisconnecteds(boolean delDiscon) {
		noDisconnecteds = delDiscon;
	}

	public void init() {
		queueTiming();
		System.out.println("There are many events: pro: " + proactiveEvents.size() + ", TO: " + timeoutEvents.size());
	}

	public void queueTiming() {
		if (!proactiveEvents.isEmpty()) {
			nextProactive = proactiveEvents.peek().T;
		} else {
			nextProactive = verynever;
			if (stopAfterPro) {
				moreEvents = false;
				nextTimeout = verynever;
				return;
			}
		}
		if (!timeoutEvents.isEmpty()) {
			nextTimeout = timeoutEvents.peek().T;
			moreEvents = true;
		} else {
			nextTimeout = verynever;
			moreEvents = false;
		}
	}

	public String stats() {
		return ("Nodes: creations: " + ncCount + "; deletions: " + ndCount + ". Edges: creations: " + ecCount
				+ "; deletions: " + edCount + " Weights: increases: " + wiCount + "; decreases: " + wdCount);
	}

	public int currentN() {
		return (nodeWeight.size());
	}

	public int currentEc() {
		return (totalEdgeCount);
	}

	public double currentEw() {
		return (totalEdgeWeight);
	}

	public int maxN() {
		return maxEverN;
	}

	public double maxEw() {
		return maxEverEdgeWeight;
	}

	public double maxEc() {
		return maxEverEdgeCount;
	}

	public int getMaxCompound() {
		return maxCompound;
	}

	public int getCurrentTimeStep() {
		return timeStepCounter;
	}

	public int getPaperCount() {
		if (paperPerTimeStep){
			return paperCount;
		} else {
			return -1;
		}
	}

	public double myRound(double in) {
		return ((double) (Math.round(in * 10000) / 10000.0));
	}

	public int size() {
		return (proactiveEvents.size() + timeoutEvents.size());
	}

	public Graph2D getCurrentGraph() {
		Graph2D g = new Graph2D();
		Entry<Integer, HashMap<Integer, Double>> currentEntry;
		Entry<Integer, Double> currentAdj;
		int u, v;
		double w;
		Node uN, vN;
		Edge uNvN;
		HashMap<Integer, Node> nidToNode = new HashMap<Integer, Node>();
		for (Iterator<Entry<Integer, HashMap<Integer, Double>>> it = adjList.entrySet().iterator(); it.hasNext();) {
			currentEntry = it.next();
			u = currentEntry.getKey();
			if (!nidToNode.containsKey(u)) {
				uN = g.createNode();
				g.setLabelText(uN, Integer.toString(u));
				nidToNode.put(u, uN);
			} else {
				uN = nidToNode.get(u);
			}
			HashMap<Integer, Double> neighbors = currentEntry.getValue();
			for (Iterator<Entry<Integer, Double>> it2 = neighbors.entrySet().iterator(); it2.hasNext();) {
				currentAdj = it2.next();
				v = currentAdj.getKey();
				if (!nidToNode.containsKey(v)) {
					vN = g.createNode();
					g.setLabelText(vN, Integer.toString(v));
					nidToNode.put(v, vN);
				} else {
					vN = nidToNode.get(v);
				}
				w = currentAdj.getValue();
				uNvN = g.createEdge(uN, vN);
				g.setLabelText(uNvN, Double.toString(myRound(w)));
			}
		}
		return g;
	}

	public LinkedList<String> nextEvents() {
		LinkedList<String> returnEvents = new LinkedList<String>();
		// System.out.println("in: " + nextProactive + ", out: " + nextTimeout);
		if (!hasMoreEvents()) {
			returnEvents.add("end reached, stop!");
			return (returnEvents);
		}
		if (nextTimeout.before(nextProactive)) { // TO-event next
			currentEvent = timeoutEvents.poll(); // get TO-event
			s = currentEvent.s;
			t = currentEvent.t;
			T = currentEvent.T;
//			id = currentEvent.id;	//not used atm
			w = currentEvent.weight;
			currentWeight = adjList.get(s).get(t) - w;
			nodeWeight.put(s, nodeWeight.get(s) - w);
			nodeWeight.put(t, nodeWeight.get(t) - w);
			totalEdgeWeight -= w;
			if (Math.abs(currentWeight) <= little) { // edge now gone
				adjList.get(s).remove(t);
				returnEvents.add("deleteEdge," + s + "," + t + "," + T + "\n");
				++edCount;
				--totalEdgeCount;
				if (noDisconnecteds) { // maybe nodes now disconnected
					if (Math.abs(nodeWeight.get(s)) <= little) {
						nodeWeight.remove(s);
						adjList.remove(s);	//sure its there!
						returnEvents.add("deleteNode," + s + "," + T + "\n");
						++ndCount;
					}
					if (Math.abs(nodeWeight.get(t)) <= little) {
						nodeWeight.remove(t);
						if (adjList.containsKey(t)){	//maybe not there
							adjList.remove(t);
						}
						returnEvents.add("deleteNode," + t + "," + T + "\n");
						++ndCount;
					}
				}
			} else { // edge still there but at -w
				adjList.get(s).put(t, currentWeight);
				if (intWeights){
					returnEvents.add("changeWeight," + s + "," + t + "," + Math.round(currentWeight) + ",down," + T + "\n");
				} else {
					returnEvents.add("changeWeight," + s + "," + t + "," + currentWeight + ",down," + T + "\n");
				}
				++wdCount;
			}
		} else { // pro-event next
			currentEvent = proactiveEvents.poll(); // get pro-event
			s = currentEvent.s;
			t = currentEvent.t;
			T = currentEvent.T;
//			id = currentEvent.id;	//not used atm
			w = currentEvent.weight;
			if (!adjList.containsKey(s)) { // s exists in adjMat?
				adjList.put(s, new HashMap<Integer, Double>()); // create s's list
			}
			if (!nodeWeight.containsKey(s)) { // s exists yet?
				nodeWeight.put(s, 0.0); // initiate
				returnEvents.add("createNode," + s + "," + T + "\n");
				++ncCount;
			}
			if (!nodeWeight.containsKey(t)) { // t exists?
				nodeWeight.put(t, 0.0); // initiate
				returnEvents.add("createNode," + t + "," + T + "\n");
				++ncCount;
			}
			if (adjList.get(s).containsKey(t)) { // edge already there?
				currentWeight = adjList.get(s).get(t) + w; // edge already there: weight increase
			} else { // no edge yet, create edge
				currentWeight = w;
				returnEvents.add("createEdge," + s + "," + t + "," + T + "\n");
				++ecCount;
				++totalEdgeCount;
			}
			if (intWeights){
				returnEvents.add("changeWeight," + s + "," + t + "," + Math.round(currentWeight) + ",up," + T + "\n");
			} else {
				returnEvents.add("changeWeight," + s + "," + t + "," + currentWeight + ",up," + T + "\n");
			}
			++wiCount;
			adjList.get(s).put(t, currentWeight); // whether already there or not, now set
			nodeWeight.put(t, nodeWeight.get(t) + w);
			nodeWeight.put(s, nodeWeight.get(s) + w);
			totalEdgeWeight += w;
		}
		queueTiming();
		if (maxEverN < nodeWeight.size()) {
			maxEverN = nodeWeight.size();
		}
		if (maxEverEdgeWeight < totalEdgeWeight) {
			maxEverEdgeWeight = totalEdgeWeight;
		}
		if (maxEverEdgeCount < totalEdgeCount) {
			maxEverEdgeCount = totalEdgeCount;
		}
		
		//when to issue a new timestep...
		T.setNanos(T.getNanos() + 1);
		if (!paperPerTimeStep){	//classic timesteps
			returnEvents.add("timeStep," + T + ",#" + timeStepCounter++ +"\n");
			T.setNanos(T.getNanos() + 1);
		} else {	//one timestep per paper
			if (!moreEvents) {	//end of list anyway...
				returnEvents.add("timeStep," + T + ",#" + timeStepCounter++ +"\n");
			} else {	//more events to come
				if (nextTimeout.before(nextProactive)) { // TO-event next
					checkEvent = timeoutEvents.peek(); // get TO-event
				}
				else {
					checkEvent = proactiveEvents.peek(); // get TO-event
				}
				if (checkEvent.id != currentEvent.id){	//end of paper!
					returnEvents.add("timeStep," + T + ",#" + timeStepCounter++ +"\n");
					paperCount++;
				}
				else {
				}
			}
			
		}
		if (maxCompound < T.getNanos()) {
			maxCompound = T.getNanos();
		}
		return returnEvents;
	}

	public void setPaperPerTimeStep(boolean ppts) {
		paperPerTimeStep = ppts;
	}

}