package de.uka.algo.dynclustering.algos;

import java.util.List;

import org.apache.commons.lang.IncompleteArgumentException;

import y.base.Edge;
import y.base.Node;
import y.base.NodeCursor;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.algorithms.Algorithm;
import de.uka.algo.clustering.algorithms.ElementalOperationMinimizer;
import de.uka.algo.clustering.algorithms.ElementalOperationOptimizer;
import de.uka.algo.clustering.algorithms.ElementalOperationSimulatedAnnealing;
import de.uka.algo.clustering.index.IndexFactory;
import de.uka.algo.dynclustering.algos.base.AlgorithmFactory;
import de.uka.algo.dynclustering.algos.base.DynamicClusteringAlgorithm;
import de.uka.algo.dynclustering.graphs.DynamicGraph;
import de.uka.algo.dynclustering.util.RunTimer.TimeUnit;

/**
 * Dynamic version of the Elemental Optimization Operation clustering algorithm
 * 
 * @author Robert Goerke
 * 
 */
public class BackupOfEOO extends DynamicClusteringAlgorithm {

	private static final String NAME = "EOO";
	
	public enum Strategy {Optimizer, SimulatedAnnealing, Minimizer};

	public static class Factory implements AlgorithmFactory {

		// parameters - set them using these public fields
		public IndexFactory quality;
		public Strategy strategy;
		public int maxOps; // maximum operations
		public boolean allowShift;
		public boolean allowMerge;
		public boolean allowSplit;

		public DynamicClusteringAlgorithm getAlgorithm(DynamicGraph dynGraph) {
			return new BackupOfEOO(dynGraph, this.quality, this.strategy, this.maxOps, this.allowShift, this.allowMerge, this.allowSplit);
		}

		public void readParams(String[] params) {
			// ignored
		}

		@Deprecated
		public void setParams(List<Object> params) throws IncompleteArgumentException {
		}


	};

	
	protected Strategy strategy;
	protected Algorithm optimizer;
	
	protected IndexFactory quality;
	protected int maxOps;
	protected boolean allowShift;
	protected boolean allowMerge;
	protected boolean allowSplit;
	

	public BackupOfEOO(DynamicGraph dynGraph, IndexFactory quality, Strategy strategy, int maxOps, boolean allowShift, boolean allowMerge, boolean allowSplit) {
		super(dynGraph);
		this.quality = quality;
		this.maxOps = maxOps;
		this.allowShift = allowShift;
		this.allowMerge = allowMerge;
		this.allowSplit = allowSplit;
		this.strategy = strategy;
		
		// set the optimizer according to the given strategy
		if (strategy == Strategy.Optimizer) {
			this.optimizer = new ElementalOperationOptimizer(quality, allowShift, allowMerge, allowSplit);
		} else if (strategy == Strategy.SimulatedAnnealing) {
			this.optimizer = new ElementalOperationSimulatedAnnealing(quality, allowShift, allowMerge, allowSplit);
		} else if (strategy == Strategy.Minimizer) {
			this.optimizer = new ElementalOperationMinimizer(quality, allowShift, allowMerge, allowSplit, maxOps);
		} else {
			throw new IllegalArgumentException("parameter strategy is wrong");
		}
	}

	@Override
	public String toString() {
		return this.id();
	}

	@Override
	protected void addEvent(boolean notificationTime, Cluster to, NodeCursor nc) {
		// ignore
	}

	@Override
	protected void deleteEvent(boolean notificationTime, Cluster deletedCluster) {
		// ignore
	}

	@Override
	protected void edgeCreationEvent(Edge newEdge) {
		// ignore
	}

	@Override
	protected void edgeRemovalEvent(boolean notificationTime, Edge edgeToRemove) {
		// ignore
	}

	@Override
	protected void extractEvent(boolean notificationTime, Cluster from, NodeCursor nc) {
		// ignore
	}

	@Override
	protected void mergeEvent(boolean notificationTime, Cluster a, Cluster b) {
		// ignore
	}

	@Override
	protected void moveEvent(boolean notificationTime, Cluster from, Cluster to, NodeCursor nc) {
		// ignore
	}

	@Override
	protected void newClusterEvent(boolean notificationTime, Cluster newCluster) {
		// ignore

	}

	@Override
	protected void newWeightsEvent() {
		throw new RuntimeException("Not implemented, should not happen!");
	}

	@Override
	protected void nodeCreationEvent(Node newNode) {
		this.clustering.newCluster(newNode);
	}

	@Override
	protected void nodeRemovalEvent(boolean notificationTime, Node nodeToRemove) {
		// ignore
	}

	@Override
	protected void policyChanged(boolean policy) {
		// ignored
	}

	@Override
	protected void resetEvent(boolean notificationTime) {
		// throw new RuntimeException("Not implemented, should not happen!");
	}

	@Override
	protected void resetLargeClusterEvent(boolean notificationTime, Cluster c) {
		// ignored
	}

	@Override
	protected void resetSingletonsEvent(boolean notificationTime) {
		// ignored
	}

	@Override
	protected void splitEvent(boolean notificationTime, Cluster a, Cluster b) {
		// ignore
	}

	@Override
	protected void weightChangedEvent(Edge e, double oldWeight, double newWeight) {
		// ignored
	}

	public void onTimeStepEvent() {
		this.timer.start();
		//	
		this.optimizer.run(this.clustering);
		//
		this.timer.stop();
		
		// DEBUG
		System.out.println(">> " + this.id() + " done after" + this.timer.read(TimeUnit.MilliSeconds) + " [ms]");
		// DEBUG
	}

	@Override
	public String id() {
		return NAME + "(" + this.strategy + "," + this.quality + "," + (allowMerge ? "Merge," : "NoMerge," ) + (allowShift ? "Shift," : "NoShift," ) + (allowSplit ? "Split," : "NoSplit," ) + this.maxOps + "):" +  this.uniqueID;
	}

}
