package de.uka.algo.dynclustering.algos;

import java.util.HashMap;
import java.util.List;

import org.apache.commons.lang.IncompleteArgumentException;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeList;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.algorithms.Algorithm;
import de.uka.algo.clustering.algorithms.ElementalOperationMinimizer;
import de.uka.algo.clustering.algorithms.ElementalOperationOptimizer;
import de.uka.algo.clustering.algorithms.ElementalOperationSimulatedAnnealing;
import de.uka.algo.clustering.algorithms.newman.util.ClusterAdjacencyMatrix;
import de.uka.algo.clustering.index.IndexFactory;
import de.uka.algo.dynclustering.algos.base.AlgorithmFactory;
import de.uka.algo.dynclustering.algos.base.DynamicClusteringAlgorithm;
import de.uka.algo.dynclustering.graphs.DynamicGraph;
import de.uka.algo.dynclustering.util.RunTimer.TimeUnit;
import de.uka.algo.graphs.GraphInterpretation;
import de.uka.algo.math.RandomArray;
import de.uka.algo.util.GYCursor;

/**
 * Dynamic version of the Elemental Optimization Operation clustering algorithm
 * 
 * @author Robert Goerke
 * 
 */
public class EOO extends DynamicClusteringAlgorithm {

	private static final String NAME = "EOO";
	
	public enum Strategy {Optimizer, SimulatedAnnealing, Minimizer};

	public static class Factory implements AlgorithmFactory {

		// parameters - set them using these public fields
		public IndexFactory quality;
		public Strategy strategy;
		public int maxOps; // maximum operations
		public boolean allowShift;
		public boolean allowMerge;
		public boolean allowSplit;

		public DynamicClusteringAlgorithm getAlgorithm(DynamicGraph dynGraph) {
			return new EOO(dynGraph, this.quality, this.strategy, this.maxOps, this.allowShift, this.allowMerge, this.allowSplit);
		}

		public void readParams(String[] params) {
			// ignored
		}

		@Deprecated
		public void setParams(List<Object> params) throws IncompleteArgumentException {
		}


	};

	
	protected Strategy strategy;
//	out, non-general: only modularity!
//	protected Algorithm optimizer;
	
	protected IndexFactory quality;
	protected int maxOps;
	protected boolean allowShift;
	protected boolean allowMerge;
	protected boolean allowSplit;
	

	public EOO(DynamicGraph dynGraph, IndexFactory quality, Strategy strategy, int maxOps, boolean allowShift, boolean allowMerge, boolean allowSplit) {
		super(dynGraph);
		this.quality = quality;
		this.maxOps = maxOps;
		this.allowShift = allowShift;
		this.allowMerge = allowMerge;
		this.allowSplit = allowSplit;
		this.strategy = strategy;
		
//		out, non-general: only modularity!
//		// set the optimizer according to the given strategy
//		if (strategy == Strategy.Optimizer) {
//			this.optimizer = new ElementalOperationOptimizer(quality, allowShift, allowMerge, allowSplit);
//		} else if (strategy == Strategy.SimulatedAnnealing) {
//			this.optimizer = new ElementalOperationSimulatedAnnealing(quality, allowShift, allowMerge, allowSplit);
//		} else if (strategy == Strategy.Minimizer) {
//			this.optimizer = new ElementalOperationMinimizer(quality, allowShift, allowMerge, allowSplit, maxOps);
//		} else {
//			throw new IllegalArgumentException("parameter strategy is wrong");
//		}
	}

	@Override
	public String toString() {
		return this.id();
	}

	@Override
	protected void addEvent(boolean notificationTime, Cluster to, NodeCursor nc) {
		// ignore
	}

	@Override
	protected void deleteEvent(boolean notificationTime, Cluster deletedCluster) {
		// ignore
	}

	@Override
	protected void edgeCreationEvent(Edge newEdge) {
		// ignore
	}

	@Override
	protected void edgeRemovalEvent(boolean notificationTime, Edge edgeToRemove) {
		// ignore
	}

	@Override
	protected void extractEvent(boolean notificationTime, Cluster from, NodeCursor nc) {
		// ignore
	}

	@Override
	protected void mergeEvent(boolean notificationTime, Cluster a, Cluster b) {
		// ignore
	}

	@Override
	protected void moveEvent(boolean notificationTime, Cluster from, Cluster to, NodeCursor nc) {
		// ignore
	}

	@Override
	protected void newClusterEvent(boolean notificationTime, Cluster newCluster) {
		// ignore

	}

	@Override
	protected void newWeightsEvent() {
		throw new RuntimeException("Not implemented, should not happen!");
	}

	@Override
	protected void nodeCreationEvent(Node newNode) {
		this.clustering.newCluster(newNode);
	}

	@Override
	protected void nodeRemovalEvent(boolean notificationTime, Node nodeToRemove) {
		// ignore
	}

	@Override
	protected void policyChanged(boolean policy) {
		// ignored
	}

	@Override
	protected void resetEvent(boolean notificationTime) {
		// throw new RuntimeException("Not implemented, should not happen!");
	}

	@Override
	protected void resetLargeClusterEvent(boolean notificationTime, Cluster c) {
		// ignored
	}

	@Override
	protected void resetSingletonsEvent(boolean notificationTime) {
		// ignored
	}

	@Override
	protected void splitEvent(boolean notificationTime, Cluster a, Cluster b) {
		// ignore
	}

	@Override
	protected void weightChangedEvent(Edge e, double oldWeight, double newWeight) {
		// ignored
	}

	public void onTimeStepEvent() {
		this.timer.start();
		//	
//		out, non-general: only modularity!
//		this.optimizer.run(this.clustering);
		double bestDelta = 1.0;
		int runs = 0;
		while ((bestDelta > 0) && (runs <= maxOps)){
			bestDelta = elementalOptimizeRun(this.clustering);
			++runs;
		}
		//
		this.timer.stop();
		
		// DEBUG
		System.out.println(">> " + this.id() + " done after" + this.timer.read(TimeUnit.MilliSeconds) + " [ms]");
		// DEBUG
	}

	private double elementalOptimizeRun(Clustering clustering) {
		// ignore options for now: all options enabled!
		GraphInterpretation gi = clustering.getInterpretation();
		double W = clustering.getOverAllWeight();
		double homeAdj;
		double bestMoveDelta, splitDelta, bestMergeDelta;
		double bestDelta = -1.0;
		Cluster bestMoveTarget, bestMergeTarget;
		HashMap<Cluster,Double> clusterWeight = new HashMap<Cluster, Double>();
		HashMap<Cluster,HashMap<Cluster,Double>> clusterAdjacency = new HashMap<Cluster, HashMap<Cluster,Double>>();
		Node[] nodeArray = clustering.getGraph().getNodeArray();
		int n = nodeArray.length;
		RandomArray ran = new RandomArray();
		int[] randomOrder = ran.nextArray(n);
		
		//prefill HashMaps for my nerves
		for (GYCursor<Cluster> cc1 = clustering.clusters(); cc1.ok(); cc1.next()){
			clusterWeight.put(cc1.active(), 0.0);
			clusterAdjacency.put(cc1.active(), new HashMap<Cluster, Double>());
			for (GYCursor<Cluster> cc2 = clustering.clusters(); cc2.ok(); cc2.next()){
//				if (cc1.active().getIndex() <= cc2.active().getIndex()){
					clusterAdjacency.get(cc1.active()).put(cc2.active(), 0.0);
//				}
			}
		}
		
		//helper calcs:
		double w;
		Cluster sc,tc;
		for (EdgeCursor ec = clustering.getGraph().edges(); ec.ok(); ec.next()){
			w = gi.getWeight(ec.edge());
			sc = clustering.getCluster(ec.edge().source());
			tc = clustering.getCluster(ec.edge().target());
			clusterWeight.put(sc, clusterWeight.get(sc) + w);
			clusterWeight.put(tc, clusterWeight.get(tc) + w);
			clusterAdjacency.get(sc).put(tc,clusterAdjacency.get(sc).get(tc) + w);
			clusterAdjacency.get(tc).put(sc,clusterAdjacency.get(tc).get(sc) + w);
		}
		
		//off we go:
		for (int i = 0; i<n; ++i){
			Node v = nodeArray[randomOrder[i]];
			double selfLoopWeight = gi.getAdjacency(v, v);
			Cluster homeCluster = clustering.getCluster(v);
			double wv = 0.0;
			for (EdgeCursor ec = v.edges(); ec.ok(); ec.next()){
				wv += gi.getWeight(ec.edge());
			}
//			//loops are twice in list of edges... good
//			wv -= gi.getAdjacency(v,v);
			//split:
			if (clustering.getCluster(v).size() == 1){
				splitDelta = 0;
			} else {
				homeAdj = adjacencyToCluster(v, homeCluster);
				splitDelta = - (homeAdj/W) + ((2.0 * wv * (clusterWeight.get(homeCluster) - wv)) / (4.0*W*W)) - ((wv*wv) / (4*W*W));
			}
			//merge:
			double currentMergeDelta;
			bestMergeDelta = -1000000.0;
			bestMergeTarget = homeCluster;
			for (GYCursor<Cluster> cc = clustering.clusters(); cc.ok(); cc.next()){
				if (cc.active().equals(homeCluster))
					continue;
				currentMergeDelta = clusterAdjacency.get(homeCluster).get(cc.active());
				currentMergeDelta = currentMergeDelta / W - ((2.0*clusterWeight.get(homeCluster) * clusterWeight.get(cc.active()))/(4.0*W*W));
				if (currentMergeDelta > bestMergeDelta){
					bestMergeDelta = currentMergeDelta;
					bestMergeTarget = cc.active();
				}
			}
			//move:
			double currentMoveDelta;
			bestMoveDelta = -1000000.0;
			bestMoveTarget = homeCluster;
			Cluster otherCluster;
			for (NodeCursor nc = v.neighbors(); nc.ok(); nc.next()){
				otherCluster = clustering.getCluster(nc.node());
				if (otherCluster.equals(homeCluster))
					continue;
				currentMoveDelta = (adjacencyToCluster(v, otherCluster) - adjacencyToCluster(v, homeCluster) - selfLoopWeight)/W; 
				currentMoveDelta += ((2.0*wv*(clusterWeight.get(homeCluster)-wv)) - (2.0*wv*clusterWeight.get(otherCluster)))/(4.0*W*W);
				if (currentMoveDelta > bestMoveDelta){
					bestMoveDelta = currentMoveDelta;
					bestMoveTarget = otherCluster;
				}
			}
			
			//do best thing and adjust data structures
			//split wins:
			if ((splitDelta > bestMoveDelta) && (splitDelta > bestMergeDelta) && (splitDelta > 0.0)){
//				System.out.println("split");
				bestDelta = splitDelta;
				//do the split
				clustering.split(homeCluster, new NodeList(v).nodes());
				//data structures
				clusterWeight.put(homeCluster,clusterWeight.get(homeCluster) - wv);
				Cluster newCluster = clustering.getCluster(v);
				clusterWeight.put(newCluster, wv);
				for (GYCursor<Cluster> cc2 = clustering.clusters(); cc2.ok(); cc2.next()){
					clusterAdjacency.get(newCluster).put(cc2.active(), 0.0);
					clusterAdjacency.get(cc2.active()).put(newCluster, 0.0);
				}
				Cluster adjCluster;
				for (EdgeCursor ec = v.edges(); ec.ok(); ec.next()){
					adjCluster = clustering.getCluster(ec.edge().opposite(v));
					clusterAdjacency.get(newCluster).put(adjCluster,clusterAdjacency.get(newCluster).get(adjCluster)+gi.getWeight(ec.edge()));
					clusterAdjacency.get(adjCluster).put(newCluster,clusterAdjacency.get(adjCluster).get(newCluster)+gi.getWeight(ec.edge()));
				}
			}
			//merge wins:
			if ((bestMergeDelta >= splitDelta) && (bestMergeDelta >= bestMoveDelta) && (bestMergeDelta > 0.0)){
//				System.out.println("merge");
				bestDelta = bestMergeDelta;
				//first do data structures, lest all is lost to demons of null
				Cluster newCluster, deadCluster;
				if (homeCluster.size() >= bestMergeTarget.size()){
					newCluster = homeCluster;
					deadCluster = bestMergeTarget;
				} else {
					newCluster = bestMergeTarget;
					deadCluster = homeCluster;
				}
//				if (deadCluster.equals(newCluster))
//					System.out.println("tot = neu!");
//				System.out.println(deadCluster + " " + newCluster);
				clusterWeight.put(newCluster, clusterWeight.get(newCluster) + clusterWeight.get(deadCluster));
				clusterWeight.remove(deadCluster);
				double tmp;
				for (GYCursor<Cluster> cc2 = clustering.clusters(); cc2.ok(); cc2.next()){
//					System.out.println(newCluster + "," + cc2.active() + "," + clusterAdjacency.get(newCluster).get(cc2.active()));
//					System.out.println(deadCluster + "," + cc2.active() + "," + clusterAdjacency.get(deadCluster).get(cc2.active()));
					tmp = clusterAdjacency.get(newCluster).get(cc2.active()) + clusterAdjacency.get(deadCluster).get(cc2.active());
					clusterAdjacency.get(newCluster).put(cc2.active(), tmp);
					clusterAdjacency.get(cc2.active()).put(newCluster, tmp);
//					clusterAdjacency.get(cc2.active()).remove(deadCluster);
				}
				for (GYCursor<Cluster> cc2 = clustering.clusters(); cc2.ok(); cc2.next()){
					clusterAdjacency.get(cc2.active()).remove(deadCluster);
				}
				clusterAdjacency.get(newCluster).remove(deadCluster);
				clusterAdjacency.remove(deadCluster);
				//do the merge
				Cluster tmpCluster = clustering.merge(homeCluster, bestMergeTarget);
//				System.out.println(deadCluster + " " + deadCluster.size());
//				if (!tmpCluster.equals(newCluster))
//					System.out.println("aaahhh der falsche ist gestorben!");
			}
			//move wins (singleton are merged instead of moved, thus dont worry about singletons here)
			if ((bestMoveDelta >= splitDelta) && (bestMoveDelta > bestMergeDelta) && (bestMoveDelta > 0.0)){
//				System.out.println("move");
				bestDelta = bestMoveDelta;
				//first do data structures, lest all is lost to demons of null
				clusterWeight.put(bestMoveTarget, clusterWeight.get(bestMoveTarget) + wv);
				clusterWeight.put(homeCluster, clusterWeight.get(homeCluster) - wv);
				double tmp;
				Cluster adjCluster;
				for (EdgeCursor ec = v.edges(); ec.ok(); ec.next()){
					if (ec.edge().opposite(v).equals(v))
						continue;
					tmp = gi.getWeight(ec.edge());
					adjCluster = clustering.getCluster(ec.edge().opposite(v));
					clusterAdjacency.get(homeCluster).put(adjCluster, clusterAdjacency.get(homeCluster).get(adjCluster) - tmp);
					clusterAdjacency.get(adjCluster).put(homeCluster, clusterAdjacency.get(adjCluster).get(homeCluster) - tmp);
					clusterAdjacency.get(bestMoveTarget).put(adjCluster, clusterAdjacency.get(bestMoveTarget).get(adjCluster) + tmp);
					clusterAdjacency.get(adjCluster).put(bestMoveTarget, clusterAdjacency.get(adjCluster).get(bestMoveTarget) + tmp);
				}
				//do the move
				clustering.move(homeCluster, bestMoveTarget, v);
			}
		}
		return bestDelta;
	}
	
	/**
	 * Returns the sum of edge weights between aNode and the nodes of aCluster
	 * 
	 * @param aNode
	 * @param aCluster
	 * @return sum of edge weights between aNode and aCluster
	 */
	/**
	 * @param aNode
	 * @param aCluster
	 * @return
	 */
	protected double adjacencyToCluster(Node aNode, Cluster aCluster) {
		double weight = 0.0;
		Clustering aClustering = aCluster.getClustering();
		GraphInterpretation gi = aClustering.getInterpretation();
		// through all in and then all out edges...
		for (EdgeCursor ec = aNode.edges(); ec.ok(); ec.next()) {
			Edge theEdge = ec.edge();
			Node neighbor = theEdge.opposite(aNode);
			if (aCluster.equals(aClustering.getCluster(neighbor))) {
				weight += gi.getWeight(theEdge);
				// dont count self loops twice, only the egde weight counts
			}
		}
		// but yfiles puts a node twice in its own set of neighbors, in case of
		// a selfloop, thus subtract
		if (aCluster.equals(aClustering.getCluster(aNode)))
			weight -= gi.getAdjacency(aNode, aNode);
		return weight;
	}

	@Override
	public String id() {
		return NAME + "(" + this.strategy + "," + this.quality + "," + (allowMerge ? "Merge," : "NoMerge," ) + (allowShift ? "Shift," : "NoShift," ) + (allowSplit ? "Split," : "NoSplit," ) + this.maxOps + "):" +  this.uniqueID;
	}

}
