package de.uka.algo.dynclustering.algos;

import java.util.List;

import org.apache.commons.lang.IncompleteArgumentException;

import y.base.Edge;
import y.base.Node;
import y.base.NodeCursor;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.ClusteringListener;
import de.uka.algo.clustering.algorithms.newman.NewmanClusterer;
import de.uka.algo.clustering.algorithms.newman.internal.CoverageQInterface;
import de.uka.algo.clustering.algorithms.newman.internal.GenericAdditiveAdapter;
import de.uka.algo.clustering.algorithms.newman.internal.GenericAdditiveDeltaQMatrix;
import de.uka.algo.dynclustering.algos.base.AlgorithmFactory;
import de.uka.algo.dynclustering.algos.base.DynamicClusteringAlgorithm;
import de.uka.algo.dynclustering.algos.newman.BacktrackAdapter;
import de.uka.algo.dynclustering.algos.newman.DynBacktrackStrategy;
import de.uka.algo.dynclustering.algos.newman.FreeingAdapter;
import de.uka.algo.dynclustering.graphs.DynamicGraph;
import de.uka.algo.dynclustering.update.SubsetStrategy;
import de.uka.algo.dynclustering.update.UpdateStrategy;
import de.uka.algo.dynclustering.util.RunTimer.TimeUnit;

/**
 * A dynamic version of the Newman algorithm. 
 * Uses @see de.uka.algo.clustering.algorithms.newman.NewmanClusterer
 * 
 * @author Christian Staudt
 *
 */
public class Newman extends DynamicClusteringAlgorithm {
	
	
	
	public static class Factory implements AlgorithmFactory {
		
		public UpdateStrategy strategy;

		public DynamicClusteringAlgorithm getAlgorithm(DynamicGraph dynGraph) {
			return new Newman(dynGraph, this.strategy);
		}

		public void setParams(List<Object> params) throws IncompleteArgumentException {
		}
		
	};
	
	
	protected NewmanClusterer newmanClusterer;
	protected UpdateStrategy updateStrategy;
	
	private static final String NAME = "Newman";

	
	public Newman(DynamicGraph dynGraph, UpdateStrategy updateStrategy) {
		super(dynGraph);
		
		CoverageQInterface modI = new CoverageQInterface();
		GenericAdditiveAdapter ma = new GenericAdditiveAdapter(true, modI, new GenericAdditiveDeltaQMatrix(modI));

		this.newmanClusterer = new NewmanClusterer(ma);
		
		this.updateStrategy = updateStrategy;
		this.updateStrategy.initialize(this);
		this.newmanClusterer.setCreatingDendro(false);
		this.newmanClusterer.setStoppingOnPeak(true);
		
		if (this.updateStrategy instanceof DynBacktrackStrategy) {
			this.newmanClusterer.setUpdateStrategy(new BacktrackAdapter((DynBacktrackStrategy) this.updateStrategy));
		} else {
			this.newmanClusterer.setUpdateStrategy(new FreeingAdapter((SubsetStrategy) this.updateStrategy));
		}
	}

	@Override
	public String toString() {
		return this.id();
	}

	@Override 
	protected void addEvent(boolean notificationTime, Cluster to, NodeCursor nc) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void deleteEvent(boolean notificationTime, Cluster deletedCluster) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void edgeCreationEvent(Edge newEdge) {
		this.timer.start();
		//
		this.updateStrategy.onEdgeCreation(newEdge);
		//
		this.timer.stop();

	}

	@Override
	protected void edgeRemovalEvent(boolean notificationTime, Edge edgeToRemove) {
		if (notificationTime == ClusteringListener.PRE) {
			this.timer.start();
			//
			this.updateStrategy.onEdgeRemoval(edgeToRemove);
			//
			this.timer.stop();
		}
	}

	@Override
	protected void extractEvent(boolean notificationTime, Cluster from, NodeCursor nc) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void mergeEvent(boolean notificationTime, Cluster a, Cluster b) {
		if (notificationTime == ClusteringListener.PRE) {
			this.timer.start();
			//
			this.updateStrategy.onMerge(a, b);
			//
			this.timer.stop();
		}

	}

	@Override
	protected void moveEvent(boolean notificationTime, Cluster from, Cluster to, NodeCursor nc) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void newClusterEvent(boolean notificationTime, Cluster newCluster) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void newWeightsEvent() {
		// TODO Auto-generated method stub

	}

	@Override
	protected void nodeCreationEvent(Node newNode) {
		this.timer.start();
		//
		this.clustering.newCluster(newNode);
		this.updateStrategy.onNodeCreation(newNode);
		//
		this.timer.stop();
	}

	@Override
	protected void nodeRemovalEvent(boolean notificationTime, Node nodeToRemove) {

		if (notificationTime == ClusteringListener.PRE) {
			this.timer.start();
			//
	 		this.updateStrategy.onNodeRemoval(nodeToRemove);
			//
			this.timer.stop();
		}

	}

	@Override
	protected void policyChanged(boolean policy) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void resetEvent(boolean notificationTime) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void resetLargeClusterEvent(boolean notificationTime, Cluster c) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void resetSingletonsEvent(boolean notificationTime) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void splitEvent(boolean notificationTime, Cluster a, Cluster b) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void weightChangedEvent(Edge e, double oldWeight, double newWeight) {
		this.timer.start();
		//
		this.updateStrategy.onWeightChange(e, oldWeight, newWeight);
		//
		this.timer.stop();
	}

	public void onTimeStepEvent() {
		this.timer.start();
		//
		if (!(this.dynGraph.isEmpty())) {
			this.newmanClusterer.run(this.clustering);
		}
		//
		this.timer.stop();
		
		// DEBUG
		System.out.println(">> " + this.id() + " done after " + this.timer.read(TimeUnit.MilliSeconds) + " [ms]");
		// DEBUG
	}

	@Override
	public String id() {
		return NAME + "(" + this.updateStrategy + "):" +  this.uniqueID;
	}

}
