package de.uka.algo.dynclustering.algos;

import java.util.List;
import java.util.Map;

import org.apache.commons.lang.IncompleteArgumentException;

import y.base.Edge;
import y.base.Node;
import y.base.NodeCursor;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.ClusteringListener;
import de.uka.algo.clustering.algorithms.EdgeHeapModularity;
import de.uka.algo.dynclustering.algos.base.AlgorithmFactory;
import de.uka.algo.dynclustering.algos.base.DynamicClusteringAlgorithm;
import de.uka.algo.dynclustering.algos.newman.DynBacktrackStrategy;
import de.uka.algo.dynclustering.graphs.DynamicGraph;
import de.uka.algo.dynclustering.update.SubsetStrategy;
import de.uka.algo.dynclustering.update.UpdateStrategy;
import de.uka.algo.dynclustering.util.RunTimer.TimeUnit;

/**
 * A dynamic version of the Newman algorithm. 
 * Uses @see de.uka.algo.clustering.algorithms.newman.NewmanClusterer
 * 
 * @author Christian Staudt
 *
 */
public class Newman2 extends DynamicClusteringAlgorithm {
	
	
	
	public static class Factory implements AlgorithmFactory {
		
		public UpdateStrategy strategy;

		public DynamicClusteringAlgorithm getAlgorithm(DynamicGraph dynGraph) {
			return new Newman2(dynGraph, this.strategy);
		}

		public void setParams(List<Object> params) throws IncompleteArgumentException {
		}
		
	};
	
	
	protected UpdateStrategy updateStrategy;
	protected EdgeHeapModularity ehm;
	
	private static final String NAME = "Newman2";

	
	public Newman2(DynamicGraph dynGraph, UpdateStrategy updateStrategy) {
		super(dynGraph);
		

		this.ehm = new EdgeHeapModularity();
		this.ehm.setRefreshCount(1);
		this.ehm.setAllowUpdate(true);
		if (updateStrategy instanceof DynBacktrackStrategy) {
			this.ehm.setBacktrackable(true);
		}
		
		this.updateStrategy = updateStrategy;
		this.updateStrategy.initialize(this);
		
	}

	@Override
	public String toString() {
		return this.id();
	}

	@Override 
	protected void addEvent(boolean notificationTime, Cluster to, NodeCursor nc) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void deleteEvent(boolean notificationTime, Cluster deletedCluster) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void edgeCreationEvent(Edge newEdge) {
		this.timer.start();
		//
		this.updateStrategy.onEdgeCreation(newEdge);
		//
		this.timer.stop();

	}

	@Override
	protected void edgeRemovalEvent(boolean notificationTime, Edge edgeToRemove) {
		if (notificationTime == ClusteringListener.PRE) {
			this.timer.start();
			//
			this.updateStrategy.onEdgeRemoval(edgeToRemove);
			//
			this.timer.stop();
		}
	}

	@Override
	protected void extractEvent(boolean notificationTime, Cluster from, NodeCursor nc) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void mergeEvent(boolean notificationTime, Cluster a, Cluster b) {
		if (notificationTime == ClusteringListener.PRE) {
			this.timer.start();
			//
			this.updateStrategy.onMerge(a, b);
			//
			this.timer.stop();
		}

	}

	@Override
	protected void moveEvent(boolean notificationTime, Cluster from, Cluster to, NodeCursor nc) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void newClusterEvent(boolean notificationTime, Cluster newCluster) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void newWeightsEvent() {
		// TODO Auto-generated method stub

	}

	@Override
	protected void nodeCreationEvent(Node newNode) {
		this.timer.start();
		//
		this.clustering.newCluster(newNode);
		this.updateStrategy.onNodeCreation(newNode);
		//
		this.timer.stop();
	}

	@Override
	protected void nodeRemovalEvent(boolean notificationTime, Node nodeToRemove) {

		if (notificationTime == ClusteringListener.PRE) {
			this.timer.start();
			//
	 		this.updateStrategy.onNodeRemoval(nodeToRemove);
			//
			this.timer.stop();
		}

	}

	@Override
	protected void policyChanged(boolean policy) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void resetEvent(boolean notificationTime) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void resetLargeClusterEvent(boolean notificationTime, Cluster c) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void resetSingletonsEvent(boolean notificationTime) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void splitEvent(boolean notificationTime, Cluster a, Cluster b) {
		// TODO Auto-generated method stub

	}

	@Override
	protected void weightChangedEvent(Edge e, double oldWeight, double newWeight) {
		this.timer.start();
		//
		this.updateStrategy.onWeightChange(e, oldWeight, newWeight);
		//
		this.timer.stop();
	}

	public void onTimeStepEvent() {
		this.timer.start();
		//
		this.prepareClustering();
		if (!(this.dynGraph.isEmpty())) {
			this.ehm.run(this.clustering);
		} else {
			System.out.println("WARNING: Graph is empty");
		}
		//
		this.timer.stop();
		
		// DEBUG
		System.out.println(">> " + this.id() + " done after " + this.timer.read(TimeUnit.MilliSeconds) + " [ms]");
		// DEBUG
	}

	@Override
	public String id() {
		return NAME + "(" + this.updateStrategy + "):" +  this.uniqueID;
	}
	
	
	/**
	 *  Nodes marked by a subset strategy to singletons.
	 */
	protected void extractNodes() {
		for (Node u :((SubsetStrategy) this.updateStrategy).getMarkedNodes()) {
			if (this.dynGraph.contains(u)) {
				Cluster C = this.clustering.getCluster(u);
				if (C != null) {
					this.clustering.extract(C, u);
				}
			}
		}
		((SubsetStrategy) this.updateStrategy).resetNodes();
	}
	
	/**
	 * Apply backtracked clustering to actual clustering.
	 */
	protected void backtrack2clustering() {
		// stop listening
		((DynBacktrackStrategy) this.updateStrategy).listening = false;
		
		this.clustering.resetSingletons();
		
		Map<Node,Node> partitionMap = ((DynBacktrackStrategy) this.updateStrategy).getPartitionMap();
		for (NodeCursor nc = this.clustering.getGraph().nodes(); nc.ok(); nc.next()) {
			Node representative = partitionMap.get(nc.node());
			this.clustering.merge(nc.node(), representative);
		}
		
		// start listening
		((DynBacktrackStrategy) this.updateStrategy).listening = true;
	}
	
	/**
	 * Prepare the clustering depending on the update strategy.
	 */
	protected void prepareClustering() {
		if (this.updateStrategy instanceof DynBacktrackStrategy) {
			this.backtrack2clustering();
		} else if (this.updateStrategy instanceof SubsetStrategy) {
			this.extractNodes();
			this.clustering.pool2Singletons();
		}
	}

}
