package de.uka.algo.dynclustering.algos;

import ilog.concert.IloException;
import ilog.concert.IloNumVar;
import ilog.cplex.IloCplex;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import lpsolve.LpSolveException;

import y.base.Edge;
import y.base.Node;
import y.base.NodeCursor;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.ClusteringListener;
import de.uka.algo.dynclustering.algos.base.AlgorithmFactory;
import de.uka.algo.dynclustering.algos.base.DynamicClusteringAlgorithm;
import de.uka.algo.dynclustering.algos.ilp.PartialCplexGenerator;
import de.uka.algo.dynclustering.algos.ilp.PartialLpSolveGenerator;
import de.uka.algo.dynclustering.algos.ilp.PartialModularity;
import de.uka.algo.dynclustering.algos.ilp.PartialObjectiveFunction;
import de.uka.algo.dynclustering.algos.ilp.PartialWeightedModularity;
import de.uka.algo.dynclustering.graphs.DynamicGraph;
import de.uka.algo.dynclustering.graphs.EmailGraph;
import de.uka.algo.dynclustering.update.SubsetStrategy;
import de.uka.algo.dynclustering.util.YUtils;
import de.uka.algo.dynclustering.util.RunTimer.TimeUnit;
import lpsolve.LpSolve;

public class PartialILP extends DynamicClusteringAlgorithm {

	// available solvers
	public static enum Solver {
		LpSolve, Cplex
	};

	// available objective functions
	public static enum Objective {
		Modularity, WeightedModularity, Performance, WeightedPerformance
	};

	public static enum Variant {
		NoMerge, Merge
	};

	protected Solver solver;
	protected Objective objective;
	protected SubsetStrategy subsetStrategy;
	protected Variant variant;
	
	private IloCplex myLp;

	private static final String NAME = "PartialILP";

	public static class Factory implements AlgorithmFactory {

		// parameters
		public Objective obj;
		public Solver solver;
		public SubsetStrategy strategy;
		public Variant variant;

		public DynamicClusteringAlgorithm getAlgorithm(DynamicGraph dynGraph) {
			PartialILP partialILP = new PartialILP(dynGraph, this.solver, this.obj, this.variant, this.strategy);
			return partialILP;
		}

	}

	public PartialILP(DynamicGraph dynGraph, Solver solver, Objective objective, Variant variant, SubsetStrategy strategy) {
		super(dynGraph);
		this.solver = solver;
		this.objective = objective;
		this.variant = variant;
		this.subsetStrategy = strategy;

		this.subsetStrategy.initialize(this);

	}

	@Override
	protected void addEvent(boolean arg0, Cluster arg1, NodeCursor arg2) {
		// ignored

	}

	@Override
	protected void deleteEvent(boolean arg0, Cluster arg1) {
		// ignored

	}

	@Override
	protected void edgeCreationEvent(Edge arg0) {
		this.timer.start();
		//
		this.subsetStrategy.onEdgeCreation(arg0);
		this.freeNodes();
		this.run();
		//
		this.timer.stop();
	}

	@Override
	protected void edgeRemovalEvent(boolean arg0, Edge arg1) {

		if (arg0 == ClusteringListener.PRE) {
			this.timer.start();
			//
			this.subsetStrategy.onEdgeRemoval(arg1);
			this.freeNodes();
			this.run();
			//
			this.timer.stop();
		}

	}

	@Override
	protected void extractEvent(boolean arg0, Cluster arg1, NodeCursor arg2) {
		// ignored

	}

	@Override
	protected void mergeEvent(boolean arg0, Cluster arg1, Cluster arg2) {
		// ignored

	}

	@Override
	protected void moveEvent(boolean arg0, Cluster arg1, Cluster arg2, NodeCursor arg3) {
		// ignored

	}

	@Override
	protected void newClusterEvent(boolean arg0, Cluster arg1) {
		// ignored

	}

	@Override
	protected void newWeightsEvent() {
		// ignored

	}

	@Override
	protected void nodeCreationEvent(Node arg0) {
		// make it a singleton
		this.clustering.newCluster(arg0);
	}

	@Override
	protected void nodeRemovalEvent(boolean arg0, Node arg1) {
		if (arg0 == PRE) {
			this.timer.start();
			//
			this.subsetStrategy.onNodeRemoval(arg1);
			//
			this.timer.stop();
		}
	}

	@Override
	protected void policyChanged(boolean arg0) {
		// ignored

	}

	@Override
	protected void resetEvent(boolean arg0) {
		// ignored

	}

	@Override
	protected void resetLargeClusterEvent(boolean arg0, Cluster arg1) {
		// ignored

	}

	@Override
	protected void resetSingletonsEvent(boolean arg0) {
		// ignored

	}

	@Override
	protected void splitEvent(boolean arg0, Cluster arg1, Cluster arg2) {
		// ignored

	}

	@Override
	protected void weightChangedEvent(Edge arg0, double arg1, double arg2) {
		this.timer.start();
		//
		this.subsetStrategy.onWeightChange(arg0, arg1, arg2);
		this.freeNodes();
		this.run();
		//
		this.timer.stop();
	}

	public void onTimeStepEvent() {
		// nothing to do

		// DEBUG
		System.out.println(">> " + this.id() + " done after " + this.timer.read(TimeUnit.MilliSeconds) + " [ms] ");
		// DEBUG
	}

	/**
	 * Select objective function and solver, then run the PartialILP algorithm
	 */
	protected void run() {

		// DEBUG
		System.out.println("<<<<<<<<<<< Running PartialILP for time " + this.dynGraph.t());
		// DEBUG

		PartialObjectiveFunction objF = null;
		switch (this.objective) {
		case Modularity:
			objF = new PartialModularity(this.dynGraph, this.clustering);
			break;
		case WeightedModularity:
			EmailGraph emailGraph = (EmailGraph) this.dynGraph;
			objF = new PartialWeightedModularity(emailGraph.getWeightedGraph(), this.clustering);
			break;
		}

		switch (this.solver) {
		case LpSolve:
			this.runLpSolve2(objF);
			break;
		case Cplex:
			this.runCplex(objF);
		}

	}

	/**
	 * Run the Cplex solver.
	 * 
	 * @param objF
	 *            an objective function
	 */
	protected void runCplex(PartialObjectiveFunction objF) {

		Map<String, IloNumVar> name2variable = new HashMap<String, IloNumVar>();
		try {
			if (this.myLp == null) {
				this.myLp = new IloCplex();
			}
//			IloCplex myLp = new IloCplex();
			myLp.clearModel();
			myLp.setOut(null);

			if (this.variant == Variant.Merge) {
				// TODO: test this

				// generate variables
				myLp = PartialCplexGenerator.generateNodePairVariables(myLp, this.clustering, true, name2variable);
				myLp = PartialCplexGenerator.generateNodeClusterVariables(myLp, this.clustering, true, name2variable);
				myLp = PartialCplexGenerator.generateClusterPairVariables(myLp, this.clustering, true, name2variable);
				// create triangle-inequality constraints for nodes
				myLp = PartialCplexGenerator.addMetricNodeTripleConstraints(myLp, this.clustering, true, name2variable);
				// create triangle-inequality constraints for free nodes and
				// clusters
				myLp = PartialCplexGenerator.addMetricNodePairClusterConstraints(myLp, this.clustering, true, name2variable);
				//
				myLp = PartialCplexGenerator.addMetricNodeClusterPairConstraints(myLp, this.clustering, true, name2variable);
				// 
				myLp = PartialCplexGenerator.addMetricClusterTripleConstraints(myLp, this.clustering, true, name2variable);
				// ROB:add objective function
				myLp = PartialCplexGenerator.addPartialObjectiveFunction(myLp, this.clustering, objF, true, name2variable, (this.variant == Variant.Merge));

				// solve here
				myLp.solve();

				// apply solution
				ArrayList<Node> nodes = YUtils.nodeCursor2ArrayList(this.clustering.poolNodes());
				ArrayList<Cluster> clusters = YUtils.gYCursor2ArrayList(this.clustering.clusters());
				int n = nodes.size();
				int k = clusters.size();

				// add pool nodes to clusters
				for (int u = 0; u < n; ++u)
					for (int c = 0; c < k; ++c) {
						IloNumVar Yuc = name2variable.get("Y(" + u + "," + c + ")");
						double value = myLp.getValue(Yuc);
						if (value < 0.001) {
							if (this.clustering.getCluster(nodes.get(u)) == null) {
								this.clustering.add(clusters.get(c), nodes.get(u));
							}
						}
					}

				// merge clusters
				// TODO: correct position for this block?
				
				Set<Integer> merged = new HashSet<Integer>(); // remember clusters which have already been merged
				
				for (int c = 0; c < k; ++c)
					for (int d = c + 1; d < k; ++d) {
						IloNumVar Zcd = name2variable.get("Z(" + c + "," + d + ")");
						double value = myLp.getValue(Zcd);
						if (value < 0.001) {
							// merge the clusters
							if (! (merged.contains(c) || merged.contains(d)) ) {
								this.clustering.merge(clusters.get(c), clusters.get(d));
								merged.add(c);
								merged.add(d);
							}

						}
					}

				// cluster remaining nodes
				if (this.clustering.poolSize() > 0) {
					this.clustering.pool2Singletons();
				}
				for (int u = 0; u < n; ++u)
					for (int v = u + 1; v < n; ++v) {
						IloNumVar Xuv = name2variable.get("X(" + u + "," + v + ")");
						double value = myLp.getValue(Xuv);
						if (value < 0.001) {
							this.clustering.merge(nodes.get(u), nodes.get(v));
						}
					}

			} else if (this.variant == Variant.NoMerge) {

				// generate variables
				myLp = PartialCplexGenerator.generateNodePairVariables(myLp, this.clustering, true, name2variable);
				myLp = PartialCplexGenerator.generateNodeClusterVariables(myLp, this.clustering, true, name2variable);
				// create triangle-inequality constraints for nodes
				myLp = PartialCplexGenerator.addMetricNodeTripleConstraints(myLp, this.clustering, true, name2variable);
				// create triangle-inequality constraints for free nodes and
				// clusters
				myLp = PartialCplexGenerator.addMetricNodePairClusterConstraints(myLp, this.clustering, true, name2variable);
				// nodes can only belong to one cluster
				myLp = PartialCplexGenerator.addMetricUniquenessConstraints(myLp, this.dynGraph, this.clustering, true, name2variable);
				// add objective function
				myLp = PartialCplexGenerator.addPartialObjectiveFunction(myLp, this.clustering, objF, true, name2variable, (this.variant == Variant.Merge));

				// solve here
				myLp.solve();

				// apply solution
				ArrayList<Node> nodes = YUtils.nodeCursor2ArrayList(this.clustering.poolNodes());
				ArrayList<Cluster> clusters = YUtils.gYCursor2ArrayList(this.clustering.clusters());
				int n = nodes.size();
				int k = clusters.size();

				// add pool nodes to clusters
				for (int u = 0; u < n; ++u)
					for (int c = 0; c < k; ++c) {
						IloNumVar Yuc = name2variable.get("Y(" + u + "," + c + ")");
						double value = myLp.getValue(Yuc);
						if (value < 0.001) {
							this.clustering.add(clusters.get(c), nodes.get(u));
						}
					}

				// cluster remaining nodes
				if (this.clustering.poolSize() > 0) {
					this.clustering.pool2Singletons();
				}
				for (int u = 0; u < n; ++u)
					for (int v = u + 1; v < n; ++v) {
						IloNumVar Xuv = name2variable.get("X(" + u + "," + v + ")");
						double value = myLp.getValue(Xuv);
						if (value < 0.001) {
							this.clustering.merge(nodes.get(u), nodes.get(v));
						}
					}

			} // end variants

		} catch (IloException e) {
			e.printStackTrace();
		}

	}

	/**
	 * Optimize the given objective function using lp_solve and update
	 * clustering.
	 * 
	 * @param objF
	 */
	@Deprecated
	protected void runLpSolve(PartialObjectiveFunction objF) {

		LpSolve lp = this.makeMetricLpSolve(this.clustering.poolSize(), this.clustering.clusterCount()); // add
		// constraints

		try {
			ArrayList<Node> poolNodes = YUtils.nodeCursor2ArrayList(this.clustering.poolNodes());
			ArrayList<Cluster> clusters = YUtils.gYCursor2ArrayList(this.clustering.clusters());

			int cn = clusters.size();
			int n = poolNodes.size();
			int offset = n * (n - 1) / 2;

			// get number of variables
			int varNumber = n * (n - 1) / 2 + n * cn;
			// declaring array that store all coefficients of the objective
			// function
			double coefficients[] = new double[varNumber + 1];
			// initializing the array
			for (int i = 0; i < coefficients.length; ++i)
				coefficients[i] = 0;
			// copy coefficient from objfct
			for (int i = 0; i < poolNodes.size(); ++i)
				for (int j = i + 1; j < poolNodes.size(); ++j) {
					int ij = getPos(i, j, n) + 1;
					coefficients[ij] = objF.getCoefficient(poolNodes.get(i), poolNodes.get(j));
				}
			for (int i = 0; i < poolNodes.size(); ++i)
				for (int j = 0; j < clusters.size(); ++j) {
					int ij = getPos2(i, j, n) + offset + 1;
					coefficients[ij] = objF.getCoefficient(poolNodes.get(i), clusters.get(j));
				}

			// set objective function
			lp.setObjFn(coefficients);
			// set type of optimization
			lp.setMinim();
			// set verbosity
			lp.setVerbose(0);

			// solve the LP
			lp.solve();

			double[] solution = lp.getPtrVariables().clone();

			// add pool nodes to clusters
			for (int i = 0; i < poolNodes.size(); ++i)
				for (int j = 0; j < clusters.size(); ++j) {
					int ij = getPos2(i, j, n) + offset;
					double XuC = solution[ij];
					Node u = poolNodes.get(i);
					Cluster C = clusters.get(j);
					if (XuC < 0.0001) {
						this.clustering.add(C, u);
					}
				}

			// cluster remaining nodes
			if (this.clustering.poolSize() > 0) {
				this.clustering.pool2Singletons();
			}

			for (int i = 0; i < poolNodes.size(); ++i)
				for (int j = i + 1; j < poolNodes.size(); ++j) {
					int ij = getPos(i, j, n);
					double Xuv = solution[ij]; // decision variable
					if (Xuv < 0.0001) {
						Node u = poolNodes.get(i);
						Node v = poolNodes.get(j);
						assert u != null;
						assert v != null;
						this.clustering.merge(u, v);
					}
				}

		} catch (LpSolveException e) {
			e.printStackTrace();
		}

	}

	/**
	 * Optimize the given objective function using lp_solve and update
	 * clustering.
	 * 
	 * @param objF
	 */
	protected void runLpSolve2(PartialObjectiveFunction objF) {

		int n = this.clustering.poolSize();
		int k = this.clustering.clusterCount();
		ArrayList<Node> nodes = YUtils.nodeCursor2ArrayList(this.clustering.poolNodes());
		ArrayList<Cluster> clusters = YUtils.gYCursor2ArrayList(this.clustering.clusters());

		if (this.variant == Variant.NoMerge) {

			int varNumber = n * (n - 1) / 2 + n * k;

			double[] coefficients = PartialLpSolveGenerator.makeCoefficientsArray(varNumber);

			try {
				// create the lp
				LpSolve lp = PartialLpSolveGenerator.makeMetricLp(varNumber, coefficients);

				// add constraints
				lp = PartialLpSolveGenerator.addMetricNodeTripleConstraints(lp, coefficients, n, k);
				lp = PartialLpSolveGenerator.addMetricNodePairClusterConstraints(lp, coefficients, n, k);
				lp = PartialLpSolveGenerator.addMetricUniquenessConstraints(lp, coefficients, n, k);

				// set objective function
				lp = PartialLpSolveGenerator.addPartialObjectiveFunction(lp, coefficients, n, objF, true, nodes, clusters);

				// set verbosity
				lp.setVerbose(0);

				// solve the LP
				lp.solve();

				// get solution
				double[] solution = lp.getPtrVariables().clone();

				int offset = n * (n - 1) / 2;
				// add pool nodes to clusters
				for (int i = 0; i < nodes.size(); ++i)
					for (int j = 0; j < clusters.size(); ++j) {
						int ij = PartialLpSolveGenerator.getPos2(i, j, n) + offset;
						double XuC = solution[ij];
						Node u = nodes.get(i);
						Cluster C = clusters.get(j);
						if (XuC < 0.0001) {
							this.clustering.add(C, u);
						}
					}

				// cluster remaining nodes
				if (this.clustering.poolSize() > 0) {
					this.clustering.pool2Singletons();
				}

				for (int i = 0; i < nodes.size(); ++i)
					for (int j = i + 1; j < nodes.size(); ++j) {
						int ij = PartialLpSolveGenerator.getPos(i, j, n);
						double Xuv = solution[ij]; // decision variable
						if (Xuv < 0.0001) {
							Node u = nodes.get(i);
							Node v = nodes.get(j);
							assert u != null;
							assert v != null;
							this.clustering.merge(u, v);
						}
					}

			} catch (LpSolveException ex) {
				ex.printStackTrace();
			}

		} else if (this.variant == Variant.Merge) {

			int varNumber = n * (n - 1) / 2 + n * k + (k - 1) * k; // additional
																	// cluster
																	// pair
																	// variables

			double[] coefficients = PartialLpSolveGenerator.makeCoefficientsArray(varNumber);

			try {
				// create the lp
				LpSolve lp = PartialLpSolveGenerator.makeMetricLp(varNumber, coefficients);

				lp = PartialLpSolveGenerator.addMetricNodeTripleConstraints(lp, coefficients, n, k);
				lp = PartialLpSolveGenerator.addMetricNodePairClusterConstraints(lp, coefficients, n, k);
				lp = PartialLpSolveGenerator.addMetricNodeClusterPairConstraints(lp, coefficients, n, k);
				lp = PartialLpSolveGenerator.addMetricClusterTripleConstraints(lp, coefficients, n, k);

				// set objective function
				lp = PartialLpSolveGenerator.addPartialObjectiveFunction(lp, coefficients, n, objF, true, nodes, clusters);

				// set verbosity
				lp.setVerbose(0);

				// solve the LP
				lp.solve();

				// get solution
				double[] solution = lp.getPtrVariables().clone();

				int offset = n * (n - 1) / 2;
				int offset2 = offset + n * k;
				// add pool nodes to clusters
				for (int i = 0; i < nodes.size(); ++i)
					for (int j = 0; j < clusters.size(); ++j) {
						int ij = PartialLpSolveGenerator.getPos2(i, j, n) + offset;
						double XuC = solution[ij];
						Node u = nodes.get(i);
						Cluster C = clusters.get(j);
						if (XuC < 0.0001) {
							this.clustering.add(C, u);
						}
					}

				for (int c = 0; c < k; ++c)
					for (int d = c + 1; d < k; ++d) {
						int cd = PartialLpSolveGenerator.getPos(c, d, k) + offset2;
						double Zcd = solution[cd];
						if (Zcd < 0.001) {
							// merge the clusters
							this.clustering.merge(clusters.get(c), clusters.get(d));
						}
					}

				// cluster remaining nodes
				if (this.clustering.poolSize() > 0) {
					this.clustering.pool2Singletons();
				}

				for (int i = 0; i < nodes.size(); ++i)
					for (int j = i + 1; j < nodes.size(); ++j) {
						int ij = PartialLpSolveGenerator.getPos(i, j, n);
						double Xuv = solution[ij]; // decision variable
						if (Xuv < 0.0001) {
							Node u = nodes.get(i);
							Node v = nodes.get(j);
							assert u != null;
							assert v != null;
							this.clustering.merge(u, v);
						}
					}

			} catch (LpSolveException ex) {
				ex.printStackTrace();
			}
		}

	}

	/**
	 * Returns the linearized position of the (i,j)-th entry in a
	 * triangle-matrix.
	 * 
	 * @param i
	 *            row position
	 * @param j
	 *            column position
	 * @param n
	 *            total size
	 * @return linearized position
	 */
	@Deprecated
	protected static int getPos(int i, int j, int n) {
		return i * n + j - (i + 1) * (i + 2) / 2;
	}

	/**
	 * Returns the linearized position of the (i,j)-th entry in a full matrix.
	 * 
	 * @param i
	 *            row position (nodes)
	 * @param k
	 *            column position (clusters)
	 * @param cols
	 *            number of columns
	 * @return linearized position
	 */
	@Deprecated
	protected static int getPos2(int i, int k, int cols) {
		return k * cols + i;
	}

	/**
	 * Add the constraints using the metric model
	 * 
	 * @param C
	 *            the clustering
	 * @return
	 * 
	 */
	@Deprecated
	protected LpSolve makeMetricLpSolve(int n, int cn) {

		int varNumber = n * (n - 1) / 2 + n * cn;
		int offset = n * (n - 1) / 2;

		LpSolve lp;
		try {
			lp = LpSolve.makeLp(0, varNumber);

			double[] coefficients = new double[varNumber + 1];
			coefficients[0] = 0;
			for (int i = 1; i < coefficients.length; ++i) {
				// set bounds and integer property
				lp.setUpbo(i, 1.0);
				lp.setLowbo(i, 0.0);
				lp.setInt(i, true);
			}

			// create triangle-inequality constraints for nodes
			for (int i = 0; i < n; ++i)
				for (int j = i + 1; j < n; ++j)
					for (int k = j + 1; k < n; ++k) {
						// get linearized positions
						int ij = getPos(i, j, n) + 1;
						int ik = getPos(i, k, n) + 1;
						int jk = getPos(j, k, n) + 1;

						// set triangle-inequality constraints
						coefficients[ij] = 1;
						coefficients[ik] = -1;
						coefficients[jk] = 1;
						lp.addConstraint(coefficients, LpSolve.GE, 0.0);
						coefficients[ij] = 1;
						coefficients[ik] = 1;
						coefficients[jk] = -1;
						lp.addConstraint(coefficients, LpSolve.GE, 0.0);
						coefficients[ij] = -1;
						coefficients[ik] = 1;
						coefficients[jk] = 1;
						lp.addConstraint(coefficients, LpSolve.GE, 0.0);

						// resets entries in the array for reuse
						coefficients[ij] = 0;
						coefficients[ik] = 0;
						coefficients[jk] = 0;
					}

			// create triangle-inequality constraints for nodes and clusters
			for (int i = 0; i < n; ++i)
				for (int j = i + 1; j < n; ++j)
					for (int k = 0; k < cn; ++k) {

						// get linearized positions
						int ij = getPos(i, j, n) + 1;
						int ik = offset + getPos2(i, k, n) + 1;
						int jk = offset + getPos2(j, k, n) + 1;

						// set triangle-inequality constraints
						coefficients[ij] = 1;
						coefficients[ik] = -1;
						coefficients[jk] = 1;
						lp.addConstraint(coefficients, LpSolve.GE, 0.0);
						coefficients[ij] = 1;
						coefficients[ik] = 1;
						coefficients[jk] = -1;
						lp.addConstraint(coefficients, LpSolve.GE, 0.0);
						coefficients[ij] = -1;
						coefficients[ik] = 1;
						coefficients[jk] = 1;
						lp.addConstraint(coefficients, LpSolve.GE, 0.0);

						// resets entries in the array for reuse
						coefficients[ij] = 0;
						coefficients[ik] = 0;
						coefficients[jk] = 0;
					}

			// // nodes can only belong to one cluster
			// for (int i = 0; i < n; ++i) {
			// for (int k = 0; k < cn; ++k) {
			// for (int h = k + 1; h < cn; ++h) {
			// int ik = getPos2(i, k, n) + offset + 1;
			// int ih = getPos2(i, h, n) + offset + 1;
			// coefficients[ik] = 1;
			// coefficients[ih] = 1;
			// lp.addConstraint(coefficients, LpSolve.GE, 1.0);
			//
			// // reset
			// coefficients[ik] = 0;
			// coefficients[ih] = 0;
			// }
			// }
			// }

			for (int i = 0; i < n; ++i) {
				for (int k = 0; k < cn; ++k) {
					int ik = getPos2(i, k, n) + offset + 1;
					coefficients[ik] = 1;
				}
				lp.addConstraint(coefficients, LpSolve.GE, cn - 1);
				// reset
				for (int k = 0; k < cn; ++k) {
					int ik = getPos2(i, k, n) + offset + 1;
					coefficients[ik] = 0;
				}
			}

			return lp;
		} catch (LpSolveException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Extract marked nodes from their clusters.
	 */
	protected void freeNodes() {
		for (Node u : this.subsetStrategy.getMarkedNodes()) {
			if (this.dynGraph.contains(u)) {
				Cluster C = this.clustering.getCluster(u);
				if (C != null) {
					this.clustering.extract(C, u);
				}
			}
		}
		this.subsetStrategy.resetNodes();
	}

	@Override
	public String id() {
		return NAME + "(" + solver + "," + objective + "," + variant + "," + subsetStrategy.toString() + "):" + this.uniqueID;
	}

}
