package de.uka.algo.dynclustering.algos;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Random;

import org.apache.commons.lang.IncompleteArgumentException;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.EdgeMap;
import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeList;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.ClusterMap;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.dynclustering.algos.Blondel.NodeOrderStrategy;
import de.uka.algo.dynclustering.algos.base.AlgorithmFactory;
import de.uka.algo.dynclustering.algos.base.DynamicClusteringAlgorithm;
import de.uka.algo.dynclustering.algos.blondel.ClusterChangeListener;
import de.uka.algo.dynclustering.graphs.DynamicGraph;
import de.uka.algo.dynclustering.util.RunTimer.TimeUnit;
import de.uka.algo.graphs.GraphInterpretation;
import de.uka.algo.graphs.WeightedGraph;
import de.uka.algo.util.GYCursor;

public class StaticBlondel extends DynamicClusteringAlgorithm {
	
	
	public static class Factory implements AlgorithmFactory {
		
		public Blondel.NodeOrderStrategy nodeOrderStrategy;
		public Blondel.StoppingCriterion stoppingCriterion;

		public DynamicClusteringAlgorithm getAlgorithm(DynamicGraph dynGraph) {
			return new StaticBlondel(dynGraph, this.nodeOrderStrategy, this.stoppingCriterion);
		}

		public void setParams(List<Object> params) throws IncompleteArgumentException {
			for (Object param : params) {
				if (param instanceof Blondel.NodeOrderStrategy) {
					this.nodeOrderStrategy = (Blondel.NodeOrderStrategy) param;
				} else if (param instanceof Blondel.StoppingCriterion) {
					this.stoppingCriterion = (Blondel.StoppingCriterion) param;
				}
			}
			if ((this.nodeOrderStrategy == null)) {
				throw new IncompleteArgumentException("Required Parameters: " +
						"Blondel.NodeOrderStrategy");
			}
		}
		
	};
	
	
	public class HierarchyLevel {
		public HierarchyLevel(Graph aGraph, Clustering aClustering) {
			this.theGraph = aGraph;
			this.theClustering = aClustering;
			this.clusterToNextLevelNode = new ClusterMap<Node>(this.theClustering);
			levelToHierarchyLevel.put(levelToHierarchyLevel.size(), this);
		}

		public HierarchyLevel(HierarchyLevel oldLevel) {
			this.theClustering = contractClustering(oldLevel);
			this.theGraph = this.theClustering.getGraph();
			this.clusterToNextLevelNode = new ClusterMap<Node>(this.theClustering);
			levelToHierarchyLevel.put(levelToHierarchyLevel.size(), this);
		}

		public Graph theGraph;
		public Clustering theClustering;
		public ClusterMap<Node> clusterToNextLevelNode;
		public ClusterChangeListener oracle;
	};


	private static final String NAME = "StaticBlondel";
	
	// parameters
	/** the chosen strategy, i.e., in which order are nodes asked */
	protected NodeOrderStrategy nodeOrderStrategy;
	
	protected Blondel.StoppingCriterion stoppingCriterion;
	
//	/** the chosen criterion for when to stop agglomeration */
//	protected StoppingCriterion stopCrit;
//	
//	/** the update policy, i.e., what is a bad node? */
//	protected UpdatePolicy updatePolicy; // needed in static case?
//	
	// algo variables
	protected NodeList initialBadNodesList;

	
	protected boolean isWeighted;
	protected double bestInnerIncrease;
	protected Graph abstractedGraph;
	protected int neighborHandShakes = 0;

	protected double incrementalModularity;
	protected HashMap<Integer, HierarchyLevel> levelToHierarchyLevel = new HashMap<Integer, HierarchyLevel>();
	protected HashSet<Node> invalidNodes = new HashSet<Node>(4);
	
	
	public StaticBlondel(DynamicGraph dynGraph, NodeOrderStrategy nodeOrderStrategy, Blondel.StoppingCriterion stoppingCriterion) {
		super(dynGraph);
		
		this.nodeOrderStrategy = nodeOrderStrategy;
		this.stoppingCriterion = stoppingCriterion;
	}


	@Override
	protected void addEvent(boolean notificationTime, Cluster to, NodeCursor nc) {
		// ignored
	}

	@Override
	protected void deleteEvent(boolean notificationTime, Cluster deletedCluster) {
		// ignored

	}

	@Override
	protected void edgeCreationEvent(Edge newEdge) {
		// ignored

	}

	@Override
	protected void edgeRemovalEvent(boolean notificationTime, Edge edgeToRemove) {
		// ignored

	}

	@Override
	protected void extractEvent(boolean notificationTime, Cluster from, NodeCursor nc) {
		// ignored

	}

	@Override
	protected void mergeEvent(boolean notificationTime, Cluster a, Cluster b) {
		// ignored

	}

	@Override
	protected void moveEvent(boolean notificationTime, Cluster from, Cluster to, NodeCursor nc) {
		// ignored

	}

	@Override
	protected void newClusterEvent(boolean notificationTime, Cluster newCluster) {
		// ignored

	}

	@Override
	protected void newWeightsEvent() {
		// ignored

	}

	@Override
	protected void nodeCreationEvent(Node newNode) {
		this.timer.start();
		this.clustering.newCluster(newNode);
		this.timer.stop();
	}

	@Override
	protected void nodeRemovalEvent(boolean notificationTime, Node nodeToRemove) {
		// ignored

	}

	@Override
	protected void policyChanged(boolean policy) {
		// ignored

	}

	@Override
	protected void resetEvent(boolean notificationTime) {
		// ignored

	}

	@Override
	protected void resetLargeClusterEvent(boolean notificationTime, Cluster c) {
		// ignored

	}

	@Override
	protected void resetSingletonsEvent(boolean notificationTime) {
		// ignored

	}

	@Override
	protected void splitEvent(boolean notificationTime, Cluster a, Cluster b) {
		// ignored

	}

	@Override
	protected void weightChangedEvent(Edge e, double oldWeight, double newWeight) {
		// ignored

	}

	public void onTimeStepEvent() {
		this.timer.start();
		//
		this.clustering.resetSingletons();
		this.init();
		//
		this.timer.stop();
		
		// DEBUG
		System.out.println(">> " + this.id() + " done after " + this.timer.read(TimeUnit.MilliSeconds) + " [ms]");
		// DEBUG
	}

	protected int[] permuteArrray(int[] order) {
		Random myRandom = new Random();
		int swapID;
		int tmp;
		for (int i = 0; i < order.length; ++i) {
			swapID = myRandom.nextInt(order.length);
			tmp = order[i];
			order[i] = order[swapID];
			order[swapID] = tmp;
		}
		return order;
	}


	/**
	 * Given the Hashmap Level->HierarchyLevel, the original clustering, and the currentClustering this procedure summarizes
	 * all arrays by iteratively substituting array values by the value found at that array index in the next higher array.
	 * Starting with singletons, in the end, this represents the final clustering. This works since new clusterings are built
	 * with oldCluster.index = newNode.index.
	 * 
	 * @param levelToClusteringVector
	 * @param outputClustering
	 * @param currentClustering
	 */
	protected void unfurlHierarchy(Clustering outputClustering) {
		Graph theGraph = outputClustering.getGraph();
		int n = theGraph.N();
		int[] nodeAssignment = new int[n];
		Cluster theCluster;
		Node theNode;
//		System.out.println("Unfurling nodes");
		for (NodeCursor nc = theGraph.nodes(); nc.ok(); nc.next()) {
			theNode = nc.node();
			for (int i = 0; i <= (levelToHierarchyLevel.size() - 2); i++) {
				theCluster = levelToHierarchyLevel.get(i).theClustering.getCluster(theNode);
				theNode = levelToHierarchyLevel.get(i).clusterToNextLevelNode.get(theCluster);
			}
			nodeAssignment[nc.node().index()] = theNode.index();
		}
//		System.out.println("Finalizing");
		if (outputClustering.poolNodes().ok()) {
			throw new RuntimeException("not ok.");
		}
		outputClustering.resetToClustering(nodeAssignment);
		if (outputClustering.poolNodes().ok()) {
			throw new RuntimeException("not ok.");
		}
	}

	/**
	 * Returns the sum of edge weights between aNode and the nodes of aCluster
	 * 
	 * @param aNode
	 * @param aCluster
	 * @return sum of edge weights between aNode and aCluster
	 */
	/**
	 * @param aNode
	 * @param aCluster
	 * @return
	 */
	protected double adjacencyToCluster(Node aNode, Cluster aCluster) {
		double weight = 0.0;
		Clustering aClustering = aCluster.getClustering();
		GraphInterpretation gi = aClustering.getInterpretation();
		// through all in and then all out edges...
		for (EdgeCursor ec = aNode.edges(); ec.ok(); ec.next()) {
			Edge theEdge = ec.edge();
			Node neighbor = theEdge.opposite(aNode);
			if (aCluster.equals(aClustering.getCluster(neighbor))) {
				weight += gi.getWeight(theEdge);
				// dont count self loops twice, only the egde weight counts
			}
		}
		// but yfiles puts a node twice in its own set of neighbors, in case of
		// a selfloop, thus subtract
		if (aCluster.equals(aClustering.getCluster(aNode)))
			weight -= gi.getAdjacency(aNode, aNode);
		return weight;
	}

	/* algo methods */
	
	
	/**
	 * calculates for a node, which of the clusters of all its neighbors is best to move to, in terms of modularity. Returns
	 * null if becoming a singleton is best. Returns himself if no improvement can be found -> old cluster
	 * 
	 * @param searchingNode
	 *            the node that is examining its neighborhood for better clusters
	 * @param newNeighbor
	 *            the neighbor which is in the cluster that is best for searchingNode to move to
	 * @return
	 */
	protected Node bestNeighbor(Node searchingNode, Clustering theClustering) {
		double currentIncrease = 0.0;
		GraphInterpretation gi = theClustering.getInterpretation();
		Node currentNeighbor = null;
		Node newNeighbor = searchingNode;
		Cluster oldCluster = theClustering.getCluster(searchingNode), newCluster;
		double covMinus, covPlus, expPlus, expMinus;
		double W = gi.getEdgeWeightSum();
		double selfLoopWeight = gi.getAdjacency(searchingNode, searchingNode);
		double nonSimpleWeight = gi.getDegree(searchingNode) + selfLoopWeight;
		for (NodeCursor nc = searchingNode.neighbors(); nc.ok(); nc.next()) {
			++neighborHandShakes;
			currentNeighbor = nc.node();
			newCluster = theClustering.getCluster(currentNeighbor);
			if (oldCluster.getClustering() == null) {
//				System.out.println("oldCluster Kaputt");
			}
			if (newCluster.getClustering() == null) {
//				System.out.println("newCluster Kaputt");
			}
			if (newCluster.equals(oldCluster))
				continue;
			// self loops automatically in this term
			covMinus = adjacencyToCluster(searchingNode, oldCluster) / W;
			// self loops will contribute to future
			covPlus = (adjacencyToCluster(searchingNode, newCluster) + selfLoopWeight) / W;
			expPlus = (nonSimpleWeight * (theClustering.getBorderWeight(oldCluster) + 2
					* theClustering.getInnerWeight(oldCluster) - nonSimpleWeight))
					/ (2.0 * W * W);
			expMinus = (nonSimpleWeight * (theClustering.getBorderWeight(newCluster) + 2 * theClustering
					.getInnerWeight(newCluster)))
					/ (2.0 * W * W);
			currentIncrease = covPlus - covMinus + expPlus - expMinus;
			if (currentIncrease > bestInnerIncrease) {
				bestInnerIncrease = currentIncrease;
				newNeighbor = currentNeighbor;
			}
		}
		// isolated node -> singleton!
		if (gi.getDegree(searchingNode) == 0) {
			newNeighbor = null;
			// System.out.print("isolated node! -> singleton, ");
		} else {
			// check goodness of becoming a simple singleton
			covMinus = adjacencyToCluster(searchingNode, oldCluster) / W;
			covPlus = selfLoopWeight / W;
			expPlus = (nonSimpleWeight * (theClustering.getBorderWeight(oldCluster) + 2
					* theClustering.getInnerWeight(oldCluster) - nonSimpleWeight))
					/ (2.0 * W * W);
			expMinus = 0.0;
			currentIncrease = covPlus - covMinus + expPlus - expMinus;
			if (currentIncrease > bestInnerIncrease) {
				bestInnerIncrease = currentIncrease;
				newNeighbor = null;
			}
		}
		return (newNeighbor);
	}

	/**
	 * iterates once through all nodes in the array and moves them to the cluster of their favorite neighbor, according to
	 * the function bestNeighbor
	 * 
	 * @param theClustering
	 * @return the maximum increase achieved in this iteration
	 */
	/**
	 * @param theClustering
	 * @return
	 */
	protected double bestNeighborsForEveryOne(HierarchyLevel hl, Node[] badNodes) {
		Clustering currentClustering = hl.theClustering;
		Graph currentGraph = hl.theGraph;
	
		for (Node aNode : badNodes) {
			if (!currentGraph.contains(aNode)) {
				System.out.println("aaaaaaaaaaaaaaahhhhhhhhhhhhhhhh");
			}
		}
	
		int n = badNodes.length;
		int[] order = orderNodes(currentGraph, badNodes);
		double maxIncrease = 0.0;
		for (int i = 0; i < n; ++i) {
			bestInnerIncrease = 0.0;
			Node currentNode = badNodes[order[i]];
			Node currentBestNeighbor = bestNeighbor(currentNode, currentClustering);
			if ((bestInnerIncrease > 1e-6) || currentNode.degree() == 0) {
				Cluster oldCluster = currentClustering.getCluster(currentNode);
				currentClustering.extract(oldCluster, currentNode);
				incrementalModularity += bestInnerIncrease;
				// System.out.println("Increase: " + bestInnerIncrease + "
				// currentBestNeighbor " + currentBestNeighbor);
				// new singleton (increase or not)
				if (currentBestNeighbor == null) {
					Cluster newCluster = currentClustering.newCluster(currentNode);
					// normal case, migration
				} else {
					Cluster newCluster = currentClustering.getCluster(currentBestNeighbor);
					currentClustering.add(newCluster, currentNode);
				}
			}
			if (bestInnerIncrease > maxIncrease) {
				maxIncrease = bestInnerIncrease;
			}
		}
		return (maxIncrease);
	}

	/**
	 * Given a Clustering, this procedure creates a new graph, with one node per given cluster (such that cluster.index =
	 * node.index!), and with edge weights corresponding to the sums of edge weights between clusters. This also creates
	 * self-loops. The graph is returned as a new clustering. Additionally this procedure takes a Hashmap of Arrays as an
	 * argument; it writes the clustering into a new array as node.index -> cluster index, and appends this array into the
	 * hashmap. The key of this entry is the level of hierarchy on which this clustering has been done, the original graph
	 * being level 0.
	 * 
	 * @param ering
	 *            input Clustering to be abstracted
	 * @return a graph with a node per input clustering, and appropriate weights
	 */
	protected Clustering contractClustering(HierarchyLevel oldLevel) {
		Clustering oldClustering = oldLevel.theClustering;
		ClusterMap<Node> oldClusterMap = oldLevel.clusterToNextLevelNode;
		abstractedGraph = new Graph();
		GraphInterpretation theInterpretation = oldClustering.getInterpretation();
		int k = oldClustering.size();
		// prepare adjacency matrix
		double[][] adjMatrix = new double[k][k];
		for (int i = 0; i < k; ++i) {
			for (int j = 0; j < k; ++j) {
				adjMatrix[i][j] = 0.0;
			}
		}
		// collect edge weights into matrix, indexed by clusterID
		Edge currentEdge = null;
		int c1, c2;
		for (EdgeCursor ec = oldClustering.getGraph().edges(); ec.ok(); ec.next()) {
			currentEdge = ec.edge();
			c1 = oldClustering.getCluster(currentEdge.source()).getIndex();
			c2 = oldClustering.getCluster(currentEdge.target()).getIndex();
			adjMatrix[c1][c2] += theInterpretation.getWeight(currentEdge);
		}
		// create k new nodes in new graph
		Node[] newNodes = new Node[k];
		for (int i = 0; i < k; ++i) {
			newNodes[i] = abstractedGraph.createNode();
		}
	
		// point from old cluster to new node, with clusterID = new node's
		// position in array newNodes[] atm.
		for (GYCursor<Cluster> cc = oldClustering.clusters(); cc.ok(); cc.next()) {
			oldClusterMap.set((Cluster) cc.current(), newNodes[((Cluster) cc.current()).getIndex()]);
		}
		// create edges, indexed by newNodes[clusterID], with matrix weight
		double currentWeight = 0.0;
		EdgeMap newEdgeMap = abstractedGraph.createEdgeMap();
		for (int i = 0; i < k; ++i) {
			for (int j = i; j < k; ++j) {
				if (i == j)
					currentWeight = adjMatrix[i][j];
				else
					currentWeight = adjMatrix[i][j] + adjMatrix[j][i];
				if (currentWeight > 0.0) {
					currentEdge = abstractedGraph.createEdge(newNodes[i], newNodes[j]);
					newEdgeMap.setDouble(currentEdge, currentWeight);
				}
			}
		}
		Clustering newClustering = new Clustering(abstractedGraph, new WeightedGraph(abstractedGraph, newEdgeMap));
		newClustering.resetSingletons();
		abstractedGraph.disposeEdgeMap(newEdgeMap);
		return (newClustering);
	}


	protected NodeList getInitialBadNodes() {
		return initialBadNodesList;
	}

	/**
	 * Calculate the clustering.
	 */
	protected void init() {
		// clear hierarchy levels
		this.levelToHierarchyLevel.clear();
		
		incrementalModularity = 0.0;
		GraphInterpretation gi = this.dynGraph.getInterpretation();
		for (NodeCursor nc = this.dynGraph.nodes(); nc.ok(); nc.next()) {
			incrementalModularity -= ((gi.getDegree(nc.node()) + gi.getAdjacency(nc.node(), nc.node())) * (gi.getDegree(nc
					.node()) + gi.getAdjacency(nc.node(), nc.node())));
		}
		incrementalModularity /= (4 * gi.getEdgeWeightSum() * gi.getEdgeWeightSum());
	
		double currentBestImprovement = 1.0;
		int oldClusterCount = this.dynGraph.N() + 1;
		int hierarchyLevel = 0;
		int it;
		Clustering currentClustering = new Clustering(this.dynGraph, gi);
		currentClustering.resetSingletons();
		HierarchyLevel currentLevel = new HierarchyLevel(this.dynGraph, currentClustering);
		while (currentClustering.size() < oldClusterCount) {
//			System.out.println("Level:" + hierarchyLevel++ + ", n:" + currentClustering.getGraph().N() + ", m:"
//					+ currentClustering.getGraph().E() + ", mod:" + incrementalModularity + ". ");
			oldClusterCount = currentClustering.size();
			currentBestImprovement = 1.0;
			it = 1;
	
			// while mod increase, migrate and merge nodes
			Node[] badNodes = currentClustering.getGraph().getNodeArray();
			while (currentBestImprovement > 1e-6) {
				currentBestImprovement = bestNeighborsForEveryOne(currentLevel, badNodes);
//				System.out.print("It:" + it++ + ", best increase:" + currentBestImprovement + " ... ");
			}
//			System.out.print("no more improvement, level done (" + currentClustering.size() + " clusters). ");
	
			currentLevel.oracle = new ClusterChangeListener(currentLevel);
			currentClustering.addListener(currentLevel.oracle);
	
			// when no more increase, possibly create next level
			if (currentClustering.size() < oldClusterCount) {
				HierarchyLevel nextLevel = new HierarchyLevel(currentLevel);
//				System.out.println();
				currentClustering = nextLevel.theClustering;
				currentLevel = nextLevel;
			}
		}
//		System.out.println("No more contractions, hierarchy done");
		unfurlHierarchy(this.clustering);
//		System.out.println("Finished at mod:" + incrementalModularity + ", with " + neighborHandShakes
//				+ " neighbors' handshakes.");
//		System.out.println("===== Done =====");
		
		for (int i = 0; i < levelToHierarchyLevel.size(); ++i) {
			if (i > 0){
//				levelToHierarchyLevel.get(i).theGraph.disposeEdgeMap(levelToHierarchyLevel.get(i).randWeights);
				levelToHierarchyLevel.get(i).oracle.disposeMaps();
			}
			else {
//				for (EdgeMap someBaseGraphEdgeMap : levelToHierarchyLevel.get(i).theGraph.getRegisteredEdgeMaps()){
//					if (!someBaseGraphEdgeMap.equals(baseGraphRandWeights)){
//						System.out.println("gotcha! get lost!");
//						levelToHierarchyLevel.get(i).theGraph.disposeEdgeMap(someBaseGraphEdgeMap);
//					}
//				}
			}
			levelToHierarchyLevel.get(i).clusterToNextLevelNode.dispose();
			levelToHierarchyLevel.get(i).theClustering.removeClusteringListener(levelToHierarchyLevel.get(i).oracle);
			if (i > -1){
				levelToHierarchyLevel.get(i).theClustering.disposeAllListeners();
				levelToHierarchyLevel.get(i).theClustering.dispose();
			}
			if (i > 0){
				for (EdgeMap deletabeEdgeMap : levelToHierarchyLevel.get(i).theGraph.getRegisteredEdgeMaps()){
					levelToHierarchyLevel.get(i).theGraph.disposeEdgeMap(deletabeEdgeMap);
				}
			}
//			System.out.print("GraphListeners clinging to graph at level " + i + ": " + levelToHierarchyLevel.get(i).theGraph.getGraphListeners().hasNext() + ": ");
//			System.out.print("EdgeMaps: " + levelToHierarchyLevel.get(i).theGraph.getRegisteredEdgeMaps().length + "\n");
//			System.out.print("NodeMaps :" + levelToHierarchyLevel.get(i).theGraph.getRegisteredNodeMaps().length + "\n");
		}
		this.levelToHierarchyLevel.clear();
		
	}

	/**
	 * @param theGraph
	 *            the graph the nodes in the array are part of
	 * @param nodeArray
	 *            the nodes to be ordered
	 * @return an integer array which is a permutation of the 0...n-1 numbers
	 */
	protected int[] orderNodes(Graph theGraph, Node[] nodeArray) {
		int n = nodeArray.length;
		int[] order = new int[n];
		switch (nodeOrderStrategy) {
		case Array:
			for (int i = 0; i < n; ++i) {
				order[i] = i;
			}
			break;
		case InvArray:
			for (int i = 0; i < n; ++i) {
				order[i] = n - 1 - i;
			}
			break;
		case Random:
			for (int i = 0; i < n; ++i) {
				order[i] = i;
			}
			order = permuteArrray(order);
			break;
		default:
			System.out.println("Error, invalid node order strategy!");
			break;
		}
	
		return (order);
	}

	@Override
	public String id() {
		return NAME + "(" + this.nodeOrderStrategy + "," + this.stoppingCriterion + "):" + this.uniqueID;
 	}

}
