package de.uka.algo.dynclustering.algos.base;



import java.util.HashMap;

import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.ClusteringListener;
import de.uka.algo.clustering.comparison.Comparator;
import de.uka.algo.clustering.comparison.ComparatorFactory;
import de.uka.algo.clustering.index.Index;
import de.uka.algo.clustering.index.IndexFactory;
import de.uka.algo.dynclustering.graphs.DynamicGraph;
import de.uka.algo.dynclustering.graphs.DynamicGraphListener;
import de.uka.algo.dynclustering.util.RunTimer;
import de.uka.algo.dynclustering.util.RunTimer.TimeUnit;


/**
 * Abstract superclass for  dynamic clustering algorithms.
 * @author Christian Staudt
 *
 */
public abstract class DynamicClusteringAlgorithm extends ClusteringListener implements DynamicGraphListener {
	
	protected DynamicGraph dynGraph;
	protected RunTimer timer; 
	/** starting and stopping the timer must be implemented by the subclasses for each relevant action they perform */
	protected Clustering previousClustering;
	protected int uniqueID = (int) (Math.random() * 1000);
	// caches for quality and distance measure objects
	private HashMap<ComparatorFactory, Comparator> factory2comparator;
	private HashMap<IndexFactory, Index> factory2index;
	
	
	
	

	public DynamicClusteringAlgorithm(DynamicGraph dynGraph) {
		// call ClusteringListener constructor
		super(new Clustering(dynGraph, dynGraph.getInterpretation()));
		this.dynGraph = dynGraph;
		this.startListening();
		this.timer = new RunTimer();
		
		this.previousClustering = null;
		
		this.factory2comparator = new HashMap<ComparatorFactory, Comparator>();
		this.factory2index = new HashMap<IndexFactory, Index>();
	}
	
	
	public long readTimer() {
		return this.timer.read();
	}
	
	/**
	 * Unregister as a ClusteringListener and DynamicGraphListener.
	 */
	public void stopListening() {
		this.clustering.removeClusteringListener(this);
		this.dynGraph.removeDynamicGraphListener(this);
	}

	/**
	 * Register as a ClusteringListener and DynamicGraphListener.
	 */
	public void startListening() {
		this.clustering.addListener(this);
		this.dynGraph.addDynamicGraphListener(this);
	}
	
	
	public double getDistance(ComparatorFactory factory) {
		
		double distance;
		if (this.previousClustering != null) {
			Comparator dM = this.factory2comparator.get(factory);
			if (dM == null) {
				dM = factory.getComparator(this.clustering, this.previousClustering);
			}
			distance = dM.getValue();
		} else {
			distance = 0;
		}

		return distance;
	}

	public double getQuality(IndexFactory factory) {
		Index qI = this.factory2index.get(factory);
		if (qI == null) {
			qI = factory.getIndex(this.clustering);
			this.factory2index.put(factory, qI);
		}
		return qI.getValue();
	}
	
	public double getRuntime(TimeUnit unit) {
		double runtime = this.timer.read(unit);
		this.timer.reset();
		return runtime;
	}
	
	
	/**
	 * Current number of clusters.
	 * @return
	 */
	public int getClusterCount() {
		return this.clustering.clusterCount();
	}
	
	
	/**
	 * Returns an identifier of the algorithm.
	 */
	public abstract String id();
	
	
	/**
	 * Clones the current clustering and saves it as previous clustering.
	 */
	public void savePreviousClustering() {
		if (this.previousClustering == null) {
			this.previousClustering = this.clustering.clone();
		}
		this.previousClustering.setToClustering(this.clustering);
	}
	
	public String toString() {
		return this.id();
	}


	public DynamicGraph getGraph() {
		return this.dynGraph;
	}
	
}
