package de.uka.algo.dynclustering.algos.blondel;

import java.util.LinkedList;
import java.util.Stack;

import y.base.Edge;
import y.base.Node;
import y.base.NodeCursor;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.ClusterMap;
import de.uka.algo.clustering.ClusteringListener;
import de.uka.algo.dynclustering.algos.Blondel;
import de.uka.algo.dynclustering.algos.StaticBlondel;
import de.uka.algo.dynclustering.algos.TDRandNStaticBlondel;
import de.uka.algo.dynclustering.algos.TDStaticBlondel;
import de.uka.algo.dynclustering.algos.Blondel.HierarchyLevel;

/**
 * maintains in invalidClusters clusters in this level, that contain nodes that
 * have been affected, in deletableAbstractNodes nodes of the next level, that
 * correspond to deleted clusters of this level, and sets entries in the map
 * clusterToNextLevelNode to null for new clusters.
 * 
 * @author robert
 * 
 */
public class ClusterChangeListener extends ClusteringListener {

	private static final Boolean VALID = new Boolean(false); // null = false
	// = valid
	private static final Boolean INVALID = new Boolean(true);

	// invalid Clusters in this hierarchy level;
	// only clusters that still exist are in this Map
	private ClusterMap<Boolean> invalidClusters; // null = false = valid
	// nodes in next level that can go, since cluster here ist gone
	private Stack<Node> deletableAbstractNodes = new Stack<Node>();
	// the mapping of this hierarchy level will be in here
	private ClusterMap<Node> clusterToNextLevelNode;

	// newClusters are implictly those of invalidClusters with null mapping

	public ClusterChangeListener(Blondel.HierarchyLevel nextHl) {
		super(nextHl.theClustering);
		invalidClusters = new ClusterMap<Boolean>(clustering);
		clusterToNextLevelNode = nextHl.clusterToNextLevelNode;
	}
	
	public ClusterChangeListener(StaticBlondel.HierarchyLevel nextHl) {
		super(nextHl.theClustering);
		invalidClusters = new ClusterMap<Boolean>(clustering);
		clusterToNextLevelNode = nextHl.clusterToNextLevelNode;
	}
	
	public ClusterChangeListener(TDStaticBlondel.HierarchyLevel nextHl) {
		super(nextHl.theClustering);
		invalidClusters = new ClusterMap<Boolean>(clustering);
		clusterToNextLevelNode = nextHl.clusterToNextLevelNode;
	}
	
	public ClusterChangeListener(TDRandNStaticBlondel.HierarchyLevel nextHl) {
		super(nextHl.theClustering);
		invalidClusters = new ClusterMap<Boolean>(clustering);
		clusterToNextLevelNode = nextHl.clusterToNextLevelNode;
	}

	public Cluster[] invalidOldClusters() {
		LinkedList<Cluster> theInvalidClusters = new LinkedList<Cluster>();
		for (Cluster c : clustering.getClusterArray()) {
			if ((invalidClusters.get(c) == INVALID) && (clusterToNextLevelNode.get(c) != null)) {
				theInvalidClusters.add(c);
			}
		}
		return theInvalidClusters.toArray(new Cluster[0]);
	}

	public Cluster[] newClusters() {
		LinkedList<Cluster> theClusters = new LinkedList<Cluster>();
		for (Cluster c : clustering.getClusterArray()) {
			if ((invalidClusters.get(c) == INVALID) && (clusterToNextLevelNode.get(c) == null)) {
				theClusters.add(c);
			}
		}
		return theClusters.toArray(new Cluster[0]);
	}

	public Node[] deletableNodes() {
		return deletableAbstractNodes.toArray(new Node[this.deletableAbstractNodes.size()]);
	}

	public void tagAsValid() {
		// reset all lists!
		for (Cluster c : clustering.getClusterArray()) {
			invalidClusters.set(c, VALID);
		}
		deletableAbstractNodes.clear();
	}

	@Override
	protected void addEvent(boolean notificationTime, Cluster to, NodeCursor nc) {
		if (notificationTime == ClusteringListener.PRE) {
			invalidClusters.set(to, INVALID);
		}
	}

	@Override
	protected void deleteEvent(boolean notificationTime, Cluster deletedCluster) {
		if (notificationTime == ClusteringListener.PRE) {
			Node node = clusterToNextLevelNode.get(deletedCluster);
			if (node != null) {
				// DEBUG
				System.out.println("stack: " + node);
				// DEBUG
				deletableAbstractNodes.push(node);
			}
		}
	}

	@Override
	protected void edgeCreationEvent(Edge newEdge) {
		if (clustering.getCluster(newEdge.source()) != null) {
			invalidClusters.set(clustering.getCluster(newEdge.source()), INVALID);
		}
		if (clustering.getCluster(newEdge.target()) != null) {
			invalidClusters.set(clustering.getCluster(newEdge.target()), INVALID);
		}
	}

	@Override
	protected void edgeRemovalEvent(boolean notificationTime, Edge edgeToRemove) {
		if (notificationTime == ClusteringListener.PRE) {
			if (clustering.getCluster(edgeToRemove.source()) != null) {
				invalidClusters.set(clustering.getCluster(edgeToRemove.source()), INVALID);
			}
			if (clustering.getCluster(edgeToRemove.target()) != null) {
				invalidClusters.set(clustering.getCluster(edgeToRemove.target()), INVALID);
			}
		}
	}

	@Override
	protected void extractEvent(boolean notificationTime, Cluster from, NodeCursor nc) {
		if (notificationTime == ClusteringListener.PRE) {
			invalidClusters.set(from, INVALID);
		}
	}

	@Override
	protected String id() {
		return (this.toString());
	}

	@Override
	protected void mergeEvent(boolean notificationTime, Cluster a, Cluster b) {
		if (notificationTime == ClusteringListener.PRE) {
			invalidClusters.set(a, INVALID);
			deletableAbstractNodes.push(clusterToNextLevelNode.get(b));
		}
	}

	@Override
	protected void moveEvent(boolean notificationTime, Cluster from, Cluster to, NodeCursor nc) {
		if (notificationTime == ClusteringListener.PRE) {
			invalidClusters.set(to, INVALID);
			invalidClusters.set(from, INVALID);
		}
	}

	@Override
	protected void newClusterEvent(boolean notificationTime, Cluster newCluster) {
		if (notificationTime == ClusteringListener.POST) {
			invalidClusters.set(newCluster, INVALID);
		}
	}

	@Override
	protected void newWeightsEvent() {
		throw new RuntimeException("Not implemented, should not happen!");
	}

	@Override
	protected void nodeCreationEvent(Node newNode) {
		// nothing to do here
	}

	@Override
	protected void nodeRemovalEvent(boolean notificationTime, Node nodeToRemove) {
		if (notificationTime == ClusteringListener.PRE) {
			if (clustering.getCluster(nodeToRemove) != null) {
				invalidClusters.set(clustering.getCluster(nodeToRemove), INVALID);
			}
		}
	}

	@Override
	protected void policyChanged(boolean policy) {
		// seen it, done nothin'
	}

	@Override
	protected void resetEvent(boolean notificationTime) {
//		throw new RuntimeException("Not implemented, should not happen!");
	}

	@Override
	protected void resetLargeClusterEvent(boolean notificationTime, Cluster c) {
//		throw new RuntimeException("Not implemented, should not happen!");
	}

	@Override
	protected void resetSingletonsEvent(boolean notificationTime) {
//		throw new RuntimeException("Not implemented, should not happen!");
	}

	@Override
	protected void splitEvent(boolean notificationTime, Cluster a, Cluster b) {
		if (notificationTime == ClusteringListener.POST) {
			invalidClusters.set(a, INVALID);
			invalidClusters.set(b, INVALID);
		}
	}

	@Override
	protected void weightChangedEvent(Edge e, double oldWeight, double newWeight) {
		if (clustering.getCluster(e.source()) != null) {
			invalidClusters.set(clustering.getCluster(e.source()), INVALID);
		}
		if (clustering.getCluster(e.target()) != null) {
			invalidClusters.set(clustering.getCluster(e.target()), INVALID);
		}
	}
	
	public void disposeMaps() {
		this.invalidClusters.dispose();
	}

}
