package de.uka.algo.dynclustering.algos.ilp;

import ilog.concert.IloException;
import ilog.concert.IloLinearNumExpr;
import ilog.concert.IloNumVar;
import ilog.cplex.IloCplex;

import java.util.ArrayList;
import java.util.Map;

import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.linearprogram.CplexGenerator;
import de.uka.algo.dynclustering.graphs.DynamicGraph;
import de.uka.algo.dynclustering.util.YUtils;

import y.base.Graph;
import y.base.Node;

/**
 * This is an extension to CplexGenerator, providing variables and constraints
 * which are needed for the dynamic version of the ILP clustering.
 * 
 * @author Christian Staudt
 * 
 */
public class PartialCplexGenerator extends CplexGenerator {

	/**
	 * Generates a decision variable for each pair of free node and remaining
	 * cluster.
	 * 
	 * @param graph
	 * @param zeta
	 * @param asILP
	 * @param name2variable
	 * @param myLP
	 * @return
	 * @throws IloException
	 */
	public static IloCplex generateNodeClusterVariables(IloCplex myLP, Clustering zeta, boolean asILP, Map<String, IloNumVar> name2variable)
			throws IloException {

		int k = zeta.clusterCount();
		int n = zeta.poolSize();

		for (int i = 0; i < n; ++i)
			for (int l = 0; l < k; ++l) {
				String name = "Y(" + i + "," + l + ")";
				IloNumVar var;
				if (asILP)
					var = myLP.boolVar(name);
				else
					var = myLP.numVar(0, 1, name);
				name2variable.put(name, var);
			}

		return myLP;
	}

	/**
	 * Create metric transitivity constraints for triples of free nodes.
	 * 
	 * @param myLp
	 * @param zeta
	 * @param asILP
	 * @param names2variables
	 * @return
	 */
	public static IloCplex addMetricNodeTripleConstraints(IloCplex myLp, Clustering zeta, boolean asILP, Map<String, IloNumVar> names2variables)
			throws IloException {
		int n = zeta.poolSize();
		IloLinearNumExpr constraint;
		for (int u = 0; u < n; ++u)
			for (int v = u + 1; v < n; ++v) {
				IloNumVar Xuv = names2variables.get("X(" + u + "," + v + ")");
				for (int w = v + 1; w < n; ++w) {
					IloNumVar Xvw = names2variables.get("X(" + v + "," + w + ")");
					IloNumVar Xuw = names2variables.get("X(" + u + "," + w + ")");

					constraint = myLp.linearNumExpr();
					constraint.addTerm(Xuv, 1);
					constraint.addTerm(Xvw, 1);
					constraint.addTerm(Xuw, -1);
					myLp.addGe(constraint, 0.0);

					constraint = myLp.linearNumExpr();
					constraint.addTerm(Xuv, -1);
					constraint.addTerm(Xvw, 1);
					constraint.addTerm(Xuw, 1);
					myLp.addGe(constraint, 0.0);

					constraint = myLp.linearNumExpr();
					constraint.addTerm(Xuv, 1);
					constraint.addTerm(Xvw, -1);
					constraint.addTerm(Xuw, 1);
					myLp.addGe(constraint, 0.0);
				}
			}
		return myLp;
	}

	/**
	 * Create metric transitivity constraints for triples of two free nodes and
	 * a cluster.
	 * 
	 * @param myLp
	 * @param graph
	 * @param zeta
	 * @param asILP
	 * @param names2variables
	 * @return
	 * @throws IloException
	 */
	public static IloCplex addMetricNodePairClusterConstraints(IloCplex myLp, Clustering zeta, boolean asILP, Map<String, IloNumVar> names2variables)
			throws IloException {

		/* create triangle-inequality constraints for nodes and clusters */

		int n = zeta.poolSize();
		int k = zeta.clusterCount();

		IloLinearNumExpr constraint;
		for (int u = 0; u < n; ++u)
			for (int v = u + 1; v < n; ++v) {
				IloNumVar Xuv = names2variables.get("X(" + u + "," + v + ")");
				assert Xuv != null;
				for (int c = 0; c < k; ++c) {

					IloNumVar Yuc = names2variables.get("Y(" + u + "," + c + ")");
					IloNumVar Yvc = names2variables.get("Y(" + v + "," + c + ")");
					assert Yuc != null;
					assert Yvc != null;

					constraint = myLp.linearNumExpr();
					constraint.addTerm(Xuv, 1);
					constraint.addTerm(Yuc, -1);
					constraint.addTerm(Yvc, 1);
					myLp.addGe(constraint, 0.0);

					constraint = myLp.linearNumExpr();
					constraint.addTerm(Xuv, 1);
					constraint.addTerm(Yuc, 1);
					constraint.addTerm(Yvc, -1);
					myLp.addGe(constraint, 0.0);

					constraint = myLp.linearNumExpr();
					constraint.addTerm(Xuv, -1);
					constraint.addTerm(Yuc, 1);
					constraint.addTerm(Yvc, 1);
					myLp.addGe(constraint, 0.0);
				}
			}

		return myLp;
	}

	/**
	 * These constraints require that a node can only be assigned to one cluster
	 * at a time.
	 * 
	 * @param myLp
	 * @param graph
	 * @param zeta
	 * @param asILP
	 * @param names2variables
	 * @return
	 * @throws IloException
	 */
	public static IloCplex addMetricUniquenessConstraints(IloCplex myLp, Graph graph, Clustering zeta, boolean asILP, Map<String, IloNumVar> names2variables)
			throws IloException {

		int n = zeta.poolSize(); // number of free nodes
		int k = zeta.clusterCount(); // number of remaining clusters

		
		IloLinearNumExpr constraint = null;

		for (int u = 0; u < n; ++u) {
			constraint = myLp.linearNumExpr();
			for (int c = 0; c < k; ++c) {
				IloNumVar YuC = names2variables.get("Y(" + u + "," + c + ")");
				constraint.addTerm(YuC, 1);
			}
			myLp.addGe(constraint, k - 1);
		}

		// equivalent, but more constraints:
//		IloLinearNumExpr constraint;
//
//		for (int u = 0; u < n; ++u)
//			for (int c = 0; c < k; ++c)
//				for (int d = c + 1; d < k; ++d) {
//
//					IloNumVar Yuc = names2variables.get("Y(" + u + "," + c + ")");
//					IloNumVar Yud = names2variables.get("Y(" + u + "," + d + ")");
//					assert Yuc != null : "Y(" + u + "," + c + ") is null \n map: " + names2variables;
//					assert Yud != null : "Y(" + u + "," + d + ") is null \n map: " + names2variables;
//
//					constraint = myLp.linearNumExpr();
//					constraint.addTerm(Yuc, 1);
//					constraint.addTerm(Yud, 1);
//					myLp.addGe(constraint, 1.0);
//				}
		
		return myLp;
	}

	public static IloCplex addPartialObjectiveFunction(IloCplex myLp, Clustering zeta, PartialObjectiveFunction objF, boolean asMinimization,
			Map<String, IloNumVar> names2variables, boolean merge) {

		// get an ordering of nodes
		ArrayList<Node> freeNodes = YUtils.nodeCursor2ArrayList(zeta.poolNodes());
		ArrayList<Cluster> clusters = YUtils.gYCursor2ArrayList(zeta.clusters());
		int n = freeNodes.size();
		int k = clusters.size();

		try {
			// create objective function for node pairs
			IloLinearNumExpr objectiveFunction = myLp.linearNumExpr();
			// assigning coefficients based on objfct
			for (int u = 0; u < n; ++u)
				for (int v = u + 1; v < n; ++v) {
					IloNumVar Xuv = names2variables.get("X(" + u + "," + v + ")");
					objectiveFunction.addTerm(Xuv, objF.getCoefficient(freeNodes.get(u), freeNodes.get(v)));
				}

			for (int u = 0; u < n; ++u)
				for (int c = 0; c < k; ++c) {
					IloNumVar Yuc = names2variables.get("Y(" + u + "," + c + ")");
					objectiveFunction.addTerm(Yuc, objF.getCoefficient(freeNodes.get(u), clusters.get(c)));
				}
			
			if (merge){
				for (int c = 0; c < k; ++c) {
					for (int d = c+1; d < k; ++d) {
						IloNumVar Zcd = names2variables.get("Z(" + c + "," + d + ")");
						objectiveFunction.addTerm(Zcd, objF.getCoefficient(clusters.get(c), clusters.get(d)));
					}
				}
			}

			// set objective function and defined if the LP is a minimization or
			// maximization problem
			// note that metricModel(variables) = asMinimization(objective)
			if (asMinimization)
				myLp.addMinimize(objectiveFunction);
			else
				myLp.addMaximize(objectiveFunction);
		} catch (IloException ex) {
			ex.printStackTrace();
		}

		return myLp;
	}

	public static IloCplex generateNodePairVariables(IloCplex myLp, Clustering zeta, boolean asILP, Map<String, IloNumVar> name2variable) throws IloException {
		int n = zeta.poolSize();

		for (int u = 0; u < n; ++u)
			for (int v = u + 1; v < n; ++v) {
				String name = "X(" + u + "," + v + ")";
				IloNumVar var;
				if (asILP)
					var = myLp.boolVar(name);
				else
					var = myLp.numVar(0, 1, name);
				name2variable.put(name, var);
			}

		return myLp;
	}
	
	
	// new approach allowing mergers
	
	/**
	 * Create decision variables Z(C_i, C_j).
	 * 
	 * @param myLP
	 * @param zeta
	 * @param asILP
	 * @param name2variable
	 * @return
	 * @throws IloException
	 */
	public static IloCplex generateClusterPairVariables(IloCplex myLP, Clustering zeta, boolean asILP, Map<String, IloNumVar> name2variable) throws IloException {
		int k = zeta.clusterCount();
		
		for (int c = 0; c < k; ++c)
			for (int d = c+1; d < k; ++d) {
				String name = "Z(" + c + "," + d + ")";
				IloNumVar var;
				if (asILP)
					var = myLP.boolVar(name);
				else
					var = myLP.numVar(0, 1, name);
				name2variable.put(name, var);
			}

		return myLP;
	}
	
	
	/**
	 * Create metric transitivity constraints for triples of a free nodes and
	 * two clusters.
	 * 
	 * @param myLp
	 * @param graph
	 * @param zeta
	 * @param asILP
	 * @param names2variables
	 * @return
	 * @throws IloException
	 */
	public static IloCplex addMetricNodeClusterPairConstraints(IloCplex myLp, Clustering zeta, boolean asILP, Map<String, IloNumVar> names2variables)
			throws IloException {

		/* create triangle-inequality constraints for nodes and clusters */

		int n = zeta.poolSize();
		int k = zeta.clusterCount();

		IloLinearNumExpr constraint;
		for (int c = 0; c < k; ++c)
			for (int d = c + 1; d < k; ++d) {
				IloNumVar Zcd = names2variables.get("Z(" + c + "," + d + ")");
				assert Zcd != null;
				for (int u = 0; u < n; ++u) {

					IloNumVar Yuc = names2variables.get("Y(" + u + "," + c + ")");
					IloNumVar Yud = names2variables.get("Y(" + u + "," + d + ")");
					assert Yuc != null;
					assert Yud != null;

					constraint = myLp.linearNumExpr();
					constraint.addTerm(Zcd, 1);
					constraint.addTerm(Yuc, -1);
					constraint.addTerm(Yud, 1);
					myLp.addGe(constraint, 0.0);

					constraint = myLp.linearNumExpr();
					constraint.addTerm(Zcd, 1);
					constraint.addTerm(Yuc, 1);
					constraint.addTerm(Yud, -1);
					myLp.addGe(constraint, 0.0);

					constraint = myLp.linearNumExpr();
					constraint.addTerm(Zcd, -1);
					constraint.addTerm(Yuc, 1);
					constraint.addTerm(Yud, 1);
					myLp.addGe(constraint, 0.0);
				}
			}

		return myLp;
	}

	
	/**
	 * 
	 * @param myLp
	 * @param zeta
	 * @param asILP
	 * @param names2variables
	 * @return
	 * @throws IloException
	 */
	public static IloCplex addMetricClusterTripleConstraints(IloCplex myLp, Clustering zeta, boolean asILP, Map<String, IloNumVar> names2variables)
	throws IloException {
		
		int k = zeta.clusterCount();
		
		IloLinearNumExpr constraint;

		for (int c = 0; c < k; ++c)
			for (int d = c+1; d < k; ++d) {
				IloNumVar Zcd = names2variables.get("Z(" + c + "," + d + ")");
				for (int e = d+1; e < k; ++e) {
					IloNumVar Zce = names2variables.get("Z(" + c + "," + e + ")");
					IloNumVar Zde = names2variables.get("Z(" + d + "," + e + ")");
					
					constraint = myLp.linearNumExpr();
					constraint.addTerm(Zcd, 1);
					constraint.addTerm(Zce, -1);
					constraint.addTerm(Zde, 1);
					myLp.addGe(constraint, 0.0);

					constraint = myLp.linearNumExpr();
					constraint.addTerm(Zcd, 1);
					constraint.addTerm(Zce, 1);
					constraint.addTerm(Zde, -1);
					myLp.addGe(constraint, 0.0);

					constraint = myLp.linearNumExpr();
					constraint.addTerm(Zcd, -1);
					constraint.addTerm(Zce, 1);
					constraint.addTerm(Zde, 1);
					myLp.addGe(constraint, 0.0);

				}
			}
		return myLp;
	}
	
	
}
