package de.uka.algo.dynclustering.algos.ilp;

import ilog.concert.IloLinearNumExpr;
import ilog.concert.IloNumVar;

import java.util.ArrayList;

import de.uka.algo.clustering.Cluster;

import y.base.Node;
import lpsolve.LpSolve;
import lpsolve.LpSolveException;

public class PartialLpSolveGenerator {

	// TODO: move LpSolve stuff here

	/**
	 * Returns the linearized position of the (i,j)-th entry in a
	 * triangle-matrix.
	 * 
	 * @param i
	 *            row position
	 * @param j
	 *            column position
	 * @param n
	 *            total size
	 * @return linearized position
	 */
	public static int getPos(int i, int j, int n) {
		return i * n + j - (i + 1) * (i + 2) / 2;
	}

	/**
	 * Returns the linearized position of the (i,j)-th entry in a full matrix.
	 * 
	 * @param i
	 *            row position (nodes)
	 * @param k
	 *            column position (clusters)
	 * @param cols
	 *            number of columns
	 * @return linearized position
	 */
	public static int getPos2(int i, int k, int cols) {
		return k * cols + i;
	}
	
	
	public static LpSolve makeMetricLp(int varNumber, double[] coefficients) throws LpSolveException {
		LpSolve lp = LpSolve.makeLp(0, varNumber);
		
		coefficients[0] = 0;
		for (int i = 1; i < coefficients.length; ++i) {
			// set bounds and integer property
			lp.setUpbo(i, 1.0);
			lp.setLowbo(i, 0.0);
			lp.setInt(i, true);
		}
		return lp;
	}
	
	
	public static double[] makeCoefficientsArray(int varNumber) {
		// create coefficients array
		double coefficients[] = new double[varNumber + 1];
		// initialize the array
		for (int i = 0; i < coefficients.length; ++i)
			coefficients[i] = 0;
		return coefficients;
	}

	/**
	 * 
	 * @param lp
	 * @param coefficients
	 * @param n
	 *            number of free nodes
	 * @param k
	 *            number of remaining clusters
	 * @return
	 * @throws LpSolveException
	 */
	public static LpSolve addMetricNodeTripleConstraints(LpSolve lp, double[] coefficients, int n, int k) throws LpSolveException {
		// create triangle-inequality constraints for nodes
		for (int u = 0; u < n; ++u)
			for (int v = u + 1; v < n; ++v)
				for (int w = v + 1; w < n; ++w) {
					// get linearized positions
					int uv = getPos(u, v, n) + 1;
					int uw = getPos(u, w, n) + 1;
					int vw = getPos(v, w, n) + 1;

					// set triangle-inequality constraints
					coefficients[uv] = 1;
					coefficients[uw] = -1;
					coefficients[vw] = 1;
					lp.addConstraint(coefficients, LpSolve.GE, 0.0);
					coefficients[uv] = 1;
					coefficients[uw] = 1;
					coefficients[vw] = -1;
					lp.addConstraint(coefficients, LpSolve.GE, 0.0);
					coefficients[uv] = -1;
					coefficients[uw] = 1;
					coefficients[vw] = 1;
					lp.addConstraint(coefficients, LpSolve.GE, 0.0);

					// resets entries in the array for reuse
					coefficients[uv] = 0;
					coefficients[uw] = 0;
					coefficients[vw] = 0;
				}
		return lp;
	}

	public static LpSolve addMetricNodePairClusterConstraints(LpSolve lp, double[] coefficients, int n, int k) throws LpSolveException {
		// create triangle-inequality constraints for nodes and clusters
		int offset = n * (n - 1) / 2;

		for (int u = 0; u < n; ++u)
			for (int v = u + 1; v < n; ++v)
				for (int c = 0; c < k; ++c) {

					// get linearized positions
					int uv = getPos(u, v, n) + 1;
					int uc = offset + getPos2(u, c, n) + 1;
					int vc = offset + getPos2(v, c, n) + 1;

					// set triangle-inequality constraints
					coefficients[uv] = 1;
					coefficients[uc] = -1;
					coefficients[vc] = 1;
					lp.addConstraint(coefficients, LpSolve.GE, 0.0);
					coefficients[uv] = 1;
					coefficients[uc] = 1;
					coefficients[vc] = -1;
					lp.addConstraint(coefficients, LpSolve.GE, 0.0);
					coefficients[uv] = -1;
					coefficients[uc] = 1;
					coefficients[vc] = 1;
					lp.addConstraint(coefficients, LpSolve.GE, 0.0);

					// resets entries in the array for reuse
					coefficients[uv] = 0;
					coefficients[uc] = 0;
					coefficients[vc] = 0;
				}
		return lp;
	}

	public static LpSolve addMetricUniquenessConstraints(LpSolve lp, double[] coefficients, int n, int k) throws LpSolveException {

		int offset = n * (n - 1) / 2;

		for (int u = 0; u < n; ++u) {
			for (int c = 0; c < k; ++c) {
				int uc = getPos2(u, c, n) + offset + 1;
				coefficients[uc] = 1;
			}
			lp.addConstraint(coefficients, LpSolve.GE, k - 1);
			// reset
			for (int c = 0; c < k; ++c) {
				int uc = getPos2(u, c, n) + offset + 1;
				coefficients[uc] = 0;
			}
		}
		return lp;
	}

	// variant Merge

	public static LpSolve addMetricClusterTripleConstraints(LpSolve lp, double[] coefficients, int n, int k) throws LpSolveException {

		int offset2 = n * (n - 1) / 2 +  n * k;

		for (int c = 0; c < k; ++c)
			for (int d = c + 1; d < k; ++d) {
				int cd = getPos(c, d, k) + offset2 + 1;
				for (int e = d + 1; e < k; ++e) {
					int ce = getPos(c, e, k) + offset2 + 1;
					int de = getPos(c, e, k) + offset2 + 1;
					
					// set triangle-inequality constraints
					coefficients[cd] = 1;
					coefficients[ce] = -1;
					coefficients[de] = 1;
					lp.addConstraint(coefficients, LpSolve.GE, 0.0);
					coefficients[cd] = 1;
					coefficients[ce] = 1;
					coefficients[de] = -1;
					lp.addConstraint(coefficients, LpSolve.GE, 0.0);
					coefficients[cd] = -1;
					coefficients[ce] = 1;
					coefficients[de] = 1;
					lp.addConstraint(coefficients, LpSolve.GE, 0.0);

					// resets entries in the array for reuse
					coefficients[cd] = 0;
					coefficients[ce] = 0;
					coefficients[de] = 0;
					
				}
			}
		return lp;
	}
	
	
	public static LpSolve addMetricNodeClusterPairConstraints(LpSolve lp, double[] coefficients, int n, int k) throws LpSolveException {
		
		int offset = n * (n - 1) / 2;
		int offset2 = offset +  n * k;
		
		for (int c = 0; c < k; ++c)
			for (int d = c + 1; d < k; ++d) {
				
				int cd = getPos(c, d, k) + offset2 + 1;

				for (int u = 0; u < n; ++u) {
					int uc = getPos2(u, c, n) + offset + 1;
					int ud = getPos2(u, d, n) + offset + 1;
					
					// set triangle-inequality constraints
					coefficients[cd] = 1;
					coefficients[uc] = -1;
					coefficients[ud] = 1;
					lp.addConstraint(coefficients, LpSolve.GE, 0.0);
					coefficients[cd] = 1;
					coefficients[uc] = 1;
					coefficients[ud] = -1;
					lp.addConstraint(coefficients, LpSolve.GE, 0.0);
					coefficients[cd] = -1;
					coefficients[uc] = 1;
					coefficients[ud] = 1;
					lp.addConstraint(coefficients, LpSolve.GE, 0.0);

					// resets entries in the array for reuse
					coefficients[cd] = 0;
					coefficients[uc] = 0;
					coefficients[ud] = 0;
				}
			}

		
		return lp;
	}
	
	
	// objective function

	public static LpSolve addPartialObjectiveFunction(LpSolve lp, double[] coefficients, int n, PartialObjectiveFunction objF, boolean minimize,
			ArrayList<Node> nodes, ArrayList<Cluster> clusters) throws LpSolveException {

		int offset = n * (n - 1) / 2;

		// copy coefficient from objfct
		for (int i = 0; i < n; ++i)
			for (int j = i + 1; j < n; ++j) {
				int ij = getPos(i, j, n) + 1;
				coefficients[ij] = objF.getCoefficient(nodes.get(i), nodes.get(j));
			}
		for (int i = 0; i < n; ++i)
			for (int j = 0; j < clusters.size(); ++j) {
				int ij = getPos2(i, j, n) + offset + 1;
				coefficients[ij] = objF.getCoefficient(nodes.get(i), clusters.get(j));
			}

		// set objective function
		lp.setObjFn(coefficients);
		// set type of optimization
		if (minimize) {
			lp.setMinim();
		} else {
			lp.setMaxim();
		}

		return lp;
	}

}
