package de.uka.algo.dynclustering.eval;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import de.uka.algo.clustering.comparison.ComparatorFactory;
import de.uka.algo.clustering.index.*;
import de.uka.algo.dynclustering.algos.base.AlgorithmFactory;
import de.uka.algo.dynclustering.algos.base.DynamicClusteringAlgorithm;
import de.uka.algo.dynclustering.graphs.DynamicGraph;
import de.uka.algo.dynclustering.graphs.DynamicGraph.GraphProperty;
import de.uka.algo.dynclustering.input.ClusteringJournalFileImporter;
import de.uka.algo.dynclustering.input.EmailFileImporter;
import de.uka.algo.dynclustering.input.GraphJournalFileImporter;
import de.uka.algo.dynclustering.util.RunTimer;
import de.uka.algo.dynclustering.util.RunTimer.TimeUnit;
import de.uka.algo.generator.standalone.graph.journaling.ClusteringJournal;

public class Evaluation2 {
	
	private static final int gcInterval = 500;
	private DynamicGraph dynGraph;
	private LinkedList<DynamicClusteringAlgorithm> activeAlgos = new LinkedList<DynamicClusteringAlgorithm>();
	
	private LinkedList<IndexFactory> activeQIs; // factories for quality indices which this evaluation keeps track of
	private LinkedList<ComparatorFactory> activeDMs; // factories for distance measures " 
	private Results results;
	private ClusteringJournal cJournal;
	
	private boolean success = false;
	private RunTimer timer = new RunTimer();
	
	public int timeStepAggregation; // the number of time steps that are aggregated for each evaluation step
	

	
	public void importGraph(String filePath) {
		if (filePath.endsWith(".graphj")) {
			this.dynGraph = (new GraphJournalFileImporter()).importFile(filePath);
		} else if (filePath.endsWith(".txt")) {
			this.dynGraph = (new EmailFileImporter(this.timeStepAggregation)).importFile(filePath);
		}

	}
	
	
	public void importClusteringJournal(String filePath) {
		this.cJournal = ClusteringJournalFileImporter.importFile(filePath);
	}
	
	
	
	
	
	
	/**
	 * Add instances of the algorithms contained in algoFactoryList
	 * to algoList and register them as listeners to the graph.
	 */
	public void registerAlgorithms(Collection<AlgorithmFactory> algoFactories) {
		if (this.dynGraph != null) {
			this.activeAlgos = new LinkedList<DynamicClusteringAlgorithm>();
			for (AlgorithmFactory algoFactory : algoFactories) {
				DynamicClusteringAlgorithm algo = algoFactory.getAlgorithm(this.dynGraph);
				this.activeAlgos.add(algo);
			}
		} else throw new IllegalStateException("A Graph must be imported first");
	}
	
	
	public void registerQualityIndices(Collection<IndexFactory> qIFactories) {
		this.activeQIs = new LinkedList<IndexFactory>();
		this.activeQIs.addAll(qIFactories);
	}
	
	public void registerDistanceMeasures(Collection<ComparatorFactory> dMFactories) {
		this.activeDMs = new LinkedList<ComparatorFactory>();
		this.activeDMs.addAll(dMFactories);
	}
	
	
	public void run() {
		this.results = new Results(this.activeQIs, this.activeDMs, this.activeAlgos);
		this.timer.start();
		//
		this.evolveGraph();
		//
		this.timer.stop();
	}
	
	@Deprecated
	public void runFor(int tMax) {
		this.results = new Results(this.activeQIs, this.activeDMs, this.activeAlgos);
		this.evolveGraph(tMax);
	}
	
	
	
	
	/**
	 * Iterate over all timesteps of the dynamic graph,
	 * let algorithms react to changes, read results.
	 */
	private void evolveGraph() {
		try {
			while (this.dynGraph.hasNextChange()) {
				boolean isTimeStep = this.dynGraph.nextChange();
				// update global vars
				GlobalVars.n = this.dynGraph.N();
				
				// algorithms react to change
				if (isTimeStep) {
					this.results.setTimeStep(this.dynGraph.t());
					this.storeResults();
					if (this.dynGraph.t() % gcInterval == 0) {
						System.gc(); // GARBAGE COLLECTION HERE
					}
					// DEBUG
					System.out.println("####### TIME STEP " + this.dynGraph.t() + " ######");
					//DEBUG
					
				}
			}
			this.success = true;
		} catch (Exception e) {
			System.out.println("EXCEPTION at time step " + this.dynGraph.t());
			System.out.println("Stack Trace:");
			e.printStackTrace();
		}
	}
	
	/**
	 * Iterate over a maximum of 
	 * @param tMax	time steps
	 * let algorithms react to changes, read results.
	 */
	private void evolveGraph(int tMax) {
		while (this.dynGraph.hasNextChange() && this.dynGraph.t() <= tMax) {
			boolean isTimeStep = this.dynGraph.nextChange();
			// algorithms react to change
			if (isTimeStep) {
				this.storeResults();
			}
		}
	}
	
	
	/**
	 * Evaluate and store the results.
	 */
	private void storeResults() {
		
		
		for (DynamicClusteringAlgorithm algo : this.activeAlgos) {
			/* evaluate and store quality indices */
			for (IndexFactory qIFactory : this.activeQIs) {
				double q = algo.getQuality(qIFactory);
				this.results.setQuality(algo, qIFactory, q);
			}
			
			
			/* evaluate and store distance measures */
			for (ComparatorFactory dMFactory : this.activeDMs) {
				double d = algo.getDistance(dMFactory);
				this.results.setDistance(algo, dMFactory, d);
			}
			
			/* store runtimes */
			this.results.setRuntime(algo, algo.getRuntime(TimeUnit.MilliSeconds));
			
			/* store clustering properties */
			this.results.setClusterCount(algo, algo.getClusterCount());
			
			algo.savePreviousClustering();
		}
		
		/* graph properties for this time step */
		this.results.setGraphProperty(GraphProperty.n, this.dynGraph.N());
		this.results.setGraphProperty(GraphProperty.m, this.dynGraph.E());
	}
	
	
	public Results getResults() {
		
		this.results.algoNames = this.getAlgoNames();
		this.results.qINames = this.getQINames();
		this.results.dMNames = this.getDMNames();
		this.results.totalRuntime = this.timer.read(RunTimer.TimeUnit.Seconds);
		this.results.clusteringHistory = new ClusteringHistory(this.cJournal);
		this.results.maxMemory = Runtime.getRuntime().maxMemory();
		
		return this.results;
	}
	
	
	
	
	
	
	public List<DynamicClusteringAlgorithm> getAlgorithms() {
		return this.activeAlgos;
	}
	
	private String[] getAlgoNames() {
		String[] names = new String[this.activeAlgos.size()];
		int i = 0;
		for (DynamicClusteringAlgorithm algo : this.activeAlgos) {
			names[i] = algo.toString();
			++i;
		}
		return names;
	}



	private String[] getQINames() {
		String[] names = new String[this.activeQIs.size()];
		int i = 0;
		for (IndexFactory qi : this.activeQIs) {
			names[i] = qi.toString();
			++i;
		}
		return names;
	}
	
	private String[] getDMNames() {
		String[] names = new String[this.activeDMs.size()];
		int i = 0;
		for (ComparatorFactory dm : this.activeDMs) {
			names[i] = dm.toString();
			++i;
		}
		return names;
	}
	
	public boolean success() {
		return this.success;
	}

}
