package de.uka.algo.dynclustering.eval;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;


import de.uka.algo.clustering.comparison.ComparatorFactory;
import de.uka.algo.clustering.index.IndexFactory;
import de.uka.algo.dynclustering.algos.base.DynamicClusteringAlgorithm;
import de.uka.algo.dynclustering.graphs.DynamicGraph.GraphProperty;
import de.uka.algo.dynclustering.util.StringUtils;

public class Results implements Serializable {
	
	
	/**
	 * 
	 */
	private static final long serialVersionUID = -3305151250773099509L;

	public String[] algoNames;
	public String[] qINames;
	public String[] dMNames;

	public ClusteringHistory clusteringHistory;
	public double totalRuntime;
	public long maxMemory;
	
	
	ArrayList<Integer> time; // an ordered list of time steps
	

	
	// a graph property is a function of time
	HashMap<Integer, HashMap<String, Number>> graphProperties;
	// a measure returns a double value which is a function of algorithm and time
	HashMap<Integer, HashMap<String, HashMap<String, Double>>> measures;
	
	/* the following are measures */
	
	ArrayList<String> qIList; // list of active quality indices
	ArrayList<String> dMList; // list of active distance measures
	ArrayList<String> algoList; // list of active algorithms
	
	ArrayList<String> clusteringProperties; // list of properties per clustering

	private int currentT; // the current time
	
	public Results(Collection<IndexFactory> activeQIs, Collection<ComparatorFactory> activeDMs, Collection<DynamicClusteringAlgorithm> activeAlgos) {
		this.time = new ArrayList<Integer>();
		this.algoList = new ArrayList<String>();
		for (DynamicClusteringAlgorithm algo : activeAlgos) {
			this.algoList.add(algo.toString());
		}
		this.dMList = new ArrayList<String>();
		for (ComparatorFactory dM : activeDMs) {
			this.dMList.add(dM.toString());
		}
		this.qIList = new ArrayList<String>();
		for (IndexFactory qI : activeQIs) {
			this.qIList.add(qI.toString());
		}
		
		this.clusteringProperties = new ArrayList<String>();
		this.clusteringProperties.add("k"); // number of clusters
		
		this.graphProperties = new HashMap<Integer, HashMap<String, Number>>();
		this.measures = new HashMap<Integer, HashMap<String, HashMap<String, Double>>>();
		this.measures = new HashMap<Integer, HashMap<String, HashMap<String, Double>>>();
	}
	
	
	public void setQuality(DynamicClusteringAlgorithm algo, IndexFactory qIFactory, double q) {
		this.measures.get(this.currentT).get(algo.toString()).put(qIFactory.toString(), q);
	}
	
	public void setDistance(DynamicClusteringAlgorithm algo, ComparatorFactory dMFactory, double d) {
		this.measures.get(this.currentT).get(algo.toString()).put(dMFactory.toString(), d);
	}
	
	/**
	 * Set the runtime in milliseconds
	 * @param algo
	 * @param t
	 */
	public void setRuntime(DynamicClusteringAlgorithm algo, double t) {
		this.measures.get(this.currentT).get(algo.toString()).put("Runtime", t);
	}
	
	
	/**
	 * Set the current number of clusters
	 * @param algo
	 * @param k
	 */
	public void setClusterCount(DynamicClusteringAlgorithm algo, int k) {
		this.measures.get(this.currentT).get(algo.toString()).put("k", (double) k);
	}
	
	
	public void setGraphProperty(GraphProperty p, Number v) {
		this.graphProperties.get(this.currentT).put(p.toString(), v);
	}
	

	
	


	/**
	 * React to time step event by incrementing t.
	 */
	public void setTimeStep(int t) {
		this.currentT = t;
		this.time.add(t);
		this.graphProperties.put(t, new HashMap<String,Number>());
		this.measures.put(t, new HashMap<String, HashMap<String, Double>>());
		this.measures.put(t, new HashMap<String, HashMap<String, Double>>());
		for (String algo : this.algoList) {
			this.measures.get(t).put(algo, new HashMap<String, Double>());
			this.measures.get(t).put(algo, new HashMap<String, Double>());
		}
		
	}
	
	public String toString() {
		String qualities = "qualities: " + this.measures.toString() + " \n";
		String measures = "measures: " + this.measures.toString() + "\n";
		String graphProperties = "graph properties: " + this.graphProperties.toString() + "\n";
		return qualities + measures + graphProperties;
	}
	
	
	public ArrayList<String> asCSVLines(String sep) {
		// DEBUG
		System.out.println(this.measures);
		// DEBUG
		
		
		ArrayList<String> result = new ArrayList<String>();
		
		/* header */
		String[] h = {"time","type","measure"};
		String header = StringUtils.concat(StringUtils.join(h, sep), StringUtils.join(this.algoList, sep), sep);
		result.add(header);
		
		/* body */
		for (Integer t : this.time) {
			String timeCol = t.toString();
			
			String typeCol = "quality";
			for (String qIName : this.qIList) {
				String measureCol = qIName;
				/* per column */
				ArrayList<String> qIValues = new ArrayList<String>();
				for (String algo : this.algoList) {
					double qI = this.measures.get(t).get(algo).get(qIName);
					qIValues.add("" + qI);
				}
				String[] l = {timeCol, typeCol, measureCol};
				String line = StringUtils.concat(StringUtils.join(l, sep), StringUtils.join(qIValues, sep), sep);
				result.add(line);
			}
			
			typeCol = "distance";
			for (String dMName: this.dMList) {
				String measureCol = dMName;
				ArrayList<String> dMValues = new ArrayList<String>();
				for (String algo : this.algoList) {
					double dM = this.measures.get(t).get(algo).get(dMName);
					dMValues.add("" + dM);
				}
				String[] l = {timeCol, typeCol, measureCol};
				String line = StringUtils.concat(StringUtils.join(l, sep), StringUtils.join(dMValues, sep), sep);
				result.add(line);
			}
			
			typeCol = "clustering";
			for (String cProp : this.clusteringProperties) {
				String measureCol = cProp;
				ArrayList<String> cPropValues = new ArrayList<String>();
				for (String algo : algoList) {
					double cPropValue = this.measures.get(t).get(algo).get(cProp);
					cPropValues.add("" + cPropValue);
				}
				String[] l = {timeCol, typeCol, measureCol};
				String line = StringUtils.concat(StringUtils.join(l, sep), StringUtils.join(cPropValues, sep), sep);
				result.add(line);
			}
			
			/* runtime */
			
			typeCol = "algorithm";
			String measureCol = "runtime";
			ArrayList<String> rtValues = new ArrayList<String>();
			for (String algo : this.algoList) {
				double rt = this.measures.get(t).get(algo).get("Runtime");
				rtValues.add("" + rt);
			}
			String[] l = {timeCol, typeCol, measureCol};
			String line = StringUtils.concat(StringUtils.join(l, sep), StringUtils.join(rtValues, sep), sep);
			result.add(line);
			
			
			
		}
		
		return result;
		
	}
	
	/**
	 * List the values of a measure ordered by time. The value is
	 * a function of algorithm and time.
	 * @param algo		name of the selected algorithm
	 * @param measure	name of the selected measure
	 * @return	list of tuples (time, measure(algorithm))
	 */
	public ArrayList<Double> listMeasure(String algo, String measure) {
		ArrayList<Double> result = new ArrayList<Double>();
		for (int t: this.time) {
			double value = this.measures.get(t).get(algo).get(measure);
			result.add(value);
		}
		return result;
	}
	
	
	/**
	 * List the values of a graph property ordered by time.
	 * @param property
	 * @return
	 */
	public List<Number> listGraphProperty(String property) {
		ArrayList<Number> result = new ArrayList<Number>();
		for (int t : this.time) {
			result.add(this.graphProperties.get(t).get(property));
		}
		return result;
	}
	
	public int getTMax() {
		return this.time.get(this.time.size()-1);
	}
	
	


}
