package de.uka.algo.dynclustering.graphs;

import de.uka.algo.generator.util.BidiHashMap;
import de.uka.algo.graphs.GraphInterpretation;
import de.uka.algo.graphs.GraphInterpretationListener;
import de.uka.algo.graphs.WeightedGraph;
import y.base.Edge;
import y.base.EdgeCursor;
import y.base.EdgeMap;
import y.base.Graph;
import y.base.GraphEvent;
import y.base.GraphListener;
import y.base.Node;
import y.base.NodeCursor;

/**
 * Couples a supergraph H = (V',E')  to a graph G = (V,E) so that
 * V is a subset of V' and E is a subset of E'.
 * Changes to G automatically write through to H. 
 * 
 * @author Christian Staudt
 * 
 * TODO: test and complete if needed
 *
 */
public class SuperGraphCoupler implements GraphListener, GraphInterpretationListener {

	Graph superGraph;
	Graph subGraph;
	
	GraphInterpretation subGraphInterpretation;
	GraphInterpretation superGraphInterpretation;
	WeightedGraph subGraphWeights;
	WeightedGraph superGraphWeights;
	boolean weighted;
	
	BidiHashMap<Node,Node> super2subNodes = new BidiHashMap<Node,Node>(); // maps nodes of supergraph to corresponding nodes of subgraph
	BidiHashMap<Edge,Edge> super2subEdges = new BidiHashMap<Edge,Edge>(); // " edges "
	
	
	public SuperGraphCoupler(Graph subGraph, GraphInterpretation subGraphInterpretation) {
		if (subGraphInterpretation.getGraph() == subGraph) {
			this.subGraph = subGraph;
			this.subGraphInterpretation = subGraphInterpretation;
			
			this.superGraph = new Graph();
			this.duplicateSubGraph();
			
			if (this.subGraphInterpretation instanceof WeightedGraph) {
				this.weighted = true;
				this.subGraphWeights = (WeightedGraph) this.subGraphInterpretation;
			} else {
				this.weighted = false;
			}
			this.duplicateSubGraphInterpretation();
			
			this.superGraph.addGraphListener(this);
			this.subGraph.addGraphListener(this);
		} else throw new IllegalArgumentException("graph interpretation does not belong to the given graph");
	}
	
	
	protected void duplicateSubGraphInterpretation() {
		// TODO: finish
		if (this.weighted) {
			this.superGraphInterpretation = this.superGraphWeights = new WeightedGraph(this.superGraph, this.superGraph.createEdgeMap());
		}
	}


	/**
	 * Create nodes and edges in the empty supergraph until
	 * it is equal to the subgraph.
	 */
	protected void duplicateSubGraph() {
		if (this.superGraph.isEmpty()) {
			for (NodeCursor nc = this.subGraph.nodes(); nc.ok(); nc.next()) {
				this.mirrorNode(nc.node());
			}
			for (EdgeCursor ec = this.subGraph.edges(); ec.ok(); ec.next()) {
				this.mirrorEdge(ec.edge());
			}
		} else throw new IllegalStateException("supergraph is not empty");
	}
	
	
	
	public void onGraphEvent(GraphEvent ev) {
		if (ev.getGraph().equals(this.subGraph)) {
			/* events in the subgraph */
			if (ev.equals(GraphEvent.NODE_CREATION)) {
				Node u_ = (Node) ev.getData();
				Node u = this.superGraph.createNode();
				this.super2subNodes.put(u_, u);
			} else if (ev.equals(GraphEvent.PRE_NODE_REMOVAL)) {
				Node u_ = (Node) ev.getData();
				Node u = this.super2subNodes.getKey(u_);
				this.superGraph.removeNode(u);
				this.super2subNodes.removeKey(u);
			} else if (ev.equals(GraphEvent.EDGE_CREATION)) {
				Edge u_v_ = (Edge) ev.getData();
				Node u = this.super2subNodes.getKey(u_v_.source());
				Node v = this.super2subNodes.getKey(u_v_.target());
				Edge uv = this.superGraph.createEdge(u, v);
				this.super2subEdges.put(uv, u_v_);
			} else if (ev.equals(GraphEvent.PRE_EDGE_REMOVAL)) {
				Edge u_v_ = (Edge) ev.getData();
				Edge uv = this.super2subEdges.getKey(u_v_);
				this.superGraph.removeEdge(uv);
				this.super2subEdges.removeKey(uv);
			}
		} else if (ev.getGraph().equals(this.superGraph)) {
			/* events in the supergraph */
			// TODO: what to write through?
		}

	}
	
	
	/**
	 * Create a node in the supergraph which
	 * corresponds to a node in the subgraph.
	 * @param u		node in subgraph
	 * @return
	 */
	protected Node mirrorNode(Node u_) {
		Node u = this.superGraph.createNode();
		this.super2subNodes.put(u, u_);
		return u;
	}
	
	/**
	 * Create a node in the supergraph but not in the subgraph.
	 * @return
	 */
	public Node addNode() {
		Node u = this.superGraph.createNode();
		this.super2subNodes.put(u, u);
		return u;
	}
	
	
	/**
	 * Create an edge in the supergraph which
	 * corresponds to an edge in the subgraph.
	 * @param uv_
	 * @return
	 */
	protected Edge mirrorEdge(Edge uv_) {
		Node u = this.super2subNodes.getKey(uv_.source());
		Node v = this.super2subNodes.getKey(uv_.target());
		Edge uv = this.superGraph.createEdge(u,v);
		this.super2subEdges.put(uv, uv_);
		return uv;
	}
	
	/**
	 * Create an edge in the supergraph but not in the subgraph;
	 * @param u
	 * @param v
	 * @return
	 */
	public Edge addEdge(Node u, Node v) {
		Edge uv = this.superGraph.createEdge(u, v);
		return uv;
	}

	
	public boolean onlyInSuperGraph(Node u) {
		return (this.super2subNodes.getValue(u) == u);
	}
	
	public boolean onlyInSuperGraph(Edge uv) {
		return (this.super2subEdges.getValue(uv) == uv);
	}
	
	public Graph getSuperGraph() {
		return this.superGraph;
	}
	
	public GraphInterpretation getSubGraphInterpretation() {
		return this.subGraphInterpretation;
	}
	
	public GraphInterpretation getSuperGraphInterpretation() {
		return this.superGraphInterpretation;
	}
	
	public Graph getSubGraph() {
		return this.subGraph;
	}

	// GraphInterpretationListener

	public void newWeightsEvent() {
		// TODO Auto-generated method stub
		
	}


	public void weightChangeEvent(Edge edge, double oldWeight, double newWeigth) {
		// TODO Auto-generated method stub
		
	}
}
