package de.uka.algo.dynclustering.util.unionfind;

import java.util.Set;

/**
 * Abstract base class for a generic union find data structure.
 * Provides template methods for the most common operations.
 * @author Christian Staudt
 *
 * @param <T>
 */
public class UnionFind<T> implements UnionFindInterface<T> {
	
	protected UnionFindImplementation<T> imp;
	protected boolean pathCompression;
	
	
	public UnionFind(UnionFindImplementation<T> imp, boolean pathCompression) {
		this.imp = imp;
		this.pathCompression = pathCompression;
	}
	
	public void add(T anElement) {
		this.imp.add(anElement);
	}
	
	public void remove(T anElement) {
		this.imp.remove(anElement);
	}
	
	/**
	 * Unites the sets containing
	 * @param anElement
	 * 		an element
	 * @param anotherElement
	 * 		another element
	 * @return
	 * 		representative of the united set
	 */
	public T union(T anElement, T anotherElement) {
		T firstRoot = this.find(anElement);
		if (firstRoot == null) throw new IllegalArgumentException("element not found: " + anElement);
		T secondRoot = this.find(anotherElement);
		if (secondRoot == null) throw new IllegalArgumentException("element not found: " + anotherElement);
		return this.imp.join(firstRoot, secondRoot);
	}
	
	/**
	 * Finds the representative of a set containing
	 * @param anElement
	 * 		an element
	 * @return
	 * 		the representative.
	 */
	public T find(T anElement) {
		// TODO: optional path compression
		T e = anElement;
		while (!(this.imp.isRoot(e))) {
			e = this.imp.getParent(e);
		}
		return e;
	}
	
	/**
	 * Gets the current number of sets.
	 * @return
	 */
	public int getSetCount() {
		// delegate to implementation
		return this.imp.getSetCount();
	}
	
	/**
	 * Gets the set of representatives.
	 * @return
	 */
	public Set<T> getRepresentatives() {
		// delegate to implementation
		return this.imp.getRoots();
	}
	
	public boolean isRepresentative(T anElement) {
		return this.imp.isRoot(anElement);
	}
	
	public boolean contains(T anElement) {
		// delegate to implementation
		return this.imp.contains(anElement);
	}
	
	public Set<T> getElements() {
		// delegate to implementation
		return this.imp.getElements();
	}
	
	
}
