package de.uka.algo.dynclustering.util.unionfind;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import de.uka.algo.generator.util.BidiHashMap;

public class UnionNodeForest<E> {

	protected abstract class TreeNode {
		public UnionNode parent = null; // TreeNode is a root by default

		public boolean isRoot() {
			return this.parent == null;
		}

		public abstract boolean isLeaf();

	};

	protected class LeafNode extends TreeNode {
		public E element;

		public LeafNode(E anElement) {
			this.element = anElement;
		}

		@Override
		public boolean isLeaf() {
			return true;
		}

		public String toString() {
			return "L(" + this.element + ")" + this.hashCode();
		}
	};

	protected class UnionNode extends TreeNode {
		public TreeNode left;
		public TreeNode right;

		public UnionNode(TreeNode left, TreeNode right) {
			this.left = left;
			this.right = right;
		}

		@Override
		public boolean isLeaf() {
			return false;
		}

		public String toString() {
			return "U" + this.hashCode();
		}
	};

	protected BidiHashMap<E, LeafNode> element2leaf = new BidiHashMap<E, LeafNode>();
	protected Set<TreeNode> roots = new HashSet<TreeNode>();
	// temporary variables
	protected List<UnionNode> markedList = new ArrayList<UnionNode>();
	protected Set<UnionNode> markedSet = new HashSet<UnionNode>();

	public void add(E anElement) {
		// make it a singleton
		LeafNode leaf = new LeafNode(anElement);
		this.element2leaf.put(anElement, leaf);
//		this.roots.add(leaf);
	}

	public void remove(E anElement) {
		assert this.element2leaf.containsKey(anElement);
		LeafNode aLeaf = this.element2leaf.getValue(anElement);
		this.remove(aLeaf);
		this.element2leaf.removeKey(anElement);

	}

	protected void remove(LeafNode aLeaf) {
		if (aLeaf.isRoot()) {
			
		} else {
			UnionNode parent = aLeaf.parent;
			TreeNode sibling = this.getSibling(aLeaf);
			if (parent.isRoot()) {
				this.makeRoot(sibling);
			} else {
				this.replace(parent.parent, parent, sibling);
			}
			// if parent.parent is null, sibling becomes a root
			this.cutOut(parent);
		}
	}

	protected TreeNode find(TreeNode aNode) {
		TreeNode n = aNode;
		while (!(n.isRoot())) {
			n = n.parent;
		}
		return n;
	}

	protected TreeNode getSibling(TreeNode aNode) {
		if (aNode.isRoot())
			throw new IllegalArgumentException(aNode + " is a root and has no sibling");
		UnionNode parent = aNode.parent;
		if (parent.left == aNode) {
			return parent.right;
		} else {
			return parent.left;
		}
	}

	protected void union(TreeNode aNode, TreeNode anotherNode) {
		TreeNode aRoot = this.find(aNode);
		TreeNode anotherRoot = this.find(anotherNode);
		UnionNode unionNode = new UnionNode(aRoot, anotherRoot);
		aRoot.parent = unionNode;
		anotherRoot.parent = unionNode;
	}

	/**
	 * Performs a union on the sets containing the given elements.
	 * 
	 * @param anElement
	 * @param anotherElement
	 */
	public void union(E anElement, E anotherElement) {
		if (this.sameSet(anElement, anotherElement)) {
			// do nothing
		} else {
			TreeNode aNode = this.element2leaf.getValue(anElement);
			TreeNode anotherNode = this.element2leaf.getValue(anotherElement);
			this.union(aNode, anotherNode);
		}
	}

	/**
	 * Backtracks one step in the union history of the set containing the given
	 * element.
	 * 
	 * @param anElement
	 */
	public void backtrack(E anElement) {
		LeafNode aLeaf = this.element2leaf.getValue(anElement);
		this.backtrack(aLeaf);
	}

	protected void backtrack(LeafNode aLeaf) {
		TreeNode aRoot = this.find(aLeaf);
		if (aRoot != aLeaf) {
			this.split((UnionNode) aRoot);
		} else {
			// TODO: cannot backtrack
		}
	}
	

	public void separate(E anElement, E anotherElement) {
		if (this.sameSet(anElement, anotherElement)) {
			LeafNode aLeaf = this.element2leaf.getValue(anElement);
			LeafNode anotherLeaf = this.element2leaf.getValue(anotherElement);
			this.separate(aLeaf, anotherLeaf);
		} else
			throw new IllegalArgumentException("separation failed: elements " + anElement + " and " + anotherElement + " are not in the same set \n" + this.getPartition());

	}

	protected void separate(LeafNode aLeaf, LeafNode anotherLeaf) {
		markedSet.clear();

		UnionNode u = aLeaf.parent;
		// go to the root and mark all visited union nodes
		while (true) {
			markedSet.add(u);
			if (u.isRoot()) {
				break;
			} else {
				u = u.parent;
			}
		}
		// go to the root and delete all marked union nodes
		u = anotherLeaf.parent;
		while (true) {
			if (markedSet.contains(u)) {
				this.split(u);
			}
			if (u.isRoot()) {
				break;
			} else {
				u = u.parent;
			}
		}
	}

	protected void cutOut(UnionNode aUnionNode) {
		aUnionNode.parent = null;
		aUnionNode.left = null;
		aUnionNode.right = null;
	}
	
	
	/**
	 * Replace parent with child as the child of grandParent.
	 * @param grandParent
	 * @param parent
	 * @param child
	 */
	protected void replace(UnionNode grandParent, UnionNode parent, TreeNode child) {
		child.parent = grandParent;
		if (grandParent.left == parent) {
			grandParent.left = child;
		} else {
			grandParent.right = child;
		}
	}

	protected void makeRoot(TreeNode aNode) {
		aNode.parent = null;
	}

	/**
	 * Splits at a union node, producing two trees.
	 * 
	 * @param aUnionNode
	 */
	protected void split(UnionNode aUnionNode) {
		this.makeRoot(aUnionNode.left);
		this.makeRoot(aUnionNode.right);
	}
	
	
	public void isolate(E anElement) {
		LeafNode aLeaf = this.element2leaf.getValue(anElement);
		this.isolate(aLeaf);
	}
	
	protected void isolate(LeafNode aLeaf) {
		
		if (aLeaf.isRoot()) {
			return;
		} else {
			markedList.clear();
			UnionNode u = aLeaf.parent;
			// go to the root and mark all visited union nodes
			while (true) {
				markedList.add(u);
				if (u.isRoot()) {
					break;
				} else {
					u = u.parent;
				}
			}
			for (int i = this.markedList.size() - 1; i >= 0; i--) {
				this.split(this.markedList.get(i));
			}
		}
 
//		while (!(aLeaf.isRoot())) {
//			this.split((UnionNode) this.find(aLeaf));
//		}
	}

	public boolean sameSet(E anElement, E anotherElement) {
		LeafNode aLeaf = this.element2leaf.getValue(anElement);
		LeafNode anotherLeaf = this.element2leaf.getValue(anotherElement);
		if (aLeaf.isRoot() || anotherLeaf.isRoot()) {
			return false;
		} else {
			return this.find(aLeaf).equals(this.find(anotherLeaf));
		}
	}

	/**
	 * Returns the current partition of the set of elements.
	 * 
	 * @return
	 */
	public HashSet<HashSet<E>> getPartition() {
		HashSet<HashSet<E>> partition = new HashSet<HashSet<E>>();
		HashMap<TreeNode, HashSet<E>> root2set = new HashMap<TreeNode, HashSet<E>>();
		for (E element : this.element2leaf.keySet()) {
			LeafNode leaf = this.element2leaf.getValue(element);
			TreeNode root = this.find(leaf);
			if (root2set.containsKey(root)) {
				root2set.get(root).add(element);
			} else {
				HashSet<E> set = new HashSet<E>();
				partition.add(set);
				root2set.put(root, set);
				set.add(element);
			}
		}
		return partition;
	}
	
	
	/**
	 * Associates each element with a representative element
	 * of its set.
	 * @return
	 */
	public HashMap<E,E> getPartitionMap() {
		HashMap<E,E> map = new HashMap<E,E>();
		HashMap<TreeNode,E> root2rep = new HashMap<TreeNode,E>();
		for (E element : this.element2leaf.keySet()) {
			LeafNode leaf = this.element2leaf.getValue(element);
			TreeNode root = this.find(leaf);
			if (root2rep.containsKey(root)) {
				map.put(element, root2rep.get(root));
			} else {
				map.put(element, element);
				root2rep.put(root, element);
			}
		}
		return map;
	}


}
