package de.uka.algo.traversal;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.Node;

/**
 * An abstract base class for BFS-like traversal algorithms.
 * They work by expanding a node, traversing its incident edges and adding
 * its neighbors to a queue. Subclasses need to implement @see SingleSourceTraversal
 * or @see MultiSourceTraversal to provide a fully functional interface.
 * 
 * @author Christian Staudt
 *
 */
public abstract class BFSLikeBase extends GraphTraversalImplementation {
	
	protected Queue<Node> queue = new LinkedList<Node>(); //  queue
	protected Set<Node> known = new HashSet<Node>(); // known nodes (encountered at least once)
	
	protected Node currentNode; // the node currently selected
	protected int nc = 0; // number of nodes which have been selected
	
	protected Map<Node,Integer> distanceMap = new HashMap<Node,Integer>();
	protected int cd = 0; // current distance

	@Override
	protected Node dequeue() {
		this.cd = this.distanceMap.get(this.queue.peek());
		// this.distanceMap.remove(this.queue.peek())
		return this.queue.remove();
	}

	@Override
	protected void enqueue(Node u) {
		// add u to the queue
		this.queue.add(u);
	}

	@Override
	protected void expand(Node u) {
		// traverse incident edges
		for (EdgeCursor ec = u.edges(); ec.ok(); ec.next()) {
			Edge uv = ec.edge();
			this.traverse(uv);
		}	
	}

	@Override
	protected void jumpTo(Node u) {
		// only called for start node
		if (! this.known.contains(u)) {
			this.known.add(u);
			this.enqueue(u);
			// distance
			this.distanceMap.put(u, 0);
		}
	}

	public void reset() {
		this.queue.clear();
		this.known.clear();
		this.nc = 0;
		this.distanceMap.clear();
		this.cd = 0;
	}

	@Override
	protected void traverse(Edge uv) {
		// enqueue unknown targets
		if (! this.known.contains(uv.target())) {
			this.known.add(uv.target());
			this.enqueue(uv.target());
			// distance
			this.distanceMap.put(uv.target(), this.distanceMap.get(uv.source()) + 1);
		}
	}

	public int count() {
		return this.nc;
	}
	
	public int distance() {
		return this.cd;
	}

	public void next() {
		this.expand(this.currentNode);
		++this.nc; 
	}

	public Node node() {
		return this.currentNode;
	}

	public boolean ok() {
		boolean ok = ! this.queue.isEmpty();
		if (ok) {
			this.currentNode = this.dequeue();
		}
		return ok;
	}

}
