/*
 * Created on Jul 5, 2009
 *
 * Copyright 2009.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */

package de.uka.algo.clustering.algorithms;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.PriorityQueue;

import y.base.EdgeCursor;
import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.graphs.GraphInterpretation;
import de.uka.algo.util.GYCursor;

/**
 * The EdgeHeapModularity Cluster Algorithm agglomerates nodes by contracting edges in the order of their modularity gain, recompute?
 * stop?
 * 
 * @author Robert
 */
public final class EdgeHeapModularity implements Algorithm {
	private ArrayList<Integer>[] clusterContent; // list of node indices inside this cluster
	private int[] c; // to which node's cluster does this node belong?
	private Double[] degSum; // sum of degrees of nodes in this cluster
	private HashMap<Integer, Double>[] adj; // adjacency of all adjacent cluster pairs
	private PriorityQueue<myEdge> edgeHeap; // the queue, only lower -> higher !
	private double weightSum; // sum of all weights in graph
	private double degsSq; // 4 * weightSum^2 = degSum^2
	private int refreshT = 1; // number of merges before queue is refreshed
	private int mergeCounter = 0; // number of performed merges 
	private boolean allowUpdate = false; // should the given clustering be used and only further coarsened, otherwise we start from singletons  
	private boolean enableBacktrack = false;  // should the merge operation from Clustering be used, otherwise quick resetToClustering is used.

	/**
	 * Does all the big decisions, basically runs the algorithm.
	 * 
	 * @param aClustering
	 *            the clustering to reset
	 */
	public void run(Clustering aClustering) {
//		System.out.println("Greedy modularity agglomerative algorithm called, updating merge-values every " + refreshT + " merges.");
//		int preCC = aClustering.clusterCount();
		GraphInterpretation gi = aClustering.getInterpretation();
		weightSum = gi.getEdgeWeightSum();
		degsSq = 4.0 * weightSum * weightSum;
		fillClusterStuff(aClustering, gi); // clusterContent and degSum and valid and adj and c
		fillQueue(aClustering); // only lower -> higher
		int it = 0;
		while ((edgeHeap.size() >= 1) && (edgeHeap.peek().getDq() > 0.0)) {
			it += merge(aClustering);
//			System.out.println("refreshT = " + refreshT + ", it = " + it);
			if (0 == (it % refreshT))
				fillQueue(aClustering);
		}
		if (edgeHeap.size() >= 1) {
//			System.out.println("Next increse: " + edgeHeap.peek().getDq() + ", stopping");

		}
		else {
//			System.out.println("Reached connected components.");
		}
		if (!enableBacktrack)
			aClustering.resetToClustering(c);
//		System.out.println("Done.");
	}

	public void setRefreshCount(int x) {
		refreshT = x;
	}

	public void setAllowUpdate(boolean allow) {
		allowUpdate = allow;
	}
	
	public void setBacktrackable(boolean backtrackable) {
		enableBacktrack = backtrackable;
	}

	/**
	 * Does what needs to be done if two clusters are merged,
	 * by polling the highest priority edge from the queue. The two incident
	 * nodes' clusters are then merged. 
	 * The source of the contracted edge is the leader of the new cluster.
	 * It steals the other leaders degSum, its adjacencies 
	 * (added up or set anew if not present before merge), its clusterContent, 
	 * and sets for all nodes i in the other's cluster c[i] = leader.
	 * If backtracking is enabled, the given clustering is updated via the merge operation.    
	 * 
	 * @param aClustering
	 * @return 1 if merge performed, 0 if not (redundant edge polled from queue)
	 */
	private int merge(Clustering aClustering) {
		myEdge me = edgeHeap.poll();
		if (c[me.s] != c[me.t]) {
//			System.out.println("(" + (++mergeCounter) +  ") merging " + me.s + " (" + c[me.s] + ") and " + me.t + " (" + c[me.t] + "), yields "
//					+ me.getDq());
			int cs = c[me.s];
			int ct = c[me.t];
			int cnb;
			double tmp;
			HashMap<Integer, Double> tmpMap = new HashMap<Integer, Double>();
			adj[cs].remove(ct);
			adj[ct].remove(cs);
			for (Integer nb : adj[cs].keySet()) { // leaders of all common neighbors of s and t
				cnb = c[nb];
				if (cnb == nb){	// only leaders, others cannot have adjacencies any more
					if (adj[ct].containsKey(cnb)) {
						tmp = adj[cs].get(cnb) + adj[ct].get(cnb);
						tmpMap.put(cnb, tmp);
						adj[cnb].put(cs, tmp);
						adj[ct].remove(cnb);
						adj[cnb].remove(ct);
					}
				}
			}
			adj[cs].putAll(tmpMap);
			for (Integer nb : adj[ct].keySet()) { // all exclusive neighbors of t
				cnb = c[nb];
				if (cnb == nb){	// only leaders, others cannot have adjacencies any more
					if (!adj[cs].containsKey(cnb)) {
						tmp = adj[ct].get(cnb);
						adj[cs].put(cnb, tmp);
						adj[cnb].put(cs, tmp);
					}
				}
				adj[cnb].remove(ct); // weg
			}
			adj[ct].clear();
			for (int tn : clusterContent[ct])
				c[tn] = cs;
			if (clusterContent[cs].size() == 0)
				System.out.println("clusterContent von cs leer");
			if (clusterContent[ct].size() == 0)
				System.out.println("clusterContent von ct leer");
			clusterContent[cs].addAll(clusterContent[ct]);
			clusterContent[ct].clear(); // weg
			degSum[cs] += degSum[ct];
			degSum[ct] = 0.0; // weg
			if (enableBacktrack){
				aClustering.merge(aClustering.getGraph().getNodeArray()[cs], aClustering.getGraph().getNodeArray()[ct]);
			}
			return 1;
		} else {
//			System.out.println("invalid merge skipped " + me.s + " and " + me.t);
			return 0;
		}
	}

	/**
	 * For all node indices which are their own leaders, we put the edges s -> t,
	 * with s > t into the priority queue for potential merges.   
	 * 
	 * @param clustering
	 */
	private void fillQueue(Clustering clustering) {
		Graph graph = clustering.getGraph();
		int count = 0;
		edgeHeap = new PriorityQueue<myEdge>(graph.E()); // zu gross
		myEdge me;
		for (int i = 0; i < adj.length; ++i) {
			if (c[i] == i) { // valid[i], i.e. i is leader
				for (int j : adj[i].keySet()) {
					if (c[j] == j) { // (valid[j]) (also leader)
						if (i > j){
							me = new myEdge(i, j);
							edgeHeap.add(me);
							++count;
						}
					}
				}
			}
		}
//		System.out.println("Queue refresh: " + count + " inputs to queue");
	}

	/**
	 * Initializes and fills the crucial data structures. First the singleton clustering is assumed, 
	 * degSum is filled, c[i] = i, clusterContent[i] = (i) and adj[i] = j's adjacencies with weight.
	 * Then, if we update a clustering, all these are updated: for each cluster we do the following:
	 * we pick an arbitrary leader and set for all nodes in the cluster c[i] = leader, 
	 * the leader further steals the others degSums and clusterContents. 
	 * Finally we iterate through all edges of the graph and if they are inside a cluster, we delete them from adj
	 * if they are between leaders they stay, if they involve a non-leader they are stolen by the respective leaders.
	 * 
	 * @param clustering the given or read clustering, possibly used for updates
	 * @param gi used for edge weights
	 */
	private void fillClusterStuff(Clustering clustering, GraphInterpretation gi) {
		Graph graph = clustering.getGraph();
		int N = graph.N();
		adj = new HashMap[N];
		c = new int[N];
		clusterContent = new ArrayList[N];
		degSum = new Double[N];
		for (Node n : graph.getNodeArray()) {
			int ind = n.index();
			degSum[ind] = 0.0;
			HashMap<Integer, Double> thisAdj = new HashMap<Integer, Double>(n.degree());
			for (EdgeCursor ec = n.edges(); ec.ok(); ec.next()) {
				degSum[ind] += gi.getWeight(ec.edge());
				thisAdj.put(ec.edge().opposite(n).index(), gi.getWeight(ec.edge()));
			}
			adj[ind] = thisAdj;
			clusterContent[ind] = new ArrayList<Integer>();
			clusterContent[ind].add(ind);
			c[ind] = ind;
		}
		if (allowUpdate){
//			System.out.println("We have an input clustering of " + clustering.clusterCount() + " clusters, I will read them and build upon them.");
			//first degSum, c, and clusterContent...
			int leader, ind;
			for (GYCursor<Cluster> cc = clustering.clusters(); cc.ok(); cc.next()){
				leader = cc.active().aMember().index();	//just anybody is good as a leader
				for (NodeCursor nc = cc.active().members(); nc.ok(); nc.next()){ //all follow...
					ind = nc.node().index(); 
					if (ind != leader){ //...who are not the leader herself
						degSum[leader] += degSum[ind];
						degSum[ind] = 0.0;
						c[ind] = leader;
						clusterContent[ind].clear();
						clusterContent[leader].add(ind);
					}
				}
			}
			//...now adjacencies
			int x,y,cx,cy;
			double weight;
			for (EdgeCursor ec = graph.edges(); ec.ok(); ec.next()){
				x = ec.edge().source().index();
				y = ec.edge().target().index();
				cx = c[x];
				cy = c[y];
				if ((cx != cy) && ((x != cx) || (y != cy))){
					weight = gi.getAdjacency(x, y);
					if (!adj[cx].containsKey(cy)){
						adj[cx].put(cy, weight);
						adj[cy].put(cx, weight);
					}
					else{
						adj[cx].put(cy, weight + adj[cx].get(cy));
						adj[cy].put(cx, weight + adj[cy].get(cx));
					}
				}
				if (cx == cy){
					adj[cx].remove(cy);
					adj[cy].remove(cx);
				}	
			}
		}
		else{
//			System.out.println("We have no input clustering, I am starting from scratch.");
		}
	}

	/**
	 * Class for an edge of a graph used for putting it into a priority queue for modularity-based
	 * agglomeration. Indices of source and target are stored, and the method compareTo bases on 
	 * how much a contraction of this edge improves a clusterings modularity value
	 * 
	 * @author robert
	 *
	 */
	public class myEdge implements Comparable<myEdge> {
		public int s;
		public int t;

		public myEdge(int i, int j) {
			s = i;
			t = j;
		}

		public double getDq() {
//			System.out.println("getting dq of " + s + " (" + c[s] + ") -> " + t + " (" + c[t] + ")");
			if (c[s] == c[t]){
//				System.out.println("Self-loop, thus manually getDq = 0.0");
				return(0.0);
			}
//			if (!adj[c[s]].containsKey(c[t]))
//				System.out.println("warning in adj, cs has no ct: " + c[s] + " -> " + c[t]);
//			if (!adj[c[t]].containsKey(c[s]))
//				System.out.println("warning in adj, ct has no cs: " + c[t] + " -> " + c[s]);
			return ((adj[c[s]].get(c[t]) / weightSum) - (2.0 * degSum[c[s]] * degSum[c[t]]) / degsSq);
		}

		@Override
		public String toString() {
			return (s + " -> " + t + "(" + getDq() + ")");
		}

		public int compareTo(myEdge other) {
			double dq1 = getDq();
			double dq2 = other.getDq();
			if (dq1 > dq2)
				return -1;
			else if (dq1 < dq2)
				return 1;
			else
				return 0;
		}
	}
}
