/*
 * Created on Aug 17, 2005
 *
 * Copyright 2004.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */

package de.uka.algo.clustering.algorithms;

import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeList;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.ClusteringListener;
import de.uka.algo.clustering.index.IndexFactory;
import de.uka.algo.math.RandomArray;

/**
 * optimises a clustering locally by moving single nodes from one cluster to
 * another, isolating single nodes or merging two clusters
 * 
 * @author Daniel Delling
 */

public class ElementalOperationOptimizer implements Algorithm {

	private IndexFactory quality;

	private int changedNodes = 0;

	private double currentIndexValue = 0.0;

	protected int maximumMovements = Integer.MAX_VALUE;

	private boolean allowShift = false;

	private boolean allowMerge = false;

	private boolean allowSplit = false;
	
	protected double startOptValue = -10000000.0; 
	
	/**
	 * initiates the LocalOptimizer.
	 * 
	 * @param quality
	 *            the IndexFactory which should be optimized
	 * @param shift true, iff shifting of nodes from one cluster to another is allowed
	 * @param merge true, iff merging of two clusters is allowed
	 * @param split true, iff splitting a single node is allowed.
	 */
	public ElementalOperationOptimizer(IndexFactory quality, boolean shift, boolean merge,
			boolean split) {
		this.quality = quality;
		this.allowMerge = merge;
		this.allowShift = shift;
		this.allowSplit = split;
	}

	/**
	 * initiates the LocalOptimizer.
	 * 
	 * @param quality
	 *            the IndexFactory which should be optimized
	 */
	public ElementalOperationOptimizer(IndexFactory quality) {
		this(quality, true, true, true);
	}

	public void run(Clustering aClustering) {
		boolean policy = aClustering.getPolicy();
		aClustering.setPolicy(ClusteringListener.UNRESTRICTED);
		boolean changed = true;
		changedNodes = 0;
		currentIndexValue = quality.getIndex(aClustering).getValue();
		Graph g = aClustering.getGraph();

		Node[] graphNodes = g.getNodeArray();

		RandomArray ran = new RandomArray();

		int n = g.N();
		while (changed && changedNodes < maximumMovements) {
			changed = false;

			// randomized order
			int[] randomOrder = ran.nextArray(n);
			int j = 0;
			// iterate all nodes in randomized order and try to optimize them
			while (j < n && changedNodes < maximumMovements) {
				if (optimizeNode(aClustering, graphNodes[randomOrder[j]])) {
					changedNodes++;
					changed = true;
				}
				j++;
			}

			afterIteration();
			// System.out.println("weitere Runde");
		}
		aClustering.setPolicy(policy);
	}

	private boolean optimizeNode(Clustering aClustering, Node node) {
		boolean changed = false;
		double tempIndexValue = startOptValue;
		double tempMaxIndexValue = startOptValue;

		int operationType = 0; // 0:move 1:merge 2:split

		Cluster from = aClustering.getCluster(node);
		Cluster to = null;
		Node aNeighbor;
		NodeList fromNodes = new NodeList(from.members());
		Cluster toMerge = null;

		// try merging
		if (allowMerge) {
			for (NodeCursor nc = node.neighbors(); nc.ok(); nc.next()) {
				from = aClustering.getCluster(node);
				aNeighbor = nc.node();
				Cluster aNeighborCluster = aClustering.getCluster(aNeighbor);
				// only try to merge when not in same cluster
				if (!aClustering.belongToSameCluster(aNeighbor, node)
						&& aNeighborCluster != toMerge) {
					Cluster mergedCluster = aClustering.merge(from,
							aNeighborCluster);
					tempIndexValue = quality.getIndex(aClustering).getValue();
					from = aClustering.split(mergedCluster, fromNodes.nodes());
					if (better(tempIndexValue, tempMaxIndexValue)) {
						tempMaxIndexValue = tempIndexValue;
						toMerge = mergedCluster;
						operationType = 1;
					}
				}
			}
		}

		// check splitting and movements only for clustersize bigger than 1
		if (from.size() > 1) {
			// check movements
			if (allowShift) {
				for (NodeCursor nc = node.neighbors(); nc.ok(); nc.next()) {
					aNeighbor = nc.node();
					// only try to move node when not in same cluster
					if (!aClustering.belongToSameCluster(aNeighbor, node)) {
						Cluster aNeighborCluster = aClustering
								.getCluster(aNeighbor);
						aClustering.move(from, aNeighborCluster, node);
						tempIndexValue = quality.getIndex(aClustering)
								.getValue();
						if (better(tempIndexValue, tempMaxIndexValue)) {
							tempMaxIndexValue = tempIndexValue;
							to = aNeighborCluster;
							operationType = 0;
						}
						aClustering.move(aNeighborCluster, from, node);
					}
				}
			}

			// check splitting
			if (allowSplit) {
				aClustering.split(from, new NodeList(node).nodes());
				tempIndexValue = quality.getIndex(aClustering).getValue();
				if (better(tempIndexValue, tempMaxIndexValue)) {
					tempMaxIndexValue = tempIndexValue;
					operationType = 2;
				}
				aClustering.merge(from, aClustering.getCluster(node));
				from = aClustering.getCluster(node);
			}
		} else {
			//merging in the sense of shifting, if cluster of n is of size 1
			if(!allowMerge && allowShift){
				// neigbors
				for (NodeCursor nc = node.neighbors(); nc.ok(); nc.next()) {
					aNeighbor = nc.node();
					// only try to move node when not in same cluster
					Cluster aNeighborCluster = aClustering.getCluster(aNeighbor);
					aClustering.merge(from, aNeighborCluster);
					tempIndexValue = quality.getIndex(aClustering).getValue();
					if (better(tempIndexValue, tempMaxIndexValue)) {
						tempMaxIndexValue = tempIndexValue;
						toMerge = aNeighborCluster;
						operationType = 1;
					}
					aClustering.split(aNeighborCluster, new NodeList(node).nodes());
					from = aClustering.getCluster(node);
				}	
			}			
		}

		if (accept(tempMaxIndexValue, currentIndexValue)) {
			// System.out.print("Was besseres: "+tempMaxIndexValue+" > " +
			// + currentIndexValue
			// + "("+(tempMaxIndexValue-currentIndexValue)+")");
			if (operationType == 0) { // movement
			// System.out.println(" by moving!");
				aClustering.move(from, to, node);
			} else if (operationType == 1) { // merging
			// System.out.println(" by merging! fromSize "+from.size()+ ",
			// toSize " + toMerge.size());
				aClustering.merge(from, toMerge);
			} else { // splitting
			// System.out.println(" by splitting!");
				aClustering.split(from, new NodeList(node).nodes());
			}
			changed = true;
			currentIndexValue = quality.getIndex(aClustering).getValue();
		}

		return changed;
	}

	/**
	 * determines, if a movement is better than another possible pne
	 * 
	 * @param a
	 *            the quality-value if doing the movement
	 * @param b
	 *            the quality-value if doing the best byfar movement
	 * @return true: movement is better
	 */
	boolean better(double a, double b) {
		return a - b > Math.pow(10, -15);
	}

	/**
	 * determines, when a movement should be accepted
	 * 
	 * @param a
	 *            the quality-value if doing the movement
	 * @param b
	 *            the quality-value if not doing the movement
	 * @return true: movement is done, false: movement is not done
	 */
	boolean accept(double a, double b) {
		return a - b > Math.pow(10, -15);
	}

	/**
	 * dummy method, may be used by subclasses executed after each iteration
	 * 
	 */
	void afterIteration() {
		// do nothing
	}

	public int getChangedNodes() {
		return changedNodes;
	}

}
