/*
 * Created on Nov 1, 2004
 * Reviewed 1.6 (Jan 9, 2005)
 *
 * Copyright 2004.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering.algorithms;

import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.ClusteringReference;
import de.uka.algo.clustering.algorithms.SpectralSingleLinkage;
import de.uka.algo.clustering.index.Index;
import de.uka.algo.clustering.index.IndexFactory;

/**
 * The GMC clustering algorithm resets a clustering using an algorithm based on
 * spectral embedding and single linkage. It performs the following steps: <br>
 * 0. Reset the clustering <br>
 * 1. For all connected components of the clustering's graph do steps 2-7 <br>
 * 2. Calculate the first n nontrivial eigenvectors of the corresponding
 * normalised adjacency matrix (Note: Instead of the normalised adjacency matrix
 * M, 0.1*I+0-9*M is used to ensure convergence of the power iteration
 * algorithm) <br>
 * 3. Embed all nodes according to their entries in those eigenvectors <br>
 * 4. Sort edges by length (according to the embedding) in ascending order <br>
 * 5. Hide all edges <br>
 * 6a. Unhide edges in order until the clustering contains sqrt(|V|) clusters.<br>
 * 6b. Gradually unhide edges (in order), compute clustering induced by the
 * connected components, compute a quality index <br>
 * 7. Return the clustering which minimizes the quality index .<br>
 * 8. Combine the connected component's clusterings <br>
 * 
 * 
 * @author Bastian Katz
 */
public class GMC implements Algorithm {

	/** internal algorithm structure */
	private AlgorithmList core;

	/**
	 * Instantiates a new GMC clustering algorithm with an embedding dimension
	 * and an index (as factory).
	 * 
	 * @param embeddingDimension
	 *            the embedding dimension, values from one to four usually lead
	 *            to best results
	 * @param qualityIndex
	 *            the index to minimize given by its factory
	 */
	public GMC(int embeddingDimension, IndexFactory qualityIndex) {

		IndexFactory quality = new ClusterCountIndex.Factory(qualityIndex);

		/* set up the refining algorithm */
		AlgorithmList refiningAlgorithm = new AlgorithmList();
		/* single linkage algorithms do only merge clusters */
		refiningAlgorithm.addAlgorithm(new Singletons());
		refiningAlgorithm.addAlgorithm(new SpectralSingleLinkage(
				embeddingDimension, quality));

		/* combine component clustering and refining algorithm */
		core = new AlgorithmList();
		core.addAlgorithm(new Components());
		core.addAlgorithm(new RefiningWrapper(refiningAlgorithm, 1));
	}

	/**
	 * Resets the clustering as described above. Postcondition: All nodes are
	 * assigned to a cluster.
	 * 
	 * @param aClustering
	 *            the clustering to reset
	 */
	public void run(Clustering aClustering) {
		core.run(aClustering);
	}

	public static class ClusterCountIndex extends ClusteringReference implements
			Index {

		private Index index = null;

		private IndexFactory factory;

		protected ClusterCountIndex(Clustering clustering, IndexFactory factory) {
			super(clustering);
			this.factory = factory;
		}

		public double getValue() {
			if (clustering.clusterCount() > Math.sqrt(clustering.getGraph()
					.nodeCount()))
				return 0;
			if (index == null)
				this.index = factory.getIndex(clustering);

			return 0.01+0.99*index.getValue();
		}

		public double getUnweightedValue() {
			if (clustering.clusterCount() > Math.sqrt(clustering.getGraph()
					.nodeCount()))
				return 0;
			if (index == null)
				this.index = factory.getIndex(clustering);

			return 0.01+0.99*index.getUnweightedValue();
		}

		public static IndexFactory factory(IndexFactory factory) {
			return new Factory(factory);
		}

		public static class Factory implements IndexFactory {

			private IndexFactory factory;

			public Factory(IndexFactory factory) {
				this.factory = factory;
			}

			public Index getIndex(Clustering clustering) {
				return new ClusterCountIndex(clustering, factory);
			}
		}
	}
}