/*
 * Created on Nov 1, 2004
 * Reviewed 1.3 (Jan 9, 2005)
 *
 * Copyright 2004.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering.algorithms;

import de.uka.algo.clustering.Clustering;
import de.uka.algo.graphs.cut.ConductanceCutter;

/**
 * The ICC clustering algorithm resets the clustering to one large cluster and
 * cuts then clusters iteratively by a minimum conductance cut if intra
 * conductance is below a threshold. For the first cut, this threshold is 1,
 * i.e. the algorithm returns at least two clusters if the graph's conductance
 * is not 1.
 * 
 * @author Bastian Katz
 */
public class ICC implements Algorithm {

	/** internal representation of the algorithm */
	private AlgorithmList algorithm;

	/**
	 * Instantiates a new ICC clustering algorithm with a certain intra cluster
	 * conductance threshold and parameters for a heuristic to find minimum
	 * conductance cuts.
	 * 
	 * @param threshold
	 *            the threshold for conductance cuts
	 * @param strategy
	 *            defines the matrix used to find minimum conductance cuts
	 *            (defined in ConductanceCutter).
	 * @param cutComplex
	 *            defines whether to find cuts using a more complex heuristic
	 *            checking more potential minimum conductance cuts.
	 * @see ConductanceCutter
	 * @see ConductanceCutter#MARKOVIAN
	 * @see ConductanceCutter#LAPLACIAN
	 */
	public ICC(double threshold, ConductanceCutter.ConductanceStrategies strategy, boolean cutComplex) {
		algorithm = new AlgorithmList();

		/* the cutting strategy */
		ConductanceCutter cc = new ConductanceCutter(strategy, threshold,
				cutComplex);

		/* the cutting strategy which is forced to cut */
		ConductanceCutter forced = new ConductanceCutter(strategy, 1,
				cutComplex);

		/* the algorithm's structure */
		algorithm.addAlgorithm(new GraphCutAlgorithm(forced));
		algorithm.addAlgorithm(new RefiningWrapper(new GraphCutAlgorithm(cc),
				Integer.MAX_VALUE));
	}

	/**
	 * Instantiates a new ICC clustering algorithm with a certain intra cluster
	 * conductance (using a heuristic).
	 * 
	 * @param threshold
	 *            the threshold for conductance cuts
	 */
	public ICC(double threshold) {
		this(threshold, ConductanceCutter.ConductanceStrategies.Markovian, false);
	}

	/**
	 * Resets to a large cluster and splits it iteratively using minimum
	 * conductance cuts. Postcondition: All nodes are assigned to clusters.
	 * 
	 * @param aClustering
	 *            the clustering to reset
	 */
	public void run(Clustering aClustering) {
		algorithm.run(aClustering);
	}

}