package de.uka.algo.clustering.algorithms;


import java.util.Iterator;
import java.util.Stack;

import y.base.Graph;
import y.base.Node;
import y.base.NodeList;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.graphs.linalg.GraphMatrix;
import de.uka.algo.graphs.linalg.MarkovianGraphMatrix;
import de.uka.algo.math.linalg.generic.Cell1D;
import de.uka.algo.math.linalg.generic.Cell2D;
import de.uka.algo.math.linalg.generic.Vector;
import de.uka.algo.math.linalg.matrices.VectorMatrix;
import de.uka.algo.math.linalg.vectors.ArrayVector;
import de.uka.algo.math.linalg.vectors.HashVector;

/**
 * The MCL algorithm is a clustering algorithm that resets a clustering using an
 * approach based on random walks. It performs the following steps: <br>
 * 0. Delete all clusters <br>
 * 1. A:=(0.01*I+0.91*M) (M is the normalised adjacency matrix) <br>
 * 2. Repeat steps 3-6 until a fixed point is reached <br>
 * 3. A:=A^x <br>
 * 4. Optionally prune entries (see below) <br>
 * 5. Emphasise entries (a_ij:=a_ij^y) <br>
 * 6. Normalise A <br>
 * 7. Build clusters for all connected components induced by A <br>
 * <p>
 * This implementation of the MCL algorithm supports four modes to prune
 * entries. They all store a maximum degree <i>d </i> for each node and keep
 * only the <i>d </i> highest entries in each row of A. The supported modes are:
 * <br>
 * <code>CLASSIC</code>: the maximum degree is the number of nodes, i.e. no
 * entries are omitted <br>
 * <code>DEGREEBOUND</code>: the maximum degree is a fixed multiple of the
 * node's degree (In the graph) <br>
 * <code>MAXDEGREEBOUND</code>: the maximum degree is a fixed multiple of the
 * maximum node degree in the graph, i.e. the same for all nodes <br>
 * <code>CONSTDEGREE</code>: the maximum degree is a constant, i.e. the same
 * for all nodes
 * 
 * @author Bastian Katz
 */
public class MCL implements Algorithm {

    /** a mode qualifier (see above), ignores strategy parameter */
    final static public int CLASSIC = 0;
    /** a mode qualifier (see above), strategy parameter = degree multiplyer */
    final static public int DEGREEBOUND = 1;
    /** a mode qualifier (see above), strategy parameter = max. degree multiplier */
    final static public int MAXDEGREEBOUND = 2;
    /**
     * a mode qualifier (see above), strategy parameter = maximum number of
     * entries
     */
    final static public int CONSTBOUND = 3;

    /** the internal representation of the algorithm */
    private AlgorithmList algorithm;

    /**
     * Instantiates a new MCL algorithm with the parameters decribed above.
     * 
     * @param matrixPower
     *            the power to set the matrix to in step 3
     * @param entryPower
     *            the power to set the entries to in step 5
     * @param strategy
     *            the strategy to prune entries
     * @param strategyParameter
     *            the additional parameter
     */
    public MCL(int matrixPower, double entryPower, int strategy,
            double strategyParameter) {

        /* set up the internal algorithm's representation */
        algorithm = new AlgorithmList();
        algorithm.addAlgorithm(new Components());
        algorithm.addAlgorithm(new RefiningWrapper(new MCLCore(matrixPower,
                entryPower, strategy, strategyParameter), 1));
    }

    /**
     * Resets the clustering as decribed above. Postcondition: All nodes are
     * assigned to clusters.
     * 
     * @param aClustering
     *            the clustering to reset
     */
    public void run(Clustering aClustering) {
        algorithm.run(aClustering);
    }

    //TODO(Bastian) code review, documentation for inner class MCLCore
    private class MCLCore implements Algorithm {

        private static final double LOOP_WEIGHT = 0.01;
		private int matrixPower;
        private double entryPower;
        private final double zero = 1E-10;
        private double strategyParameter;
        private int strategy;

        private MCLCore(int matrixPower, double entryPower, int strategy,
                double strategyParameter) {
            this.matrixPower = matrixPower;
            this.entryPower = entryPower;
            this.strategyParameter = strategyParameter;
            this.strategy = strategy;
        }

        public void run(Clustering aClustering) {
            aClustering.reset();
            Graph graph = aClustering.getGraph();

            Node[] nodesAsArray = graph.getNodeArray();
            boolean[] nodesDone = new boolean[nodesAsArray.length];

            GraphMatrix normalizedAdjacency = new MarkovianGraphMatrix(graph,
                    aClustering.getInterpretation(), LOOP_WEIGHT);
            VectorMatrix fastMatrixAdapter = null;

            int[] degreeBoundary = new int[graph.nodeCount()];
            switch (strategy) {
            case MCL.CONSTBOUND:
                fastMatrixAdapter = new HashVector.Matrix(normalizedAdjacency);
                for (int index = 0; index < graph.nodeCount(); index++) {
                    degreeBoundary[index] = (int) Math.round(strategyParameter);
                }
                break;
            case MCL.DEGREEBOUND:
                fastMatrixAdapter = new HashVector.Matrix(normalizedAdjacency);
                for (int index = 0; index < graph.nodeCount(); index++) {
                    int degree = nodesAsArray[index].degree();
                    degreeBoundary[index] = (int) Math.round(degree
                            * strategyParameter);
                }
                break;
            case MCL.MAXDEGREEBOUND:
                fastMatrixAdapter = new HashVector.Matrix(normalizedAdjacency);
                int maxFoundDeg = 0;
                for (int index = 0; index < graph.nodeCount(); index++) {
                    int degree = nodesAsArray[index].degree();
                    maxFoundDeg = Math.max(degree, maxFoundDeg);
                }
                for (int index = 0; index < graph.nodeCount(); index++) {
                    degreeBoundary[index] = (int) Math.round(maxFoundDeg
                            * strategyParameter);
                }
                break;
            default:
                fastMatrixAdapter = new ArrayVector.Matrix(normalizedAdjacency);
                for (int index = 0; index < graph.nodeCount(); index++) {
                    degreeBoundary[index] = graph.nodeCount();
                }
                break;
            }

            VectorMatrix recallLastIteration = (VectorMatrix) fastMatrixAdapter.getLike();
            do {
                recallLastIteration.assign(fastMatrixAdapter);
                fastMatrixAdapter.exp(matrixPower);
                fastMatrixAdapter.deg(degreeBoundary);
                fastMatrixAdapter.emphFixNormalize(entryPower, zero);
            } while (!eq(recallLastIteration, fastMatrixAdapter, zero));
            VectorMatrix transposedAccess = (VectorMatrix) fastMatrixAdapter.getLike();
            transposedAccess.assignTransposed(fastMatrixAdapter);

            NodeList componentNodes = new NodeList();
            Stack<Integer> nodesToTraverse = new Stack<Integer>();

            for (int i = 0; i < nodesAsArray.length; i++) {
                if (!nodesDone[i]) {
                    nodesToTraverse.push(i);
                    componentNodes.push(nodesAsArray[i]);
                    nodesDone[i] = true;
                    while (!nodesToTraverse.isEmpty()) {
                        int index = nodesToTraverse.pop()
                                .intValue();
                        Vector row = fastMatrixAdapter.getRow(index);
                        for (Iterator<Cell1D> rowCells = row.nonZeros(); rowCells
                                .hasNext();) {
                            Cell1D aCell = rowCells.next();
                            int cellIndex = aCell.getIndex();
							if (aCell.value != 0 && !nodesDone[cellIndex]) {
                                nodesToTraverse.push(new Integer(cellIndex));
                                componentNodes.push(nodesAsArray[cellIndex]);
                                nodesDone[cellIndex] = true;
                            }
                        }
                        row = transposedAccess.getRow(index);
                        for (Iterator<Cell1D> rowCells = row.nonZeros(); rowCells
                                .hasNext();) {
                            Cell1D aCell = rowCells.next();
							int cellIndex = aCell.getIndex();
                            if (aCell.value != 0 && !nodesDone[cellIndex]) {
                                nodesToTraverse.push(new Integer(cellIndex));
                                componentNodes.push(nodesAsArray[cellIndex]);
                                nodesDone[cellIndex] = true;
                            }
                        }
                    }
                    aClustering.newCluster(componentNodes.nodes());
                    componentNodes.clear();
                }
            }
        }

        private boolean eq(VectorMatrix compareTo, VectorMatrix m, double useAsZero) {
            for (Iterator<Cell2D> i = compareTo.nonZeros(); i.hasNext();) {
                Cell2D c = i.next();
                double other = m.get(c.row, c.column);
                if (c.value - other > useAsZero)
                    return false;
                if (other - c.value > useAsZero)
                    return false;
            }
            for (Iterator<Cell2D> i = m.nonZeros(); i.hasNext();) {
                Cell2D c = i.next();
                double other = compareTo.get(c.row, c.column);
                if (c.value - other > useAsZero)
                    return false;
                if (other - c.value > useAsZero)
                    return false;
            }
            return true;
        }
    }
}