/*
 * Created on Feb 18, 2008
 *
 * Copyright 2008. --- All rights resevered. 
 * Algorithmic Group, Karlsruhe, Germany.
 * 
 * project: betweennessClustering
 * package: de.uka.algo.clustering.algorithms.betweenness
 */
package de.uka.algo.clustering.algorithms.betweenness;

import y.base.EdgeCursor;
import y.base.EdgeList;
import y.base.Graph;
import y.base.NodeCursor;
import y.base.NodeList;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.algorithms.Algorithm;
import de.uka.algo.clustering.algorithms.Components;
import de.uka.algo.clustering.algorithms.betweenness.edgechoosing.EdgeChooser;
import de.uka.algo.clustering.index.IndexFactory;
import de.uka.algo.graphs.centralities.betweenness.BetweennessBFS;
import de.uka.algo.graphs.centralities.betweenness.sampling.ExactSampler;
import de.uka.algo.graphs.centralities.betweenness.sampling.Sampler;
import de.uka.algo.graphs.traverse.bfs.BFS;

/**
 * Clustering a graph using the betweenness centrality:
 * Edges are removed from the graph depending on their
 * betweenness centrality. The resulting compontent clustering
 * is evaluated. When there are no more edges left in the
 * graph, the best clustering will be returned.
 * 
 * @author Ulrich Geilmann
 */

public class BetweennessClusterer implements Algorithm {

	/* **** members **** */
	// an index as quality measure 
	IndexFactory qualityMeasure; 
	// a strategy to select edges 
	EdgeChooser selectEdges; 
	// a strategy for (node-)sampling 
	Sampler selectNodes; 
	// cluster a graph using its connected components
	Components componentClusterer;
	// edges that were hidden during clustering
	EdgeList hiddenEdges;

	/**
	 * The graph will be clustered using the exact betweenness calculation
	 * @param fac
	 * 	quality measure for the created clusterings
	 * @param chooser
	 * 	strategy for choosing the edges to remove
	 */
	public BetweennessClusterer(IndexFactory fac, EdgeChooser chooser) {
		this(fac, chooser, null);
	}
	
	/**
	 * The graph will be clustered using an approximation of the betweenness centrality
	 * @param fac
	 * 	quality measure for the created clusterings
	 * @param chooser
	 * 	strategy for choosing the edges to remove
	 * @param sampler
	 * 	(node-)sampling strategy for approximating the betweenness-calculation
	 */
	public BetweennessClusterer(IndexFactory fac, EdgeChooser chooser, Sampler sampler) {
		qualityMeasure = fac;
		selectEdges = chooser;
		selectNodes = sampler;
		componentClusterer = new Components();
		hiddenEdges = new EdgeList();
		if (selectNodes == null)
			selectNodes = new ExactSampler();
	}
	
	/* (non-Javadoc)
	 * @see de.uka.algo.clustering.algorithms.Algorithm#run(de.uka.algo.clustering.Clustering)
	 */
	public void run(Clustering arg0) {
		Graph graph = arg0.getGraph();
	
		BFS bfs = new BFS(graph);
		BetweennessBFS betweenness = new BetweennessBFS(false);
		bfs.addModule(betweenness);
		
		double bestIndex = Double.NEGATIVE_INFINITY;
		Clustering currentClustering = new Clustering(graph);
		
		while (graph.edgeCount() > 0) {

			/* create clustering */
			componentClusterer.run(currentClustering);

			/* evaluate clustering, store only the best one */
			for (EdgeCursor c = hiddenEdges.edges(); c.ok(); c.next())
				graph.unhide(c.edge());
			double index = qualityMeasure.getIndex(currentClustering).getValue();
			for (EdgeCursor c = hiddenEdges.edges(); c.ok(); c.next())
				graph.hide(c.edge());

			if (index > bestIndex) {
				arg0.setToClustering(currentClustering);
				bestIndex = index;
			}
			
			bfs.initAll();
			/* run the search to calculate the bc */
			NodeList samples = selectNodes.sample(graph);
			for (NodeCursor nc = samples.nodes(); nc.ok(); nc.next()) {
				bfs.run(nc.node());
			}

			/* hide edges */
			EdgeList hide = selectEdges.chooseEdge (graph, betweenness.getEdgeCentrality(), currentClustering);
			hiddenEdges.addAll(hide);
		}
		for (EdgeCursor c = hiddenEdges.edges(); c.ok(); c.next())
			graph.unhide(c.edge());
		
		currentClustering.dispose();
		hiddenEdges.clear();
		bfs.cleanUp();
	}

}
