/*
 * Created on Feb 5, 2009 by Robert Goerke
 *
 * Copyright 2009.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering.algorithms.blondel;

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Random;
import java.util.Map.Entry;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.EdgeMap;
import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeList;
import y.layout.orthogonal.f.e;
import y.view.Graph2D;
import yext.graphml.graph2D.GraphMLIOHandler;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.ClusterMap;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.algorithms.Algorithm;
import de.uka.algo.clustering.index.Modularity;
import de.uka.algo.graphs.GraphInterpretation;
import de.uka.algo.graphs.WeightedGraph;
import de.uka.algo.util.GYCursor;

/**
 * Starting from singletons, nodes are iteratively asked, to which of their neighbors to associate, with the one yielding the
 * highest increase in modularity winning. This is repeated until no node wants to change his cluster. Then clusters are
 * contracted to super nodes and the process starts over.
 * 
 * @author Robert Goerke
 */
public class Blondel implements Algorithm {

	public enum NodeOrderStrategies {
		array, invArray, random
	}

	public enum StoppingCriteria {
		firstPeak
	}

	public enum UpdatePolicy {
		affected, neighbors
	}

	/** the chosen strategy, i.e., in which order are nodes asked */
	private NodeOrderStrategies nodeOrderStrategy = NodeOrderStrategies.random;
	/** the chosen criterion for when to stop agglomeration */
	private StoppingCriteria stoppingCriterionCriterion = StoppingCriteria.firstPeak;
	/** the update policy, i.e., what is a bad node? */
	private UpdatePolicy updatePolicy = UpdatePolicy.affected;

	private boolean isWeighted;
	private double bestInnerIncrease;
	private Graph abstractedGraph;
	private BlondelBaseListener theBaseListener;
	private Clustering outputClustering;
	private int neighborHandShakes = 0;

	double incrementalModularity;
	HashMap<Integer, HierarchyLevel> levelToHierarchyLevel = new HashMap<Integer, HierarchyLevel>();
	HashSet<Node> invalidNodes = new HashSet<Node>(4);

	public class HierarchyLevel {
		public HierarchyLevel(Graph aGraph, Clustering aClustering) {
			this.theGraph = aGraph;
			this.theClustering = aClustering;
			this.clusterToNextLevelNode = new ClusterMap<Node>(this.theClustering);
			levelToHierarchyLevel.put(levelToHierarchyLevel.size(), this);
		}

		public HierarchyLevel(HierarchyLevel oldLevel) {
			this.theClustering = contractClustering(oldLevel);
			this.theGraph = this.theClustering.getGraph();
			this.clusterToNextLevelNode = new ClusterMap<Node>(this.theClustering);
			levelToHierarchyLevel.put(levelToHierarchyLevel.size(), this);
		}

		public Graph theGraph;
		public Clustering theClustering;
		public ClusterMap<Node> clusterToNextLevelNode;
		public ClusterChangeListener oracle;
	}

	public Blondel(NodeOrderStrategies calledNodeOrderStrategy, StoppingCriteria calledStoppingCriterion,
			UpdatePolicy calledUpdatePolicy) {
		nodeOrderStrategy = calledNodeOrderStrategy;
		stoppingCriterionCriterion = calledStoppingCriterion;
		updatePolicy = calledUpdatePolicy;
	}
	
	public Blondel() {
		this(NodeOrderStrategies.random, StoppingCriteria.firstPeak, UpdatePolicy.affected);
	}

	/**
	 * calculates for a node, which of the clusters of all its neighbors is best to move to, in terms of modularity. Returns
	 * null if becoming a singleton is best. Returns himself if no improvement can be found -> old cluster
	 * 
	 * @param searchingNode
	 *            the node that is examining its neighborhood for better clusters
	 * @param newNeighbor
	 *            the neighbor which is in the cluster that is best for searchingNode to move to
	 * @return
	 */
	public Node bestNeighbor(Node searchingNode, Clustering theClustering) {
		double currentIncrease = 0.0;
		GraphInterpretation gi = theClustering.getInterpretation();
		Node currentNeighbor = null;
		Node newNeighbor = searchingNode;
		Cluster oldCluster = theClustering.getCluster(searchingNode), newCluster;
		if (oldCluster == null) {
			System.out.println("oldCluster Kaputt, is null");
			oldCluster = theClustering.newCluster(searchingNode);	//darf gar nicht soweit kommen... versuch...
		}
		double covMinus, covPlus, expPlus, expMinus;
		double W = gi.getEdgeWeightSum();
		double selfLoopWeight = gi.getAdjacency(searchingNode, searchingNode);
		double nonSimpleWeight = gi.getDegree(searchingNode) + selfLoopWeight;
		for (NodeCursor nc = searchingNode.neighbors(); nc.ok(); nc.next()) {
			++neighborHandShakes;
			currentNeighbor = nc.node();
			newCluster = theClustering.getCluster(currentNeighbor);
			if (oldCluster.getClustering() == null) {
				System.out.println("oldClustering Kaputt");
			}
			if (newCluster.getClustering() == null) {
				System.out.println("newCluster Kaputt");
			}
			if (newCluster.equals(oldCluster))
				continue;
			// self loops automatically in this term
			covMinus = adjacencyToCluster(searchingNode, oldCluster) / W;
			// self loops will contribute to future
			covPlus = (adjacencyToCluster(searchingNode, newCluster) + selfLoopWeight) / W;
			expPlus = (nonSimpleWeight * (theClustering.getBorderWeight(oldCluster) + 2
					* theClustering.getInnerWeight(oldCluster) - nonSimpleWeight))
					/ (2.0 * W * W);
			expMinus = (nonSimpleWeight * (theClustering.getBorderWeight(newCluster) + 2 * theClustering
					.getInnerWeight(newCluster)))
					/ (2.0 * W * W);
			currentIncrease = covPlus - covMinus + expPlus - expMinus;
			if (currentIncrease > bestInnerIncrease) {
				bestInnerIncrease = currentIncrease;
				newNeighbor = currentNeighbor;
			}
		}
		// isolated node -> singleton!
		if (gi.getDegree(searchingNode) == 0) {
			newNeighbor = null;
			// System.out.print("isolated node! -> singleton, ");
		} else {
			// check goodness of becoming a simple singleton
			covMinus = adjacencyToCluster(searchingNode, oldCluster) / W;
			covPlus = selfLoopWeight / W;
			expPlus = (nonSimpleWeight * (theClustering.getBorderWeight(oldCluster) + 2
					* theClustering.getInnerWeight(oldCluster) - nonSimpleWeight))
					/ (2.0 * W * W);
			expMinus = 0.0;
			currentIncrease = covPlus - covMinus + expPlus - expMinus;
			if (currentIncrease > bestInnerIncrease) {
				bestInnerIncrease = currentIncrease;
				newNeighbor = null;
			}
		}
		return (newNeighbor);
	}

	/**
	 * Returns the sum of edge weights between aNode and the nodes of aCluster
	 * 
	 * @param aNode
	 * @param aCluster
	 * @return sum of edge weights between aNode and aCluster
	 */
	/**
	 * @param aNode
	 * @param aCluster
	 * @return
	 */
	public double adjacencyToCluster(Node aNode, Cluster aCluster) {
		double weight = 0.0;
		Clustering aClustering = aCluster.getClustering();
		GraphInterpretation gi = aClustering.getInterpretation();
		// through all in and then all out edges...
		for (EdgeCursor ec = aNode.edges(); ec.ok(); ec.next()) {
			Edge theEdge = ec.edge();
			Node neighbor = theEdge.opposite(aNode);
			if (aCluster.equals(aClustering.getCluster(neighbor))) {
				weight += gi.getWeight(theEdge);
				// dont count self loops twice, only the egde weight counts
			}
		}
		// but yfiles puts a node twice in its own set of neighbors, in case of
		// a selfloop, thus subtract
		if (aCluster.equals(aClustering.getCluster(aNode)))
			weight -= gi.getAdjacency(aNode, aNode);
		return weight;
	}

	/**
	 * @param theGraph
	 *            the graph the nodes in the array are part of
	 * @param nodeArray
	 *            the nodes to be ordered
	 * @return an integer array which is a permutation of the 0...n-1 numbers
	 */
	public int[] orderNodes(Graph theGraph, Node[] nodeArray) {
		int n = nodeArray.length;
		int[] order = new int[n];
		switch (nodeOrderStrategy) {
		case array:
			for (int i = 0; i < n; ++i) {
				order[i] = i;
			}
			break;
		case invArray:
			for (int i = 0; i < n; ++i) {
				order[i] = n - 1 - i;
			}
			break;
		case random:
			for (int i = 0; i < n; ++i) {
				order[i] = i;
			}
			order = permuteArrray(order);
			break;
		default:
			System.out.println("Error, invalid node order strategy!");
			break;
		}

		return (order);
	}

	private int[] permuteArrray(int[] order) {
		Random myRandom = new Random();
		int swapID;
		int tmp;
		for (int i = 0; i < order.length; ++i) {
			swapID = myRandom.nextInt(order.length);
			tmp = order[i];
			order[i] = order[swapID];
			order[swapID] = tmp;
		}
		return order;
	}

	/**
	 * 
	 */
	public void run(Clustering theClustering) {
		outputClustering = theClustering;
		Graph inputGraph = theClustering.getGraph();
		GraphInterpretation gi = theClustering.getInterpretation();
		theBaseListener = new BlondelBaseListener(theClustering);
		theClustering.addListener(theBaseListener);

		incrementalModularity = 0.0;
		for (NodeCursor nc = inputGraph.nodes(); nc.ok(); nc.next()) {
			incrementalModularity -= ((gi.getDegree(nc.node()) + gi.getAdjacency(nc.node(), nc.node())) * (gi.getDegree(nc
					.node()) + gi.getAdjacency(nc.node(), nc.node())));
		}
		incrementalModularity /= (4 * gi.getEdgeWeightSum() * gi.getEdgeWeightSum());

		double currentBestImprovement = 1.0;
		int oldClusterCount = inputGraph.N() + 1;
		int hierarchyLevel = 0;
		int it;
		Clustering currentClustering = new Clustering(inputGraph, gi);
		currentClustering.copyClusterAssignmentFrom(theClustering);
		currentClustering.resetSingletons();
		HierarchyLevel currentLevel = new HierarchyLevel(inputGraph, currentClustering);
		while (currentClustering.size() < oldClusterCount) {
			System.out.println("Level:" + hierarchyLevel++ + ", n:" + currentClustering.getGraph().N() + ", m:"
					+ currentClustering.getGraph().E() + ", mod:" + incrementalModularity + ". ");
			oldClusterCount = currentClustering.size();
			currentBestImprovement = 1.0;
			it = 1;

			// while mod increase, migrate and merge nodes
			Node[] badNodes = currentClustering.getGraph().getNodeArray();
			while (currentBestImprovement > 0.00000001) {
				currentBestImprovement = bestNeighborsForEveryOne(currentLevel, badNodes);
				System.out.print("It:" + it++ + ", best increase:" + currentBestImprovement + " ... ");
			}
			System.out.print("no more improvement, level done (" + currentClustering.size() + " clusters). ");

			currentLevel.oracle = new ClusterChangeListener(currentLevel);
			currentClustering.addListener(currentLevel.oracle);

			// when no more increase, possibly create next level
			if (currentClustering.size() < oldClusterCount) {
				HierarchyLevel nextLevel = new HierarchyLevel(currentLevel);
				System.out.println();
				currentClustering = nextLevel.theClustering;
				currentLevel = nextLevel;
			}
		}
		System.out.println("No more contractions, hierarchy done");
		unfurlHierarchy(outputClustering);
		System.out.println("Finished at mod:" + incrementalModularity + ", with " + neighborHandShakes
				+ " neighbors' handshakes.");
		System.out.println("===== Done =====");
	}

	/**
	 * Given the Hashmap Level->HierarchyLevel, the original clustering, and the currentClustering, this procedure summarizes
	 * all arrays by iteratively substituting array values by the value found at that array index in the next higher array.
	 * Starting with singletons, in the end, this represents the final clustering. This works since new clusterings are built
	 * with oldCluster.index = newNode.index.
	 * 
	 * @param levelToClusteringVector
	 * @param outputClustering
	 * @param currentClustering
	 */
	private void unfurlHierarchy(Clustering outputClustering) {
		Graph theGraph = outputClustering.getGraph();
		int n = theGraph.N();
		int[] nodeAssignment = new int[n];
		Cluster theCluster;
		Node theNode;
		System.out.println("Unfurling nodes");
		for (NodeCursor nc = theGraph.nodes(); nc.ok(); nc.next()) {
			theNode = nc.node();
			for (int i = 0; i <= (levelToHierarchyLevel.size() - 2); i++) {
				theCluster = levelToHierarchyLevel.get(i).theClustering.getCluster(theNode);
				theNode = levelToHierarchyLevel.get(i).clusterToNextLevelNode.get(theCluster);
			}
			nodeAssignment[nc.node().index()] = theNode.index();
		}
		System.out.println("Finalizing");
		// if (outputClustering.poolNodes().ok()) {
		// System.out.println("poolsize: " + outputClustering.poolSize());
		// throw new RuntimeException("not ok.");
		// }
		outputClustering.resetToClustering(nodeAssignment);
		if (outputClustering.poolNodes().ok()) {
			System.out.println("poolsize: " + outputClustering.poolSize());
			throw new RuntimeException("not ok.");
		}
	}

	/**
	 * Given a Clustering, this procedure creates a new graph, with one node per given cluster (such that cluster.index =
	 * node.index!), and with edge weights corresponding to the sums of edge weights between clusters. This also creates
	 * self-loops. The graph is returned as a new clustering. Additionally this procedure takes a Hashmap of Arrays as an
	 * argument; it writes the clustering into a new array as node.index -> cluster index, and appends this array into the
	 * hashmap. The key of this entry is the level of hierarchy on which this clustering has been done, the original graph
	 * being level 0.
	 * 
	 * @param ering
	 *            input Clustering to be abstracted
	 * @return a graph with a node per input clustering, and appropriate weights
	 */
	private Clustering contractClustering(HierarchyLevel oldLevel) {
		Clustering oldClustering = oldLevel.theClustering;
		ClusterMap<Node> oldClusterMap = oldLevel.clusterToNextLevelNode;
		abstractedGraph = new Graph();
		GraphInterpretation theInterpretation = oldClustering.getInterpretation();
		int k = oldClustering.size();
		// prepare adjacency matrix
		double[][] adjMatrix = new double[k][k];
		for (int i = 0; i < k; ++i) {
			for (int j = 0; j < k; ++j) {
				adjMatrix[i][j] = 0.0;
			}
		}
		// collect edge weights into matrix, indexed by clusterID
		Edge currentEdge = null;
		int c1, c2;
		for (EdgeCursor ec = oldClustering.getGraph().edges(); ec.ok(); ec.next()) {
			currentEdge = ec.edge();
			c1 = oldClustering.getCluster(currentEdge.source()).getIndex();
			c2 = oldClustering.getCluster(currentEdge.target()).getIndex();
			adjMatrix[c1][c2] += theInterpretation.getWeight(currentEdge);
		}
		// create k new nodes in new graph
		Node[] newNodes = new Node[k];
		for (int i = 0; i < k; ++i) {
			newNodes[i] = abstractedGraph.createNode();
		}

		// point from old cluster to new node, with clusterID = new node's
		// position in array newNodes[] atm.
		for (GYCursor<Cluster> cc = oldClustering.clusters(); cc.ok(); cc.next()) {
			oldClusterMap.set((Cluster) cc.current(), newNodes[((Cluster) cc.current()).getIndex()]);
		}
		// create edges, indexed by newNodes[clusterID], with matrix weight
		double currentWeight = 0.0;
		EdgeMap newEdgeMap = abstractedGraph.createEdgeMap();
		for (int i = 0; i < k; ++i) {
			for (int j = i; j < k; ++j) {
				if (i == j)
					currentWeight = adjMatrix[i][j];
				else
					currentWeight = adjMatrix[i][j] + adjMatrix[j][i];
				if (currentWeight > 0.0) {
					currentEdge = abstractedGraph.createEdge(newNodes[i], newNodes[j]);
					newEdgeMap.setDouble(currentEdge, currentWeight);
				}
			}
		}
		Clustering newClustering = new Clustering(abstractedGraph, new WeightedGraph(abstractedGraph, newEdgeMap));
		newClustering.resetSingletons();
		abstractedGraph.disposeEdgeMap(newEdgeMap);
		return (newClustering);
	}

	private void writeGraph(Clustering clustering, String fileName) {
		Graph graph = clustering.getGraph();
		Graph2D g2d = new Graph2D();
		for (int i = 0; i < graph.N(); ++i)
			g2d.createNode();
		Node[] nodeArray = g2d.getNodeArray();
		Edge currentEdge;
		for (EdgeCursor ec = graph.edges(); ec.ok(); ec.next()) {
			currentEdge = g2d.createEdge(nodeArray[ec.edge().source().index()], nodeArray[ec.edge().target().index()]);
			g2d.setLabelText(currentEdge, Double.toString(clustering.getInterpretation().getAdjacency(ec.edge().source(),
					ec.edge().target())));
		}
		GraphMLIOHandler gio = new GraphMLIOHandler();
		try {
			gio.write(g2d, fileName);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * iterates once through all nodes in the array and moves them to the cluster of their favorite neighbor, according to
	 * the function bestNeighbor
	 * 
	 * @param theClustering
	 * @return the maximum increase achieved in this iteration
	 */
	/**
	 * @param theClustering
	 * @return
	 */
	private double bestNeighborsForEveryOne(HierarchyLevel hl, Node[] badNodes) {
		Clustering currentClustering = hl.theClustering;
		Graph currentGraph = hl.theGraph;

		for (Node aNode : badNodes) {
			if (!currentGraph.contains(aNode)) {
				System.out.println("aaaaaaaaaaaaaaahhhhhhhhhhhhhhhh");
			}
		}

		int n = badNodes.length;
		int[] order = orderNodes(currentGraph, badNodes);
		double maxIncrease = 0.0;
		for (int i = 0; i < n; ++i) {
			bestInnerIncrease = 0.0;
			Node currentNode = badNodes[order[i]];
			Node currentBestNeighbor = bestNeighbor(currentNode, currentClustering);
			if ((bestInnerIncrease > 0.00000001) || currentNode.degree() == 0) {
				Cluster oldCluster = currentClustering.getCluster(currentNode);
				currentClustering.extract(oldCluster, currentNode);
				incrementalModularity += bestInnerIncrease;
				// System.out.println("Increase: " + bestInnerIncrease + "
				// currentBestNeighbor " + currentBestNeighbor);
				// new singleton (increase or not)
				if (currentBestNeighbor == null) {
					Cluster newCluster = currentClustering.newCluster(currentNode);
					// normal case, migration
				} else {
					Cluster newCluster = currentClustering.getCluster(currentBestNeighbor);
					currentClustering.add(newCluster, currentNode);
				}
			}
			if (bestInnerIncrease > maxIncrease) {
				maxIncrease = bestInnerIncrease;
			}
		}
		return (maxIncrease);
	}

	public Clustering refreshClustering() {
		return (refreshClustering(theBaseListener.getInitialBadNodes()));
	}

	/**
	 * Receives a list of badNodes in the level 0 graph. Then these nodes are again reclustered by neighbor-search. If this
	 * changes the clustering, this change is listened to by a ClusterChangeListener which the next level can ask.
	 * 
	 * @param initialBadNodesList
	 */
	private Clustering refreshClustering(NodeList initialBadNodesList) {
		// check given badNodes, compound actions might mess things up, wegen
		// deletion order
		for (NodeCursor nc = initialBadNodesList.nodes(); nc.ok(); nc.next()) {
			if (!levelToHierarchyLevel.get(0).theGraph.contains(nc.node())) {
				initialBadNodesList.remove(nc.node());
			}
		}
		// new nodes not yet in clusters
		outputClustering.pool2Singletons();
		levelToHierarchyLevel.get(0).theClustering.pool2Singletons();
		neighborHandShakes = 0;
		if (updatePolicy == UpdatePolicy.neighbors) {
			NodeList neighbors = new NodeList();
			for (NodeCursor nc = initialBadNodesList.nodes(); nc.ok(); nc.next()) {
				for (NodeCursor nnc = nc.node().neighbors(); nnc.ok(); nnc.next()) {
					neighbors.add(nnc.node());
				}
			}
			for (NodeCursor nc = neighbors.nodes(); nc.ok(); nc.next()) {
				if (!initialBadNodesList.contains(nc.node())) {
					initialBadNodesList.add(nc.node());
				}
			}

		}
		System.out.println("Refresh called with initial bad nodes: " + initialBadNodesList.size()
				+ " at old (new graph, old clustering) mod:"
				+ new Modularity.Factory().getIndex(outputClustering).getValue());
		Node[] currentBadNodes = initialBadNodesList.toNodeArray();
		double iterationBestImprovement;
		int it;
		int oldClusterCount = levelToHierarchyLevel.get(0).theGraph.N();
		throughLevels: for (int level = 0; level < levelToHierarchyLevel.size(); ++level) {
			System.out.println("refreshing level " + level + " n/m = " + levelToHierarchyLevel.get(level).theGraph.N() + "/" + levelToHierarchyLevel.get(level).theGraph.E());
			HierarchyLevel hl = levelToHierarchyLevel.get(level);
			Clustering currentClustering = hl.theClustering;
			currentClustering.pool2Singletons();
			if (level > 0) {
				System.out.print("Level:" + level + " cleanup and updating. ");
				HashSet<Node> badNodesSet = new HashSet<Node>();
				HierarchyLevel lastHl = levelToHierarchyLevel.get(level - 1);
				ClusterChangeListener myOracle = lastHl.oracle;
				WeightedGraph currentGraph = (WeightedGraph) hl.theClustering.getInterpretation();
				// deletable nodes due to deleted clusters on lastHL
				for (Node toDelete : myOracle.deletableNodes()) {
					if (toDelete != null) {
						// TODO policy entscheidung, affected or neighbors
						if (updatePolicy == UpdatePolicy.neighbors) {
							for (NodeCursor nc = toDelete.neighbors(); nc.ok(); nc.next()) {
								// if (!badNodesSet.contains(nc.node())) {
								badNodesSet.add(nc.node());
								// }
							}
							// if (badNodesSet.contains(toDelete)) {
							badNodesSet.remove(toDelete);
							// }
						}
						hl.theGraph.removeNode(toDelete);
					}
				}
				
				
				//new version
				// new clusters/nodes
				double tmp;
				Node otherNode;
				//first create all new nodes
				for (Cluster insertedCluster : myOracle.newClusters()) {
					Node newNode = hl.theGraph.createNode();
					lastHl.clusterToNextLevelNode.set(insertedCluster, newNode);
				}
				
				//unnecessary ?
//				//nodes are there, now create all nec. new edges ==================================================
//				for (Cluster insertedCluster : myOracle.newClusters()) {
//					Node newNode = lastHl.clusterToNextLevelNode.get(insertedCluster);
//					if (newNode == null){
//						System.out.println("there is no such supernode!");
//					}
//					HashMap<Node, Double> adjacenciesN = new HashMap<Node, Double>();
//					// collect weight to all others
//					for (NodeCursor nc = insertedCluster.members(); nc.ok(); nc.next()) {
//						for (EdgeCursor ec = nc.node().edges(); ec.ok(); ec.next()) {
//							tmp = lastHl.theClustering.getInterpretation().getWeight(ec.edge());
//							Cluster otherCluster = lastHl.theClustering.getCluster(ec.edge().opposite(nc.node()));
//							otherNode = lastHl.clusterToNextLevelNode.get(otherCluster);
//							// self loops are counted twice, as are edges within the insertedCluster
//							if (insertedCluster.equals(otherCluster)) {
//								tmp /= 2.0;
//							}
//							if (adjacenciesN.containsKey(otherNode)) {
//								adjacenciesN.put(otherNode, adjacenciesN.get(otherNode) + tmp);
//							} else {
//								adjacenciesN.put(otherNode, tmp);
//							}
//						}
//					}
//					// create all those new edges, all are new, but might be tried twice - between two new nodes ! 
//					Entry<Node, Double> currentNeighNEntry;
//					for (Iterator<Entry<Node, Double>> neighNIt = adjacenciesN.entrySet().iterator(); neighNIt.hasNext();) {
//						currentNeighNEntry = neighNIt.next();
//						otherNode = currentNeighNEntry.getKey();
//						if (!currentGraph.getGraph().containsEdge(newNode, otherNode) && !currentGraph.getGraph().containsEdge(otherNode, newNode)){
//							Edge newEdge = currentGraph.getGraph().createEdge(newNode, otherNode);
//							currentGraph.setWeight(newEdge, currentNeighNEntry.getValue());
//						}
//					}
//					// newNode clustered, otherwise trouble with neighbors
//					hl.theClustering.newCluster(newNode);
//					badNodesSet.add(newNode);
//					// TODO policy entscheidung, affected or neighbors
//					if (updatePolicy == UpdatePolicy.neighbors) {
//						for (NodeCursor nc = newNode.neighbors(); nc.ok(); nc.next()) {
//							// if (!badNodesSet.contains(nc.node())) {
//							badNodesSet.add(nc.node());
//							// }
//						}
//					}
//				}
//				//end of new version =========================================
				
//				//old version ==========================================================================
//				// new clusters/nodes
//				double tmp;
//				Node otherNode;
//				for (Cluster insertedCluster : myOracle.newClusters()) {
//					Node newNode = hl.theGraph.createNode();
//					lastHl.clusterToNextLevelNode.set(insertedCluster, newNode);
//					// collect weight to all others
//					double[] adjacencies = new double[lastHl.theClustering.size()];
//					for (NodeCursor nc = insertedCluster.members(); nc.ok(); nc.next()) {
//						for (EdgeCursor ec = nc.node().edges(); ec.ok(); ec.next()) {
//							tmp = lastHl.theClustering.getInterpretation().getWeight(ec.edge());
//							// self loops are counted twice, as are edges within the insertedCluster
//							if (lastHl.theClustering.getCluster(ec.edge().source()).equals(
//									lastHl.theClustering.getCluster(ec.edge().target()))) {
//								tmp /= 2.0;
//							}
//							adjacencies[lastHl.theClustering.getCluster(ec.edge().opposite(nc.node())).getIndex()] += tmp;
//						}
//					}
//					// create edges to all existing nodes, with weights as collected
//					for (GYCursor<Cluster> cc = lastHl.theClustering.clusters(); cc.ok(); cc.next()) {
//						// only if node already created! Several might be created at once.
//						if (lastHl.clusterToNextLevelNode.get((Cluster) cc.current()) != null) {
//							Edge newEdge = hl.theGraph.createEdge(newNode, lastHl.clusterToNextLevelNode.get((Cluster) cc
//									.current()));
//							currentGraph.setWeight(newEdge, adjacencies[((Cluster) cc.current()).getIndex()]);
//						}
//					}
//					// newNode clustered, otherwise trouble with neighbors
//					hl.theClustering.newCluster(newNode);
//					badNodesSet.add(newNode);
//					// TODO policy entscheidung, affected or neighbors
//					if (updatePolicy == UpdatePolicy.neighbors) {
//						for (NodeCursor nc = newNode.neighbors(); nc.ok(); nc.next()) {
//							// if (!badNodesSet.contains(nc.node())) {
//							badNodesSet.add(nc.node());
//							// }
//						}
//					}
//				}
//				//end of old version ==========================================================================

				//new version
				// invalid clusters, recompute adjacencies from level below (one could save adj. to new clusters, as these
				// are already correct, but never mind...do it again)
				// take all invalid clusters from lower HL...
				for (Cluster invalidCluster : myOracle.invalidOldClusters()) {
					Node invalidNode = lastHl.clusterToNextLevelNode.get(invalidCluster);
					HashMap<Node, Double> adjacenciesN = new HashMap<Node, Double>();
					for (NodeCursor nc = invalidCluster.members(); nc.ok(); nc.next()) {
						for (EdgeCursor ec = nc.node().edges(); ec.ok(); ec.next()) {
							tmp = lastHl.theClustering.getInterpretation().getWeight(ec.edge());
							Cluster otherCluster = lastHl.theClustering.getCluster(ec.edge().opposite(nc.node()));
							otherNode = lastHl.clusterToNextLevelNode.get(otherCluster);
							// self loops are counted twice, and the same is true for edges within the invalidCluster
							if (invalidCluster.equals(otherCluster)) {
								tmp /= 2.0;
							}
							if (adjacenciesN.containsKey(otherNode)) {
								adjacenciesN.put(otherNode, adjacenciesN.get(otherNode) + tmp);
							} else {
								adjacenciesN.put(otherNode, tmp);
							}
						}
					}
					// first go through existing edges and edit/remove them, in order not to delete and remake all...
					// clean out corresp. entries from HashMap
					boolean selfLoopSeen = false;
					Entry<Node, Double> currentNeighNEntry;
					for (EdgeCursor ec = invalidNode.edges(); ec.ok(); ec.next()) {
						otherNode = ec.edge().opposite(invalidNode);
						if (otherNode.equals(invalidNode)) {	//self loop only once! comes twice in edge cursor...
							if (!selfLoopSeen) {	//dont do this twice for self-loops
								if (adjacenciesN.containsKey(otherNode)) { // in map and already there -> update
									currentGraph.setWeight(ec.edge(), adjacenciesN.get(otherNode));
									adjacenciesN.remove(otherNode);	//and kick out for later new adding
								} else { // there but not in map -> delete
									currentGraph.getGraph().removeEdge(ec.edge());
								}
							}
							selfLoopSeen = true;
						} else {	//it is not a self loop
							if (adjacenciesN.containsKey(otherNode)) { // in map and already there -> update
								currentGraph.setWeight(ec.edge(), adjacenciesN.get(otherNode));
								adjacenciesN.remove(otherNode);	//and kick out for later new adding
							} else { // there but not in map -> delete
								currentGraph.getGraph().removeEdge(ec.edge());
							}							
						}
					}
					// now add all new adjacencies not yet in graph but still in map, if any
					for (Iterator<Entry<Node, Double>> neighNIt = adjacenciesN.entrySet().iterator(); neighNIt.hasNext();) {
						currentNeighNEntry = neighNIt.next();
						otherNode = currentNeighNEntry.getKey();
						Edge newEdge = currentGraph.getGraph().createEdge(invalidNode, otherNode);
						currentGraph.setWeight(newEdge, currentNeighNEntry.getValue());
					}
					badNodesSet.add(invalidNode);
					// TODO policy entscheidung, affected or neighbors
					if (updatePolicy == UpdatePolicy.neighbors) {
						for (NodeCursor nc = invalidNode.neighbors(); nc.ok(); nc.next()) {
							// if (!badNodesSet.contains(nc.node())) {
							badNodesSet.add(nc.node());
							// }
						}
					}
				}
				//end of new version
				
//				//old version ==========================================================================
//				// invalid clusters, recompute adjacencies from level below (one could save adj. to new clusters, as these
//				// are already correct, but never mind...do it again)
//				// take all invalid clusters from lower HL...
//				for (Cluster invalidCluster : myOracle.invalidOldClusters()) {
//					Node invalidNode = lastHl.clusterToNextLevelNode.get(invalidCluster);
//					// create an array of adjacencies from lower HL
//					double[] adjacencies = new double[lastHl.theClustering.size()];
//					for (NodeCursor nc = invalidCluster.members(); nc.ok(); nc.next()) {
//						for (EdgeCursor ec = nc.node().edges(); ec.ok(); ec.next()) {
//							tmp = lastHl.theClustering.getInterpretation().getWeight(ec.edge());
//							// self loops are counted twice, as are edges within the insertedCluster
//							if (lastHl.theClustering.getCluster(ec.edge().source()).equals(
//									lastHl.theClustering.getCluster(ec.edge().target()))) {
//								tmp /= 2.0;
//							}
//							adjacencies[lastHl.theClustering.getCluster(ec.edge().opposite(nc.node())).getIndex()] += tmp;
//						}
//					}
//					// use the above array to renovate all adjacencies of this HL's invalid supernode
//					for (GYCursor<Cluster> cc = lastHl.theClustering.clusters(); cc.ok(); cc.next()) {
//						otherNode = lastHl.clusterToNextLevelNode.get((Cluster) cc.current());
//						Edge invalidEdge = invalidNode.getEdge(otherNode);
//						if (invalidEdge == null) {
//							invalidEdge = otherNode.getEdge(invalidNode);
//						}
//						if (invalidEdge == null) {
//							invalidEdge = hl.theGraph.createEdge(invalidNode, otherNode);
//						}
//						currentGraph.setWeight(invalidEdge, adjacencies[((Cluster) cc.current()).getIndex()]);
//					}
//					badNodesSet.add(invalidNode);
//					// TODO policy entscheidung, affected or neighbors
//					if (updatePolicy == UpdatePolicy.neighbors) {
//						for (NodeCursor nc = invalidNode.neighbors(); nc.ok(); nc.next()) {
//							// if (!badNodesSet.contains(nc.node())) {
//							badNodesSet.add(nc.node());
//							// }
//						}
//					}
//				}
//				//end of old version ==========================================================================
				
				currentBadNodes = badNodesSet.toArray(new Node[0]);
				myOracle.tagAsValid(); // the oracle of the last HL
				// endif (level > 0)
			}
			//now comes stuff for all levels, also for level 0
			// level consistent, bad nodes collected -> partial refresh!
			System.out.println("Refresh, n:" + hl.theGraph.N() + ", m:" + hl.theGraph.E());
			iterationBestImprovement = 1.0;
			it = 1;
			for (Node bn : currentBadNodes){
				if (hl.theClustering.getCluster(bn) == null){
					hl.theClustering.newCluster(bn);
//					System.out.println("emergency inserted new cluster");
				}
			}
			while (iterationBestImprovement > 0.0) {
//				for (Node bn : currentBadNodes){
//					System.out.print(hl.theClustering.getCluster(bn) + ", ");
//				}
//				System.out.println();
				iterationBestImprovement = bestNeighborsForEveryOne(hl, currentBadNodes);
				System.out.print("It:" + it++ + ", best increase:" + iterationBestImprovement + " ... ");
			}
			System.out.println("no more improvement, level done (" + currentClustering.size() + " clusters). ");
			if (oldClusterCount == currentClustering.clusterCount()) {
				int originalSize = levelToHierarchyLevel.size();
				for (int higherLevel = originalSize - 1; higherLevel > level; --higherLevel) {

					// dispose of listener
					levelToHierarchyLevel.get(higherLevel).theClustering.removeClusteringListener(levelToHierarchyLevel
							.get(higherLevel).oracle);
					// dispose clustering
					levelToHierarchyLevel.get(higherLevel).theClustering.dispose();
					//dispose map from lower level
					for (GYCursor<Cluster> cc = levelToHierarchyLevel.get(higherLevel - 1).theClustering.clusters(); cc.ok(); cc
							.next()) {
						levelToHierarchyLevel.get(higherLevel - 1).clusterToNextLevelNode.set(cc.active(), null);
					}
					levelToHierarchyLevel.remove(higherLevel);
					System.out.print("removing level " + higherLevel);
				}
				System.out.println();
				break throughLevels;
			} else if (level == levelToHierarchyLevel.size() - 1) {
				HierarchyLevel nextHl = new HierarchyLevel(hl);
				nextHl.oracle = new ClusterChangeListener(nextHl);
				nextHl.theClustering.addListener(nextHl.oracle);
				System.out.println("adding level " + (level + 1));
			}
			oldClusterCount = currentClustering.clusterCount();
		}
		levelToHierarchyLevel.get(levelToHierarchyLevel.size() - 1).oracle.tagAsValid();
		unfurlHierarchy(outputClustering);
		theBaseListener.tagAsValid();
		System.out.println("Refresh finished at current mod:"
				+ new Modularity.Factory().getIndex(outputClustering).getValue() + ", with " + neighborHandShakes
				+ " neighbors' handshakes.");
		System.out.println("===== Done =====");
		return (outputClustering);
	}
}