/*
 * Created on Nov 3, 2005
 *
 * Copyright 2005.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
*/
package de.uka.algo.clustering.algorithms.newman;

import java.awt.BorderLayout;

import javax.swing.JFrame;

import y.base.Graph;
import y.view.Graph2D;
import y.view.Graph2DView;
import y.view.NavigationMode;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.algorithms.Algorithm;
import de.uka.algo.clustering.algorithms.newman.internal.DendrogrammStorage;
import de.uka.algo.clustering.algorithms.newman.internal.MeasurementAdapter;
import de.uka.algo.clustering.algorithms.newman.internal.UpdateStrategy;
import de.uka.algo.clustering.algorithms.newman.util.MaximumPair;
import de.uka.algo.clustering.algorithms.newman.util.UnionFindBacktrack;
import de.uka.algo.graphs.GraphInterpretation;
import de.uka.algo.graphs.WeightedGraph;

public class NewmanClusterer implements Algorithm{
	
	static org.apache.log4j.Logger logger = org.apache.log4j.Logger
			.getLogger(NewmanClusterer.class);
	
	MeasurementAdapter measure;
	boolean stopOnPeak = true;
	DendrogrammStorage dendroStore;
	boolean createDendro = false;
	int desiredClusters = 0;
	UpdateStrategy us = null;
	WeightedGraph oldGraph = null;
	
	Clustering ourClustering;
	MaximumPair maxPair;
	
	public NewmanClusterer(MeasurementAdapter measure) {
		this.measure = measure;
	}
	
	public void run(Clustering aClustering) {
		logger.info("doCluster started.");

		 this.ourClustering = aClustering.clone();
		
		if (createDendro){
			logger.debug("dendrogram will be created.");
			if(dendroStore == null)
				dendroStore = new DendrogrammStorage();
		}
		else {
			dendroStore = null;
		}
		
		logger.info("preparing clustering");
		if(us != null && oldGraph != null) {
			us.prepareClustering(ourClustering, oldGraph);
		} else {
			ourClustering.resetSingletons();
			if (us != null)
				us.init(ourClustering);
		}
		oldGraph = new WeightedGraph(ourClustering.getInterpretation());
		
		logger.info("clustering has " + ourClustering.clusterCount() + " clusters.");
		
		if(dendroStore != null)
			dendroStore.init(ourClustering);

		int mergeCount = 0;
		int remainingClusters = desiredClusters;
		if (remainingClusters > ourClustering.clusterCount()){
			remainingClusters = ourClustering.clusterCount();
		} else{ 
			if (remainingClusters < 1){
				remainingClusters = 1;
			}
		}
		
		measure.initialize(ourClustering);
		this.maxPair = new MaximumPair();
		boolean resultCopied = false;
		
		while(true){
			logger.info("started next pass");
			logger.info("remaining clusters: " + ourClustering.clusterCount());
			logger.info("measure: " + measure.getMeasure());
			measure.getMaximumPair(maxPair);
			
			
			if(!maxPair.exists){
				logger.info("no further pair exists.");
				if (!resultCopied) aClustering.setToClustering(ourClustering);
				resultCopied = true;
				break;
			}
			
			if (stopOnPeak && measure.hasPeak()){
				logger.info("measure reached its peak.");
				if (!resultCopied){
					aClustering.setToClustering(ourClustering);
					resultCopied = true;
					if (createDendro)
						dendroStore.setHasPeak();
					else
						break;
				}
			}
			
			if (ourClustering.clusterCount()==remainingClusters){
				logger.info("Desired fixed number of clusters reached.");
				if (!resultCopied){
					aClustering.setToClustering(ourClustering);

					resultCopied = true;
					if (createDendro)
						dendroStore.setHasPeak();
					else
						break;
				}
			}

			if(us != null)
				us.mergingClusters(maxPair.c1, maxPair.c2);
			measure.mergeClusters(maxPair.c1, maxPair.c2);
			mergeCount++;
			logger.info("merges so far:" + mergeCount);
			if (createDendro)
				dendroStore.setLatestModularity(measure.getMeasure());
		}
		
		if (createDendro) {
			logger.info("Now creating dendrogram.");
			
			dendroStore.detach();
			// dendroStore.createDendrogrammDiagonalEdges();
			// dendroStore.createDendrogrammRectangularEdgesAtSide();
			new DendrogramWindow(dendroStore.createDendrogrammRectangularEdgesDown()).setVisible(true);
		}
		
		ourClustering.dispose();
		
	}
	
public class DendrogramWindow extends JFrame {
		
		private static final long serialVersionUID = 3258409551756734777L;
		private Graph2D dendro;
		private Graph2DView view;
		
		public DendrogramWindow(Graph2D dendro) {
			super("Dendrogramwindow");
			this.dendro = dendro;
			
			setLayout(new BorderLayout());
			Graph2DView view = new Graph2DView(dendro);
			add(view, BorderLayout.CENTER);
			view.fitContent();
			view.addViewMode(new NavigationMode());
			pack();
		}
	}

public boolean isCreatingDendro() {
	return createDendro;
}



public void setCreatingDendro(boolean createDendro) {
	this.createDendro = createDendro;
}



public int getDesiredClusters() {
	return desiredClusters;
}



public void setDesiredClusters(int desiredClusters) {
	this.desiredClusters = desiredClusters;
}



public boolean isStoppingOnPeak() {
	return stopOnPeak;
}



public void setStoppingOnPeak(boolean stopOnPeak) {
	this.stopOnPeak = stopOnPeak;
}



public UpdateStrategy getUpdateStrategy() {
	return us;
}



public void setUpdateStrategy(UpdateStrategy us) {
	this.us = us;
}

}
