/*
 * Created on Apr 21, 2005
 *
 * Copyright 2005. --- All rights resevered. 
 * Algorithmic Group, Karlsruhe, Germany.
 * 
 * project: newman
 * package: de.uka.algo.modularityClustering
 */
package de.uka.algo.clustering.algorithms.newman.internal;

import java.util.HashMap;
import java.util.Map;

/**
 * Heap.
 * 
 * @author Robert Goerke
 * 
 */
public class BoundedHeap<T> {

	dataType<T> heapData[];

	Map<T, Integer> indexToHeapPosition;

	int numberOfElements;

	/**
	 * Default constructor that initialize a heap where the maximum number of
	 * elements is given.
	 * 
	 * @param capacity
	 *            maximum number of allowed elements
	 */
	public BoundedHeap(int capacity) {
		heapData = (dataType<T>[]) new dataType[capacity];
		for (int i = 0; i < capacity; ++i)
			heapData[i] = null;
		indexToHeapPosition = new HashMap<T, Integer>();
		numberOfElements = 0;
	}

	/**
	 * Inserts a given value-index pair.
	 * 
	 * @param value
	 *            the value of the given pair
	 * @param index
	 *            the index of the given pair
	 */
	public void insertElement(double value, T aCluster) {
		if (numberOfElements < heapData.length) {
			int pos = numberOfElements;
			while (pos > 0 && (heapData[parent(pos)].value < value)) {
				heapData[pos] = heapData[parent(pos)];
				indexToHeapPosition.remove(heapData[parent(pos)].aCluster);
				indexToHeapPosition.put(heapData[parent(pos)].aCluster, pos);
				pos = parent(pos);
			}
			heapData[pos] = new dataType<T>(value, aCluster);
			indexToHeapPosition.put(aCluster, pos);
			++numberOfElements;
		} else {
			throw new ArrayIndexOutOfBoundsException("heap is already full");
		}
	}

	/**
	 * Doesn't update indexToHeapPosition
	 * 
	 * @param value
	 * @param aCluster
	 */public void insertElement2(double value, T aCluster) {
		if (numberOfElements < heapData.length) {
			int pos = numberOfElements;
			heapData[pos] = new dataType<T>(value,aCluster);
			while (pos > 0 && (heapData[parent(pos)].value < value)) {
				swap(pos, parent(pos));
				pos = parent(pos);
			}
			++numberOfElements;
		} else {
			throw new ArrayIndexOutOfBoundsException("heap is already full");
		}
	}
	
	public boolean existsElement(T aCluster){
		return indexToHeapPosition.containsKey(aCluster);
	}

	/**
	 * Returns the index of the maximum value in the heap.
	 * 
	 * @return the index of the maximum value
	 */
	public T getMaximumElement() {
		if (numberOfElements <= 0) {
			throw new ArrayIndexOutOfBoundsException("no elements available");
		}
		return heapData[0].aCluster;
	}

	/**
	 * Returns the maximum value in the heap.
	 * 
	 * @return the value of the maximum value
	 */
	public double getMaximumValue() {
		if (numberOfElements <= 0) {
			throw new ArrayIndexOutOfBoundsException("no elements available");
		}
		return heapData[0].value;
	}

	/**
	 * Removes the maximum value and returns the corresponding T.
	 * 
	 * @return the index of the maximum value
	 */
	public T removeMaximum() {
		if (numberOfElements <= 0) {
			throw new ArrayIndexOutOfBoundsException(
					"no more elements available");
		}
		T indexOfMaximum = heapData[0].aCluster;
		indexToHeapPosition.remove(indexOfMaximum);
		if (numberOfElements > 1) {
			heapData[0] = heapData[numberOfElements - 1];
			indexToHeapPosition.remove(heapData[0].aCluster);
			indexToHeapPosition.put(heapData[0].aCluster, 0);
			heapData[numberOfElements - 1] = null;
			--numberOfElements;
			heapify(0);
		} else {
			heapData[0] = null;
			numberOfElements = 0;
		}
		return indexOfMaximum;
	}

	/**
	 * Returns the size of the heap.
	 * 
	 * @return heap size
	 */
	public int size() {
		return numberOfElements;
	}

	/**
	 * Returns the array-representation of the heap;
	 */
	public String toString() {
		StringBuffer cache = new StringBuffer();
		for (int i = 0; i < numberOfElements; ++i)
			cache.append("(").append(heapData[i].value).append(", ").append(
					heapData[i].aCluster).append(") ");  //macht im Moment keinen Sinn !
		return cache.toString();
	}

	/**
	 * gives the value stored in the record with given index
	 * 
	 * @param aCluster
	 *            given T to look up
	 * @return value of the T
	 */
	public double valueOfElement(T aCluster) {
		if (!indexToHeapPosition.containsKey(aCluster)) {
			//throw new ArrayIndexOutOfBoundsException(
			//		"T does not exist");
			
			// System.out.println("Returned NaN.");
			return Double.NaN;
		}
		return (heapData[indexToHeapPosition.get(aCluster)].value);
	}

	/**
	 * Removes the element from the heap that has a given T (not heap
	 * position!) and restore the heap property.
	 * 
	 * @param aCluster
	 *            given T
	 */
	public void removeElement(T aCluster) {
		if (indexToHeapPosition.get(aCluster) != null) {
			int pos = indexToHeapPosition.remove(aCluster);
			if (pos < numberOfElements - 1) {
				heapData[pos] = heapData[numberOfElements - 1];
				indexToHeapPosition.remove(heapData[pos].aCluster);
				indexToHeapPosition.put(heapData[pos].aCluster, pos);
				heapData[numberOfElements - 1] = null;
				--numberOfElements;
				heapify(pos);
			} else {
				heapData[pos] = null;
				--numberOfElements;
			}
		}
	}

	/**
	 * Checks and restores (if necessary) the heap propertie of the given T.
	 * 
	 * @param index
	 *            heap position which needs to be checked
	 */
	protected void heapify(int index) {
		int indexOfLeftChild = leftChild(index);
		if (indexOfLeftChild >= numberOfElements) {
			// final element - no further children!
			return;
		} else {
			// do the dirty work
			int indexOfLocalMaximum = 0;
			if (heapData[indexOfLeftChild].value > heapData[index].value) {
				indexOfLocalMaximum = indexOfLeftChild;
			} else {
				indexOfLocalMaximum = index;
			}
			int indexOfRightChild = rightChild(index);
			if (indexOfRightChild < numberOfElements
					&& heapData[indexOfRightChild].value > heapData[indexOfLocalMaximum].value) {
				indexOfLocalMaximum = indexOfRightChild;
			}
			if (indexOfLocalMaximum != index) {
				swap(index, indexOfLocalMaximum);
				heapify(indexOfLocalMaximum);
			}
		}
	}

	/**
	 * Swaps two given heap elements.
	 * 
	 * @param index
	 *            position of first element
	 * @param otherIndex
	 *            position of swap element
	 */

	private void swap(int index, int otherIndex) {
		dataType<T> tmp = heapData[index];
		heapData[index] = heapData[otherIndex];
		heapData[otherIndex] = tmp;
		indexToHeapPosition.remove(heapData[index].aCluster);
		indexToHeapPosition.remove(heapData[otherIndex].aCluster);
		indexToHeapPosition.put(heapData[index].aCluster, index);
		indexToHeapPosition.put(heapData[otherIndex].aCluster, otherIndex);
	}

	/**
	 * Returns the position of the parent of the given heap element.
	 * 
	 * @param index
	 *            position in the heap
	 * @return the position of the parent
	 */
	public int parent(int index) {
		return (index - 1) >> 1;
	}

	/**
	 * Returns the position of the (possible) left child of the given heap
	 * element.
	 * 
	 * @param index
	 *            position in the heap
	 * @return the position of the left child
	 */
	public int leftChild(int index) {
		return 2 * index + 1;
	}

	/**
	 * Returns the position of the (possible) right child of the given heap
	 * element.
	 * 
	 * @param index
	 *            position in the heap
	 * @return the position of the right child
	 */
	public int rightChild(int index) {
		return 2 * index + 2;
	}

	public void clear() {
		for(int i=0; i<numberOfElements; ++i)
			heapData[i] = null;
		indexToHeapPosition.clear();
		numberOfElements = 0;
	}
	
	private class dataType <D> {
		public double value;

		public D aCluster;

		public dataType(double value, D aCluster) {
			this.value = value;
			this.aCluster = aCluster;
		}
	}

/*	public static void main(String args[]) {
		BoundedHeapOfCluster bh = new BoundedHeapOfCluster(10);
		for (int i = 1; i < 6; ++i)
			bh.insertElement(3 * i, i);
		System.out.println(bh.removeMaximum() + "\t" + bh);
		bh.removeIndex(3);
		System.out.println(bh);
		while (bh.size() > 0)
			System.out.println(bh.removeMaximum() + "\t" + bh);
	} */
}