/*
 * Created on Nov 4, 2005
 *
 * Copyright 2005.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
*/
package de.uka.algo.clustering.algorithms.newman.internal;

import org.apache.log4j.Logger;

import y.base.Graph;

import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.algorithms.newman.util.ClusterAdjacencyMatrix;
import de.uka.algo.clustering.algorithms.newman.util.FractionOfEdgeWeights;
import de.uka.algo.clustering.algorithms.newman.util.MaximumPair;
import de.uka.algo.graphs.GraphInterpretation;


/*
 * Diese Klasse soll eine Generisierung des Modularity-Adapters für
 * den Newman-Clusterer sein, die mit beliebigen Maßen klarkommt, die
 * bestimmte Anforderungen erfüllen.
 * 
 * Das Maß muß insofern "ähnlich" Coverage-Modularity sein, als
 * Updates beim Mergen aus den schon berechneten Updates berechnet
 * werden sollen.
 * 
 * TODO: wirklich nur additiv?
 */
public class GenericAdditiveAdapter implements MeasurementAdapter{

	Clustering clustering;
	Graph graph;
	GraphInterpretation graphInt;
	boolean weighted;
	boolean considerWeightsOverride;
	double edgeWeightSum;
	
	/**
     * this flag is set when the peak has been reached
     */
    boolean peakIsReached = false;
    
	
    /*
     * ausgelagerter Krams
     */
    ClusterAdjacencyMatrix adj;
    GenericDeltaQMatrix deltaQ;
    FractionOfEdgeWeights frac;
    QInterface measurementInterface;
    
	public GenericAdditiveAdapter(boolean considerEdgeWeights, QInterface modI, GenericDeltaQMatrix matrix){
    	weighted = considerEdgeWeights;
    	considerWeightsOverride = true;
    	this.measurementInterface = modI;
    	this.deltaQ = matrix;
    }
    
	public GenericAdditiveAdapter(){};
	
    public void initialize(Clustering aClustering) {
    	peakIsReached = false;
    	/*
    	 * Infos uebers Clustering / den Graphen
    	 */
    	clustering = aClustering;
    	graph = aClustering.getGraph();
    	graphInt = aClustering.getInterpretation();
    	if (!considerWeightsOverride)
    		weighted = graphInt.isWeighted();
    	edgeWeightSum = graphInt.getEdgeWeightSum();
    	
    	this.frac = new FractionOfEdgeWeights();
    	frac.init(aClustering, weighted);  	
		    	
    	
    	this.adj = new ClusterAdjacencyMatrix();
    	adj.init(aClustering);
    	
    	
    	/*
    	 * (DONE): make this more generic, i.e., pluggable
    	 */
    	// this.deltaQ = new GenericAdditiveDeltaQMatrix(measurementInterface);
    	deltaQ.init(aClustering, adj, frac, weighted);
 	}


		
	public void getMaximumPair(MaximumPair maxPair) {
		Logger logger = Logger.getLogger(this.getClass());
		
		/* The _whole_ matrix needs to be computed, when the peak has
		 * been reached, i.e., when the maximum value in the
		 * (incomplete) matrix is negative.
		 * 
		 * So, when the value we are about to return is negative,
		 * and the peakIsReached has not yet been set, compute the
		 * whole matrix and set the flag.
		 */
		boolean recompute = false;
		
		if (clustering.clusterCount()==1){
			maxPair.exists = false;
			return;
		}
		
		if (deltaQ.isEmpty()){
			// Matrix leer: neu berechnen
			// onlyDisconnectedClusters = true;
			peakIsReached = true;
			deltaQ.setPeakIsReached();
			
			recompute = true;
		} else {
			
			// Cluster c1 = bestOfEachRow.getMaximumCluster();
			double maxValue = deltaQ.getMaximumValue();
			if (maxValue < 0 && !peakIsReached){
				// Peak erreicht/ueberschritten: neu berechnen
				peakIsReached = true;
				deltaQ.setPeakIsReached();
				
				recompute = true;
			}
		}
			 
		
		if (recompute){
			deltaQ.recomputeForAll();
		}
		
		
		
		
		if (!deltaQ.isEmpty()){
			Cluster c1 = deltaQ.getMaximumCluster1();
			Cluster c2 = deltaQ.getMaximumCluster2();
			
			maxPair.c1 = c1;
			maxPair.c2 = c2;
			
			maxPair.value = deltaQ.getMaximumValue();
			logger.info("maximum deltaQ: " + maxPair.value);
			
			maxPair.exists = true;
			
			return;
		} else {
			/* eigentlich kann dieser Fall nicht vorkommen
			 */
			
			throw new RuntimeException("Something has gone terribly wrong ...");
		}
	}
	
	
	
	
	

	public void mergeClusters(Cluster c1, Cluster c2) {
		// updateModularity(c1, c2);
		
	    // mergeBestClusters
		Cluster combinedCluster = clustering.merge(c1, c2);
	    
		
		deltaQ.merge(c1, c2, combinedCluster);
	    adj.merge(c1, c2, combinedCluster);
        frac.merge(c1, c2, combinedCluster);
	}

	
	public boolean hasPeak(){
		return (deltaQ.getMaximumValue() <= 0.0);
	}
    
	public double getMeasure(){
    	return measurementInterface.getModularity();
    }
}
