package de.uka.algo.clustering.algorithms.newman.internal;

import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.graphs.GraphInterpretation;


/**
 * An UpdateStrategy defines a way to use an existing clustering when
 * running Newman's algorithm. It provides methods to prepare the given
 * clustering so that it can serve as input for Newman's algorithm. It
 * also updates the dendrogram created by previous runs of Newman's
 * algorithm if this is possible.
 * 
 * This class is abstract. Concrete subclasses are for example
 * @see ResetUpdateStrategy
 *  
 * @author maillard
 *
 */
public abstract class UpdateStrategy {
	public void init(Clustering cl) {}
	public abstract void prepareClustering(Clustering cl, GraphInterpretation oldGraph);
	public void mergingClusters(Cluster c1, Cluster c2) {}
	
	private static String[] descriptions = new String[] {
		"reset to singletons", "stone pit", "backtrack", "random backtrack"};
	private static Class[] classes = new Class[] {
		ResetUpdateStrategy.class, StonePitUpdateStrategy.class,
		BacktrackUpdateStrategy.class, RandomBacktrackUpdateStrategy.class};
	
	public static int numStrategies() {return descriptions.length;}
	public static String[] getDescriptions() {return descriptions;}
	public static UpdateStrategy create(int index) {
		if(index < 0 || index >= numStrategies())
			throw new IllegalArgumentException();
		try {
			return (UpdateStrategy) classes[index].newInstance();
		} catch(Exception e) {
			e.printStackTrace();
			return null;
		}
	}
}
