package de.uka.algo.clustering.algorithms.newman.util;

/** A data structure that manages partitions of the set {0,...,N-1}.
 * It supports:
 *   1. finding the set that contains a given element (in O(log N))
 *   2. unifying two sets (in O(1))
 *   3. splitting a set into the two sets it was built from (in O(1))
 * 
 * and more...
 * 
 * The space requirement is in O(n).
 *
 * @author maillard
 */
public class UnionFindBacktrack {
	private int N = 0;
	/* UF forest. 2 integers for each element e:
	 * First integer: -1 for root nodes, pointer to parent for non-root nodes
	 * Second integer: The number of elements in the subtree induced by e
	 */
	private int [] ufForest;
	
	/* implementation of the union forest that stores the information
	 * about the unions performed
	 * There are 2 integers for each element e, whereby the following
	 * invariant is true:
	 *   1. If e is a root node in the uf forest, then e has at most
	 *      one child c in the union forest which is pointed to by
	 *      the first integer. c is the element that was attached to e
	 *      in the uf forest during the last union.
	 *   2. If e was made child of a root node r in the uf forest
	 *      during an union, then the second integer of e points
	 *      to the former children of this node in the union forest.
	 * As a consequence of this invariant, splitting a set with root node e,
	 * where c is the first child of e in the union forest, is equivalent to
	 *   - in the uf forest: cutting away c and the subtree induced by it
	 *   - in the union forest; setting e's first child in the union forest
	 *     to c's second and setting the latter to -1
	 */
	private int [] unionForest;
	
	private int setCount;
	
	public UnionFindBacktrack(int numElements) {
		init(numElements);
	}
	
	public void init(int numElements) {
		N = numElements;
		setCount = N;
		ufForest = new int[N*2];
		unionForest = new int[N*2];
		for(int i=0; i<N; i++) {
			ufForest[2*i] = -1;
			ufForest[2*i+1] = 1;
			unionForest[2*i] = -1;
			unionForest[2*i+1] = -1;
		}
	}
	
	public boolean isRoot(int elem) {
		assert(0 <= elem);
		assert(elem < N);
		
		return ufForest[2*elem] < 0;
	}
	
	/** Finds the set a given element is contained in.
	 * 
	 * @param e The aforesaid element.
	 * @return The root of the set.
	 */
	public int find(int e) {
		assert(0 <= e);
		assert(e < N);
		while(ufForest[2*e] >= 0)
			e = ufForest[2*e];
		return e;
	}
	
	/** Merges two sets.
	 * @param root1 Root of the first set.
	 * @param root2 Root of the second set.
	 * @return Root of the new set (either root1 or root2)
	 */
	public int union(int root1, int root2) {
		assert(isRoot(root1));
		assert(isRoot(root2));
		
		int w1 = ufForest[2*root1+1];
		int w2 = ufForest[2*root2+1];
		if(w1 < w2) {
			int temp = root1;
			root1 = root2;
			root2 = temp;
		}
		
		ufForest[2*root2] = root1;
		ufForest[2*root1+1] = w1+w2;
		unionForest[2*root2+1] = unionForest[2*root1]; 
		unionForest[2*root1] = root2;
		
		setCount--;
		
		return root1;
	}
	
	/** Splits a set into the two sets it was built from.
	 * 
	 * @param root The root of the set to split.
	 * @return The element that was made root or -1 if the set consisted
	 *   of only one element. 
	 */
	public int split(int root) {
		assert(isRoot(root));
		
		int newroot = unionForest[2*root];
		if(newroot >= 0) {
			ufForest[2*newroot] = -1;
			ufForest[2*root+1] -= ufForest[2*newroot+1];
			unionForest[2*root] = unionForest[2*newroot+1];
			unionForest[2*newroot+1] = -1;
			setCount++;
		}
		
		return newroot;
	}
	
	public void isolate(int elem) {
		// FIXME: this is not the operation we want
		while(!isRoot(elem)) {
			split(find(elem));
		}
	}
	
	/**
	 * Splits a set and its subsets repeatedly until there are only
	 * singletons left.
	 * 
	 * @param root The root of the set to split totally.
	 */
	public void splitTotally(int root) {
		assert(isRoot(root));
		
		if(unionForest[2*root] >= 0) {
			int newroot = split(root);
			splitTotally(root);
			splitTotally(newroot);
		}
	}

	public int getSetCount() {
		return setCount;
	}

	public int getNumElements() {
		return N;
	}
}
