package de.uka.algo.clustering.generators.attractors;

import y.base.Graph;
import y.base.NodeCursor;
import y.base.NodeList;
import de.uka.algo.clustering.generators.GeometricClusteredGraph;

/**
 * Attractors are pre clustered graphs using geometric properties to create an
 * inital clustering. At first, k nodes -- so called attractors -- are added to
 * the plain, each representing a single cluster. after that, n-k satellite
 * nodes are added, each assigned to the cluster of its nearest attractor node.
 * The probability that a satellite is added is dependent on the distance to its
 * nearest attractor node. f indicates the density of the graph and therefore
 * the significance of the created clustering.
 * 
 * @author Daniel Delling
 * 
 */
public class Attractor extends GeometricClusteredGraph {
	
	protected NodeList attractorNodes;


	/**
	 * creates an attractor graph with n nodes and uniformly at random between
	 * log_10(n) and sqrt(2) clusters and density f. a density value of 0 will
	 * lead to a starlike graph, which means, that only the satellites are
	 * connected to the attractors and the attractors are connected pairwise.
	 * Use values around 1 for good clusterings
	 * 
	 * @param graph
	 *            graph that the attractor is built in
	 * @param numberOfNodes
	 *            number of nodes
	 * @param densityThreshold
	 *            density of the graph
	 */
	public Attractor(Graph graph, int numberOfNodes, double densityThreshold) {
		super(graph);

		int numberOfClusters = getRandomNumberOfClusters(numberOfNodes);

		double internalThreshold = Math.sqrt(2.0 / (Math.PI * numberOfClusters))
				* stretchFactor;

		attractorNodes = createVoronoiEmbedding(numberOfClusters, numberOfNodes-numberOfClusters, internalThreshold);
		
		// connect satellite nodes with attractors 
		for(NodeCursor c = attractorNodes.nodes(); c.ok(); c.next()) {
			for(NodeCursor c2 = getCluster(c.node()).members(); c2.ok(); c2.next()) {
				if(c2.node()!=c.node())
					graph.createEdge(c.node(), c2.node());
			}
		}

		// connect attractors 
		createEdgeset(attractorNodes, 2.5 * internalThreshold);
		
		// connect everything 
		createEdgeset(new NodeList(graph.nodes()), 0.5 * internalThreshold * densityThreshold);
		
		rescale();

	
	}
	@Override
	public String getID() {
		return "Attractor";
	}

}
