/*
 * Copyright 2006.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering.generators.randomCluster;

import java.util.Random;

import y.base.Graph;
import y.base.Node;
import de.uka.algo.clustering.generators.ClusteredGraph;

/**
 * RandomClustergraphGenerator generates a graph with a clustering according to
 * a given partition. IntraclusterEdges are created with probability pIn and
 * InterclusterEdges are created with probability pOut
 * 
 * @author Daniel Delling
 * 
 */
public class RandomClusteredGraph extends ClusteredGraph {

	protected int[] clusterSizes = null;

	protected double pIn;

	protected double pOut;

	private Random random = new Random();
	

	/**
	 * Creates a randomcluster in the given graph with a clustering to the given
	 * clusterSizes.
	 * 
	 * @param g
	 *            the Graph
	 * @param clusterSizes
	 *            entries of the array set the size of each cluster
	 * @param pIn
	 *            probability that an IntraclusterEdge is created
	 * @param pOut
	 *            probability that an InterclusterEdge is created
	 */
	public RandomClusteredGraph(Graph g, int[] clusterSizes, double pIn,
			double pOut) {
		super(g);
		this.createGraph(clusterSizes, pIn, pOut);
	}

	RandomClusteredGraph(Graph g) {
		super(g);
	}

	protected void createGraph(int[] clusterSizes, double pIn,
			double pOut) { 
//		super.clear();
		this.clusterSizes = clusterSizes;
		this.pIn = pIn;
		this.pOut = pOut;
		
		
		// nodecount
		int nodeCount = 0;
		for (int i = 0; i < clusterSizes.length; i++) {
			if (clusterSizes[i] > 0) {
				nodeCount += clusterSizes[i];
			}
		}

		// NodeArray for creating edges
		Node tmpNodes[] = new Node[nodeCount];

		int nodeIndex = 0;

		// Initial for Interclusteredges, according to Brandes
		double rOut = 0.0; //
		int wOut = 0; // pointer for next edges
		if (pOut == 0) {
			rOut = 1;
			wOut = Integer.MIN_VALUE;
		} else if (pOut == 1) {
			rOut = 0;
			wOut = -1;
		} else {
			rOut = 1 / Math.log(1 - pOut);
			wOut = (int) (1 - rOut * Math.log(random.nextDouble())) - 1;
		}

		// Initial for Intraclusteredges, according to Brandes
		double rIn = 0.0; //
		int wIn = 0; // pointer for next edges
		if (pIn == 0) {
			rIn = 1;
			wIn = Integer.MAX_VALUE;
		} else if (pIn == 1) {
			rIn = 0;
			wIn = 0;
		} else {
			rIn = 1 / Math.log(1 - pIn);
			wIn = (int) (rIn * Math.log(random.nextDouble()));
		}

		Graph graph = getGraph();
		for (int clusterNumber = 0; clusterNumber < clusterSizes.length; clusterNumber++) {			
			// create Nodes for each cluster
			for (int i = 0; i < clusterSizes[clusterNumber]; i++) {
				tmpNodes[nodeIndex] = graph.createNode();

				// create interClusterEdges with probability pIn
				while (wIn < i) {
					graph.createEdge(tmpNodes[nodeIndex], tmpNodes[nodeIndex - i
							+ wIn]);
					wIn = wIn + 1 + (int) (rIn * Math.log(random.nextDouble()));
				}
				wIn -= i;

				// create interClusterEdges with probability pOut
				if (wOut < 0) {
					wOut = wOut + nodeIndex - i;
				}
				while (wOut >= 0) {
					graph.createEdge(tmpNodes[nodeIndex], tmpNodes[wOut]);
					wOut = wOut - 1
							- (int) (rOut * Math.log(random.nextDouble()));
				}

				nodeIndex++;
			}
			// after adding nodes create new cluster with new nodes
			if(poolSize() > 0){
				this.newCluster(poolNodes());				
			}
		}
	}
	

	
	@Override
	public String getID() {
		return "RandomClusterGraphGenerator";
	}

}
