/*
 * Copyright 2006.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering.generators.utils;

import java.util.Random;
import java.util.Stack;

import de.uka.algo.math.RandomArray;

/**
 * This class provides generates several partitions which are used by the
 * {@link de.uka.algo.generators.randomCluster.RandomClusteredGraph RandomClustergraphGenerators}
 * 
 * @author Daniel Delling
 * 
 */

public class PartitionGenerator {

	/**
	 * generates an integer Equipartition
	 * 
	 * @param size
	 *            amount of entries in the partition
	 * @param min
	 *            minimum value of the partition
	 * @param max
	 *            maximum value of the partition
	 * @return the partition as integer array
	 */
	public static int[] equiPartition(int size, int min, int max) {
		Random random = new Random();
		int[] part = new int[size];
		for (int i = 0; i < part.length; i++) {
			part[i] = random.nextInt(max - min + 1) + min;
		}
		return part;
	}

	/**
	 * generates an integer gaussian Partition
	 * 
	 * @param size
	 *            amount of entries in the partition
	 * @param average
	 *            the average of gaussian
	 * @param deviation
	 *            the deviation of gaussian
	 * @return the partition as integer array
	 */
	public static int[] gaussPartition(int size, int average, int deviation) {
		Random random = new Random();
		double temp = 0;
		int[] part = new int[size];
		for (int i = 0; i < part.length; i++) {
			temp = random.nextGaussian();
			temp *= deviation;
			temp += average;
			part[i] = (int) temp;
		}

		return part;
	}

	
	/**
	 * generates an integer gaussian Partition with size uniform random between log_10(sum) and sqrt(sum).
	 * the average is floor(sum/size), the deviation is average/4.
	 * 
	 * @param sum
	 *            sum of all antries
	 * @return the partition as integer array
	 */
	public static int[] standardGaussPartition(int sum) {
		Random random = new Random();
		double temp = 0;
		int size = random.nextInt((int)(Math.floor(Math.sqrt(sum))+1-Math.floor(Math.log10(sum))))+(int)Math.floor(Math.log10(sum));
		int average = (int)Math.floor((double)sum/(double)size);
		int deviation = (int)Math.floor(average/4.0);
		int[] part = new int[size];
		for (int i = 0; i < part.length; i++) {
			temp = random.nextGaussian();
			temp *= deviation;
			temp += average;
			part[i] = (int) temp;
		}

		return part;
	}
	
	
	/**
	 * generates an integer MultiGaussian Partition
	 * 
	 * @param n
	 *            amount of entries in the partition
	 * @param average[]
	 *            the averages of gaussian
	 * @param deviation[]
	 *            the deviations of gaussian
	 * @return the partition as integer array
	 */
	public static int[] multiGaussPartition(int n, int[] average,
			int[] deviation) {
		Random random = new Random();
		double temp = 0;
		int[] part = new int[n];
		int j = 0;
		for (int i = 0; i < part.length; i++) {
			temp = random.nextGaussian();
			j = random.nextInt(average.length);
			temp *= deviation[j];
			temp += average[j];
			part[i] = (int) temp;
		}

		return part;
	}

	/**
	 * generates an random integerarray with values minimum min and maximum g.
	 * n/g random arrays of size g with values between 1 and g are construced.
	 * The entries of the partition equal all cyclelengths of the generated
	 * array
	 * 
	 * @param n
	 *            sum of all entries if min is 1
	 * @param g
	 *            maximum size of an entry
	 * @param min
	 *            minimum size of an entry
	 * @return
	 */
	public static int[] generatePermutationPartition(int n, int g, int min) {
		int k = n / g;
		int r = n - g * k;
		Stack<Integer> s = new Stack<Integer>();
		s.empty();

		RandomArray randomArray = new RandomArray();
		for (int i = 0; i < k; i++) {
			cycleLenghts(s, randomArray.nextArray(g));
		}

		if (r > 0) {
			cycleLenghts(s, randomArray.nextArray(r));
		}
		int[] part = new int[s.size()];
		for (int i = 0; i < part.length; i++) {
			part[i] = s.pop();
			if (part[i] < min) {
				part[i] = 0;
			}
		}
		return part;
	}

	/**
	 * calculates the cycles length of the given array and pushes the calculated
	 * length on the given stack
	 * 
	 * @param s
	 * @param array
	 */
	private static void cycleLenghts(Stack<Integer> s, int[] array) {
		boolean[] done = new boolean[array.length];
		int total = 0;
		int start = 0;
		int i = array[0];
		int l = 1;
		while (total < array.length) {
			while (done[start] == true) {
				start++;
			}
			i = array[start];
			l = 1;
			done[i] = true;
			while (i != start) {
				i = array[i];
				done[i] = true;
				l++;
			}
			start = 0;
			s.push(l);
			total += l;
		}
	}

}
