/*
 * Created on 07.10.2004
 * 
 * Copyright 2004.  --- All rights resevered. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering.index;

import y.base.Graph;
import de.uka.algo.clustering.Cluster;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.addon.Subgraphs;
import de.uka.algo.graphs.SimpleGraph;
import de.uka.algo.graphs.cut.ConductanceCut;
import de.uka.algo.graphs.cut.ConductanceCutter;
import de.uka.algo.graphs.cut.ConductanceCutter.ConductanceStrategies;
import de.uka.algo.util.GYCursor;

/**
 * The class IntraClusterConductance implements a clustering quality index
 * called inner-cluster-conductance. This index minimizes the ratio of a cut
 * cutting a cluster to the sum of inner edges in the smaller half over all cuts
 * and all clusters. The smaller this ratio, the worse the clustering (as there
 * is a cluster which can be cut into two large parts compared to the capacity
 * of the cut). <br>
 * Since minimizing the cut over all cuts in a given cluster is done by a
 * heuristic, this index might rate a clustering 'too good'. <br>
 * It is a singleton (for an instance of Clustering) being created on demand by
 * either the factory method or the factory class. If there is no reference to
 * this singleton instance, it may be garbage collected.
 * 
 * @author Bastian Katz
 */
public final class IntraClusterConductance extends
		ClusterBufferedIndex<ConductanceCut[]> {
	final static String INDEX_ID = "IntraClusterConductance";
	static String getID(ConductanceStrategies strategy) {
		return INDEX_ID+"("+strategy+")";
	}

	private Subgraphs subgraphs;
	private ConductanceStrategies strategy;
	private ConductanceCutter cc;

	protected IntraClusterConductance(Clustering clustering, ConductanceStrategies strategy) {
		super(clustering);
		subgraphs = Subgraphs.getInstance(clustering);
		clustering.addListener(this);
		cc = new ConductanceCutter(strategy, 1, false);
		this.strategy = strategy;
	}

	@Override
	protected ConductanceCut[] getData(Cluster c) {
		ConductanceCut[] result = new ConductanceCut[2];
		Graph subgraph = subgraphs.getGraph(c);
		result[0] = cc.getConductanceCut(subgraph, subgraphs.getInterpretation(c));

		result[1] = cc.getConductanceCut(subgraph, new SimpleGraph(subgraph));
		return result;
	}

	@Override
	protected double value(GYCursor<ConductanceCut[]> infos) {
		double result = 1.0;
		for (; infos.ok(); infos.next()) {
			if (infos.active()[0].getConductance()<result){
				result = infos.active()[0].getConductance(); 
			}
		}
		return result;
	}

	@Override
	protected double unweightedValue(GYCursor<ConductanceCut[]> infos) {
		double result = 1.0;
		for (; infos.ok(); infos.next()) {
			if (infos.active()[1].getConductance()<result){
				result = infos.active()[1].getConductance(); 
			}
		}
		return result;
	}

	@Override
	public String toString() {
		return getID(strategy);
	}
	
	public String id() {
		return getID(strategy);
	}

	
	public static class Factory implements IndexFactory {

		private ConductanceStrategies strategy;

		public Factory(ConductanceStrategies strategy) {
			this.strategy = strategy;
		}

		public Index getIndex(Clustering cl) {
			IntraClusterConductance icc;
			icc = (IntraClusterConductance) cl.getClusteringListener(id());
			if (icc == null) {
				return new IntraClusterConductance(cl, strategy);
			}
			return icc;
		}
		
		public String toString() {
			return getID(strategy);
		}
		
		public String id() {
			return getID(strategy);
		}

	}
	
	
}