/*
 * Created on May 28, 2008. 
 *  
 * Copyright 2008. --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany. 
 * 
 * Project: extension
 */
package de.uka.algo.clustering.linearprogram;

import ilog.concert.IloException;
import ilog.concert.IloNumVar;
import ilog.cplex.IloCplex;

import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

import y.base.Graph;
import de.uka.algo.clustering.Clustering;
import de.uka.algo.clustering.algorithms.Algorithm;
import de.uka.algo.graphs.GraphInterpretation;

/**
 * This class is a template for clustering algorithms that are based on (I)LP
 * solutations. The original method {@linkplain Algorithm#run(Clustering)} is
 * implemented as a wrapper for {@linkplain #run(Clustering, double[])} that
 * takes the clustering and a solution of the LP. The (I)LP is created via the
 * {@linkplain #createLP(Graph)}.
 * 
 * @author Marco Gaertler
 * 
 */
public abstract class CplexAlgorithm implements Algorithm {

	public static Logger logger = Logger.getLogger(CplexAlgorithm.class);

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.clustering.algorithms.Algorithm#run(de.uka.algo.clustering.Clustering)
	 */
	public void run(Clustering clustering) {
		// get graph
		Graph myGraph = clustering.getGraph();
		GraphInterpretation myGraphInt = clustering.getInterpretation();
		// create (I)LP;
		IloCplex myLP;
		Map<String, IloNumVar> name2variable = new HashMap<String, IloNumVar>();
		try {
			// create LP
			myLP = createLP(myGraph, myGraphInt, name2variable);
			// solve LP
			myLP.solve();
			// call run with the solution
			run(clustering, myLP, name2variable);
			// clean up license and memory 
			myLP.end();
		} catch (IloException e) {
			// if an exception occurs, send a warning to the Logger
			StringBuffer cache = new StringBuffer();
			cache.append("could not process (I)LP: " + e.toString());
			for (StackTraceElement ste : e.getStackTrace())
				cache.append(ste.toString());
			logger.warn(cache.toString());
			clustering.resetSingletons();
		}
	}

	/**
	 * Create the associated (I)LP to the given graph.
	 * 
	 * @param graph
	 *            corresponding graph
	 * @param name2variable
	 *            mapping of names to variables
	 * @return associated (I)LP
	 */
	protected abstract IloCplex createLP(Graph graph,
			GraphInterpretation graphInt, 
			Map<String, IloNumVar> name2variable) throws IloException;

	/**
	 * Performs the algorithm to change the clustering.
	 * 
	 * @param clustering
	 *            associated clustering
	 * @param name2variable
	 *            mapping of names to variables
	 */
	protected abstract void run(Clustering clustering, IloCplex cplex, 
			Map<String, IloNumVar> name2variable) throws IloException;

	
}
