/*
 * Created on May 05, 2008
 *
 * Copyright 2008.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering.linearprogram;

import ilog.concert.IloConstraint;
import ilog.concert.IloException;
import ilog.concert.IloLinearNumExpr;
import ilog.concert.IloNumVar;
import ilog.cplex.IloCplex;

import java.util.Map;

import lpsolve.LpSolveException;
import y.base.Graph;
import y.base.Node;

/**
 * This class provides basic methods to describe a clustering with linear
 * constraints using (binary) variables. More precisely, a clustering can be
 * modeled as equivalence relation (1 means that both end-nodes belong to the
 * same equivalence class) or as a (discrete) metric (0 means that both
 * end-nodes belong to the same cluster).
 * 
 * <b>Warning:</b> The graph is to be considered a static one, i.e., the
 * ordering of the nodes (obtained via
 * {@link Graph#getNodeArray() Graph#getNodeArray()}) may not change.
 * 
 * @author Marco Gaertler
 * 
 */
public class CplexGenerator {

	public static IloCplex generateEdgeVariables(Graph graph, boolean asILP, Map<String, IloNumVar> name2variable)
			throws IloException {
		if (graph == null)
			throw new IllegalArgumentException("cannot create an LP for an empty graph");
		if (name2variable == null)
			throw new IllegalArgumentException("cannot create an LP without storing name-variable relations");
		name2variable.clear();

		// creating new LP
		IloCplex myLP = new IloCplex();

		// get an ordering of the nodes
		Node vertices[] = graph.getNodeArray();

		// create \binom(n, 2) binary variables named x(i,j) with i\le j
		for (int i = 0; i < vertices.length; ++i)
			for (int j = i + 1; j < vertices.length; ++j) {
				String name = "X(" + i + "," + j + ")";
				IloNumVar var;
				if (asILP)
					var = myLP.boolVar(name);
				else
					var = myLP.numVar(0, 1, name);
				name2variable.put(name, var);
			}

		return myLP;
	}

	/**
	 * if no clustercount is predefined
	 * 
	 * @param graph
	 * @param asILP
	 * @param name2variable
	 * @param myLP
	 * @return
	 * @throws IloException
	 */
	public static IloCplex generateAdditionalNodeClusterAssocVariables(Graph graph, boolean asILP,
			Map<String, IloNumVar> name2variable, IloCplex myLP) throws IloException {
		return generateAdditionalNodeClusterAssocVariables(graph, asILP, name2variable, myLP, graph.N());
	}

	public static IloCplex generateAdditionalNodeClusterAssocVariables(Graph graph, boolean asILP,
			Map<String, IloNumVar> name2variable, IloCplex myLP, int maxClustCount) throws IloException {
		if (graph == null)
			throw new IllegalArgumentException("cannot add n-c-constraints to an LP for an empty graph");
		if (name2variable == null)
			throw new IllegalArgumentException("cannot add n-c-constraints without storing name-variable relations");
		if (name2variable.size() == 0)
			throw new IllegalArgumentException("cannot add n-c-constraints without having any name-variable relations");

		// get an ordering of the nodes
		Node vertices[] = graph.getNodeArray();
		int k = Math.min(maxClustCount, vertices.length);

		// create \binom(n, 2) binary variables named x(i,j) with i\le j
		for (int i = 0; i < graph.N(); ++i)
			for (int l = 0; l < k; ++l) {
				String name = "Y(" + i + "," + l + ")";
				IloNumVar var;
				if (asILP)
					var = myLP.boolVar(name);
				else
					var = myLP.numVar(0, 1, name);
				name2variable.put(name, var);
			}

		return myLP;
	}

	/**
	 * Returns an (I)LP with [0,1]-variables for each node pair that has
	 * transitivity as constraints set.
	 * 
	 * @param graph
	 *            given graph
	 * @param asILP
	 *            flag indicating, if the program should be an ILP
	 * @return an LP that models a clustering as equivalence relation
	 * @throws IloException
	 */
	public static IloCplex addEquivalenceClusteringConstraints(IloCplex myLP, Graph graph, boolean asILP,
			Map<String, IloNumVar> names2variables) throws IloException {

		// get an ordering of the nodes
		Node vertices[] = graph.getNodeArray();

		// add transitivity constraints
		IloLinearNumExpr constraint;
		for (int i = 0; i < vertices.length; ++i)
			for (int j = i + 1; j < vertices.length; ++j) {
				IloNumVar Xij = names2variables.get("X(" + i + "," + j + ")");
				for (int k = j + 1; k < vertices.length; ++k) {
					IloNumVar Xik = names2variables.get("X(" + i + "," + k + ")");
					IloNumVar Xjk = names2variables.get("X(" + j + "," + k + ")");
					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, 1);
					constraint.addTerm(Xik, -2);
					constraint.addTerm(Xjk, 1);
					myLP.addLe(constraint, 1.0);

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, 1);
					constraint.addTerm(Xik, 1);
					constraint.addTerm(Xjk, -2);
					myLP.addLe(constraint, 1.0);

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, -2);
					constraint.addTerm(Xik, 1);
					constraint.addTerm(Xjk, 1);
					myLP.addLe(constraint, 1.0);
				}
			}

		return myLP;
	}

	/**
	 * Returns an (I)LP with [0,1]-variables for each node pair that has
	 * transitivity as constraints set.
	 * 
	 * @param graph
	 *            given graph
	 * @param asILP
	 *            flag indicating, if the program should be an ILP
	 * @return an LP that models a clustering as equivalence relation
	 * @throws IloException
	 */
	public static IloCplex addTighterEquivalenceClusteringConstraints(IloCplex myLP, Graph graph, boolean asILP,
			Map<String, IloNumVar> names2variables) throws IloException {

		// get an ordering of the nodes
		Node vertices[] = graph.getNodeArray();

		// add transitivity constraints
		IloLinearNumExpr constraint;
		for (int i = 0; i < vertices.length; ++i)
			for (int j = i + 1; j < vertices.length; ++j) {
				IloNumVar Xij = names2variables.get("X(" + i + "," + j + ")");
				for (int k = j + 1; k < vertices.length; ++k) {
					IloNumVar Xik = names2variables.get("X(" + i + "," + k + ")");
					IloNumVar Xjk = names2variables.get("X(" + j + "," + k + ")");
					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, 1);
					constraint.addTerm(Xik, -1);
					constraint.addTerm(Xjk, 1);
					myLP.addLe(constraint, 1.0);

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, 1);
					constraint.addTerm(Xik, 1);
					constraint.addTerm(Xjk, -1);
					myLP.addLe(constraint, 1.0);

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, -1);
					constraint.addTerm(Xik, 1);
					constraint.addTerm(Xjk, 1);
					myLP.addLe(constraint, 1.0);
				}
			}

		return myLP;
	}

	/**
	 * Returns an (I)LP with [0,1]-variables for each node pair that has
	 * transitivity as constraints set.
	 * 
	 * @param graph
	 *            given graph
	 * @param asILP
	 *            flag indicating, if the program should be an ILP
	 * @return an LP that models a clustering as metric
	 * @throws LpSolveException
	 */
	public static IloCplex addMetricClusteringConstraints(IloCplex myLP, Graph graph, boolean asILP,
			Map<String, IloNumVar> names2variables) throws IloException {
		// get an ordering of the nodes
		Node vertices[] = graph.getNodeArray();

		// add constraints for triangle inequality
		IloLinearNumExpr constraint;
		for (int i = 0; i < vertices.length; ++i)
			for (int j = i + 1; j < vertices.length; ++j) {
				IloNumVar Xij = names2variables.get("X(" + i + "," + j + ")");
				for (int k = j + 1; k < vertices.length; ++k) {
					IloNumVar Xik = names2variables.get("X(" + i + "," + k + ")");
					IloNumVar Xjk = names2variables.get("X(" + j + "," + k + ")");

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, 1);
					constraint.addTerm(Xik, -1);
					constraint.addTerm(Xjk, 1);
					myLP.addGe(constraint, 0.0);

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, 1);
					constraint.addTerm(Xik, 1);
					constraint.addTerm(Xjk, -1);
					myLP.addGe(constraint, 0.0);

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, -1);
					constraint.addTerm(Xik, 1);
					constraint.addTerm(Xjk, 1);
					myLP.addGe(constraint, 0.0);
				}
			}

		return myLP;
	}

	/**
	 * Returns an (I)LP with [0,1]-variables for each node pair that has
	 * transitivity as constraints set - configured as user cuts, explicitly
	 * redundant constraints, that need not be checked by the ILP! Normally you
	 * should use the method without declaring constaints as user cuts.
	 * 
	 * @param graph
	 *            given graph
	 * @param asILP
	 *            flag indicating, if the program should be an ILP
	 * @return an LP that models a clustering as metric
	 * @throws LpSolveException
	 */
	public static IloCplex addUserCutMetricClusteringConstraints(IloCplex myLP, Graph graph, boolean asILP,
			Map<String, IloNumVar> names2variables) throws IloException {
		// get an ordering of the nodes
		Node vertices[] = graph.getNodeArray();

		// add constraints for triangle inequality
		IloLinearNumExpr constraint;
		int numOfUserCuts = vertices.length * (vertices.length-1) * (vertices.length-2) / 2;
		IloConstraint[] myUserCuts = new IloConstraint[numOfUserCuts];
		int userCutCounter = 0;
		for (int i = 0; i < vertices.length; ++i){
			for (int j = i + 1; j < vertices.length; ++j) {
				IloNumVar Xij = names2variables.get("X(" + i + "," + j + ")");
				for (int k = j + 1; k < vertices.length; ++k) {
					IloNumVar Xik = names2variables.get("X(" + i + "," + k + ")");
					IloNumVar Xjk = names2variables.get("X(" + j + "," + k + ")");

//					IloConstraint myUserCut;

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, 1);
					constraint.addTerm(Xik, -1);
					constraint.addTerm(Xjk, 1);
//					myUserCut = myLP.addGe(constraint, 0.0);
//					myLP.addUserCut(myUserCut);
					myUserCuts[userCutCounter++] = myLP.addGe(constraint, 0.0);
					
					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, 1);
					constraint.addTerm(Xik, 1);
					constraint.addTerm(Xjk, -1);
//					myUserCut = myLP.addGe(constraint, 0.0);
//					myLP.addUserCut(myUserCut);
					myUserCuts[userCutCounter++] = myLP.addGe(constraint, 0.0);

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, -1);
					constraint.addTerm(Xik, 1);
					constraint.addTerm(Xjk, 1);
//					myUserCut = myLP.addGe(constraint, 0.0);
//					myLP.addUserCut(myUserCut);
					myUserCuts[userCutCounter++] = myLP.addGe(constraint, 0.0);
				}
			}
		}
		System.out.println("Adding user cuts to the pool");
		myLP.addUserCuts(myUserCuts); // userCuts!
		System.out.println("Done with adding user cuts to the pool");		
		return myLP;
	}

	/**
	 * Returns an (I)LP with [0,1]-variables for each node pair that has
	 * transitivity as constraints set. The last of the three transitivity
	 * constraints each is declared lazy, since a vast number of these
	 * constraints will be redundant.
	 * 
	 * @param graph
	 *            given graph
	 * @param asILP
	 *            flag indicating, if the program should be an ILP
	 * @return an LP that models a clustering as metric
	 * @throws LpSolveException
	 */
	public static IloCplex addLazyMetricClusteringConstraints(IloCplex myLP, Graph graph, boolean asILP,
			Map<String, IloNumVar> names2variables) throws IloException {
		// get an ordering of the nodes
		Node vertices[] = graph.getNodeArray();

		// add constraints for triangle inequality
		IloLinearNumExpr constraint;
		int numOfLazies = vertices.length * (vertices.length-1) * (vertices.length-2) / 6;
		IloConstraint[] myLazyConstraints = new IloConstraint[numOfLazies];
		int lazyCounter = 0;
		for (int i = 0; i < vertices.length; ++i){
			for (int j = i + 1; j < vertices.length; ++j) {
				IloNumVar Xij = names2variables.get("X(" + i + "," + j + ")");
				for (int k = j + 1; k < vertices.length; ++k) {
					IloNumVar Xik = names2variables.get("X(" + i + "," + k + ")");
					IloNumVar Xjk = names2variables.get("X(" + j + "," + k + ")");

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, 1);
					constraint.addTerm(Xik, -1);
					constraint.addTerm(Xjk, 1);
					myLP.addGe(constraint, 0.0);

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, 1);
					constraint.addTerm(Xik, 1);
					constraint.addTerm(Xjk, -1);
					myLP.addGe(constraint, 0.0);

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, -1);
					constraint.addTerm(Xik, 1);
					constraint.addTerm(Xjk, 1);
//					IloConstraint myLazyConstraint = myLP.addGe(constraint, 0.0);
					myLazyConstraints[lazyCounter++] = myLP.addGe(constraint, 0.0);
//					System.out.println("lazy defined");
//					myLP.addLazyConstraint(myLazyConstraint); // lazy!
//					System.out.println("lazy added");
				}
//				System.out.println("inner loop done");
			}
//			System.out.println("outer loop done");
		}
		System.out.println("Adding lazy constraints to the pool");
		myLP.addLazyConstraints(myLazyConstraints); // lazy!
		System.out.println("Done with adding lazy constraints to the pool");
		return myLP;
	}

	/**
	 * Adds constraint for number of clusters if not given
	 * 
	 * @param myLP
	 * @param graph
	 * @param asILP
	 * @param names2variables
	 * @return
	 * @throws IloException
	 */
	public static IloCplex addPerClusterMetricClusteringConstraints(IloCplex myLP, Graph graph, boolean asILP,
			Map<String, IloNumVar> names2variables) throws IloException {
		return (addPerClusterMetricClusteringConstraints(myLP, graph, asILP, names2variables, graph.N()));
	}

	/**
	 * Returns an (I)LP with [0,1]-variables for each node pair, and additional
	 * [0,1] variables for node-cluster associations. Transitivity is guaranteed
	 * via triples node1, node2, cluster. An additional transitivity constraint
	 * as in addMetricClusteringConstraints is introduced as a user cut.
	 * 
	 * @param graph
	 *            given graph
	 * @param asILP
	 *            flag indicating, if the program should be an ILP
	 * @return an LP that models a clustering as metric
	 * @throws LpSolveException
	 */
	public static IloCplex addPerClusterMetricClusteringConstraints(IloCplex myLP, Graph graph, boolean asILP,
			Map<String, IloNumVar> names2variables, int maxClustCount) throws IloException {
		// get an ordering of the nodes
		Node vertices[] = graph.getNodeArray();
		int k = Math.min(maxClustCount, vertices.length);

		// add constraints for triangle inequality
		IloLinearNumExpr constraint;
		for (int i = 0; i < vertices.length; ++i)
			for (int j = i + 1; j < vertices.length; ++j) {
				IloNumVar Xij = names2variables.get("X(" + i + "," + j + ")");
				for (int l = 0; l < k; ++l) {
					IloNumVar Yil = names2variables.get("Y(" + i + "," + l + ")");
					IloNumVar Yjl = names2variables.get("Y(" + j + "," + l + ")");

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, 1);
					constraint.addTerm(Yil, -1);
					constraint.addTerm(Yjl, 1);
					myLP.addGe(constraint, 0.0);

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, 1);
					constraint.addTerm(Yil, 1);
					constraint.addTerm(Yjl, -1);
					myLP.addGe(constraint, 0.0);

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, -1);
					constraint.addTerm(Yil, 1);
					constraint.addTerm(Yjl, 1);
					// IloConstraint myLazyConstraint =
					myLP.addGe(constraint, 0.0);
					// myLP.addLazyConstraint(myLazyConstraint);
				}
			}
		
		//add constraints that no node is in more than one cluster!
		for (int i = 0; i < vertices.length; ++i){
			constraint = myLP.linearNumExpr();
			for (int l = 0; l < k; ++l) {
				IloNumVar Yil = names2variables.get("Y(" + i + "," + l + ")");
				constraint.addTerm(Yil, 1);
			}
			myLP.addEq(constraint, k-1);
		}

		return myLP;
	}

	/**
	 * Adds constraint for number of clusters if not given
	 * 
	 * @param myLP
	 * @param graph
	 * @param asILP
	 * @param names2variables
	 * @return
	 * @throws IloException
	 */
	public static IloCplex addLazyPerClusterMetricClusteringConstraints(IloCplex myLP, Graph graph, boolean asILP,
			Map<String, IloNumVar> names2variables) throws IloException {
		return (addLazyPerClusterMetricClusteringConstraints(myLP, graph, asILP, names2variables, graph.N()));
	}

	/**
	 * As addPerClusterMetricClusteringConstraints but with one constraint lazy
	 * 
	 * @param graph
	 *            given graph
	 * @param asILP
	 *            flag indicating, if the program should be an ILP
	 * @return an LP that models a clustering as metric
	 * @throws LpSolveException
	 */
	public static IloCplex addLazyPerClusterMetricClusteringConstraints(IloCplex myLP, Graph graph, boolean asILP,
			Map<String, IloNumVar> names2variables, int maxClustCount) throws IloException {
		// get an ordering of the nodes
		Node vertices[] = graph.getNodeArray();
		int k = Math.min(maxClustCount, vertices.length);

		// add constraints for triangle inequality
		IloLinearNumExpr constraint;
		int numOfLazies = vertices.length * (vertices.length-1) / 2 * k;
		IloConstraint[] myLazyConstraints = new IloConstraint[numOfLazies];
		int lazyCounter = 0;
		for (int i = 0; i < vertices.length; ++i){
			for (int j = i + 1; j < vertices.length; ++j) {
				IloNumVar Xij = names2variables.get("X(" + i + "," + j + ")");
				for (int l = 0; l < k; ++l) {
					IloNumVar Yil = names2variables.get("Y(" + i + "," + l + ")");
					IloNumVar Yjl = names2variables.get("Y(" + j + "," + l + ")");

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, 1);
					constraint.addTerm(Yil, -1);
					constraint.addTerm(Yjl, 1);
					myLP.addGe(constraint, 0.0);

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, 1);
					constraint.addTerm(Yil, 1);
					constraint.addTerm(Yjl, -1);
					myLP.addGe(constraint, 0.0);

					constraint = myLP.linearNumExpr();
					constraint.addTerm(Xij, -1);
					constraint.addTerm(Yil, 1);
					constraint.addTerm(Yjl, 1);
//					IloConstraint myLazyConstraint = myLP.addGe(constraint, 0.0);
					myLazyConstraints[lazyCounter++] = myLP.addGe(constraint, 0.0);
//					System.out.println("lazy defined");
//					myLP.addLazyConstraint(myLazyConstraint); // lazy!
//					System.out.println("lazy added");
				}
//				System.out.println("inner loop done");
			}
//			System.out.println("outer loop done");
		}
		System.out.println("Adding lazy constraints to the pool");
		myLP.addLazyConstraints(myLazyConstraints); // lazy!
		System.out.println("Done with adding lazy constraints to the pool");
		
		//add constraints that no node is in more than one cluster!
		for (int i = 0; i < vertices.length; ++i){
			constraint = myLP.linearNumExpr();
			for (int l = 0; l < k; ++l) {
				IloNumVar Yil = names2variables.get("Y(" + i + "," + l + ")");
				constraint.addTerm(Yil, 1);
			}
			myLP.addEq(constraint, k-1);
		}
		
		return myLP;
	}

	/**
	 * Inserts the number of MaxClusters if not given as n
	 * 
	 * @param myLP
	 * @param graph
	 * @param asILP
	 * @param names2variables
	 * @return
	 * @throws IloException
	 */
	public static IloCplex breakMetricNodeClusterSymmetries(IloCplex myLP, Graph graph, boolean asILP,
			Map<String, IloNumVar> names2variables) throws IloException {
		return breakMetricNodeClusterSymmetries(myLP, graph, asILP, names2variables, graph.N());
	}

	/**
	 * Breaks Symmetries of Yil-Variables by ensuring that Yil==1 for all l>i
	 * 
	 * @param myLP
	 * @param graph
	 * @param asILP
	 * @param names2variables
	 * @param maxClustCount
	 * @return
	 * @throws IloException
	 */
	public static IloCplex breakMetricNodeClusterSymmetries(IloCplex myLP, Graph graph, boolean asILP,
			Map<String, IloNumVar> names2variables, int maxClustCount) throws IloException {
		// get an ordering of the nodes
		Node vertices[] = graph.getNodeArray();
		int k = Math.min(maxClustCount, vertices.length);
		IloLinearNumExpr constraint;

		//Node must not be in cluster of smaller index! i.e.: $\forall v_i \in C_l: i \ge l$
		for (int i = 0; i < vertices.length; ++i)
			for (int l = 0; (l < k) && (l <= i-1); ++l) {
				IloNumVar Yil = names2variables.get("Y(" + i + "," + l + ")");
				constraint = myLP.linearNumExpr();
				constraint.addTerm(Yil, 1);
				myLP.addEq(constraint, 1.0);
			}

		return myLP;
	}

	/**
	 * Adds a objective function to a given LP. The objective function is given
	 * through an interface that provides the required coefficients per node
	 * pair.
	 * 
	 * @param myLP
	 *            given LP
	 * @param graph
	 *            associated graph
	 * @param objfct
	 *            objective function defined by
	 *            {@linkplain GraphObjectiveFunction}
	 * @param asMinimization
	 *            flag to indicate whether it is a minimization or maximization
	 *            problem
	 * @param names2variables
	 *            map that stores the name-variables relation
	 * @return the associated LP object
	 * @throws IloException
	 */
	public static IloCplex addObjectiveFunction(IloCplex myLP, Graph graph, GraphObjectiveFunction objfct,
			boolean asMinimization, Map<String, IloNumVar> names2variables) throws IloException {
		// get an ordering of nodes
		Node nodes[] = graph.getNodeArray();

		// create objective function
		IloLinearNumExpr objectiveFunction = myLP.linearNumExpr();
		// assigning coefficients based on objfct
		for (int i = 0; i < nodes.length; ++i)
			for (int j = i + 1; j < nodes.length; ++j) {
				IloNumVar var = names2variables.get("X(" + i + "," + j + ")");
				objectiveFunction.addTerm(var, objfct.getCoefficient(nodes[i], nodes[j]));
			}

		// set objective function and defined if the LP is a minimization or
		// maximization problem
		// note that metricModel(variables) = asMinimization(objective)
		if (asMinimization)
			myLP.addMinimize(objectiveFunction);
		else
			myLP.addMaximize(objectiveFunction);

		return myLP;
	}
}
