/*
 * Created on May 05, 2008
 *
 * Copyright 2008.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.clustering.linearprogram;

import y.base.Graph;
import y.base.Node;
import lpsolve.LpSolve;
import lpsolve.LpSolveException;

/**
 * This class provides basic methods to describe a clustering with linear
 * constraints using (binary) variables. More precisely, a clustering can be
 * modeled as equivalence relation (1 means that both end-nodes belong to the
 * same equivalence class) or as a (discrete) metric (0 means that both
 * end-nodes belong to the same cluster).
 * 
 * <b>Warning:</b> The graph is to be considered a static one, i.e., the
 * ordering of the nodes (obtained via
 * {@link Graph#getNodeArray() Graph#getNodeArray()}) may not change.
 * 
 * @author Marco Gaertler
 * 
 */
public class LPSolveGenerator {

	/**
	 * Returns the linearized position of the (i,j)-th entry in a
	 * triangle-matrix.
	 * 
	 * @param i
	 *            row position
	 * @param j
	 *            column position
	 * @param n
	 *            total size
	 * @return linearized position
	 */
	public static int getPos(int i, int j, int n) {
		return i * n + j - (i + 1) * (i + 2) / 2;
	}

	/**
	 * Returns an (I)LP with [0,1]-variables for each node pair that has
	 * transitivity as constraints set.
	 * 
	 * @param graph
	 *            given graph
	 * @param asILP
	 *            flag indicating, if the program should be an ILP
	 * @return an LP that models a clustering as equivalence relation
	 * @throws LpSolveException
	 */
	public static LpSolve makeEquivalenceClustering(Graph graph, boolean asILP)
			throws LpSolveException {
		// get number of nodes
		int n = graph.nodeCount();
		// get number of variables
		int numberOfVariables = n * (n - 1) / 2;
		// get an ordering of the nodes
		Node nodes[] = graph.getNodeArray();
		// create new LP with the corresponding number of variables.
		LpSolve myLP = LpSolve.makeLp(0, numberOfVariables);

		// create array for constraints and range constraints
		double constraint[] = new double[numberOfVariables + 1];
		for (int i = 0; i < constraint.length; ++i) {
			constraint[i] = 0;
			if (i > 0) {
				// set range to [0, 1]
				myLP.setUpbo(i, 1.0);
				myLP.setLowbo(i, 0.0);
				// set binary, if a ILP is required
				if (asILP)
					myLP.setInt(i, true);
			}
		}

		// create transitivity constraints
		for (int i = 0; i < nodes.length; ++i)
			for (int j = i + 1; j < nodes.length; ++j)
				for (int k = j + 1; k < nodes.length; ++k) {
					// get the linearized positions of the pairs
					int ij = getPos(i, j, n) + 1;
					int ik = getPos(i, k, n) + 1;
					int jk = getPos(j, k, n) + 1;

					constraint[ij] = 1;
					constraint[ik] = -2;
					constraint[jk] = 1;
					myLP.addConstraint(constraint, LpSolve.LE, 1.0);
					constraint[ij] = 1;
					constraint[ik] = 1;
					constraint[jk] = -2;
					myLP.addConstraint(constraint, LpSolve.LE, 1.0);
					constraint[ij] = -2;
					constraint[ik] = 1;
					constraint[jk] = 1;
					myLP.addConstraint(constraint, LpSolve.LE, 1.0);

					// resets entries in the array for reuse
					constraint[ij] = 0;
					constraint[ik] = 0;
					constraint[jk] = 0;
				}
		// return (I)LP
		return myLP;
	}

	/**
	 * Returns an (I)LP with [0,1]-variables for each node pair that has
	 * transitivity as constraints set.
	 * 
	 * @param graph
	 *            given graph
	 * @param asILP
	 *            flag indicating, if the program should be an ILP
	 * @return an LP that models a clustering as metric
	 * @throws LpSolveException
	 */
	public static LpSolve makeMetricClustering(Graph graph, boolean asILP)
			throws LpSolveException {
		// get number of nodes
		int n = graph.nodeCount();
		// get number of variables
		int numberOfVariables = n * (n - 1) / 2;
		// get an ordering of nodes
		Node nodes[] = graph.getNodeArray();
		// create new LP with given number of variables
		LpSolve myLP = LpSolve.makeLp(0, numberOfVariables);

		// create array for constraints and range constraints
		double constraint[] = new double[numberOfVariables + 1];
		for (int i = 0; i < constraint.length; ++i) {
			constraint[i] = 0;
			if (i > 0) {
				// set range to [0, 1]
				myLP.setUpbo(i, 1.0);
				myLP.setLowbo(i, 0.0);
				// set binary, if a ILP is required
				if (asILP)
					myLP.setInt(i, true);
			}
		}

		// create triangle-inequality constraints
		for (int i = 0; i < nodes.length; ++i)
			for (int j = i + 1; j < nodes.length; ++j)
				for (int k = j + 1; k < nodes.length; ++k) {
					// get linearized positions
					int ij = getPos(i, j, n) + 1;
					int ik = getPos(i, k, n) + 1;
					int jk = getPos(j, k, n) + 1;

					// set triangle-inequality constraints
					constraint[ij] = 1;
					constraint[ik] = -1;
					constraint[jk] = 1;
					myLP.addConstraint(constraint, LpSolve.GE, 0.0);
					constraint[ij] = 1;
					constraint[ik] = 1;
					constraint[jk] = -1;
					myLP.addConstraint(constraint, LpSolve.GE, 0.0);
					constraint[ij] = -1;
					constraint[ik] = 1;
					constraint[jk] = 1;
					myLP.addConstraint(constraint, LpSolve.GE, 0.0);

					// resets entries in the array for reuse
					constraint[ij] = 0;
					constraint[ik] = 0;
					constraint[jk] = 0;
				}
		// return (I)LP
		return myLP;
	}

	/**
	 * Adds a objective function to a given LP. The objective function is given
	 * through an interface that provides the required coefficients per node
	 * pair.
	 * 
	 * @param myLP
	 *            given LP
	 * @param graph
	 *            associated graph
	 * @param objfct
	 *            objective function defined by
	 *            {@linkplain GraphObjectiveFunction}
	 * @param asMinimization
	 *            flag to indicate whether it is a minimization or maximization
	 *            problem
	 * @return the associated LP object
	 */
	public static LpSolve addObjectiveFunction(LpSolve myLP, Graph graph,
			GraphObjectiveFunction objfct, boolean asMinimization)
			throws LpSolveException {
		// get number of nodes
		int n = graph.nodeCount();
		// get number of variables
		int numberOfVariables = n * (n - 1) / 2;
		// get an ordering of nodes
		Node nodes[] = graph.getNodeArray();
		// declaring array that store all coefficients of the objective function
		double coefficients[] = new double[numberOfVariables + 1];
		// initializing the array
		for (int i = 0; i < coefficients.length; ++i)
			coefficients[i] = 0;
		// copy coefficient form objfct
		for (int i = 0; i < nodes.length; ++i)
			for (int j = i + 1; j < nodes.length; ++j) {
				int ij = getPos(i, j, n) + 1;
				coefficients[ij] = objfct.getCoefficient(nodes[i], nodes[j]);
			}
		// set objective function
		myLP.setObjFn(coefficients);
		// set type of optimization
		if (asMinimization)
			myLP.setMinim();
		else
			myLP.setMaxim();
		// return (I)LP
		return myLP;
	}

}
