package de.uka.algo.graphs.centralities.betweenness;


import java.util.Stack;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.EdgeMap;
import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeList;
import y.base.NodeMap;
import de.uka.algo.graphs.centralities.betweenness.sampling.ExactSampler;
import de.uka.algo.graphs.centralities.betweenness.sampling.Sampler;
import de.uka.algo.graphs.traverse.bfs.BFS;
import de.uka.algo.graphs.traverse.bfs.BFSModule;
import de.uka.algo.graphs.traverse.bfs.BFS.EdgeType;

/**
 * Calculate the betweenness centrality using the breadth first search
 * There are two ways to calculate the edge- or node betweenness:
 * I) Add this module to a BFS instace and run the search once for every
 *    node as root. After that the centralities are returned by
 *    <code>getNodeCentrality()</code> and <code>getEdgeCentrality()</code>
 * II) Use one of the following methods: <code>calcNodeBetweenness()</code>
 *     and <code>calcEdgeBetweenness()</code>.
 * 
 * @author Ulrich Geilmann
 */
public class BetweennessBFS implements BFSModule {
	
	/**
	 * Number of shortest paths starting at the root-node for every node in the graph
	 */
	private NodeMap sp;
	
	/**
	 * List of predecessors on shortest paths for every node in the graph
	 */
	private NodeMap pred;
	
	/**
	 * Nodes ordered by their distance to the root node
	 */
	private Stack<Node> stack;
	
	/**
	 * The calculated dependencies of the root node to the other nodes in the graph
	 */
	private NodeMap nodeDep;

	/**
	 * The dependencies of the root node to the edges in the graph
	 */
	private EdgeMap edgeDep;
	
	/**
	 * The betweenness-centrality for the edges and nodes
	 */
	private NodeMap nodeCentrality;
	private EdgeMap edgeCentrality;
	private double edgeSum;
	private double nodeSum;

	/**
	 * Whether the centralities should be normalized or not
	 */
	private boolean normalize;
	
	/**
	 * The graph the BFS is performed in
	 */
	private Graph graph;
	
	/**
	 * Standard Constructor
	 * @param norm
	 * 	Whether the centralities should be normalized 
	 */
	public BetweennessBFS (boolean norm) {
		graph = null;
		normalize = norm;
	}
	
	/**
	 * When the BFS has been run once with every node as root,
	 * this method will return the correct node betweenness centrality
	 * @return the betweenness centrality for every node
	 */
	public NodeMap getNodeCentrality() {
		if (normalize) {
			divAll (nodeCentrality, nodeSum);
			nodeSum = 0;
		}
		return nodeCentrality;
	}
	
	/**
	 * When the BFS has been run once with every node as root,
	 * this method will return the correct node betweenness centrality
	 * @return the betweenness centrality for the given node
	 */
	public double getNodeCentrality(Node n) {
		double c = nodeCentrality.getDouble(n);
		if (normalize && nodeSum != 0)
			c = c / nodeSum;
		return c;
	}

	/**
	 * When the BFS has been run once with every node as root,
	 * this method will return the correct edge betweenness centrality
	 * @return the betweenness-centrality for every edge
	 */
	public EdgeMap getEdgeCentrality() {
		if (normalize) {
			divAll (nodeCentrality, edgeSum);
			edgeSum = 0;
		}
		return edgeCentrality;
	}
	
	/**
	 * When the BFS has been run once with every node as root,
	 * this method will return the correct edge betweenness centrality
	 * @return the betweenness-centrality for the given edge
	 */
	public double getEdgeCentrality(Edge e) {
		double c = edgeCentrality.getDouble(e);
		if (normalize && edgeSum != 0)
			c = c / edgeSum;
		return c;
	}

	/**
	 * (Re-)Initializing data 
	 */
	public void init ( Graph g ) {
		cleanUp();
		graph = g;
		sp = graph.createNodeMap();
		pred = graph.createNodeMap();
		nodeDep = graph.createNodeMap();
		edgeDep = graph.createEdgeMap();
		nodeCentrality = graph.createNodeMap();
		edgeCentrality = graph.createEdgeMap();
		for (NodeCursor c = graph.nodes(); c.ok(); c.next()) {
			nodeCentrality.setDouble(c.node(), 0);
		}
		for (EdgeCursor c = graph.edges(); c.ok(); c.next()) {
			edgeCentrality.setDouble(c.edge(), 0);
		}
		stack = new Stack<Node>();
		edgeSum = 0;
		nodeSum = 0;
	}
	
	/**
	 * Preparing for the search
	 */
	public void root ( Node s ) {
		stack.clear();
		for (NodeCursor c = graph.nodes(); c.ok(); c.next()) {
			nodeDep.setDouble(c.node(), 0);
			pred.set(c.node(), new NodeList());
		}
		for (EdgeCursor c = graph.edges(); c.ok(); c.next())
			edgeDep.setDouble(c.edge(), 0);
		sp.setInt(s, 1);
	}
	
	/**
	 * Updating sp and pred, filling the stack
	 */
	public void traverse ( Node v, int v_bfs, Edge e, EdgeType marking, Node w, int w_bfs ) {
		if (marking == EdgeType.TreeEdge) {
			addPred (w, v);
			sp.setInt(w, sp.getInt(v));
			stack.push(w);
		} else if (marking == EdgeType.ForwardEdge) {
			addPred ( w, v );
			sp.setInt(w, sp.getInt(w) + sp.getInt(v));
		}
	}
	
	/**
	 * Empty the stack, calculating the dependencies, updating the centralities
	 */
	public void done ( Node r ) {
		while ( !stack.empty() ) {
			Node w = stack.pop();
			NodeList l = (NodeList)pred.get(w);
			for (NodeCursor nc = l.nodes(); nc.ok(); nc.next()) {
				Node v = nc.node();
				double d = sp.getInt(v)*(1 + nodeDep.getDouble(w)) / sp.getInt(w);
				nodeDep.setDouble(v, nodeDep.getDouble(v) + d);
				for (EdgeCursor c = v.edges(); c.ok(); c.next()) {
					Edge e = c.edge();
					if ( e.opposite(v) == w )
						edgeDep.setDouble(e, d);						
				}
			}
		}
		/* Updating the centralities */
		for (NodeCursor c = graph.nodes(); c.ok(); c.next())
			if (c.node() != r) {
				nodeCentrality.setDouble(c.node(), nodeCentrality.getDouble(c.node()) + nodeDep.getDouble(c.node()));
				nodeSum += nodeDep.getDouble(c.node());
			}
		for (EdgeCursor c = graph.edges(); c.ok(); c.next()) {
			edgeCentrality.setDouble(c.edge(), edgeCentrality.getDouble(c.edge()) + edgeDep.getDouble(c.edge()));
			edgeSum += edgeDep.getDouble(c.edge());
		}
	}
	
	/**
	 * Add a Node to a list in pred
	 * @param key
	 * 	The node the list is associated with in pred
	 * @param n
	 * 	The node to add
	 */
	private void addPred ( Node key, Node n ) {
		NodeList l = (NodeList)pred.get(key);
		l.add (n);
	}
	
	/**
	 * Divide all (double-)entries in a NodeMap
	 * @param map
	 * 	The map whose entries should be divided
	 * @param div
	 * 	The divisor
	 */
	private void divAll (NodeMap map, double div) {
		if (div == 0) return;
		for (NodeCursor c = graph.nodes(); c.ok(); c.next()) {
			map.setDouble(c.node(), map.getDouble(c.node())/div);
		}
	}
	/**
	 * Divide all (double-)entries in an EdgeMap
	 * @param map
	 * 	The map whose entries should be divided
	 * @param div
	 * 	The divisor
	 */
	private void divAll (EdgeMap map, double div) {
		if (div == 0) return;
		for (EdgeCursor c = graph.edges(); c.ok(); c.next()) {
			map.setDouble(c.edge(), map.getDouble(c.edge())/div);
		}
	}
	
	/**
	 * Cleaning up the mess of previously run searches
	 */
	public void cleanUp() {
		/* Destroy the NodeMaps within the Graph */
		if (graph != null) {
			if ( nodeDep != null ) {
				graph.disposeNodeMap(nodeDep);
				nodeDep = null;
			}
			if ( edgeDep != null ) {
				graph.disposeEdgeMap(edgeDep);
				edgeDep = null;
			}
			if ( pred != null ) {
				graph.disposeNodeMap(pred);
				pred = null;
			}
			if ( sp != null ) {
				graph.disposeNodeMap(sp);
				sp = null;
			}
			if ( nodeCentrality != null ) {
				graph.disposeNodeMap(nodeCentrality);
				nodeCentrality = null;
			}
			if ( edgeCentrality != null ) {
				graph.disposeEdgeMap(edgeCentrality);
				edgeCentrality = null;
			}
			graph = null;
		}
	}
	
	/**
	 * Cleaning up
	 * @see java.lang.Object#finalize()
	 */
	@Override
	protected void finalize() throws Throwable {
		cleanUp();
		super.finalize();
	}

	/**
	 * Calculate the betweenness-centrality for the nodes of the given
	 * graph. This method will run the BFS once with every node as root
	 * @param map
	 * 	the centralities will be stored in this map. It does not need
	 * 	to be initialized
	 */
	public void calcNodeBetweenness (Graph graph, NodeMap map) {
		calcNodeBetweenness (graph, map, null);
	}
	
	/**
	 * Calculate the betweenness-centrality for the edges of the given
	 * graph. This method will run the BFS once with every node as root
	 * @param map
	 * 	the centralities will be stored in this map. It does not need
	 *  to be initialized
	 */
	public void calcEdgeBetweenness (Graph graph, EdgeMap map) {
		calcEdgeBetweenness(graph, map, null);
	}
	
	/**
	 * Calculate an approximation of the betweenness centrality for the
	 * nodes of the given graph. The method will run the BFS with the
	 * nodes chosen by the sampler as root
	 * @param map
	 * 	the approximated centralities will be stored in this map. It does
	 * 	not need to be initialized
	 */
	public void calcNodeBetweenness (Graph graph, NodeMap map, Sampler sampler) {
		if ( graph == null ) return;
		if (sampler == null)
			sampler = new ExactSampler();
		BFS bfs = new BFS ( graph );
		bfs.addModule (this);
		
		NodeMap back = nodeCentrality;
		nodeCentrality = map;
		for (NodeCursor c = graph.nodes(); c.ok(); c.next())
			nodeCentrality.setDouble(c.node(), 0);
		
		for (NodeCursor nc = sampler.sample(graph).nodes(); nc.ok(); nc.next())
			bfs.run (nc.node());
		if (normalize)
			divAll (nodeCentrality, nodeSum);
		
		nodeCentrality = back;
	}
	
	/**
	 * Calculate an approximation of the betweenness centrality for the
	 * edges of the given graph. The method will run the BFS with the
	 * nodes chosen by the sampler as root
	 * @param map
	 * 	the approximated centralities will be stored in this map. It does
	 * 	not need to be initialized
	 */
	public void calcEdgeBetweenness (Graph graph, EdgeMap map, Sampler sampler) {
		if (graph == null) return;
		if (sampler == null)
			sampler = new ExactSampler();
		BFS bfs = new BFS (graph);
		bfs.addModule (this);
		
		EdgeMap back = edgeCentrality;
		edgeCentrality = map;
		for (EdgeCursor c = graph.edges(); c.ok(); c.next())
			edgeCentrality.setDouble(c.edge(), 0);
		
		for (NodeCursor c = sampler.sample(graph).nodes(); c.ok(); c.next())
			bfs.run(c.node());
		if (normalize)
			divAll (edgeCentrality, edgeSum);
		
		edgeCentrality = back;
	}
	

}
