/**
 * Created on 29.11.2007
 */
package de.uka.algo.graphs.centralities.betweenness.sampling;

import java.util.Random;

import y.base.Graph;
import y.base.Node;
import y.base.NodeList;

/**
 * Choose nodes randomly where the probability of a node to be chosen
 * is proportional to its degree.
 * The number of nodes to be chosen is either a fixed number or a
 * fixed fraction of the nodes in the graph.
 * 
 * @author Ulrich Geilmann
 *
 */
public class DegreeSampler implements Sampler {
	
	private Random random;
	private int n;
	private double f;

	/**
	 * Choose a fixed number of nodes
	 * @param n
	 * 	The number of nodes to choose
	 */
	public DegreeSampler (int n) {
		random = new Random();
		this.n = n;
		f = -1;
	}
	
	/**
	 * Choose a fixed fraction of the nodes in the graph
	 * @param f
	 * 	The fraction of nodes to choose
	 */
	public DegreeSampler (double f) {
		random = new Random();
		this.f = f;
		n = -1;
	}

	/*
	 * @see de.uka.algo.clustering.betweenness.util.sampling.Sampler#sample(y.base.Graph)
	 */
	public NodeList sample(Graph g) {
		// TODO: With f > 0.5, random choose the nodes NOT being sampled
		
		/* Determine how many nodes to choose */
		int z = n;
		if (f != -1) {
			z = (int)(g.nodeCount() * f + 0.5);
			if (z <= 0)
				z = 1;
		}
		
		/* Set up an array to guarantee the proportionality to the nodes' degree */
		Node[] nodes = g.getNodeArray();
		int m = nodes.length;
		int[] s = new int[m];
		int sum = 0;
		for (int i = 0; i < nodes.length; i++) {
			sum += nodes[i].degree();
			s[i] = sum;
		}
		
		/* Choose nodes, abort when all nodes have degree 0 */
		NodeList list = new NodeList ();
		while (z > 0 && m > 0 && sum > 0) {
			int r = random.nextInt(sum);
			/* Calculate the to r corresponding element in */
			int p = binarySearch (s, m, r);
			list.add (nodes[p]);
			/* Recreate s to prevent a node to be chosen more than one time */
			int degree = nodes[p].degree(); 
			sum -= degree;
			m--; z--;
			for (int i = p; i < m; i++) {
				s[i] = s[i+1] - degree;
				nodes[i] = nodes[i+1];
			}
			
		}
		
		return list;
	}

	/**
	 * Perform binary search
	 * @param s
	 * 	The array to search in. The elements in the array must be in ascending order.
	 * @param l
	 * 	The length of the array: l <= s.length
	 * @param k
	 *	The key to search for. The key must not be bigger than the largest element in the array.
	 * @return
	 * 	the minimal index i with s[i] >= k
	 */
	private int binarySearch (int[] s, int l, int k) {
		int left = 0;
		int right = l-1;
		while (left != right) {
			int p = (left+right)/2;
			if (s[p] < k) {
				/* Reject the part left to p including p */
				left = p + 1;
			} else {
				/* Reject the part right to p excluding p */
				right = p;
			}
		}
		return left;
	}
}
