/*
 * Created on May 22, 2004 
 * Reviewed 1.16 (Mar 27, 2005)
 *
 * Copyright 2004.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.graphs.cut;

import y.algo.GraphConnectivity;
import y.base.EdgeCursor;
import y.base.Graph;
import y.base.Node;
import y.base.NodeList;
import de.uka.algo.graphs.GraphInterpretation;
import de.uka.algo.graphs.SimpleGraph;
import de.uka.algo.graphs.linalg.GraphMatrix;
import de.uka.algo.graphs.linalg.LaplacianGraphMatrix;
import de.uka.algo.graphs.linalg.MarkovianGraphMatrix;

/**
 * The class ConductanceCutter implements algorithms finding cuts with low
 * conductance heuristically.
 * 
 * @author Bastian Katz
 */
public class ConductanceCutter extends GraphCutter {

	public enum ConductanceStrategies {
		/**
		 * strategy finding the cut using a node order induced by the largest
		 * non-trivial eigenvalue of the normalized adjacency matrix.
		 */
		Markovian,
		/**
		 * strategy finding the cut using a node order induced by the smalles
		 * non-trivial eigenvalue of the Laplacian matrix.
		 */
		Laplacian
	}

	/** the chosen strategy */
	private ConductanceStrategies strategy = ConductanceStrategies.Markovian;

	/** cutting mode (see below) */
	private boolean complex = false;

	/** the threshold (see below) */
	private double threshold;

	/**
	 * Instantiates a Cutter calculating the minimum conductance cut according
	 * to a certain strategy. Nodes are ordered by a given strategy, the
	 * returned cut has a conductance below a given threshold and is minimal
	 * under those with the following properties: It is an interval of the node
	 * order, starting with the first node in the simple and anywhere in the
	 * complex node. I.e. it is minimal under O(|V|) cuts in the simple and
	 * O(|V|^2) cuts in the complex mode. If no cut has a conductance below the
	 * threshold, the cut, i.e. its set of inner nodes, is empty.
	 * 
	 * @param strategy
	 *            node-ordering strategy
	 * @param threshold
	 *            threshold value
	 * @param complex
	 *            true for complex mode
	 */
	public ConductanceCutter(ConductanceStrategies strategy, double threshold,
			boolean complex) {
		this.strategy = strategy;
		this.complex = complex;
		this.threshold = threshold;
	}

	/**
	 * Same as ConductanceCutter(int strategy, double threshold, boolean
	 * complex), but only simple mode.
	 * 
	 * @param strategy
	 *            node-ordering strategy
	 * @param threshold
	 *            threshold value
	 */
	public ConductanceCutter(ConductanceStrategies strategy, double threshold) {
		this(strategy, threshold, false);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.graphs.cut.Cutter#getCut(y.base.Graph,
	 *      de.uka.algo.graphs.GraphInterpretation)
	 */
	@Override
	public Cut getCut(Graph aGraph, GraphInterpretation anInterpretation) {
		return getConductanceCut(aGraph, anInterpretation);
	}

	/**
	 * Performs the algorithm to find a cut of the give graph
	 * 
	 * @param aGraph
	 *            the graph
	 * @param anInterpretation
	 *            the interpretation
	 * @return the cut
	 */
	public ConductanceCut getConductanceCut(Graph aGraph,
			GraphInterpretation anInterpretation) {
		return getCut(aGraph, anInterpretation, strategy, threshold, complex);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.graphs.cut.Cutter#getCut(y.base.Graph)
	 */
	@Override
	public Cut getCut(Graph graph) {
		return getCut(graph, new SimpleGraph(graph));
	}

	/**
	 * Basic method to find cuts given the graph with an interpretation a
	 * node-ordering strategy and a threshold
	 * 
	 * @param aGraph
	 *            the graph
	 * @param anInterpretation
	 *            the interpretation
	 * @param strategy
	 *            the node-ordering strategy (one of MARKOVIAN or LAPLACIAN)
	 * @param threshold
	 *            the threshold
	 * @param complex
	 *            the flag to indicate whether the cut has to minimize all
	 *            intervals (true) of the node order or only intervals beginning
	 *            with the first node.
	 * @return a minimum cut under all inspected cuts (this is not necessarily a
	 *         global minimum), an empty cut if no cut below the threshold is
	 *         found.
	 */
	private static ConductanceCut getCut(Graph aGraph,
			GraphInterpretation anInterpretation,
			ConductanceStrategies strategy, double threshold, boolean complex) {
		calls++; // for algorithm evaluation

		/* eliminate trivial cuts */
		if (!GraphConnectivity.isConnected(aGraph)) {
			trivialCuts++; // for algorithm evaluation
			NodeList[] comp = GraphConnectivity.connectedComponents(aGraph);
			return new ConductanceCut(comp[0].nodes(), 0);
		}

		/* get node ordering */
		NodeList spectralNodeList = nodeOrder(aGraph, anInterpretation,
				strategy);

		/* get minimal cut */
		ConductanceCut linearCut = linearCut(aGraph, anInterpretation,
				spectralNodeList.toNodeArray(), complex, threshold);

		// for algorithm evaluation
		if (linearCut.innerNodes.size() != aGraph.nodeCount()
				&& linearCut.innerNodes.size() != 0) {
			nontrivialCuts++;
		}

		return linearCut;
	}

	/**
	 * Returns the node ordering according to the given strategy
	 * 
	 * @param aGraph
	 *            the graph to order nodes
	 * @param anInterpretation
	 *            the graph's interpretation
	 * @param strategy
	 *            the strategy
	 * @return a node order
	 */
	private static NodeList nodeOrder(Graph aGraph,
			GraphInterpretation anInterpretation, ConductanceStrategies strategy) {

		/* instantiate matrix to take most significant eigenvectors from */
		GraphMatrix strategyMatrix = null;
		switch (strategy) {
			case Markovian : {
				strategyMatrix = new MarkovianGraphMatrix(aGraph,
						anInterpretation);
				break;
			}
			case Laplacian : {
				strategyMatrix = new LaplacianGraphMatrix(aGraph,
						anInterpretation);
				break;
			}
		}
		return strategyMatrix.getSpectralNodeList();
	}

	/**
	 * Basic method finding a subsequence of a given array of nodes which
	 * induces a cut minimizing the term <i>cutted/(cutted+2*min(inner, outer))
	 * </i> for <i>cutted </i>:=weight of cutted edges, <i>inner </i>:=weight of
	 * inner edges and <i>outer </i>:=weight of outer edges.
	 * 
	 * @param aGraph
	 *            the graph
	 * @param anInterpretation
	 * @param orderedNodes
	 *            the array of ordered nodes
	 * @param threshold
	 *            the threshold
	 * @param complex
	 *            the flag to choose whether to inspect only subsequences
	 *            starting with the first node or all subsequences
	 * @return the cut (by 'inner' nodes), empty if no cut with conductance
	 *         below threshold was found
	 */
	private static ConductanceCut linearCut(Graph aGraph,
			GraphInterpretation anInterpretation, Node[] orderedNodes,
			boolean complex, double threshold) {

		double edgeSum = anInterpretation.getEdgeWeightSum();

		// store the nodes positions for reverse lookup
		int[] indices = new int[orderedNodes.length];
		for (int c = 0; c < indices.length; c++) {
			indices[orderedNodes[c].index()] = c;
		}

		// variables containing the best cut found
		double bestIndex = threshold; // all values are in (0, 1)
		int bestFirst = 0;
		int bestLast = 0;

		int firstTo = complex ? aGraph.nodeCount() : 0;
		for (int first = 0; first <= firstTo; first++) {

			double inner = 0;
			double cutted = 0;
			double outer = edgeSum;

			for (int last = first; last < orderedNodes.length - 1; last++) {
				Node currentNode = orderedNodes[last];

				for (EdgeCursor ec = currentNode.edges(); ec.ok(); ec.next()) {
					Node opposite = ec.edge().opposite(currentNode);
					int index = indices[opposite.index()];
					double weight = anInterpretation.getWeight(ec.edge());
					if (index < first | index > last) { // edge to outer node
						outer -= weight;
						cutted += weight;
					} else if (index != last) { // edge to different inner node
						cutted -= weight;
						inner += weight;
					} else { // loop from underlying graph come twice!
						outer -= weight / 2;
						inner += weight / 2;
					}
				}
				double min = ((inner < outer) ? inner : outer);
				if (!anInterpretation.isDirected())
					min *= 2;
				double denom = cutted + min;
				// check whether this cut improves the index
				if (denom > 0) {
					double index = cutted / denom;
					if (index < bestIndex) {
						bestFirst = first;
						bestLast = last;
						bestIndex = index;
					}
				}
			}
		}

		// construct resulting lists
		NodeList resultList = null;

		if (bestIndex == threshold) {
			return new ConductanceCut(aGraph.nodes(), 1);
		}

		NodeList[] result = new NodeList[2];
		result[0] = new NodeList();
		result[1] = new NodeList();
		for (int i = 0; i < bestFirst; i++) {
			result[1].add(orderedNodes[i]);
		}
		for (int i = bestFirst; i <= bestLast; i++) {
			result[0].add(orderedNodes[i]);
		}
		for (int i = bestLast + 1; i < orderedNodes.length; i++) {
			result[1].add(orderedNodes[i]);
		}

		resultList = result[0];
		if ((result[1].size() < resultList.size() && result[1].size() != 0)
				|| resultList.size() == 0) {
			resultList = result[1];
		}

		return new ConductanceCut(resultList.nodes(), bestIndex);
	}

	/* additional information */

	/** counts calls to static getCut method */
	private static long calls = 0;

	/** counts trivial cuts (not connected) */
	private static long trivialCuts = 0;

	/** counts number of computed cuts below threshold */
	private static long nontrivialCuts = 0;

	/**
	 * @return calls to any 'getCut' method
	 */
	public static long getCalls() {
		return calls;
	}

	/**
	 * @return number of non-trivial cuts below threshold
	 */
	public static long getNontrivialCuts() {
		return nontrivialCuts;
	}

	/**
	 * @return number of trivial cuts (not connected)
	 */
	public static long getTrivialCuts() {
		return trivialCuts;
	}

	/**
	 * @return number of calculations
	 */
	public static long getComputations() {
		return calls - trivialCuts;
	}
}