/*
 * Created on May 18, 2004
 * Reviewed 1.2 (Mar 29, 2005)
 *
 * Copyright 2004.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.graphs.linalg;

import java.util.Comparator;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.EdgeList;
import y.base.EdgeMap;
import y.base.Graph;
import de.uka.algo.math.linalg.generic.Vector;

/**
 * A EdgeMapVector is a Vector with an EdgeMap as underlying structure. Since the
 * EdgeMap is not copied nor cloned, changing the graph or the EdgeMap
 * externally will change the Vector's behavior as well.
 * 
 * @author Bastian Katz
 */
public class EdgeMapVector extends Vector implements Comparator<Edge> {

	/** the underlying EdgeMap */
	private EdgeMap edgeMap = null;

	/** the graph the EdgeMap belongs to */
	private Graph graph = null;

	/**
	 * Instantiates a new EdgeMapVector for the given graph and EdgeMap.
	 * 
	 * @param anEdgeMap
	 *            the underlying EdgeMap
	 * @param aGraph
	 *            the graph the EdgeMap belongs to
	 */
	public EdgeMapVector(Graph aGraph) {
		if (aGraph == null)
			throw new IllegalArgumentException();
		this.graph = aGraph;
		this.edgeMap = aGraph.createEdgeMap();
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#finalize()
	 */
	@Override
	protected void finalize() throws Throwable {	
		graph.disposeEdgeMap(edgeMap);
		super.finalize();
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.Vector#assign(int, double)
	 */
	@Override
	public void assign(int index, double value) {
		assign(graph.getEdgeArray()[index], value);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.VectorAdapter#size()
	 */
	@Override
	public int size() {
		return graph.edgeCount();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.VectorAdapter#get(int)
	 */
	@Override
	public double get(int index) {
		return edgeMap.getDouble(graph.getEdgeArray()[index]);
	}

	/**
	 * Constructs a list containing all the graph's edges sorted by this
	 * vector's values.
	 * 
	 * @return the ordered list
	 */
	public EdgeList orderedList() {
		EdgeList result = new EdgeList(graph.edges());
		result.sort(this);
		return result;
	}

	/**
	 * Constructs a list containing all the graph's edges sorted by this
	 * vector's values.
	 * 
	 * @return the ordered list's cursor
	 */
	public EdgeCursor orderedCursor() {
		return orderedList().edges();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
	 */
	public int compare(Edge arg0, Edge arg1) {
		double v1 = edgeMap.getDouble(arg0);
		double v2 = edgeMap.getDouble(arg1);
		return Double.compare(v1, v2);
	}

	/**
	 * @return Returns the underlying EdgeMap.
	 */
	public EdgeMap getEdgeMap() {
		return edgeMap;
	}

	/**
	 * @param edge
	 * @param value
	 */
	public void assign(Edge edge, double value) {
		edgeMap.setDouble(edge, value);
	}
}