/*
 * Created on May 7, 2004
 * Reviewed 1.21 (Mar 31, 2005)
 *
 * Copyright 2004.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.graphs.linalg;

import java.util.Iterator;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.EdgeList;
import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeList;
import y.base.YCursor;
import y.base.YList;
import de.uka.algo.graphs.GraphInterpretation;
import de.uka.algo.math.linalg.IterationFailedException;
import de.uka.algo.math.linalg.WrongDimensionException;
import de.uka.algo.math.linalg.generic.Cell2D;
import de.uka.algo.math.linalg.generic.CellAggregator;
import de.uka.algo.math.linalg.generic.Matrix;
import de.uka.algo.math.linalg.generic.MatrixAdapter;
import de.uka.algo.math.linalg.generic.Vector;
import de.uka.algo.math.linalg.generic.VectorAdapter;
import de.uka.algo.math.linalg.vectors.ArrayVector;
import de.uka.algo.math.linalg.vectors.HashVector;

/**
 * The class GraphMatrix is an adapter class wrapping graphs to extend them by
 * some operations of 2D-matrices such as matrix-vector-multiplication. It
 * dynamically changes w/o notice if the graph changes, i. e. adding a node may
 * result in changing the size of the GraphMatrix, changing the node order in
 * permuting the rows and/or columns.
 * 
 * @author Bastian Katz
 */
public abstract class GraphMatrix extends MatrixAdapter {

	/** the underlying graph */
	protected Graph graph = null;

	/** the graph's interpretation */
	protected GraphInterpretation interpretation = null;
	
	protected double coefficientEV;
	protected double offsetEV;
	protected VectorAdapter scalarProductEV;
	protected YList prohibitedEV;

	/**
	 * Instantiates a GraphMatrix of the given underlying graph.
	 * 
	 * @param aGraph
	 *            the graph
	 * @param anInterpretation
	 *            the graph's interpretation
	 */
	public GraphMatrix(Graph aGraph, GraphInterpretation anInterpretation) {
		this.graph = aGraph;
		this.interpretation = anInterpretation;
		coefficientEV = 0.0; 
		offsetEV = 0.0; 
		scalarProductEV = null;
		prohibitedEV = null;
	}
	
	public GraphMatrix(GraphMatrix other) {
		this.graph = other.graph;
		this.interpretation = other.interpretation;
		this.coefficientEV = other.coefficientEV; 
		this.offsetEV = other.offsetEV; 
		this.scalarProductEV = other.scalarProductEV;
		this.prohibitedEV = other.prohibitedEV;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.matrix.MatrixAdapter#mult(de.uka.algo.math.linalg.vectors.VectorAdapter,
	 *      de.uka.algo.math.linalg.vectors.Vector)
	 */
	@Override
	public void mult(VectorAdapter preimage, Vector image) {
		if (preimage == null || image == null)
			throw new NullPointerException();
		if (preimage.size() != columns() || image.size() != rows())
			throw new WrongDimensionException();

		double[] preimageAsArray = preimage.getArray();
		double[] resultAsArray = new double[image.size()];
		for (Iterator<Cell2D> ci = nonZeros(); ci.hasNext();) {
			Cell2D c = ci.next();
			resultAsArray[c.row] += preimageAsArray[c.column] * c.value;
		}
		image.assign(resultAsArray);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.matrix.MatrixAdapter#toArray()
	 */
	@Override
	public double[][] toArray() {
		double[][] result = new double[rows()][];
		for (int initRow = 0; initRow < rows(); initRow++) {
			result[initRow] = new double[columns()];
		}
		for (Iterator<Cell2D> ci = nonZeros(); ci.hasNext();) {
			Cell2D cell = ci.next();
			result[cell.row][cell.column] = cell.value;
		}
		return result;
	}

	/**
	 * Returns the graph's nodes in an order induced by the most significant
	 * eigenvector. This eigenvector is taken from
	 * <code>getEigenVectors(1)</code>, s.t. it depends on this method's
	 * implementation what 'most significant' means. Nodes are ordered by this
	 * vector's entries.
	 * 
	 * @return the nodes in 'spectral order'
	 */
	public NodeList getSpectralNodeList() {
		if (graph.nodeCount() <= 1)
			return new NodeList(graph.nodes());
		YList eigenVectors = getEigenVectors(1);
		NodeMapVector eigenVector = new NodeMapVector(graph);
		if (eigenVectors.isEmpty()) {
			throw new IterationFailedException(
					"Could not determine an eigenvector ");
		}
		eigenVector.assign((Vector) eigenVectors.first());
		NodeList orderedList = eigenVector.orderedList();
		return orderedList;
	}

	/**
	 * Returns the graph's edges sorted by their length in a spectral embedding.
	 * The spectral embedding is done by taking up to <code>dimension</code>
	 * most significant eigenvectors (from
	 * <code>getEigenVectors(dimension)</code>.
	 * 
	 * @param dimension
	 *            the spectral embedding dimension (may be lowered to the number
	 *            of significant eigenvalues)
	 * @return the edges in 'spectral order'
	 */
	public EdgeList getSpectralEdgeList(int dimension) {
		if (graph.edgeCount() <= 1)
			return new EdgeList(graph.edges());
		EdgeMapVector spectralDistances = new EdgeMapVector(graph);		
		YList eigenVectors = getEigenVectors(dimension);
		if (eigenVectors.isEmpty()) {
			throw new IterationFailedException(
					"Could not determine an eigenvector");
		}
		for (EdgeCursor ec = graph.edges(); ec.ok(); ec.next()) {
			Edge edge = ec.edge();
			Node source = edge.source();
			Node target = edge.target();
			int sourceIndex = source.index();
			int targetIndex = target.index();
			double overallDist = 0;
			for (YCursor c = eigenVectors.cursor(); c.ok(); c.next()) {
				VectorAdapter current = (VectorAdapter) c.current();
				double thisDist = Math.abs(current.get(sourceIndex)
						- current.get(targetIndex));
				overallDist += thisDist * thisDist;
			}
			spectralDistances.assign(edge, overallDist);
		}

		EdgeList orderedList = spectralDistances.orderedList();
		return orderedList;
	}

	/* (non-Javadoc)
	 * @see de.uka.algo.math.linalg.matrix.MatrixAdapter#getFast()
	 */
	@Override
	public Matrix getFastMatrix() {
		if (graph.edgeCount() <= graph.nodeCount()
				* Math.sqrt(graph.nodeCount())) {
			return new HashVector.Matrix(this);
		}
		return new ArrayVector.Matrix(this);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.matrix.MatrixAdapter#getEigenVectors(int)
	 */
	@Override
	public YList getEigenVectors(int dimension) {
		YList orthogonalizeTo = prohibitedEV;
		if(orthogonalizeTo==null)
			orthogonalizeTo = new YList();
		return getEigenVectors(coefficientEV, offsetEV, scalarProductEV, orthogonalizeTo, dimension);
	}
	
	protected void setEigenvectorParameters(double coefficient, double offset, VectorAdapter scalarProduct, YList prohibited) {
		coefficientEV = coefficient;
		offsetEV = offset;
		scalarProductEV = scalarProduct;
		prohibitedEV = prohibited;
	}
	
	
	/**
	 * Returns an iterator over all pairs of nodes with an adjacency different
	 * from zero (as 2D-cells).
	 * 
	 * @return the iterator
	 */
	public Iterator<Cell2D> getAdjacencies() {
		return new AdjacencyCellAggregator();
	}

	/**
	 * Helper class to collect adjacencies
	 * 
	 * @author Bastian Katz
	 */
	private class AdjacencyCellAggregator extends CellAggregator {
		public AdjacencyCellAggregator() {
			super(graph.nodeCount(), graph.nodeCount(), graph.edgeCount()
					/ graph.nodeCount());
			for (NodeCursor nc = graph.nodes(); nc.ok(); nc.next()) {
				Node source = nc.node();
				int sourceIndex = source.index();
				for (EdgeCursor ec = source.outEdges(); ec.ok(); ec.next()) {
					Edge e = ec.edge();
					Node target = e.opposite(source);
					int targetIndex = target.index();
					put(sourceIndex, targetIndex, interpretation.getWeight(e));
					if (!interpretation.isDirected()) {
						put(targetIndex, sourceIndex, interpretation.getWeight(e));
					}
				}
			}
			load();
		}
	}

}