/*
 * Created on May 5, 2004
 * Reviewed 1.9 (Mar 31, 2005)
 *
 * Copyright 2004.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.graphs.linalg;

import java.util.Iterator;

import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import y.base.YList;
import de.uka.algo.graphs.GraphInterpretation;
import de.uka.algo.math.linalg.generic.Cell2D;
import de.uka.algo.math.linalg.generic.CellAggregator;
import de.uka.algo.math.linalg.vectors.ConstantVectorAdapter;

/**
 * The LaplacianGraphMatrix is a GraphMatrix representing the Laplacian matrix
 * of the underlying graph. The Laplacian matrix is defined as D-A (for
 * undirected graphs) resp. D-A-A^T (for directed graphs). In both cases the
 * degree matrix can optionally be weighted.
 * 
 * @author Bastian Katz
 */
public class LaplacianGraphMatrix extends GraphMatrix {

	/** the weight of the degree matrix, 1 by default */
	double diagWeight = 1.0;

	/**
	 * Instantiates a LaplacianMatrix of the given underlying graph.
	 * 
	 * @param aGraph
	 *            the graph
	 * @param anInterpretation
	 *            the graph's interpretation
	 * @param diagWeight
	 *            a scalar for the degree matrix
	 */
	public LaplacianGraphMatrix(Graph aGraph,
			GraphInterpretation anInterpretation, double diagWeight) {
		this(aGraph, anInterpretation);
		this.diagWeight = diagWeight;
	}

	/**
	 * Instantiates a LaplacianMatrix of the given underlying graph.
	 * 
	 * @param aGraph
	 *            the graph
	 * @param anInterpretation
	 *            the graph's interpretation
	 */
	public LaplacianGraphMatrix(Graph aGraph,
			GraphInterpretation anInterpretation) {
		super(aGraph, anInterpretation);

		YList orthogonalizeTo = new YList();
		orthogonalizeTo.add(new ConstantVectorAdapter(columns(), 1));
		setEigenvectorParameters(-1.0, 2 * interpretation.getMaxOutDegree() + 1,
				null, orthogonalizeTo);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.MatrixAdapter#get(int, int)
	 */
	@Override
	public double get(int row, int column) {
		double result = 0;
		if (interpretation.isDirected()) {
			result -= interpretation.getAdjacency(row, column)
					+ interpretation.getAdjacency(column, row);
		} else {
			result -= interpretation.getAdjacency(row, column);
		}

		if (row == column) {
			result += diagWeight
					* interpretation.getOutDegree(graph.getNodeArray()[row]);
		}
		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.matrix.MatrixAdapter#nonZeros()
	 */
	@Override
	public Iterator<Cell2D> nonZeros() {
		return new LaplacianCellAggregator();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.MatrixAdapter#rows()
	 */
	@Override
	public int rows() {
		return graph.nodeCount();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.MatrixAdapter#columns()
	 */
	@Override
	public int columns() {
		return graph.nodeCount();
	}

	/**
	 * Helper class to aggregate the matrix' entries.
	 * 
	 * @author Bastian Katz
	 */
	private class LaplacianCellAggregator extends CellAggregator {
		public LaplacianCellAggregator() {
			super(rows(), columns(), graph.edgeCount() / graph.nodeCount());
			for (Iterator<Cell2D> ci = getAdjacencies(); ci
					.hasNext();) {
				Cell2D cell = ci.next();
				put(cell.row, cell.column, -cell.value);
				if (interpretation.isDirected()) {
					put(cell.column, cell.row, -cell.value);
				}
			}
			for (NodeCursor nc = graph.nodes(); nc.ok(); nc.next()) {
				Node node = nc.node();
				int nodeIndex = node.index();
				put(nodeIndex, nodeIndex, diagWeight
						* interpretation.getOutDegree(node));
			}
			load();
		}
	}
}