/*
 * Created on May 7, 2004
 * Reviewed 1.9 (Mar 31, 2005)
 *
 * Copyright 2004.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.graphs.linalg;

import java.util.Iterator;

import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import y.base.YList;
import de.uka.algo.graphs.GraphInterpretation;
import de.uka.algo.graphs.SimpleGraph;
import de.uka.algo.math.linalg.generic.Cell2D;
import de.uka.algo.math.linalg.generic.CellAggregator;
import de.uka.algo.math.linalg.vectors.ArrayVector;
import de.uka.algo.math.linalg.vectors.ConstantVectorAdapter;

/**
 * The class MarkovianGraphMatrix represents the normalized adjacency matrix,
 * optionally slightly changed to x*I+(1-x)*M to ensure ergodicity.
 * 
 * @author Bastian Katz
 */
public class MarkovianGraphMatrix extends GraphMatrix {

	/** a scalar to change M to unitWeight*I+(1-unitWeight)*M */
	protected double unitWeight = 0.1;

	/**
	 * Instantiates the matrix 0.1*I+0.9*M with M being a graph's normalized
	 * adjacency matrix.
	 * 
	 * @param aGraph
	 *            the graph
	 * @param anInterpretation
	 *            the graph's interpretation
	 */
	public MarkovianGraphMatrix(Graph aGraph,
			GraphInterpretation anInterpretation) {
		super(aGraph, anInterpretation);

		YList orthogonalizeTo = new YList();
		orthogonalizeTo.add(new ConstantVectorAdapter(columns(), 1));
		setEigenvectorParameters(1.0, 1.0, new ArrayVector(new OutDegreeVector(
				graph, interpretation)), orthogonalizeTo);
	}

	/**
	 * Instantiates the matrix unitWeight*I+(1-unitWeight)*M with M being a
	 * graph's normalized adjacency matrix.
	 * 
	 * @param aGraph
	 *            the graph
	 * @param anInterpretation
	 *            the graph's interpretation
	 * @param unitWeight
	 *            the weight of the unit matrix. Ensures ergodicity w/o changing
	 *            the matrix' spectrum
	 */
	public MarkovianGraphMatrix(Graph aGraph,
			GraphInterpretation anInterpretation, double unitWeight) {
		this(aGraph, anInterpretation);
		this.unitWeight = unitWeight;
	}

	/**
	 * Instantiates the matrix 0.1*I+0.9*M with M being a graph's normalized
	 * adjacency matrix interpreting the graph as simple, i.e. undirected and
	 * not weighted.
	 * 
	 * @param aGraph
	 *            the graph
	 */
	public MarkovianGraphMatrix(Graph aGraph) {
		this(aGraph, new SimpleGraph(aGraph));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.MatrixAdapter#get(int, int)
	 */
	@Override
	public double get(int row, int column) {
		Node[] nodes = graph.getNodeArray();

		double result = interpretation.getAdjacency(row, column);

		result /= interpretation.getOutDegree(nodes[row]);
		result *= (1 - unitWeight);
		result += unitWeight;
		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.MatrixAdapter#nonZeros()
	 */
	@Override
	public Iterator<Cell2D> nonZeros() {
		return new MarkovianCellAggregator();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.MatrixAdapter#rows()
	 */
	@Override
	public int rows() {
		return graph.nodeCount();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.MatrixAdapter#columns()
	 */
	@Override
	public int columns() {
		return graph.nodeCount();
	}

	/**
	 * Helper class to aggregate the matrix' entries.
	 * 
	 * @author Bastian Katz
	 */
	private class MarkovianCellAggregator extends CellAggregator {
		double rowSum[] = new double[rows()];

		public MarkovianCellAggregator() {
			super(rows(), columns(), graph.edgeCount() / graph.nodeCount());

			for (Iterator<Cell2D> ci = getAdjacencies(); ci
					.hasNext();) {
				Cell2D cell = ci.next();
				put(cell.row, cell.column, cell.value);
				rowSum[cell.row] += cell.value;
			}

			for (NodeCursor nc = graph.nodes(); nc.ok(); nc.next()) {
				Node node = nc.node();
				int nodeIndex = node.index();
				if (rowSum[nodeIndex] == 0.0) {
					put(nodeIndex, nodeIndex, 1.0);
				} else {
					put(nodeIndex, nodeIndex, unitWeight * rowSum[nodeIndex]
							/ (1 - unitWeight));
				}
			}
			load();
		}

		@Override
		public Cell2D next() {
			Cell2D result = super.next();
			if (rowSum[result.row] == 0.0) {
				if ((result.row == result.column)) {
					result.value = 1.0;
				}
			} else {
				result.value *= ((1 - unitWeight) / rowSum[result.row]);
			}
			return result;
		}
	}
}