/*
 * Created on May 18, 2004
 * Reviewed 1.5 (Mar 31, 2005)
 *
 * Copyright 2004.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.graphs.linalg;

import java.util.Comparator;
import java.util.Iterator;

import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeList;
import y.base.NodeMap;
import de.uka.algo.math.linalg.generic.Cell1D;
import de.uka.algo.math.linalg.generic.Vector;
import de.uka.algo.math.linalg.generic.VectorAdapter;

/**
 * A NodeMapVector is an Vector with a NodeMap as underlying structure. 
 * 
 * @author Bastian Katz, bastian.katz@kit.edu
 * 
 */
public class NodeMapVector extends Vector implements Comparator<Node> {

	/** the underlying NodeMap */
	private NodeMap nodeMap = null;

	/** the graph the NodeMap belongs to */
	private Graph graph = null;

	/**
	 * Instantiates a new NodeMapVector for the given graph.
	 * 
	 * @param nodeMap
	 *            the underlying NodeMap
	 * @param aGraph
	 *            the graph the NodeMap belongs to
	 */
	public NodeMapVector(Graph aGraph) {
		if (aGraph == null)
			throw new IllegalArgumentException();
		this.graph = aGraph;
		this.nodeMap = graph.createNodeMap();
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#finalize()
	 */
	@Override
	protected void finalize() throws Throwable {
		super.finalize();
		graph.disposeNodeMap(nodeMap);
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.vectors.Vector#assign(de.uka.algo.math.linalg.vectors.VectorAdapter)
	 */
	@Override
	public void assign(VectorAdapter aVector) {
		Node[] nodes = graph.getNodeArray();
		for (Iterator<Cell1D> ci = aVector.nonZeros(); ci.hasNext();) {
			Cell1D cell = ci.next();
			nodeMap.setDouble(nodes[cell.getIndex()], cell.value);
		}
	}
	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.Vector#assign(int, double)
	 */
	@Override
	public void assign(int row, double value) {
		nodeMap.setDouble(graph.getNodeArray()[row], value);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.VectorAdapter#size()
	 */
	@Override
	public int size() {
		return graph.nodeCount();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.uka.algo.math.linalg.VectorAdapter#get(int)
	 */
	@Override
	public double get(int index) {
		return nodeMap.getDouble(graph.getNodeArray()[index]);
	}

	/**
	 * Constructs a list containing all the graph's nodes sorted by their
	 * coefficients.
	 * 
	 * @return the ordered list
	 */
	public NodeList orderedList() {
		NodeList result = new NodeList(graph.nodes());
		result.sort(this);
		return result;
	}

	/**
	 * Constructs a list containing all the graph's nodes sorted by their
	 * coefficients.
	 * 
	 * @return the ordered list's cursor
	 */
	public NodeCursor orderedCursor() {
		return orderedList().nodes();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
	 */
	public int compare(Node arg0, Node arg1) {
		double v1 = nodeMap.getDouble(arg0);
		double v2 = nodeMap.getDouble(arg1);
		return Double.compare(v1, v2);
	}

	/**
	 * @return Returns the underlying NodeMap.
	 */
	public NodeMap getNodeMap() {
		return nodeMap;
	}
}