/**
 * Created on Okt 21, 2007
 */
package de.uka.algo.graphs.traverse.bfs;

import java.util.Collection;
import java.util.Queue;
import java.util.Vector;
import java.util.concurrent.ConcurrentLinkedQueue;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.EdgeMap;
import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeMap;

/**
 * The class implements the (undirected) breadth first search.
 * Run the search using <code>run()</code>
 * To use the BFS, implement the root, traverse and done methods
 * of the BFSModule Interface and register your class via <code>addModule()</code>.
 * This way it is possible to run several calculations implemented in different
 * classes during one search.
 * When the BFS class is no longer needed, the <code>cleanUp()</code> method has
 * to be called in order to destroy all registered maps in the graph.
 * 
 * @author Ulrich Geilmann
 */
public class BFS {
	
	public enum EdgeType {
		NotVisited, TreeEdge, ForwardEdge, BackwardEdge, CrossEdge
	};
	
	/**
	 * The Graph to search in
	 */
	protected Graph graph;

	/**
	 * Storing the marking of all edges
	 */
	private EdgeMap marking;
	
	/**
	 * The BFS-Layer of a node is its distance from the root-node
	 * The root-node has layer 0, successors of it 1 and so on 
	 */
	private NodeMap bfsLayer;
	
	/**
	 * The queue used for the search
	 */
	private Queue<Node> q;
	
	/**
	 * The modules whose root, traverse and done methods are called during the search
	 */
	private Collection<BFSModule> modules;
	
	/**
	 * @param g
	 *	The Graph to perform the breadth first search in
	 */
	public BFS ( Graph g ) {
		graph = g;
		marking = graph.createEdgeMap ();
		bfsLayer = graph.createNodeMap ();
		modules = new Vector<BFSModule>();
		q = new ConcurrentLinkedQueue<Node>();
	}
	
	/**
	 * Initialization of all modules. This method resets all results of
	 * previously run searches.
	 */
	public void initAll () {
		for ( BFSModule m : modules )
			m.init (graph);
	}
	
	/**
	 * Add an implementation of root, traverse and done.
	 * The new module will be initialized.
	 * @param m
	 * 	The module
	 */
	public void addModule ( BFSModule m ) {
		modules.add(m);
		m.init(graph);
	}
	
	/**
	 * Unregister a module
	 * @param m
	 * 	The module
	 */
	public void removeModule (BFSModule m) {
		modules.remove(m);
	}
	
	/**
	 * Run the breadth-first search
	 * @param s
	 * 	The root node for the BFS
	 */
	public void run ( Node s ) {
		for (NodeCursor c = graph.nodes(); c.ok(); c.next())
			bfsLayer.setInt(c.node(), -1);
		for (EdgeCursor c = graph.edges(); c.ok(); c.next())
			marking.set(c.edge(), EdgeType.NotVisited);
		
		q.clear();
		q.add ( s );
		bfsLayer.setInt (s, 0);
		for ( BFSModule m : modules )
			m.root (s);
		
		while ( !q.isEmpty() ) {
			Node v = q.poll();
			for ( EdgeCursor c = v.edges(); c.ok(); c.next() ) {
				Edge e = c.edge();
				
				if ( getMarking(e) == EdgeType.NotVisited ) {
					Node w = e.opposite(v);
					mark ( v, e, w );
					if ( bfsLayer.getInt(w) < 0 ) {
						bfsLayer.setInt(w, bfsLayer.getInt(v) + 1);
						q.add ( w );
					}
					
					for ( BFSModule i : modules )
						i.traverse ( v, bfsLayer.getInt(v), e, getMarking(e), w, bfsLayer.getInt(w) );
				}
			}
		}
		for ( BFSModule m : modules )
			m.done ( s );
	}
	
	/**
	 * Mark an edge as TreeEdge, ForwardEdge, BackwardEdge or CrossEdge,
	 * depending on the current state of the search
	 * 
	 * @param source
	 * 	The source node of the edge
	 * @param e
	 * 	The edge to be marked
	 * @param target
	 * 	The target node of the edge
	 */
	private void mark (Node source, Edge e, Node target) {
		int sourceLayer = bfsLayer.getInt(source);
		int targetLayer = bfsLayer.getInt(target);
		if (getMarking(e) == EdgeType.NotVisited) {
			if (targetLayer < 0) {
				marking.set(e, EdgeType.TreeEdge);
			} else {
				if (sourceLayer < targetLayer) {
					marking.set(e, EdgeType.ForwardEdge);
				} else if (sourceLayer == targetLayer) {
					marking.set(e, EdgeType.CrossEdge);
				} else
					marking.set(e, EdgeType.BackwardEdge);
			}
		}
	}
	
	/**
	 * @return
	 * 	The marking of the given edge
	 */
	protected EdgeType getMarking ( Edge e ) {
		return (EdgeType)marking.get(e);
	}
	
	/**
	 * The BFS-Layer of a node is its distance from the root-node
	 * The root-node has layer 0, successors of it 1 and so on
	 * @return
	 * 	The BFS-Layer of the given node calculated in the last search performed
	 */
	public int getLayer (Node n) {
		return bfsLayer.getInt(n);
	}
	
	public void cleanUp() {
		/* Destroy the Edge and NodeMap within the Graph */
		if (graph != null) {
			if (marking != null) {
				graph.disposeEdgeMap(marking);
				marking = null;
			}
			if (bfsLayer != null) {
				graph.disposeNodeMap(bfsLayer);
				bfsLayer = null;
			}
			graph = null;
		}
		for ( BFSModule m : modules )
			m.cleanUp();
	}
	
	/**
	 * Cleaning up
	 * @see java.lang.Object#finalize()
	 */
	protected void finalize() throws Throwable {
		cleanUp();
		super.finalize();
	}
}
