/*
 * Created on Aug 28, 2004
 * Reviewed 1.2 (Mar 27, 2005)
 *
 * Copyright 2004.  --- All rights reserved. 
 * Algorithmic Group, Karlsruhe, Germany.
 */
package de.uka.algo.graphs.util;

import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeMap;

/**
 * A MappedNodeCursor is a NodeCursor that iterates over a set of nodes given by
 * a NodeCursor and a node-to-node mapping. One can use it whereever a set of
 * nodes is passed as a NodeCursor without explicitly constructing the set as a
 * list.
 * 
 * @author Bastian Katz
 */
public class MappedNodeCursor implements NodeCursor {

	/** The underlying NodeCursor */
	private NodeCursor preimages;

	/** the node-to-node mapping */
	private NodeMap mapping;

	/**
	 * Instantiates a new MappedNodeCursor for a given NodeCursor iterating over
	 * a set of nodes and a mapping of those nodes to nodes.
	 * 
	 * @param preimages
	 *            the cursor iterating over the preimages. This cursor will be
	 *            wrapped, i.e. it will not be set to the first element, unless
	 *            <code>toFirst()</code> is called. Concurrent changes will
	 *            take effect onto the MappedNodeCursor and changes to the
	 *            MappedNodeCursor will change the underlying NodeCursor as
	 *            well.
	 * @param mapping a mapping from the cursor's nodes to nodes 
	 */
	public MappedNodeCursor(NodeCursor preimages, NodeMap mapping) {
		this.preimages = preimages;
		this.mapping = mapping;
	}

	/* (non-Javadoc)
	 * @see y.base.NodeCursor#node()
	 */
	public Node node() {
		return (Node) mapping.get(preimages.node());
	}

	/* (non-Javadoc)
	 * @see y.base.NodeCursor#cyclicNext()
	 */
	public void cyclicNext() {
		preimages.cyclicNext();
	}

	/* (non-Javadoc)
	 * @see y.base.NodeCursor#cyclicPrev()
	 */
	public void cyclicPrev() {
		preimages.cyclicPrev();
	}

	/* (non-Javadoc)
	 * @see y.base.YCursor#ok()
	 */
	public boolean ok() {
		return preimages.ok();
	}

	/* (non-Javadoc)
	 * @see y.base.YCursor#next()
	 */
	public void next() {
		preimages.next();
	}

	/* (non-Javadoc)
	 * @see y.base.YCursor#prev()
	 */
	public void prev() {
		preimages.prev();
	}

	/* (non-Javadoc)
	 * @see y.base.YCursor#toFirst()
	 */
	public void toFirst() {
		preimages.toFirst();
	}

	/* (non-Javadoc)
	 * @see y.base.YCursor#toLast()
	 */
	public void toLast() {
		preimages.toLast();
	}

	/* (non-Javadoc)
	 * @see y.base.YCursor#current()
	 */
	public Object current() {
		return mapping.get(preimages.current());
	}

	/* (non-Javadoc)
	 * @see y.base.YCursor#size()
	 */
	public int size() {
		return preimages.size();
	}
}
