/****************************************************************************
 **
 ** This file is part of yEd Lite 2.6.0.1.
 **
 ** yWorks proprietary/confidential. Use is subject to license terms.
 **
 ** Unauthorized redistribution of this file or of an byte-code version
 ** of this file is strictly forbidden.
 **
 ** Copyright (c) 2008 by yWorks GmbH, Vor dem Kreuzberg 28, 
 ** 72070 Tuebingen, Germany. All rights reserved.
 **
 ***************************************************************************/

package yed.feature.modules;

import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import y.base.YCursor;
import y.module.YModule;
import y.module.YPackage;
import y.option.OptionHandler;
import y.util.ResourceResolver;
import yed.AbstractApplicationFeature;
import yed.GraphEditorApplication;
import yed.feature.applicationsettings.ApplicationSettingsFeature;
import yed.ui.UIActionNode;
import yed.ui.UIApplicationNode;
import yed.ui.UIGroupNode;
import yed.ui.option.ContributableOptionHandler;

/**
 * Feature that makes diverse {@link YModule} implementations available to the application. Modules are divided up into a tool and layout section.
 * Each section has its own top level menu in the application. The tools modules are specified in the resource file <code>resource/tools-modules.xml</code>.
 * The layout modules are specified in the resource file <code>resource/layout-modules.xml</code>.  
 */
public class ModuleFeature extends AbstractApplicationFeature {

  public static final String SECTION_MODULES = "MODULES";
  public static final String LAYOUT_MORPHING = "LAYOUT_MORPHING";
  public static final String SHOW_OPTIONHANDLER = "SHOW_OPTIONHANDLER";
  public static final String NONMODAL_DIALOG = "NONMODAL_DIALOG";

  List packageList;

  public ModuleFeature() {
    packageList = new ArrayList();
  }

  public void installFeature() {
    installDefaultPackages();

    ApplicationSettingsFeature settingsFeature = (ApplicationSettingsFeature) GraphEditorApplication.getInstance().getFeature(ApplicationSettingsFeature.class);
    settingsFeature.getApplicationOptionHandler().addContributor(new ModuleApplicationSetting());

  }

  public void installActions(UIApplicationNode uiNode) {
    installActionResources();
    ModuleFeature moduleFeature = (ModuleFeature) GraphEditorApplication.getInstance().getFeature(ModuleFeature.class);
    for (Iterator iter = moduleFeature.getModulePackages(); iter.hasNext();) {
      YPackage modulePackage = (YPackage) iter.next();
      UIGroupNode packageNode = createUIPackageNode(modulePackage, modulePackage.getName());
      uiNode.getMenubarNode().addNode(packageNode);
    }
  }

  private UIGroupNode createUIPackageNode(YPackage modulePackage, String prefix) {
    UIGroupNode packageNode = new UIGroupNode(prefix);
    
    for (YCursor cursor = modulePackage.getPackages(); cursor.ok(); cursor.next()) {
      YPackage subPackage = (YPackage) cursor.current();
      UIGroupNode subPackageNode = createUIPackageNode(subPackage, prefix + '.' + subPackage.getName());
      packageNode.addNode(subPackageNode);
    }

    for (YCursor cursor = modulePackage.getModuleNames(); cursor.ok(); cursor.next()) {
      String name = (String) cursor.current();
      YModule module = modulePackage.getModule(name);
      if (module == null) {
    	  throw new RuntimeException("Error during initialization of "+prefix+"."+name);
      }
    	  
      GraphEditorApplication.getInstance().getResourceManager().addResource(module.getClass().getName());
      ExecuteModuleAction action = new ExecuteModuleAction(getGraphEditor(), module, prefix + '.' + name);
      packageNode.addNode(new UIActionNode(action));
    }
    return packageNode;
  }

  public void addModulePackage(YPackage modulePackage) {
    packageList.add(modulePackage);
  }

  public void removeModulePackage(YPackage modulePackage) {
    packageList.remove(modulePackage);
  }

  public Iterator getModulePackages() {
    return Collections.unmodifiableList(packageList).iterator();
  }

  protected void installDefaultPackages() {
    final String prefix = GraphEditorApplication.APPLICATION_DIRECTORY_NAME;
    URL url, url2;
    url = ResourceResolver.getResource("layout-modules.xml", prefix + "/layout-modules.xml", "yed/feature/modules/resource/layout-modules.xml");
    if (url != null) {
      addModulePackage(new YPackage(url));
    }
    url = ResourceResolver.getResource("tools-modules.xml", prefix + "/tools-modules.xml", "yed/feature/modules/resource/tools-modules.xml");
    if (url != null) {
      addModulePackage(new YPackage(url));
    }
    url = ResourceResolver.getResource("clustering-modules.xml", prefix + "/clustering-modules.xml", "yed/feature/modules/resource/clustering-modules.xml");
    if (url != null) {
      addModulePackage(new YPackage(url));
    }
  }

  class ModuleApplicationSetting implements ContributableOptionHandler.Contributor {

    public void installItems(OptionHandler oh) {
      oh.useSection(SECTION_MODULES);
      oh.addBool(SHOW_OPTIONHANDLER, showModuleOptionsEnabled);
      oh.addBool(NONMODAL_DIALOG, nonModalDialogEnabled);
      oh.addBool(LAYOUT_MORPHING, layoutMorphingEnabled);
    }

    public void commitValues(OptionHandler oh) {
      layoutMorphingEnabled = oh.getBool(LAYOUT_MORPHING);
      nonModalDialogEnabled = oh.getBool(NONMODAL_DIALOG);
      showModuleOptionsEnabled = oh.getBool(SHOW_OPTIONHANDLER);
    }

    public void updateValues(OptionHandler oh) {
      //nop
    }
  }

  boolean layoutMorphingEnabled = true;
  public boolean isLayoutMorphingEnabled() {
    return layoutMorphingEnabled;
  }

  boolean nonModalDialogEnabled = true;
  public boolean isNonModalDialogEnabled() {
    return nonModalDialogEnabled;
  }

  boolean showModuleOptionsEnabled = true;
  public boolean isShowModuleSettingsEnabled() {
    return showModuleOptionsEnabled;
  }


}
